/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgVisitor;
import ru.ispras.retrascope.model.cfg.walker.CfgWalker;

import java.util.Arrays;
import java.util.Map;

/**
 * Abstract engine for control flow graph model elaboration.
 *
 * <p>Runs internal walkers on the control flow graph model.</p>
 *
 * <p>Note that while executed, walkers will be used
 * in the same order they appeared in constructor.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class CfgEngine extends Engine {

  private CfgWalker[] walkers;

  /**
   * Constructs object with the specified identifier, input/output types,
   * walkers.
   *
   * @param engineId Engine identifier.
   * @param inputType Engine model input type.
   * @param outputType Engine model output type.
   * @param walkers Model walkers.
   *
   * @throws IllegalArgumentException when walkers array is {@code null}.
   *
   */
  public CfgEngine(
      final String engineId,
      final EntityType inputType,
      final EntityType outputType,
      final CfgWalker... walkers) {

    super(engineId);

    InvariantChecks.checkNotNull(walkers);

    this.addInputType(inputType);
    this.setOutputType(outputType);

    this.walkers = walkers;
  }

  /**
   * Initializes engine with the specified inputs.
   *
   * @param inputs Engine inputs.
   */
  public abstract void initialize(final Map<EntityType, Entity> inputs);

  @Override
  public Entity start(final Map<EntityType, Entity> inputs) {

    initialize(inputs);

    final CfgModel cfgModel = (CfgModel) inputs.get(CfgModel.TYPE);

    for (final CfgWalker walker : getWalkers()) {
      walker.start(cfgModel);
    }
    return getOutput();
  }

  /**
   * Returns engine output entity.
   *
   * @return Output entity.
   */
  public abstract Entity getOutput();

  /**
   * Return engine walkers.
   *
   * @return Internal walkers.
   */
  public CfgWalker[] getWalkers() {
    return Arrays.copyOf(this.walkers, this.walkers.length);
  }

  /**
   * Sets the specified walker sequence.
   *
   * @param walkers New sequence of walkers.
   */
  public final void setWalkers(final CfgWalker... walkers) {
    this.walkers = walkers;
  }

  /**
   * Returns initial visitor.
   *
   * @return Visitor.
   */
  public final CfgVisitor getInitVisitor() {
    return this.walkers[0].getVisitor();
  }
}
