/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.extractor.iface;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.result.cfginterface.CfgInterface;
import ru.ispras.retrascope.result.cfginterface.CfgSignalInterface;

import java.util.Collection;
import java.util.Map;

/**
 * Interface extractor for control flow graph model.
 *
 * <p>Engine that extracts {@link ru.ispras.retrascope.result.cfginterface.CfgInterface} interface
 * from {@link ru.ispras.retrascope.model.cfg.CfgModel} control flow graph model.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgInterfaceExtractor extends Engine {

  /* The engine identifier. */
  public static final String ENGINE_ID = "cfg-cfginterface-extractor";

  /**
   * Default constructor.
   */
  public CfgInterfaceExtractor() {
    super(CfgInterfaceExtractor.ENGINE_ID);

    addInputType(CfgModel.TYPE);
    setOutputType(CfgInterface.TYPE);
  }

  @Override public Entity start(final Map<EntityType, Entity> inputs) {
    final CfgModel cfgModel = (CfgModel)inputs.get(CfgModel.TYPE);

    return getInterface(cfgModel);
  }

  private CfgInterface getInterface(final CfgModel cfgModel) {
    InvariantChecks.checkNotNull(cfgModel);

    final CfgInterface cfgInterface = new CfgInterface();

    final Collection<CfgModule> modules = cfgModel.getModules();
    final int size = modules.size();
    int moduleCounter = 0;
    if (!modules.isEmpty()) {
      for (final CfgModule module : modules) {
        addSignalInterface(cfgInterface, module);

        /* The module has sub-modules. */
        if (!module.getModules().isEmpty()) {
          getSubModulesInterface(cfgInterface, module.getModules());
        }
        moduleCounter++;

        setProgress((moduleCounter / size) * 100);
      }
    }
    return cfgInterface;
  }

  /* Fills interface object by modules' data. */
  private void getSubModulesInterface(
      final CfgInterface cfgInterface,
      final Collection<CfgModule> modules) {

    for (final CfgModule module : modules) {

      addSignalInterface(cfgInterface, module);

      final Collection<CfgModule> subModules = module.getModules();
      if (!subModules.isEmpty()) {
        getSubModulesInterface(cfgInterface, subModules);
      }
    }
  }

  private void addSignalInterface(
      final CfgInterface cfgInterface,
      final CfgModule module) {

    cfgInterface.addInterface(module.getFullName(),
        new CfgSignalInterface(module.getInputs(), module.getOutputs()));
  }
}
