/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.generator.test.random;

import org.apache.commons.cli.ParseException;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.randomizer.Randomizer;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.basis.TestGeneratorEngine;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.IntegerRange;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.result.cfginterface.CfgInterface;
import ru.ispras.retrascope.result.cfginterface.CfgSignalInterface;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.TestModel;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.Vector;
import ru.ispras.retrascope.util.LogMessage;

import java.io.File;
import java.math.BigInteger;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Random test generator for Control Flow Graph representation model (precisely, it's interface
 * representation).
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgRandomTestGenerator extends TestGeneratorEngine {

  class ClockInfo {
    private final String name;
    private final int activeLevel;

    ClockInfo(
      final String name,
      final int level) {

      this.name = name;
      this.activeLevel = level;
    }

    public String getName() {
      return this.name;
    }

    public int getLevel() {
      return this.activeLevel;
    }
  }

  class ResetInfo extends ClockInfo {
    private final int delay;

    ResetInfo(
      final String name,
      final int level,
      final int delay) {
      super(name, level);
      this.delay = delay;
    }

    public int getDelay() {
      return this.delay;
    }
  }

  /**
   * Engine identifier.
   */
  public static final String ENGINE_ID = "cfg-random-test-generator";

  /**
   * Generated test sequence length.
   */
  public static final Parameter TEST_LEN =
      new Parameter("test-len", true, "Length of generated test sequence.");

  /**
   * Random seed value.
   */
  public static final Parameter TEST_SEED = new Parameter("test-seed", true, "Test seed value.");

  /**
   * Name of clock signal.
   */
  public static final Parameter CLK_NAME = new Parameter("clk-name", true, "Name of clock signal.");

  /**
   * Active level of clock signal.
   */
  public static final Parameter CLK_LVL =
      new Parameter("clk-lvl", true, "Clock signal active level");

  /**
   * Name of reset signal.
   */
  public static final Parameter RST_NAME = new Parameter("rst-name", true, "Name of reset signal.");

  /**
   * Active level of reset signal.
   */
  public static final Parameter RST_LVL =
    new Parameter("rst-lvl", true, "Reset signal active level");

  /**
   * Delay in ticks through which the reset signal should be kept active for completed resetting.
   */
  public static final Parameter RST_DELAY = new Parameter("rst-delay", true, "Reset signal delay.");

  /* Input signal names and their possible value ranges. */
  public static final Parameter INPUT_VALUES =
      new Parameter("input-values", true, "Input signals and their possible values.");

  /**
   * Default constructor.
   */
  public CfgRandomTestGenerator() {
    super(ENGINE_ID, CfgInterface.TYPE);

    addParameter(TEST_LEN);
    addParameter(TEST_SEED);
    addParameter(CLK_NAME);
    addParameter(CLK_LVL);
    addParameter(RST_NAME);
    addParameter(RST_LVL);
    addParameter(RST_DELAY);
    addParameter(INPUT_VALUES);
  }

  @Override
  public Entity start(Map<EntityType, Entity> inputs) {

    final CfgInterface cfgInterface = (CfgInterface) inputs.get(CfgInterface.TYPE);
    final TestModel testModel = new TestModel();

    try {

      /* Initialise command line arguments. */
      final Arguments arguments =
          parseCommandLine(((Arguments) inputs.get(Arguments.TYPE)).getCommandLine());

      final int testLength = Integer.parseInt(arguments.getValue(TEST_LEN));

      final ClockInfo clockInfo =
          new ClockInfo(
              arguments.getValue(CLK_NAME, "clk"),
              Integer.parseInt(arguments.getValue(CLK_LVL, "1")));

      final ResetInfo resetInfo =
          new ResetInfo(
              arguments.getValue(RST_NAME, "rst"),
              Integer.parseInt(arguments.getValue(RST_LVL, "1")),
              Integer.parseInt(arguments.getValue(RST_DELAY, "1")));

      final Map<String, CfgSignalInterface> signalInterfaces = cfgInterface.getMapping();

      if (signalInterfaces.isEmpty()) {
        return testModel;
      }

      Randomizer.get().setSeed(Integer.parseInt(arguments.getValue(TEST_SEED, "0")));

      final Map<String, IntegerRange> inputValues =
          getInputValues(arguments.getValue(INPUT_VALUES, ""), signalInterfaces.values());

      for (final String name : signalInterfaces.keySet()) {

        final CfgSignalInterface cfgSignalInterface = signalInterfaces.get(name);
        final Test test = new Test();
        testModel.addTest(test);

        final Sequence testSequence = new Sequence();
        test.addSequence(testSequence);

        addResetVectors(testSequence, cfgSignalInterface, clockInfo, resetInfo);

        for (int i = 0; i < testLength; i++) {
          addRandomTestVector(testSequence, cfgSignalInterface, clockInfo, inputValues);
        }
      }

    } catch (ParseException e) {
      throw new RetrascopeRuntimeException(e);
    }
    return testModel;
  }

  private Map<String, IntegerRange> getInputValues(
      final String value,
      final Collection<CfgSignalInterface> interfaces) {

    final Map<String, IntegerRange> inputValues = new LinkedHashMap<>();

    if (value.length() > 0) {
      final String[] pairs = value.split(File.pathSeparator);

      for (final String inputValuePair : pairs) {
        final String[] keyValue = inputValuePair.split("=");

        final String signalName = keyValue[0];
        final String values = keyValue[1];

        final DataType signalType = getDataType(signalName, interfaces);

        if (signalType != null) {
          final IntegerRange range = getRange(values, signalType);
          inputValues.put(signalName, range);
        }
      }
    }
    return inputValues;
  }

  /* String 'values' has the following form: "[x,y]" */
  private IntegerRange getRange(String values, DataType signalType) {

    final int firstBracketIndex = values.indexOf('[');
    final int lastBracketIndex = values.indexOf(']');

    final String valuesWithoutBrackets = values.substring(firstBracketIndex + 1, lastBracketIndex);
    final String[] ranges = valuesWithoutBrackets.split(",");
    final int radix = Integer.parseInt(ranges[2]);
    return new IntegerRange(
        signalType.valueOf(ranges[0], radix).getInteger(),
        signalType.valueOf(ranges[1], radix).getInteger());
  }

  private DataType getDataType(
      final String signalName,
      final Collection<CfgSignalInterface> interfaces) {

    for (final CfgSignalInterface signalInterface : interfaces) {
      if (signalInterface.containsInput(signalName)
          || signalInterface.containsOutput(signalName)) {

        return signalInterface.getSignal(signalName).getDataType();
      }
    }
    return null;
  }

  private void addResetVectors(
    final Sequence sequence,
    final CfgSignalInterface signalInterface,
    final ClockInfo clockInfo,
    final ResetInfo resetInfo) {

    final String clkName = clockInfo.getName();
    final String rstName = resetInfo.getName();

    final NodeValue rstActiveValue = NodeValue.newInteger(resetInfo.getLevel());
    final NodeValue rstInActiveValue = NodeValue.newInteger(resetInfo.getLevel() == 0 ? 1 : 0);

    if (!containsSignal(signalInterface, resetInfo.getName())) {
      return;
    }

    final NodeVariable clkVariable = getVariable(signalInterface, clkName);
    final boolean clkExists = clkVariable != null;
    final Event enableClockEvent =
        clkExists ? getClkEvent(clkVariable, clockInfo.getLevel(), true) : null;

    /* Activate a reset for one tick. */
    final Transaction enResetTransaction = new Transaction();
    enResetTransaction.addVariable(rstName, rstActiveValue);
    addZeroOther(enResetTransaction, signalInterface, clkName, rstName);
    final Vector enResetVector = new Vector(enResetTransaction);
    if (clkExists) {
      enResetVector.addEvent(enableClockEvent);
    }
    sequence.addVector(enResetVector);

    /* Make a flip-flop till delay. */
    for (int i = 0; i < resetInfo.getDelay() - 1; i++) {

      final Transaction waitTransaction = new Transaction();
      addZeroOther(waitTransaction, signalInterface, clkName, rstName);
      final Vector enableClkVector = new Vector(waitTransaction);
      if (clkExists) {
        enableClkVector.addEvent(enableClockEvent);
      }
      sequence.addVector(enableClkVector);
    }

    /* Deactivate reset. */
    final Transaction disableResetTransaction = new Transaction();
    disableResetTransaction.addVariable(rstName, rstInActiveValue);
    addZeroOther(disableResetTransaction, signalInterface, clkName, rstName);
    final Vector disableResetVector = new Vector(disableResetTransaction);
    disableResetVector.addEvent(enableClockEvent);
    sequence.addVector(disableResetVector);
  }

  private void addZeroOther(
      final Transaction transaction,
      final CfgSignalInterface signalInterface,
      final String ... otherNames) {

    final Collection<NodeVariable> inputs = signalInterface.getInputs();
    for (final NodeVariable input : inputs) {
      if (!hasNameAmong(input, otherNames)) {
        transaction.addVariable(input.getName(), getZeroValue(input.getDataType()));
      }
    }
  }

  private boolean hasNameAmong(NodeVariable variable, String[] names) {

    for (final String name : names) {
      if (variable.getName().equals(name)) {
        return true;
      }
    }
    return false;
  }

  private Event getClkEvent(
      final NodeVariable clkVariable,
      final int activeLevel,
      final boolean enableEvent) {

    final EventType type =
        activeLevel == 0 ?
            (enableEvent ? EventType.NEGATIVE_EDGE : EventType.POSITIVE_EDGE)
            : (enableEvent ? EventType.POSITIVE_EDGE : EventType.NEGATIVE_EDGE);
    return new Event(new RangedVariable(clkVariable), type);
  }

  private NodeVariable getVariable(
      final CfgSignalInterface signalInterface,
      final String name) {
    for (final NodeVariable inputSignal : signalInterface.getInputs()) {
      if (inputSignal.getName().equals(name)) {
        return inputSignal;
      }
    }
    return null;
  }

  private void addRandomTestVector(
      final Sequence sequence,
      final CfgSignalInterface signalInterface,
      final ClockInfo clockInfo,
      final Map<String, IntegerRange> permissible) {

    final Transaction transaction = new Transaction();
    for (final NodeVariable input : signalInterface.getInputs()) {
      transaction.addVariable(input.getName(), getRandomValue(input, permissible));
    }

    final Vector vector = new Vector(transaction);
    final NodeVariable clkVariable = getVariable(signalInterface, clockInfo.getName());
    final boolean clkExists = clkVariable != null;
    if (clkExists) {
      vector.addEvent(getClkEvent(clkVariable, clockInfo.getLevel(), true));
    }
    sequence.addVector(vector);

  }

  private NodeValue getRandomValue(
      final NodeVariable variable,
      final Map<String, IntegerRange> permissible) {

    final DataType dataType = variable.getDataType();
    final String varName = variable.getName();
    final boolean hasPermissibleValues = permissible.containsKey(varName);

    BigInteger min = null;
    BigInteger max = null;
    if (hasPermissibleValues) {
      min = permissible.get(varName).getMin();
      max = permissible.get(varName).getMax();
    }

    switch (dataType.getTypeId()) {
      case BIT_VECTOR:
        final BitVector bitVector = BitVector.newEmpty(dataType.getSize());
        Randomizer.get().fill(bitVector);
        return NodeValue.newBitVector(
            hasPermissibleValues
                ? BitVector.valueOf(
                    Randomizer.get().nextBigIntegerRange(min, max),
                    dataType.getSize())
                : bitVector);

      case LOGIC_BOOLEAN:
        return NodeValue.newBoolean(Randomizer.get().nextBoolean());
      case LOGIC_INTEGER:
        return hasPermissibleValues
            ? NodeValue.newInteger(Randomizer.get().nextBigIntegerRange(min, max))
            : NodeValue.newInteger(Randomizer.get().next());
      default:
        throw new RetrascopeRuntimeException(
            LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, dataType));
    }
  }

  private NodeValue getZeroValue(final DataType dataType) {

    switch (dataType.getTypeId()) {
      case BIT_VECTOR:
        return NodeValue.newBitVector(BitVector.valueOf(0, dataType.getSize()));
      case LOGIC_BOOLEAN:
        return NodeValue.newBoolean(false);
      case LOGIC_INTEGER:
        return NodeValue.newInteger(0);
      default:
        throw new RetrascopeRuntimeException(
          LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, dataType));
    }
  }

  private boolean containsSignal(
      final CfgSignalInterface signalInterface,
      final String signalName) {

    for (final NodeVariable inputSignal : signalInterface.getInputs()) {
      if (inputSignal.getName().equals(signalName)) {
        return true;
      }
    }
    return false;
  }
}
