/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.cgaa;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.engine.basis.TransformUtils;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgUtils;

import java.util.Collections;
import java.util.Deque;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Clocked guarded atomic actions decision diagram builder.
 *
 * @see CfgCgaaTransformer
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgCgaaTransformerBuilder {

  /* The mapping of new module to old processes transformations. */
  private final Map<CfgModule, Map<CfgProcess, List<CfgTransform>>> moduleMap;

  /**
   * Constructs object with the specified mapping.
   *
   * @param map Mapping of modules to their processes and generated transformations.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public CfgCgaaTransformerBuilder(final Map<CfgModule, Map<CfgProcess, List<CfgTransform>>> map) {
    InvariantChecks.checkNotNull(map);

    this.moduleMap = map;
  }

  private Map<CfgModule, Map<CfgProcess, List<CfgTransform>>> getModuleMap() {
    return Collections.unmodifiableMap(this.moduleMap);
  }

  /**
   * Creates clocked guarded atomic action model.
   *
   * <p>Builds clocked guarded atomic actions decision diagram
   * from {@link CfgModel} control flow graph model representation.</p>
   *
   * @return Guarded actions decision diagram.
   */
  public CfgModel build() {
    final CfgModel model = new CfgModel();

    for (final CfgModule module : getModuleMap().keySet()) {
      buildModule(module);
      if (module.getParent() == null) {
        model.addModule(module);
      }
    }
    return model;
  }

  private void buildModule(final CfgModule module) {

    /* The mapping for "old" case nodes and linked sequences of block nodes. */
    final Map<CfgCaseStatement, List<CfgBlockStatement>> terminations = new LinkedHashMap<>();

    /* The mapping for old cases and new cases. */
    final Map<CfgCaseStatement, List<CfgCaseStatement>> caseMap = new LinkedHashMap<>();

    /* Apply transformations of the current module's process. */
    final Map<CfgProcess, List<CfgTransform>> ptList = getModuleMap().get(module);
    applyTransformations(terminations, caseMap, module, ptList);
  }

  private void applyTransformations(
      final Map<CfgCaseStatement, List<CfgBlockStatement>> terminations,
      final Map<CfgCaseStatement, List<CfgCaseStatement>> caseMap,
      final CfgModule module,
      final Map<CfgProcess, List<CfgTransform>> ptMap) {

    for (final Map.Entry<CfgProcess, List<CfgTransform>> entry : ptMap.entrySet()) {

      /* Create new process container. */
      final CfgProcess oldProcess = entry.getKey();
      final CfgProcess process =
          new CfgProcess(oldProcess.isInitial(), oldProcess.getSensitivityList());
      process.setVariablesMapping(oldProcess.getDeclarationMap());
      module.add(process);

      /* Build the control flow graph from the list of transformations. */
      buildCfgControlPart(process, terminations, caseMap, entry.getValue());

      /* Continue terminal case nodes with sequences of basic blocks. */
      buildCfgDataPart(terminations);
    }
  }

  private void buildCfgDataPart(final Map<CfgCaseStatement, List<CfgBlockStatement>> map) {

    for (final Map.Entry<CfgCaseStatement, List<CfgBlockStatement>> entry : map.entrySet()) {

      final CfgCaseStatement newCase = entry.getKey();

      if (CfgUtils.getSuccessors(newCase).isEmpty()) {
        final List<CfgBlockStatement> termBlocks = entry.getValue();

        if (!termBlocks.isEmpty()) {
          addBlocksAfterCase(newCase, termBlocks);
        }
      }
    }
  }

  private void addBlocksAfterCase(
      final CfgCaseStatement caseNode,
      final List<CfgBlockStatement> blocks) {

    for (final CfgBlockStatement block : blocks) {
      final CfgBlockStatement blockCopy = (CfgBlockStatement) block.deepCopy();
      caseNode.add(blockCopy);
    }
  }

  private void buildCfgControlPart(
      final CfgProcess process,
      final Map<CfgCaseStatement, List<CfgBlockStatement>> terminals,
      final Map<CfgCaseStatement, List<CfgCaseStatement>> caseMap,
      final List<CfgTransform> transforms) {

    for (final CfgTransform transform : transforms) {

      if (transform.containsSwitch()) {

        applySwitchTransform(process, caseMap, transform);
      } else {

        applyNonSwitchTransform(process, transform);
      }
      if (transform.containsMap()) {
        applyMapTransform(terminals, caseMap, transform);
      }
    }
  }

  private void applyMapTransform(
      final Map<CfgCaseStatement, List<CfgBlockStatement>> terminals,
      final Map<CfgCaseStatement, List<CfgCaseStatement>> caseMap,
      final CfgTransform transform) {

    final Map<CfgCaseStatement, List<CfgBlockStatement>> map = transform.getMap();

    for (final Map.Entry<CfgCaseStatement, List<CfgBlockStatement>> entry : map.entrySet()) {

      final List<CfgCaseStatement> newCases = caseMap.get(entry.getKey());

      for (final CfgCaseStatement newCase : newCases) {
        if (CfgUtils.getSuccessors(newCase).isEmpty() && !terminals.containsKey(newCase)) {
          terminals.put(newCase, entry.getValue());
        }
      }
    }
  }

  private void applyNonSwitchTransform(
      final CfgProcess process,
      final CfgTransform transform) {

    final List<CfgBlockStatement> blocks = transform.getBlocks();

    /* The basic block-only control flow graph - no transformation needed. */
    for (int i = 0; i < blocks.size(); i++) {
      final CfgBlockStatement oldBlock = blocks.get(i);
      final CfgBlockStatement newBlock =
          new CfgBlockStatement(oldBlock.getAssignments(), oldBlock.isConcurrent());
      process.add(newBlock, i);
    }
  }

  private void applySwitchTransform(
      final CfgProcess process,
      final Map<CfgCaseStatement, List<CfgCaseStatement>> caseMap,
      final CfgTransform transform) {

    final CfgSwitchStatement oldSwitch = transform.getSwitchNode();
    final List<CfgBlockStatement> blocks = getBlocking(transform.getBlocks());
    final Deque<CfgCaseStatement> cases = transform.getCases();

    /* Create new switch with the modified expression. */
    final Node condition = oldSwitch.getCondition();
    final CfgSwitchStatement newSwitch = oldSwitch.hasEvents()
        ? new CfgSwitchStatement(oldSwitch.getEventList().getEvents().iterator().next())
        : new CfgSwitchStatement(TransformUtils.substituteBackward(blocks, condition));

    /* Add all the cases of old switch to terminals and create new case nodes. */
    for (int i = 0; i < oldSwitch.getStatementNum(); i++) {

      final CfgCaseStatement oldCase = (CfgCaseStatement) oldSwitch.getStatement(i);
      final CfgCaseStatement newCase =
          oldCase.isDefault()
              ? new CfgCaseStatement(oldCase.isDefault())
              : new CfgCaseStatement(oldCase.getValues());

      newSwitch.add(newCase);

      /* Add new case nodes to old-new mapping. */
      if (caseMap.containsKey(oldCase)) {
        caseMap.get(oldCase).add(newCase);
      } else {
        caseMap.put(oldCase, new LinkedList<>(Collections.singleton(newCase)));
      }
    }

    /* Determine the parent node for the node switch. */
    if (cases.isEmpty()) {
      newSwitch.setParent(process);
    } else {

      /* The new incarnation of old case is already created. */
      final CfgCaseStatement tCasesPeek = cases.peek();

      final List<CfgCaseStatement> terminalCases = caseMap.get(tCasesPeek);

      /* Get cases which does not have children. */
      final List<CfgCaseStatement> noChildrenCases = new LinkedList<>();
      for (final CfgCaseStatement termCase : terminalCases) {
        if (CfgUtils.getSuccessors(termCase).isEmpty()) {
          noChildrenCases.add(termCase);
        }
      }

      /* No-child-having cases should be the parents of the new switch node. */
      if (noChildrenCases.size() == 1) {
        newSwitch.setParent(noChildrenCases.get(0));
      } else if (noChildrenCases.size() > 1) {

        final CfgSwitchStatement switchParent =
            (CfgSwitchStatement) noChildrenCases.get(0).getParent();
        final CfgStatement parentOfSwitch = switchParent.getParent();
        parentOfSwitch.add(newSwitch, parentOfSwitch.indexOf(switchParent) + 1);
      }
    }
  }

  private List<CfgBlockStatement> getBlocking(final List<CfgBlockStatement> list) {
    final List<CfgBlockStatement> blocking = new LinkedList<>();

    for (final CfgBlockStatement block : list) {
      if (!block.isConcurrent()) {
        blocking.add(block);
      }
    }

    return blocking;
  }

}
