/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.cgaa;

import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.ArrayDeque;
import java.util.Collections;
import java.util.Deque;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Visitor component of transformer to clocked guarded atomic actions model.
 *
 * @see CfgCgaaTransformer
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgCgaaTransformerVisitor extends CfgDefaultVisitor {

  /* The current new module. */
  private final Deque<CfgModule> modules;

  /* The current basic blocks collection. */
  private final List<CfgBlockStatement> blocks;

  /* The current cases stack. */
  private final Deque<CfgCaseStatement> cases;

  /* The current process transformation. */
  private List<CfgTransform> transformations;

  /* The mapping of module data to transformations. */
  private final Map<CfgModule, Map<CfgProcess, List<CfgTransform>>> moduleMap;

  private final Map<CfgModule, CfgModule> oldNewModuleMap;

  private static final String LOOP_FOUND = "Loop has been found.";

  /**
   * Default constructor.
   */
  public CfgCgaaTransformerVisitor() {
    super();
    this.modules = new ArrayDeque<>();
    this.blocks = new LinkedList<>();
    this.cases = new ArrayDeque<>();
    this.moduleMap = new LinkedHashMap<>();
    this.oldNewModuleMap = new LinkedHashMap<>();
  }

  private Deque<CfgModule> getModules() {
    return this.modules;
  }

  private List<CfgBlockStatement> getBlocks() {
    return this.blocks;
  }

  private Deque<CfgCaseStatement> getCases() {
    return this.cases;
  }

  private List<CfgTransform> getTransformations() {
    return this.transformations;
  }

  /**
   * Returns mapping that keeps the transformed control flow graph model.
   *
   * @return Transformed control flow graph model.
   */
  public Map<CfgModule, Map<CfgProcess, List<CfgTransform>>> getModuleMap() {
    return Collections.unmodifiableMap(this.moduleMap);
  }

  @Override
  public void onModuleBegin(final CfgModule module) {
    final CfgModule newModule = new CfgModule(module.getFullName());
    newModule.setVariablesMapping(module.getDeclarationMap());
    final CfgStatement moduleParent = module.getParent();
    if (moduleParent != null && moduleParent instanceof CfgModule
        && this.oldNewModuleMap.containsKey(moduleParent)) {

      newModule.setParent(this.oldNewModuleMap.get(moduleParent));
    }
    this.moduleMap.put(newModule, new LinkedHashMap<CfgProcess, List<CfgTransform>>());
    this.oldNewModuleMap.put(module, newModule);
    this.modules.push(newModule);
  }

  @Override
  public void onModuleEnd(final CfgModule module) {
    this.modules.pop();
  }

  @Override
  public void onProcessBegin(final CfgProcess process) {
    this.transformations = new LinkedList<>();
  }

  @Override
  public void onProcessEnd(final CfgProcess process) {
    this.moduleMap.get(getModules().peek()).put(process, getTransformations());
  }

  @Override
  public void onBasicBlockBegin(final CfgBlockStatement block) {

    if (getBlocks().contains(block)) {
      Log.getLogger().log(LogLevel.WARNING, LOOP_FOUND);
    }
    this.blocks.add(block);
    finalizeForTerminalStatement(block);
  }

  @Override
  public void onBasicBlockEnd(final CfgBlockStatement block) {

    /* Remove the last block. */
    final int lastBlockNum = getBlocks().size() - 1;
    this.blocks.remove(lastBlockNum);
  }

  @Override
  public void onCaseBegin(final CfgCaseStatement caseNode) {
    if (getCases().contains(caseNode)) {
      Log.getLogger().log(LogLevel.WARNING, LOOP_FOUND);
    }
    this.cases.push(caseNode);
    finalizeForTerminalStatement(caseNode);
  }

  @Override
  public void onCaseEnd(final CfgCaseStatement caseNode) {
    this.cases.pop();
  }

  @Override
  public void onSwitchBegin(final CfgSwitchStatement switchNode) {
    addTransformation(getBlocks(), getCases(), switchNode);
  }

  private void addTransformation(
      final List<CfgBlockStatement> blocks,
      final Deque<CfgCaseStatement> cases,
      final CfgSwitchStatement switchNode) {

    final CfgTransform transformation =
        new CfgTransform(
            cases,
            blocks,
            switchNode,
            new LinkedHashMap<CfgCaseStatement, List<CfgBlockStatement>>());

    this.transformations.add(transformation);
  }

  private void finalizeForTerminalStatement(CfgStatement statement) {
    if (!CfgUtils.hasSuccessors(statement)) {

      if (getCases().isEmpty()) {

        /* There is no switch\case nodes. */
        this.transformations.add(new CfgTransform(getCases(), getBlocks(), null, null));
      } else {

        /* Get last control flow graph transformation. */
        final int lastProcessTrSize = getTransformations().size();
        final CfgTransform lastTransform = getTransformations().get(lastProcessTrSize - 1);

        /* Create a copy of current blocks list. */
        final List<CfgBlockStatement> terminalBlocks = new LinkedList<>(getBlocks());

        lastTransform.addTermination(getCases().peek(), terminalBlocks);
      }
    }
  }
}
