/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.cgaa;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * CGAA constructor component that filters (removes) unreachable paths from the constructed model.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CgaaFilterPathBackend extends CfgDefaultVisitor {

  private final Map<CfgSwitchStatement, CfgCaseStatement> switchCasePairs;

  private CfgCaseStatement skipped;

  private Status status;

  public CgaaFilterPathBackend() {
    super();
    this.switchCasePairs = new LinkedHashMap<>();
  }

  private Map<CfgSwitchStatement, CfgCaseStatement> getPath() {
    return this.switchCasePairs;
  }

  @Override
  public Status getStatus() {
    return this.status;
  }

  private void setStatus(final Status status) {
    InvariantChecks.checkNotNull(status);

    this.status = status;
  }

  private CfgCaseStatement getSkipped() {
    return this.skipped;
  }

  private void setSkipped(final CfgCaseStatement caseStatement) {
    this.skipped = caseStatement;
  }

  @Override
  public void onSwitchBegin(final CfgSwitchStatement statement) {

  }

  @Override
  public void onSwitchEnd(final CfgSwitchStatement statement) {
    if (getStatus() == Status.SKIP) {
      statement.remove(getSkipped());
      setStatus(Status.OK);
    }
  }

  @Override
  public void onCaseBegin(final CfgCaseStatement statement) {
    this.switchCasePairs.put((CfgSwitchStatement) statement.getParent(), statement);

    if (!pathReachable()) {
      setStatus(Status.SKIP);
      setSkipped(statement);
    }
  }

  private boolean pathReachable() {

    final Node pathConstraint = getPathConstraint(getPath());
    return SolverUtils.isSat(pathConstraint);
  }

  private static Node getPathConstraint(Map<CfgSwitchStatement, CfgCaseStatement> controlPath) {

    final List<Node> conditions = new LinkedList<>();

    for (final Map.Entry<CfgSwitchStatement, CfgCaseStatement> entry : controlPath.entrySet()) {
      if (entry.getKey().getCondition() != null) {
        conditions.add(CfgUtils.getExpression(entry.getValue(), entry.getKey()));
      }
    }

    return conditions.isEmpty()
      ? SolverUtils.TRUE
      : ExprUtils.getConjunction(conditions.toArray(new Node[conditions.size()]));
  }

  @Override
  public void onCaseEnd(final CfgCaseStatement statement) {
    /* Do nothing. */
  }
}
