/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * Component that detects Clock-like variables in the clocked guarded atomic actions model.
 *
 * <p>Visitor of clocked guarded actions model {@link ru.ispras.retrascope.model.cgaa.CgaaModel}
 * for clock-like signals detection.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaClockDetector extends MapCfgVisitor {

  /* The clock-like variables for current process. */
  private List<NodeVariable> currentClockList;

  private void setClockList(final List<NodeVariable> clockList) {
    this.currentClockList = new LinkedList<>(clockList);
  }

  private List<NodeVariable> getClockList() {
    return Collections.unmodifiableList(this.currentClockList);
  }

  /**
   * Checks whether the specified variable is a one-bit length signal.
   *
   * @param variable variable
   *
   * @return {@code true} if variable is one bit long signal, {@code false} otherwise
   *
   * @throws RetrascopeRuntimeException in the following situations:
   *         (1) when any of arguments is {@code null};
   *         (2) when argument's user data is not of {@link VariableDescriptor} type
   */
  private static boolean isOneBitSignal(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    final Object variableData = variable.getUserData();
    if (variableData instanceof VariableDescriptor) {
      final boolean isSignal =
          VariableType.isSignal(((VariableDescriptor) variableData).getVariableType());
      final DataType dataType = variable.getDataType();

      return isSignal
            && (dataType.equals(DataType.BOOLEAN) || dataType.equals(DataType.BIT_VECTOR(1)));
    } else {
      throw RetrascopeRuntimeException.makeException(
          LogMessage.WRONG_TYPE,
          variableData.getClass());
    }
  }

  @Override
  public void onProcessBegin(final CfgProcess process) {

    setClockList(new LinkedList<NodeVariable>());

    /* Analyse sensitivity list. */
    final EventList eventList = process.getSensitivityList();
    if (eventList != null && !eventList.isEmpty()) {
      final Collection<Event> events = eventList.getEvents();
      for (final Event event : events) {
        final NodeVariable variable = event.getRangedVariable().getVariable();
        if (isOneBitSignal(variable) && !getClockList().contains(variable)) {
          this.currentClockList.add(variable);
        }
      }
    }
  }

  @Override
  public void onProcessEnd(final CfgProcess process) {

    final Set<NodeVariable> clocks = new LinkedHashSet<>();
    clocks.addAll(getClockList());
    this.getClockMap().put(process, clocks);
    this.currentClockList.clear();
  }

  @Override
  public void onBasicBlockBegin(final CfgBlockStatement block) {
    final List<Assignment> statements = block.getAssignments();
    for (final Assignment statement : statements) {

      /* Analyse the signals are used in basic blocks assignments. */
      final Collection<NodeVariable> stmtUses = statement.getUses();
      if (stmtUses != null && !stmtUses.isEmpty()) {
        for (final NodeVariable stmtUse : stmtUses) {
          if (getClockList().contains(stmtUse)) {
            this.currentClockList.remove(stmtUse);
          }
        }
      }

      /* Analyse the signals are defined in basic blocks assignments. */
      final Collection<NodeVariable> stmtDefines = statement.getDefines();
      if (stmtDefines != null && !stmtDefines.isEmpty()) {
        for (final NodeVariable stmtDefine : stmtDefines) {
          if (getClockList().contains(stmtDefine)) {
            this.currentClockList.remove(stmtDefine);
          }
        }
      }
    }
  }
}
