/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
* Component that extracts data for constructing EFSM states
* and transitions from clocked guarded atomic actions model.
*
* @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
*/
final class CgaaEfsmDataExtractor extends MapCfgVisitor {

  /* Processes and corresponding EFSM data containers. */
  private Map<CfgProcess, List<EfsmDataContainer>> processDataMap;

  /* Fields that are used throughout walking process. */

  /* The current process. */
  private CfgProcess currentProcess;

  /* The data containers for current process. */
  private List<EfsmDataContainer> dataList;

  /* The current state expression. */
  private List<Node> currentState;

  /* The current list of events. */
  private List<Event> currentEventList;

  /* The current guard. */
  private List<Node> currentGuard;

  /* The current action. */
  private List<CfgBlockStatement> currentAction;

  /* Be set to true when guard processing has been finished. */
  private boolean guardFinished;

  /* Be set to true when guard & state are satisfiable & compatible. */
  private boolean actionReachable;

  /* Cases that used into state-like variables' depending expressions. */
  private List<CfgCaseStatement> stateCases;

  /* The current clock. */
  private Event clock;
  private boolean clockEnabled;

  /* The current case. */
  private CfgCaseStatement clockCase;
  private boolean clockCaseEnabled;

  /* Diagram paths number (for statistics keeping). */
  private int pathNum;

  public CgaaEfsmDataExtractor() {
    super();
  }

  /**
   * Returns mapping of processes into data containers.
   *
   * @return mapping of processes into data containers
   */
  public Map<CfgProcess, List<EfsmDataContainer>> getProcessDataMap() {
    Log.printInfo("Number of GADD paths: " + this.pathNum);
    return this.processDataMap;
  }

  @Override
  public void onRootBegin() {
    this.processDataMap = new LinkedHashMap<>();
    this.currentState = new LinkedList<>();
    this.currentEventList = new LinkedList<>();
    this.currentGuard = new LinkedList<>();
    this.currentAction = new LinkedList<>();
    this.guardFinished = false;
    this.actionReachable = true;
    this.stateCases = new LinkedList<>();
    this.pathNum = 0;
  }

  @Override
  public void onProcessBegin(final CfgProcess process) {
    this.currentProcess = process;
    this.dataList = new LinkedList<>();
    this.currentEventList.addAll(process.getSensitivityList().getEvents());
  }

  @Override
  public void onProcessEnd(final CfgProcess process) {
    this.processDataMap.put(this.currentProcess, this.dataList);
    this.currentEventList.clear();
  }

  @Override
  public void onCaseBegin(final CfgCaseStatement caseNode) {

    final CfgSwitchStatement caseParent = (CfgSwitchStatement) caseNode.getParent();

    /* Construct a condition expression for the current case node. */
    final Node expr = CfgUtils.getExpression(caseNode, caseParent);

    if (isStateVarFreeExpr(expr)) {
      this.currentGuard.add(expr);
    } else {
      this.currentState.add(expr);
      this.stateCases.add(caseNode);
    }

    /* If switch has event, save it as clock. */
    if (caseParent.hasEvents()) {
      if (!this.clockEnabled) {
        this.clock = caseParent.getEventList().getEvents().iterator().next();
        this.clockEnabled = true;
        this.clockCase = caseNode;
        this.clockCaseEnabled = true;
      } else {
        Log.printWarn(
            String.format(
                "Unreachable path for: event1=%s, event2=%s.",
                this.clock.toString(),
                caseParent.getEventList().getEvents().iterator().next().toString()));
      }
    }
  }

  private boolean isStateVarFreeExpr(final Node expr) {
    return CollectionUtils.intersectSets(
        new LinkedHashSet<>(ExprUtils.getVariables(expr)),
        this.getStateMap().get(this.currentProcess)).isEmpty();
  }

  @Override
  public void onCaseEnd(final CfgCaseStatement caseNode) {

    processIfLast(caseNode);

    if (this.stateCases.contains(caseNode)) {
      this.currentState.remove(this.currentState.size() - 1);
      this.stateCases.remove(caseNode);
    } else {
      this.currentGuard.remove(this.currentGuard.size() - 1);
    }

    /* Visitor is ready again to process guard part elaboration. */
    if (this.guardFinished) {
      this.guardFinished = false;
    }

    if (this.clockCaseEnabled && this.clockCase.equals(caseNode)) {
      this.clockCaseEnabled = false;
      this.clockEnabled = false;
    }
  }

  @Override
  public void onBasicBlockBegin(final CfgBlockStatement block) {

    checkConditionsIfNot();

    if (this.actionReachable) {
      this.currentAction.add(block);
    }

    processIfLast(block);
  }

  private void checkConditionsIfNot() {
    if (!this.guardFinished) {
      this.actionReachable = checkActionReachable();
      this.guardFinished = true;
    }
  }

  private boolean checkActionReachable() {
    return checkInvCompliance(this.currentState) && checkInvCompliance(this.currentGuard);
  }

  /* Check the compliance of nodes from the specified collection to the invariants
  * on used variables at the specified process.*/
  private boolean checkInvCompliance(final List<Node> list) {
    InvariantChecks.checkNotNull(list);

    if (list.isEmpty()) {
      return true;
    }

    Node conjunction = ExprUtils.getConjunction(list.toArray(new Node[list.size()]));

    if (ExprUtils.areCompatible(conjunction)) {
      final Collection<Node> invariantNodes =
          CgaaEfsmTransformerUtils.getInvariants(
              this.currentProcess,
              ExprUtils.getVariables(conjunction));

      for (final Node invNode : invariantNodes) {
        conjunction = ExprUtils.getConjunction(invNode, conjunction);
      }
      return SolverUtils.areCompatible(conjunction);

    } else {
      return false;
    }
  }

  private void processIfLast(CfgStatement statement) {
    if (!CfgUtils.hasSuccessors(statement)) {
      this.pathNum++;
      if (this.actionReachable) {

        Action action;

        if (this.currentAction.isEmpty()) {
          action = new Action(new LinkedList<Assignment>());
        } else {
          final List<Assignment> assignments = new LinkedList<>();
          for (final CfgBlockStatement block : this.currentAction) {
            assignments.addAll(block.getAssignments());
          }
          action = new Action(assignments);
        }

        final Set<Node> states = getNodes(this.currentState);
        final Set<Node> guard = getNodes(this.currentGuard);

        if (states.isEmpty()) {
          states.add(SolverUtils.TRUE);
        }

        /* It is supposed that guard is reachable. */
        final Guard guardExpr = new Guard(guard);
        final GuardedAction guardedAction = new GuardedAction(guardExpr, action);
        if (!this.clockEnabled) {
          guardedAction.addClocks(this.currentEventList);
        } else {
          guardedAction.addClock(this.clock.deepCopy());
        }
        final EfsmDataContainer container = new EfsmDataContainer(states, guardedAction);
        this.dataList.add(container);
      }
      this.currentAction.clear();
    }
  }

  private Set<Node> getNodes(final List<Node> list) {
    final Set<Node> nodes = new LinkedHashSet<>();
    if (list.isEmpty()) {
      nodes.add(SolverUtils.TRUE);
    } else {
      nodes.addAll(list);
    }
    return nodes;
  }
}
