/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import org.apache.commons.cli.ParseException;
import org.apache.commons.collections15.BidiMap;
import org.apache.commons.collections15.bidimap.DualHashBidiMap;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.logic.Clause;
import ru.ispras.fortress.logic.NormalForm;
import ru.ispras.fortress.logic.Orthogonalizer;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.basis.TransformUtils;
import ru.ispras.retrascope.engine.cfg.CfgEngine;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.VariableDeclaration;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgDeclarationStatement;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgVisitor;
import ru.ispras.retrascope.model.cfg.walker.CfgWalker;
import ru.ispras.retrascope.model.cfg.walker.CfgWalkerFactory;
import ru.ispras.retrascope.model.cfg.walker.CfgWalkerMode;
import ru.ispras.retrascope.model.cgaa.CgaaModel;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmUtils;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.SolverUtils;
import ru.ispras.retrascope.util.TransformerUtils;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The transformer of clocked guarded actions container
 * {@link ru.ispras.retrascope.model.cgaa.CgaaModel} into
 * extended finite state machine {@link ru.ispras.retrascope.model.efsm.EfsmModel}.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CgaaEfsmTransformer extends CfgEngine {

  /**
   * The engine identifier.
   */
  public static final String ENGINE_ID = "cgaa-efsm-transformer";

  /**
   * State-like variables names command line parameter.
   */
  public static final Parameter STATE_VARIABLES =
      new Parameter("state-vars", true, "Names of state-like variables");

  /* The CGAA model that is to be transformed into EFSM. */
  private CfgModel cfgModel;

  /* The clock-like variables detector. */
  private CgaaClockDetector clockDetector;

  /* The state-like variables extractor. */
  private CgaaStateDetector stateDetector;

  /* The state-like variables statistics keeping visitor. */
  private CgaaStatisticsKeeper statKeeper;

  /* The depending-on-states switches splitter. */
  private CgaaSwitchSplitter switchSplitter;

  /* The states expressions' extractor. */
  private CgaaEfsmDataExtractor efsmDataExtractor;

  /* Mapping for state expressions and compliant sub-expressions */
  private final Map<Node, Set<Node>> compatible = new LinkedHashMap<>();

  /* Debug mode flag. */
  private boolean debugMode;

  /* File for debug output. */
  private File debugFile;

  /* Debug file default name. */
  private static final String DEFAULT_FILE_NAME = ENGINE_ID + ".log";

  /* The default charset (for printing to debug file). */
  private static final Charset DEFAULT_CHARSET = Charset.defaultCharset();

  private static final String EXTRA_RST = "Extra resetting transition has been found: ";
  private static final String CANT_FIND_RST = "Can't find resetting transition.";

  /**
   * Default constructor.
   */
  public CgaaEfsmTransformer() {
    super(
      CgaaEfsmTransformer.ENGINE_ID,
      CgaaModel.TYPE,
      EfsmModel.TYPE,
      CfgWalkerFactory.newWalker(new CgaaClockDetector(), CfgWalkerMode.DFS_NO_RPT),
      CfgWalkerFactory.newWalker(new CgaaStateDetector(), CfgWalkerMode.DFS),
      CfgWalkerFactory.newWalker(new CgaaStatisticsKeeper(), CfgWalkerMode.DFS),
      CfgWalkerFactory.newWalker(new CgaaSwitchSplitter(), CfgWalkerMode.DFS),
      CfgWalkerFactory.newWalker(new CgaaEfsmDataExtractor(), CfgWalkerMode.DFS));

    /* Engine command line parameters. */
    addParameter(Retrascope.TOOL_DEBUG_PARAM);
    addParameter(STATE_VARIABLES);
  }

  private CfgModel getCfgModel() {
    return this.cfgModel;
  }

  private void setCfgModel(final CfgModel cfgModel) {
    InvariantChecks.checkNotNull(cfgModel);

    this.cfgModel = cfgModel;
  }

  private CgaaClockDetector getClockDetector() {
    return this.clockDetector;
  }

  private void setClockDetector(final CgaaClockDetector clockDetector) {
    InvariantChecks.checkNotNull(clockDetector);

    this.clockDetector = clockDetector;
  }

  private CgaaStateDetector getStateDetector() {
    return this.stateDetector;
  }

  private void setStateDetector(final CgaaStateDetector stateDetector) {
    InvariantChecks.checkNotNull(stateDetector);

    this.stateDetector = stateDetector;
  }

  private CgaaStatisticsKeeper getStatKeeper() {
    return this.statKeeper;
  }

  private void setStatKeeper(final CgaaStatisticsKeeper statKeeper) {
    InvariantChecks.checkNotNull(statKeeper);

    this.statKeeper = statKeeper;
  }

  private CgaaSwitchSplitter getSwitchSplitter() {
    return this.switchSplitter;
  }

  private void setSwitchSplitter(final CgaaSwitchSplitter switchSplitter) {
    InvariantChecks.checkNotNull(switchSplitter);

    this.switchSplitter = switchSplitter;
  }

  private CgaaEfsmDataExtractor getEfsmDataExtractor() {
    return this.efsmDataExtractor;
  }

  private void setEfsmDataExtractor(final CgaaEfsmDataExtractor efsmDataExtractor) {
    InvariantChecks.checkNotNull(efsmDataExtractor);

    this.efsmDataExtractor = efsmDataExtractor;
  }

  private boolean isDebugMode() {
    return this.debugMode;
  }

  private void setDebugMode(final boolean debugMode) {
    this.debugMode = debugMode;
  }

  private File getDebugFile() {
    return this.debugFile;
  }

  private void setDebugFile(final File file) {
    this.debugFile = file;
  }

  @Override
  public void initialize(final Map<EntityType, Entity> inputs) {

    /* Extract control flow graph from wrapper. */

    final CgaaModel cgaaModel = (CgaaModel) inputs.get(CgaaModel.TYPE);
    inputs.remove(CgaaModel.TYPE);
    final CfgModel internalCfgModel = cgaaModel.getModel();
    setCfgModel(internalCfgModel);
    inputs.put(CfgModel.TYPE, internalCfgModel);

    initWalkers();

    /* Create mappings for internal components. */
    createClockMap();
    createStateMap();

    /* Elaborate command line parameters. */
    final Arguments arguments = (Arguments)inputs.get(Arguments.TYPE);

    try {
      final Arguments engineArgs = parseCommandLine(arguments.getCommandLine());
      final String value = engineArgs.getValue(STATE_VARIABLES);
      if (value != null) {
        final String[] stateNames = value.split(File.pathSeparator);
        getStateDetector().setStateNames(stateNames);
      }

      if (engineArgs.contains(Retrascope.TOOL_DEBUG_PARAM)) {
        setDebugMode(true);
        setDebugFile(new File(engineArgs.getValue(Retrascope.TOOL_DEBUG_PARAM, DEFAULT_FILE_NAME)));
      }
    } catch (ParseException e) {
      throw new RetrascopeRuntimeException(e);
    }
  }

  private void initWalkers() {

    setClockDetector(new CgaaClockDetector());
    setStateDetector(new CgaaStateDetector());
    setStatKeeper(new CgaaStatisticsKeeper());
    setSwitchSplitter(new CgaaSwitchSplitter());
    setEfsmDataExtractor(new CgaaEfsmDataExtractor());

    setWalkers(CfgWalkerFactory.newWalker(getClockDetector(), CfgWalkerMode.DFS_NO_RPT),
        CfgWalkerFactory.newWalker(getStateDetector(), CfgWalkerMode.DFS),
        CfgWalkerFactory.newWalker(getStatKeeper(), CfgWalkerMode.DFS),
        CfgWalkerFactory.newWalker(getSwitchSplitter(), CfgWalkerMode.DFS),
        CfgWalkerFactory.newWalker(getEfsmDataExtractor(), CfgWalkerMode.DFS));

    setDebugMode(false);
  }

  private void createStateMap() {
    final Map<CfgProcess, Set<NodeVariable>> map = new LinkedHashMap<>();
    getStateDetector().setStateMap(map);
    getStatKeeper().setStateMap(map);
    getSwitchSplitter().setStateMap(map);
    getEfsmDataExtractor().setStateMap(map);
  }

  private void createClockMap() {
    final Map<CfgProcess, Set<NodeVariable>> map = new LinkedHashMap<>();
    getClockDetector().setClockMap(map);
    getEfsmDataExtractor().setClockMap(map);
  }

  @Override
  public Entity getOutput() {

    final EfsmModel model = new EfsmModel();

    /* Elaborate CGAA modules and their sub-modules. */

    /* Create CGAA model visitor that transforms modules. */
    final CfgVisitor moduleVisitor = new CgaaModuleVisitor(model);

    final CfgWalker walker = CfgWalkerFactory.newWalker(moduleVisitor, CfgWalkerMode.DFS_NO_RPT);
    walker.start(getCfgModel());

    final Map<CfgModule, EfsmModule> moduleEfsmMap =
        ((CgaaModuleVisitor)moduleVisitor).getModuleMap();

    final Map<CfgProcess, List<EfsmDataContainer>> processDataMap =
        getEfsmDataExtractor().getProcessDataMap();

    Log.printInfo("======================================");
    int efsmCount = 0;

    int stateCount = 0;
    int transitionCount = 0;
    for (final Map.Entry<CfgProcess, List<EfsmDataContainer>> entry : processDataMap.entrySet()) {

      final CfgProcess keyProcess = entry.getKey();
      final CfgStatement keyParent = keyProcess.getParent();

      /* Check that Process parent is Module object. */
      assert keyParent instanceof CfgModule
          : LogMessage.makeMessage(LogMessage.WRONG_TYPE, keyParent.getClass());

      final EfsmModule parentContainer = moduleEfsmMap.get(keyParent);

      /* Print clock-like variables. */

      final List<String> clockNames = new LinkedList<>();

      final Set<NodeVariable> clockVars = getClockDetector().getClockMap().get(keyProcess);

      for (final NodeVariable var : clockVars) {
        clockNames.add(var.getName());
      }

      Log.printInfo("Clock-like variables: " + clockNames.toString() + ".");
      Log.printInfo("Transforming the process of module: "
          + ((CfgModule)keyParent).getFullName() + ".");

      final List<EfsmDataContainer> list = entry.getValue();

      final Efsm efsm = new Efsm(parentContainer, keyProcess.getId());
      efsm.declareVariables(keyProcess.getDeclarationMap());

      for (final NodeVariable clockVariable : clockVars) {
        efsm.addClock(clockVariable);
      }

      /* Construct EFSM states. */
      stateCount += addStates(efsm, keyProcess, list);

      if (efsm.containsStates()) {
        transitionCount += addTransitions(efsm, list);
        findInitialState(keyProcess, efsm);
        if (keyProcess.isInitial()) {
          parentContainer.addInitialEfsm(efsm);
        } else {
          parentContainer.addNonInitialEfsm(efsm);
        }
        efsmCount++;
      }
    }
    Log.printInfo(efsmCount + " EFSM(s) are extracted.");

    /* If debug log mode is enabled, write statistical info to log file. */
    if (isDebugMode()) {
      writeToDebugFile("states=" + stateCount + " ", false);
      writeToDebugFile("transitions=" + transitionCount + " ", false);
      writeToDebugFile("efsms=" + efsmCount, true);
    }

    return model;
  }

  private void findInitialState(final CfgProcess process, final Efsm efsm) {

    final Map<NodeVariable, Node> initValueMap = getInitialValues(process);

    if (allStateVariablesAreInitialized(process)) {

      final Node initValueExpr = createInitValueState(initValueMap);
      final Set<EfsmState> efsmStates = efsm.getStates();

      EfsmState initialState = null;
      for (final EfsmState state : efsmStates) {

        final Node conjunction = ExprUtils.getConjunction(initValueExpr, state.getExpression());

        if (SolverUtils.isSat(conjunction)) {
          initialState = state;
          break;
        }
      }

      if (initialState == null) {
        Log.printError("Can't find initial state.");
      } else {

        efsm.setInitialState(initialState);

        final Set<EfsmTransition> outTransitions = efsm.getOutgoingTransitions(initialState);
        final Set<EfsmTransition> rstTransitions = getResetTransitions(outTransitions);

        if (rstTransitions.isEmpty()) {
          Log.printWarn(CANT_FIND_RST);
        } else {

          final Iterator<EfsmTransition> iterator = rstTransitions.iterator();
          efsm.addResetGuardedAction(iterator.next().getGuardedAction());

          if (iterator.hasNext()) {
            /* More than one resetting transition has been found. */

            while (iterator.hasNext()) {
              Log.printWarn(EXTRA_RST + iterator.next().toString());
            }
          }
        }
      }
    } else {

      efsm.addState(Efsm.UNINIT_STATE);
      efsm.setInitialState(Efsm.UNINIT_STATE);

      final Set<EfsmTransition> rstTransitions = getResetTransitions(efsm.getTransitions());

      if (rstTransitions.isEmpty()) {
        Log.printWarn(CANT_FIND_RST);
      } else {

        final Iterator<EfsmTransition> iterator = rstTransitions.iterator();
        final EfsmTransition next = iterator.next();

        efsm.createTransition(
            Efsm.UNINIT_STATE,
            next.getDestinationState(),
            next.getGuardedAction());

        efsm.addResetGuardedAction(next.getGuardedAction());

        if (iterator.hasNext()) {

          /* More than one resetting transition has been found. */
          while (iterator.hasNext()) {
            Log.printWarn(EXTRA_RST + iterator.next().toString());
          }
        }
      }
    }
  }

  private Map<NodeVariable, Node> getInitialValues(final CfgProcess process) {
    final Map<NodeVariable, Node> initValueMap = new LinkedHashMap<>();

    CfgStatement container = process;

    do {
      if (container instanceof CfgDeclarationStatement) {

        final Map<NodeVariable, VariableDeclaration> declarationMap =
            ((CfgDeclarationStatement) container).getDeclarationMap();

        for (final Map.Entry<NodeVariable, VariableDeclaration> entry : declarationMap.entrySet()) {

          if (!initValueMap.containsKey(entry.getKey())
              && entry.getValue().containsInitialValue()) {

            initValueMap.put(entry.getKey(), entry.getValue().getInitialValue());
          }
        }

        container = container.getParent();
      }
    } while (container != null);

    return initValueMap;
  }

  private boolean allStateVariablesAreInitialized(final CfgProcess process) {

    boolean result;

    final Set<NodeVariable> stateVariables = getStateDetector().getStateMap().get(process);
    final Set<NodeVariable> initialized = new LinkedHashSet<>();

    if (stateVariables.isEmpty()) {
      result = false;
    } else {

      CfgStatement container = process;

      do {

        if (container instanceof CfgDeclarationStatement) {

          final CfgDeclarationStatement varContainer = (CfgDeclarationStatement)container;
          for (final NodeVariable variable : stateVariables) {
            if (varContainer.contains(variable)
                && varContainer.getDeclaration(variable).containsInitialValue()) {
              initialized.add(variable);
            }
          }
          container = container.getParent();
        }
      } while (container != null);

      result = initialized.size() == stateVariables.size();
    }
    return result;
  }

  private Node createInitValueState(final Map<NodeVariable, Node> map) {

    Node stateCondition = null;

    for (final Map.Entry<NodeVariable, Node> entry : map.entrySet()) {
      final Node equation = NodeUtils.getEquation(entry.getKey(), entry.getValue());
      if (stateCondition == null) {
        stateCondition = equation;
      } else {
        stateCondition = ExprUtils.getConjunction(stateCondition, equation);
      }
    }

    return stateCondition;
  }

  private Set<EfsmTransition> getResetTransitions(final Set<EfsmTransition> outTransitions) {
    final Set<EfsmTransition> rstTransitions = new LinkedHashSet<>();

    for (final EfsmTransition transition : outTransitions) {
      if (isResetTransition(transition)) {
        rstTransitions.add(transition);
      }
    }
    return rstTransitions;
  }

  private boolean isResetTransition(final EfsmTransition transition) {
    return hasInputDependGuard(transition.getGuardedAction()) && hasConstantAction(transition);
  }

  /* Check whether guard depends on exactly one in-inout signal
  . or guard is trivial and clocks collection is not empty. */
  private boolean hasInputDependGuard(final GuardedAction guardedAction) {

    final Guard guard = guardedAction.getGuard();

    final Set<NodeVariable> guardUses = guard.getUses();

    boolean oneInputDepend = false;
    if (guardUses.size() == 1) {
      final NodeVariable use = guardUses.iterator().next();

      final Object useData = use.getUserData();
      if (useData instanceof VariableDescriptor
          && VariableType.isInputSignal(((VariableDescriptor) useData).getVariableType())) {
        oneInputDepend = true;
      }
    }

    if (isTrivialGuard(guardedAction, oneInputDepend)) {
      oneInputDepend = true;
    }

    return oneInputDepend;
  }

  private boolean isTrivialGuard(
      final GuardedAction guardedAction,
      final boolean oneInputDependGuard) {

    return guardedAction.getGuard().getUses().isEmpty()
        && !oneInputDependGuard
        && !guardedAction.getClocks().isEmpty();
  }

  private boolean hasConstantAction(final EfsmTransition transition) {

    /* Check whether action contains only assignments like var := constant. */

    final Action action = transition.getGuardedAction().getAction();
    boolean checkAction = false;
    for (final Assignment assignment : action.getAssignments()) {
      checkAction = assignment.getUses().isEmpty();
      if (!checkAction) {
        break;
      }
    }
    return checkAction;
  }

  private int addStates(
      final Efsm efsm,
      final CfgProcess process,
      final Collection<EfsmDataContainer> list) {

    /* Create the convenience collection of pre-states. */
    final Set<Set<Node>> preStates = getContainerStates(list);

    final List<Node> states = new LinkedList<>();

    /* Create a bi-directional mapping for literals and expressions. */
    final BidiMap<Integer, Node> literalMap = new DualHashBidiMap<>();

    if (!preStates.isEmpty() && preStates.size() > 1) {

      /* Transform data containers' pre-states into clauses. */
      final List<Clause> clauses = createClauses(preStates, literalMap);

      /* Detect conflicts between literals. */
      final Set<Clause> conflicts = getConflicts(literalMap);

      /* Make clauses orthogonal. */
      final NormalForm normalForm = new NormalForm(NormalForm.Type.DNF, clauses);
      final NormalForm orthogonalForm = Orthogonalizer.orthogonalize(normalForm, conflicts);
      final List<Clause> ortClauses = orthogonalForm.getClauses();

      /* Transform orthogonal clauses into EFSM states. */
      createStates(process, states, literalMap, ortClauses);
    }

    /* If states are not extracted, a trivial EFSM should be created. */
    if (states.isEmpty()) {
      states.add(SolverUtils.TRUE);
    }

    /* Add states to EFSM model. */
    final List<String> stateVarNames = new LinkedList<>();
    int stateCount = 0;
    for (final Node state : states) {

      /* Store state-like variables names (for statistics only). */
      final Collection<NodeVariable> stateVariables = ExprUtils.getVariables(state);

      for (final NodeVariable stateVar : stateVariables) {
        final String name = stateVar.getName();
        if (!stateVarNames.contains(name)) {
          stateVarNames.add(name);
        }
      }
      efsm.createState(state);
      stateCount++;
    }

    Log.printInfo(stateCount + " states are extracted.");
    Log.printInfo("The state-like variables are: " + stateVarNames.toString() + ".");
    return stateCount;
  }

  private Set<Clause> getConflicts(final BidiMap<Integer, Node> literalMap) {

    final Set<Clause> conflicts = new LinkedHashSet<>();
    final Set<Integer> literalSet = literalMap.keySet();
    final Integer[] variables = literalSet.toArray(new Integer[literalSet.size()]);

    for (int i = 0; i < variables.length - 1; i++) {
      for (int j = i + 1; j < variables.length; j++) {
        final int variableI = variables[i];
        final int variableJ = variables[j];

        final Node nodeI = literalMap.get(variableI);
        final Node nodeJ = literalMap.get(variableJ);
        InvariantChecks.checkFalse(nodeI.equals(nodeJ));

        for (final boolean signI : new boolean[] {false, true}) {
          final Node signedNodeI = signI ? ExprUtils.getNegation(nodeI) : nodeI;

          for (final boolean signJ : new boolean[] {false, true}) {
            final Node signedNodeJ = signJ ? ExprUtils.getNegation(nodeJ) : nodeJ;

            if (!SolverUtils.areCompatible(signedNodeI, signedNodeJ)) {
              final Clause.Builder conflictBuilder = new Clause.Builder();

              conflictBuilder.add(variableI, signI);
              conflictBuilder.add(variableJ, signJ);

              conflicts.add(conflictBuilder.build());
            }
          }
        }
      }
    }
    return conflicts;
  }

  private Set<Set<Node>> getContainerStates(final Collection<EfsmDataContainer> list) {
    final Set<Set<Node>> preStates = new LinkedHashSet<>();

    for (final EfsmDataContainer data : list) {
      preStates.add(data.getState());
    }
    return preStates;
  }

  private void createStates(
      final CfgProcess process,
      final Collection<Node> states,
      final BidiMap<Integer, Node> literalMap,
      final Collection<Clause> ortClauses) {

    for (final Clause clause : ortClauses) {

      Set<Node> compatibleNodes = new LinkedHashSet<>();

      Node state = null;
      final Set<Integer> clauseLiterals = clause.getVars();
      for (final Integer literal : clauseLiterals) {
        Node node = literalMap.get(literal);

        if (clause.getSign(literal)) {
          node = ExprUtils.getNegation(node);
        } else {
          compatibleNodes.add(node);
        }
        state = state == null ? node : ExprUtils.getConjunction(state, node);
        state = TransformerUtils.standardize(state);
      }

      /* Add variable invariants to state constraint. */

      final Collection<NodeVariable> stateVars = ExprUtils.getVariables(state);
      final Collection<Node> stateInvariants =
          CgaaEfsmTransformerUtils.getInvariants(process, stateVars);
      final Node invariants = stateInvariants.isEmpty()
          ? SolverUtils.TRUE :
          ExprUtils.getConjunction(stateInvariants.toArray(new Node[stateInvariants.size()]));

      /* If state expression is satisfiable, then it is a correct state. */
      if (SolverUtils.areCompatible(state, invariants)) {
        states.add(state);

        if (this.compatible.containsKey(state)) {
          this.compatible.get(state).addAll(compatibleNodes);
        } else {
          this.compatible.put(state, compatibleNodes);
        }
      }
    }
  }

  private List<Clause> createClauses(
      final Collection<Set<Node>> preStates,
      final BidiMap<Integer, Node> literalMap) {

    final List<Clause> clauses = new LinkedList<>();

    for (final Set<Node> preState : preStates) {

      final Clause.Builder clauseBuilder = new Clause.Builder();

      for (Node node : preState) {

        boolean negated = false;

        if (node.getKind() == Node.Kind.OPERATION
            && ((NodeOperation)node).getOperationId().equals(StandardOperation.NOT)) {

          node = ((NodeOperation) node).getOperand(0);
          negated = true;
        }

        int literal;

        if (literalMap.containsValue(node)) {
          literal = literalMap.getKey(node);
        } else {
          literal = literalMap.keySet().size();
          literalMap.put(literal, node);
        }
        clauseBuilder.add(literal, negated);
      }
      clauses.add(clauseBuilder.build());
    }
    return clauses;
  }

  /**
   * Adds transitions from the specified collection to the specified {@link Efsm} object.
   *
   * @param efsm model the transitions to be added to
   * @param list collection of transition candidates
   *
   * @return number of transitions that are constructed
   */
  private int addTransitions(
      final Efsm efsm,
      final Collection<EfsmDataContainer> list) {

    /* Variables for EFSM extraction statistics keeping. */
    int transitionCount = 0;

    /* Create EFSM transitions. */
    for (final EfsmDataContainer data : list) {

      final GuardedAction guardedAction = data.getGuardedAction();

      final List<CfgBlockStatement> blocks =
          Collections.singletonList(guardedAction.getAction().toBasicBlock());

      EfsmState srcState;

      final Node dataStateExpr =
          ExprUtils.getConjunction(data.getState().toArray(new Node[data.getState().size()]));

      final Set<EfsmState> efsmStates = efsm.getStates();
      for (final EfsmState efsmState : efsmStates) {

        final Node efsmStateExpr = efsmState.getExpression();
        final Guard guard = guardedAction.getGuard();
        final Collection<Event> clocks = guardedAction.getClocks();

        final boolean stateCompatible = areCompatible(dataStateExpr, efsmStateExpr);
        final Node clockNode = clocks.isEmpty() ? null : EfsmUtils.eventsToNode(clocks);
        final boolean guardSat =
            clockNode == null
                ? SolverUtils.isSat(guard.toNode()) :
                SolverUtils.isSat(ExprUtils.getConjunction(guard.toNode(), clockNode));

        if (stateCompatible && guardSat) {

          /* Detect source state. */
          srcState = efsmState;

          /* Detect possible destination states. */
          final List<EfsmState> possible = getDestinationStates(blocks, srcState, efsmStates);

          /* Add possible destination states. */
          if (possible.isEmpty()) {
            Log.printWarn(
                String.format(
                    "Lockup found for: src-state=%s, guarded action=%s.",
                    srcState.toString(), guardedAction.toString()));

          }
          if (possible.size() == 1) {

            /* Get the only possible destination state. */
            final EfsmState dst = possible.get(0);

            efsm.createTransition(srcState, dst, guardedAction);

            transitionCount++;
          } else {

            /* More than one possible destination state. */
            for (final EfsmState dst : possible) {

              final Node refinedGuard = refineGuard(guardedAction, dst.getExpression());

              if (clockNode == null
                  ? SolverUtils.isSat(refinedGuard)
                  : ExprUtils.areCompatible(refinedGuard, clockNode)) {

                final GuardedAction transition =
                  new GuardedAction(new Guard(refinedGuard), guardedAction.getAction());
                transition.addClocks(guardedAction.getClocks());

                efsm.createTransition(srcState, dst, transition);

              /* Increment transition counter. */
                transitionCount++;
              }
            }
          }
        }
      }
    }
    Log.printInfo(transitionCount + " transitions are extracted.");

    return transitionCount;
  }

  private List<EfsmState> getDestinationStates(
      final List<CfgBlockStatement> blocks,
      final EfsmState srcState,
      final Set<EfsmState> efsmStates) {

    final List<EfsmState> possible = new LinkedList<>();
    for (final EfsmState possibleDst : efsmStates) {

      /* Calculate a strongest post-condition. */
      final Node node = applySubstitutions(possibleDst.getExpression(), blocks);

      if (SolverUtils.areCompatible(srcState.getExpression(), node)) {
        possible.add(possibleDst);
      }
    }
    return possible;
  }

  private static Node applySubstitutions(
      final Node dstNode,
      final List<CfgBlockStatement> blocks) {

    final List<NodeBinding.BoundVariable> boundVariables = new LinkedList<>();

    for (final CfgBlockStatement block : blocks) {
      for (final Assignment assignment : block.getAssignments()) {
        boundVariables.add(
            NodeBinding.bindVariable(assignment.getVariable(), assignment.getValue()));
      }
    }
    return Transformer.substituteAllBindings(new NodeBinding(dstNode, boundVariables));
  }

  private boolean areCompatible(
      final Node dataStateExpr,
      final Node efsmStateExpr) {

    return this.compatible.containsKey(efsmStateExpr)
        && this.compatible.get(efsmStateExpr).contains(dataStateExpr)
        || dataStateExpr.equals(SolverUtils.TRUE)
        || SolverUtils.areCompatible(dataStateExpr, efsmStateExpr);
  }

  private Node refineGuard(
      final GuardedAction guardedAction,
      final Node dstExprCopy) {

    final List<CfgBlockStatement> actionBlocks =
        Collections.singletonList(guardedAction.getAction().toBasicBlock());

    return TransformerUtils.standardize(
        ExprUtils.getConjunction(
            guardedAction.getGuard().toNode(),
            Transformer.reduce(
                TransformUtils.substituteBackward(actionBlocks, dstExprCopy))));
  }

  private void writeToDebugFile(final String string, final boolean asLine) {
    try {
      final BufferedWriter writer =
          new BufferedWriter(
              new OutputStreamWriter(new FileOutputStream(getDebugFile(), true), DEFAULT_CHARSET));
      writer.write(string);
      if (asLine) {
        writer.newLine();
      }
      writer.close();

    } catch (IOException e) {
      throw new RetrascopeRuntimeException(e);
    }
  }
}
