/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgDeclarationStatement;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * Component that detects state-like variables in clocked guarded atomic actions model.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaStateDetector extends MapCfgVisitor {

  /* The control part of current path. */
  private List<CfgSwitchStatement> controlPath;

  /* The data part of current path. */
  private List<CfgBlockStatement> dataPath;

  /* The switch nodes expressions. */
  private Set<Node> switchNodes;

  /* The state-like variables. */
  private List<NodeVariable> stateVariables;

  /* Variables that cannot be state-like. */
  private Set<NodeVariable> forbidden;

  /* Names of user-defined state variables. */
  private String[] userDefinedNames;

  private Status status;

  private static final String LOOP_DETECTED = "Loop has been found.";

  /**
   * The default constructor.
   */
  public CgaaStateDetector() {
    super();
    setStatus(Status.OK);
  }

  private List<CfgSwitchStatement> getControlPath() {
    return this.controlPath;
  }

  private void setControlPath(final List<CfgSwitchStatement> controlPath) {
    InvariantChecks.checkNotNull(controlPath);

    this.controlPath = controlPath;
  }

  private List<CfgBlockStatement> getDataPath() {
    return this.dataPath;
  }

  private void setDataPath(final List<CfgBlockStatement> dataPath) {
    InvariantChecks.checkNotNull(dataPath);

    this.dataPath = dataPath;
  }

  private Set<Node> getSwitchNodes() {
    return this.switchNodes;
  }

  private List<NodeVariable> getStateVariables() {
    return this.stateVariables;
  }

  private void setStateVariables(final List<NodeVariable> stateVariables) {
    InvariantChecks.checkNotNull(stateVariables);

    this.stateVariables = stateVariables;
  }

  private String[] getUserDefinedNames() {
    return userDefinedNames;
  }

  private void setStatus(final Status status) {
    InvariantChecks.checkNotNull(status);

    this.status = status;
  }

  /**
   * Sets user-defined names of state-like variables.
   *
   * @param names state-like variables names
   */
  public void setStateNames(final String[] names) {
    this.userDefinedNames = names;
  }

  @Override
  public Status getStatus() {
    return this.status;
  }

  @Override
  public void onRootBegin() {
    setControlPath(new LinkedList<CfgSwitchStatement>());
    setDataPath(new LinkedList<CfgBlockStatement>());
    setStateVariables(new LinkedList<NodeVariable>());
  }

  @Override
  public void onRootEnd() {
    setStatus(Status.OK);
  }

  @Override
  public void onProcessBegin(final CfgProcess process) {

    this.forbidden = new LinkedHashSet<>();

    if (getUserDefinedNames() != null && getUserDefinedNames().length > 0) {
      final Set<NodeVariable> states = new LinkedHashSet<>();

      for (final String name : getUserDefinedNames()) {
        final CfgDeclarationStatement variableContainer =
            CfgUtils.getVariableContainer(process, name);

        if (variableContainer != null) {
          states.add(variableContainer.getVariable(name));
        }
      }
      this.getStateMap().put(process, states);
      this.status = Status.SKIP;

    } else {
      this.switchNodes = new LinkedHashSet<>();
    }
  }

  @Override
  public void onProcessEnd(final CfgProcess process) {

    if (this.userDefinedNames == null || this.userDefinedNames.length == 0) {

      /* Remove forbidden variables from the collection of state-like variables. */
      if (!this.forbidden.isEmpty()) {
        for (final NodeVariable forbiddenVar : this.forbidden) {
          if (this.stateVariables.contains(forbiddenVar)) {
            this.stateVariables.remove(forbiddenVar);
          }
        }
      }

      /* Check that all the module switch nodes
       * that use state-like variables does not use non-state-like variables. */
      checkStateVariables();

      final Set<NodeVariable> states = new LinkedHashSet<>();
      states.addAll(getStateVariables());
      this.getStateMap().put(process, states);
      this.stateVariables.clear();
    }
  }

  private void checkStateVariables() {
    for (final Node node : getSwitchNodes()) {
      final Collection<NodeVariable> nodeVariables = ExprUtils.getVariables(node);

      boolean hasState = false;
      boolean hasNonState = false;

      for (final NodeVariable variable : nodeVariables) {
        if (getStateVariables().contains(variable)) {
          hasState = true;
        } else if (!hasNonState) {
          hasNonState = true;
        }
      }
      if (hasState && hasNonState) {
        for (final NodeVariable variable : nodeVariables) {
          if (getStateVariables().contains(variable)) {
            this.stateVariables.remove(variable);
          }
        }
      }
    }
  }

  @Override
  public void onBasicBlockBegin(final CfgBlockStatement block) {
    if (getDataPath().contains(block)) {
      Log.printWarn(LOOP_DETECTED);
    }
    this.dataPath.add(block);

    processTerminalStatement(block);
  }

  @Override
  public void onBasicBlockEnd(final CfgBlockStatement block) {
    if (getDataPath().contains(block)) {
      this.dataPath.remove(block);
    } else {
      throw RetrascopeRuntimeException.makeException(LogMessage.CANT_FIND, block.getDescription());
    }
  }

  @Override
  public void onSwitchBegin(final CfgSwitchStatement switchNode) {
    if (getControlPath().contains(switchNode)) {
      Log.printWarn(LOOP_DETECTED);
    }
    this.controlPath.add(switchNode);
    this.switchNodes.add(switchNode.getCondition());
  }

  @Override
  public void onSwitchEnd(final CfgSwitchStatement switchNode) {
    if (getControlPath().contains(switchNode)) {
      this.controlPath.remove(switchNode);
    } else {
      throw RetrascopeRuntimeException.makeException(
          LogMessage.CANT_FIND,
          switchNode.getDescription());
    }
  }

  @Override
  public void onCaseBegin(final CfgCaseStatement caseStatement) {
    processTerminalStatement(caseStatement);
  }

  private void processTerminalStatement(final CfgStatement statement) {
    if (!CfgUtils.hasSuccessors(statement)) {

      /* Get variables that are defined in data parts of graph paths. */
      final Set<NodeVariable> definedDataVars = getDefinedVariables(getDataPath());

    /* Get variables that are used in control parts of graph paths. */
      final List<NodeVariable> usedControlVars = getUsedVariables(getControlPath());

     /* The weak heuristic is applied. */
      for (final NodeVariable var : definedDataVars) {

        if (var.getUserData() != null && var.getUserData() instanceof VariableDescriptor) {

          final VariableType type = ((VariableDescriptor) var.getUserData()).getVariableType();
          if (!VariableType.isSignal(type)
              && usedControlVars.contains(var)
              &&  !getStateVariables().contains(var)) {
            this.stateVariables.add(var);
          }
        }
      }
    }
  }

  private Set<NodeVariable> getDefinedVariables(final List<CfgBlockStatement> dataPath) {

    /* Get those defined variables that do not depend on input signals. */
    final Set<NodeVariable> defined = new LinkedHashSet<>();

    final Set<NodeVariable> forbidden = new LinkedHashSet<>();

    for (final CfgBlockStatement block : dataPath) {
      final List<Assignment> assignments = block.getAssignments();

      for (final Assignment assignment : assignments) {
        final Collection<NodeVariable> stmtUses = assignment.getUses();

        final Set<NodeVariable> defines = assignment.getDefines();
        if (areSignalDependent(stmtUses)) {

          /* Variables that are defined here, cannot be returned. */
          forbidden.addAll(defines);
        } else {
          defined.addAll(CollectionUtils.complementSets(defines, forbidden));
        }
      }

      final Collection<NodeVariable> blockDefines = block.getDefines();
      for (final NodeVariable blockDefined : blockDefines) {
        if (!forbidden.contains(blockDefined)) {
          defined.add(blockDefined);
        }
      }
    }
    this.forbidden.addAll(forbidden);
    return defined;
  }

  private boolean areSignalDependent(final Collection<NodeVariable> stmtUses) {
    boolean dependOnSignal = false;
    for (final NodeVariable use : stmtUses) {
      if (VariableType.isInputSignal(((VariableDescriptor) use.getUserData()).getVariableType())) {
        dependOnSignal = true;
        break;
      }
    }
    return dependOnSignal;
  }

  private List<NodeVariable> getUsedVariables(final List<CfgSwitchStatement> controlPath) {
    final List<NodeVariable> used = new LinkedList<>();

    for (final CfgSwitchStatement node : controlPath) {
      final Collection<NodeVariable> nodeUses = node.getUses();
      for (final NodeVariable nodeUse : nodeUses) {
        if (!used.contains(nodeUse)) {
          used.add(nodeUse);
        }
      }
    }
    return used;
  }
}
