/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * Component that creates rules of switch nodes splitting in clocked guarded atomic actions model.
 *
 * <p>Visitor for splitting switch nodes, that are depend on state-like variables
 * and can be split into many-nodes-containing sub-graph. Such splitting
 * can be done when top-level operation identifier in switch node's expression
 * is of Boolean type (like NOT, OR, AND, IML, XOR). If the switch node's expression
 * also contains internal Boolean identifiers, the splitting procedure will be applied
 * to them also only if one of their sub-expressions depends on state-like variable.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaSwitchSplitter extends MapCfgVisitor {

  /* The process to be transformed. */
  private CfgProcess current;

  @Override
  public void onProcessBegin(final CfgProcess process) {
    this.current = process;
  }

  @Override
  public void onSwitchEnd(final CfgSwitchStatement switchStatement) {

    CfgSwitchStatement topLevelSwitch = switchStatement;
    Node condition = topLevelSwitch.getCondition();
    while (isSplittable(condition)) {

      final Node stateDepOperand = getStateDependOperand((NodeOperation)condition);
      final Node rest = getRestOperand((NodeOperation)condition, stateDepOperand);

      final StandardOperation operationType =
          (StandardOperation) ((NodeOperation) condition).getOperationId();

      /* Splittable switch node should have boolean condition, so try to detect true/false cases */
      CfgCaseStatement trueCase = null;
      CfgCaseStatement falseCase = null;

      for (int i = 0; i < topLevelSwitch.getStatementNum(); i++) {
        final CfgCaseStatement child = (CfgCaseStatement) topLevelSwitch.getStatement(i);
        final Set<NodeValue> childValues = child.getValues();
        if (childValues.size() == 1) {
          final NodeValue value = childValues.iterator().next();
          if (value.equals(SolverUtils.TRUE)) {
            trueCase = child;
          }
          if (value.equals(SolverUtils.FALSE)) {
            falseCase = child;
          }
        }
      }

      if (trueCase == null || falseCase == null) {
        throw new RetrascopeRuntimeException(
            String.format("switch with condition: %s does not have true or false case", condition));
      }

      final CfgSwitchStatement restSwitch = new CfgSwitchStatement(rest);
      final CfgCaseStatement restTrueCase = new CfgCaseStatement(SolverUtils.TRUE);
      final CfgCaseStatement restFalseCase = new CfgCaseStatement(SolverUtils.FALSE);
      restSwitch.add(restTrueCase);
      restSwitch.add(restFalseCase);

      final CfgSwitchStatement trueStateSwitch = new CfgSwitchStatement(stateDepOperand);
      restTrueCase.add(trueStateSwitch);
      final CfgSwitchStatement falseStateSwitch = new CfgSwitchStatement(stateDepOperand);
      restFalseCase.add(falseStateSwitch);

      /* Add children nodes depending on operation type */
      final CfgCaseStatement ttCase = new CfgCaseStatement(SolverUtils.TRUE);
      final CfgCaseStatement tfCase = new CfgCaseStatement(SolverUtils.FALSE);
      final CfgCaseStatement ftCase = new CfgCaseStatement(SolverUtils.TRUE);
      final CfgCaseStatement ffCase = new CfgCaseStatement(SolverUtils.FALSE);

      trueStateSwitch.add(ttCase);
      trueStateSwitch.add(tfCase);
      falseStateSwitch.add(ftCase);
      falseStateSwitch.add(ffCase);

      final List<CfgStatement> trueCaseStatements = getStatements(trueCase);
      final List<CfgStatement> falseCaseStatements = getStatements(falseCase);

      switch (operationType) {
        case OR:
          ttCase.addAll(copy(trueCaseStatements));
          tfCase.addAll(copy(trueCaseStatements));
          ftCase.addAll(copy(trueCaseStatements));
          ffCase.addAll(copy(falseCaseStatements));
          break;
        case AND:
          ttCase.addAll(copy(trueCaseStatements));
          tfCase.addAll(copy(trueCaseStatements));
          ftCase.addAll(copy(trueCaseStatements));
          ffCase.addAll(copy(falseCaseStatements));
          break;
        default:
          throw new UnsupportedOperationException(
              LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, operationType));
      }

      condition = rest;
      topLevelSwitch = restSwitch;
    }
  }

  private List<CfgStatement> copy(List<CfgStatement> statements) {
    final List<CfgStatement> copies = new LinkedList<>();
    for (final CfgStatement statement : statements) {
      copies.add(statement.deepCopy());
    }
    return copies;
  }

  private List<CfgStatement> getStatements(final CfgCaseStatement caseStatement) {
    final List<CfgStatement> statements = new LinkedList<>();

    for (int i = 0; i < caseStatement.getStatementNum(); i++) {
      statements.add(caseStatement.getStatement(i));
    }
    return statements;
  }

  private Node getRestOperand(
      final NodeOperation operation,
      final Node stateDepOperand) {

    /* Exclude state-dependent operands from operation */
    final int operandCount = operation.getOperandCount();
    final List<Node> operands = new LinkedList<>();

    for (int i = 0; i < operandCount; i++) {
      final Node operand = operation.getOperand(i);
      if (!operand.equals(stateDepOperand)) {
        operands.add(operand);
      }
    }

    return new NodeOperation(operation.getOperationId(), operands);
  }

  private Node getStateDependOperand(NodeOperation operation) {

    for (int i = 0; i < operation.getOperandCount(); i++) {
      if (dependFromOneStateVar(operation.getOperand(i))) {
        return operation.getOperand(i);
      }
    }
    return null;
  }


  private boolean isSplittable(final Node expr) {

    if (expr.getKind() == Node.Kind.OPERATION) {

      final int operandCount = ((NodeOperation)expr).getOperandCount();
      for (int i = 0; i < operandCount; i++) {
        if (dependFromOneStateVar(((NodeOperation) expr).getOperand(i))) {
          return isSplittableOperation(((NodeOperation) expr).getOperationId());
        }
      }
      return false;
    } else {
      return false;
    }
  }

  private boolean dependFromOneStateVar(final Node operation) {

    final Collection<NodeVariable> opVariables = ExprUtils.getVariables(operation);
    final Set<NodeVariable> stateVariables = getStateMap().get(this.current);

    return opVariables.size() == 1
        && CollectionUtils.areIntersectedSets(new LinkedHashSet<>(opVariables), stateVariables);
  }

  private boolean isSplittableOperation(final Enum<?> opId) {
    return opId == StandardOperation.AND
        || opId == StandardOperation.OR;
  }
}
