/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.hldd;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.hldd.Hldd;
import ru.ispras.retrascope.model.hldd.HlddModel;
import ru.ispras.retrascope.model.hldd.HlddModule;
import ru.ispras.retrascope.model.hldd.HlddNode;
import ru.ispras.retrascope.model.hldd.HlddNode.NodeType;
import ru.ispras.retrascope.model.hldd.HlddProcess;
import ru.ispras.retrascope.util.NodeUtils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Stack;

/**
 * This class represents a visitor which is used to construct a HLDD model from a CGAA model.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
final class CgaaHlddVisitor extends CfgDefaultVisitor {

  // HLDD being constructed at the moment.
  private Hldd currentHldd;
  // Resulting model.
  private final HlddModel hlddModel;
  // HlddNode stack.
  private LinkedList<HlddNode> nodeStack;
  // Switch - corresponding variable mapping.
  private final Map<Node, NodeVariable> switchesAsVariables;
  // Counter of guard conditions.
  private int guardCounter;
  // CFG node - HLDD node mapping.
  private final Map<CfgStatement, HlddNode> nodeMapping;
  // Current module variables.
  private Collection<NodeVariable> currentVariables;
  // Contains resulting process HLDDs.
  private HlddProcess processHldds;
  // Current node level.
  private int currentLevel;
  // HlddModule stack.
  private final Stack<HlddModule> moduleStack;


  /**
   * Constructs a visitor.
   */
  public CgaaHlddVisitor() {
    super();
    currentHldd = new Hldd();
    hlddModel = new HlddModel();
    nodeStack = new LinkedList<>();
    switchesAsVariables = new HashMap<>();
    guardCounter = 0;
    nodeMapping = new HashMap<>();
    currentLevel = 0;
    moduleStack = new Stack<>();
  }

  /**
   * Returns the constructed HLDD model.
   * 
   * @return The {@link HlddModel}.
   */
  public HlddModel getModel() {
    return hlddModel;
  }

  @Override
  public void onModuleBegin(final CfgModule module) {
    final HlddModule hlddModule = new HlddModule();
    hlddModule.setModule(module);
    moduleStack.push(hlddModule);
    currentVariables = new ArrayList<>();
    final Collection<NodeVariable> moduleVariables = module.getVariables();
    final Collection<NodeVariable> moduleInputs = module.getInputs();
    currentVariables.addAll(moduleVariables);
    currentVariables.removeAll(moduleInputs);
  }

  @Override
  public void onModuleEnd(final CfgModule module) {
    final HlddModule hlddModule = moduleStack.pop();
    if (moduleStack.isEmpty()) {
      hlddModel.addModule(hlddModule);
    } else {
      moduleStack.peek().addNestedModule(hlddModule);
    }
  }

  @Override
  public void onProcessBegin(final CfgProcess process) {
    currentHldd = new Hldd();
    processHldds = new HlddProcess();
    processHldds.setProcessId(process.getId());
    processHldds.setCfgProcess(process);

    nodeStack = new LinkedList<>();
    final HlddNode node = new HlddNode(NodeType.OTHER);
    node.setRoot(true);
    node.setLevel(currentLevel);
    currentLevel++;
    nodeMapping.put(process, node);
    currentHldd.addNode(node);
    currentHldd.setRootNode(node);
    nodeStack.push(node);

  }

  @Override
  public void onProcessEnd(final CfgProcess process) {
    nodeStack.pop();

    final Collection<NodeVariable> variables = new ArrayList<>(process.getVariables());
    variables.addAll(currentVariables);

    for (NodeVariable variable : variables) {
      final Hldd hldd = removeWrongVariableNodes(currentHldd.getCopy(), variable);
      if (!hldd.getTerminalNodes().isEmpty()) {
        if (checkOneValue(hldd)) {
          final HlddNode singleNode = new HlddNode(NodeType.ASSIGNMENT);
          singleNode.addExpression(hldd.getTerminalNodes().iterator().next().getExpressions());
          singleNode.setRoot(true);
          singleNode.setLevel(0);
          final Hldd singleNodeHldd = new Hldd(variable);
          singleNodeHldd.addNode(singleNode);
          singleNodeHldd.setRootNode(singleNode);
          this.processHldds.add(singleNodeHldd);
        } else {
          this.processHldds.add(hldd);
        }
      }
    }

    currentLevel--;

    for (Entry<Node, NodeVariable> entry : switchesAsVariables.entrySet()) {
      final Node switchNode = entry.getKey();
      final NodeVariable var = entry.getValue();
      this.processHldds.add(createHlddFromSwitch(switchNode, var));
    }
    moduleStack.peek().add(processHldds);

  }

  /*
   * Removes all the nodes which are not associated with the specified variable.
   * 
   * @param hldd - the specified HLDD.
   * 
   * @param variable - the specified variable.
   * 
   * @return The {@link Hldd} containing only nodes with the specified variable.
   */
  private Hldd removeWrongVariableNodes(final Hldd hldd, final NodeVariable variable) {
    final Hldd newHldd = new Hldd(variable);
    for (HlddNode node : hldd.getNodes()) {
      if (node.getType() == NodeType.ASSIGNMENT) {
        if (node.getExpressions().get(0).equals(variable)) {
          newHldd.addNode(node);
        } else {
          final HlddNode parent = node.getParent();
          parent.removeFromChildren(node);
        }
      } else {
        newHldd.addNode(node);
      }
    }
    newHldd.setRootNode(hldd.getRootNode());
    return newHldd;
  }

  /*
   * Checks whether terminal nodes of the specified HLDD contain identical value.
   * 
   * @param hldd - the specified {@link Hldd}.
   * 
   * @return <code>true</code> if terminal nodes contain identical value, <code>false</code>
   * otherwise.
   */
  private boolean checkOneValue(final Hldd hldd) {
    if (hldd.getTerminalNodes().isEmpty()) {
      return false;
    } else {
      final HlddNode singleValue = hldd.getTerminalNodes().iterator().next();
      for (HlddNode node : hldd.getTerminalNodes()) {
        if (!node.getExpressions().get(1).equals(singleValue.getExpressions().get(1))) {
          return false;
        }
      }
    }
    return true;
  }

  @Override
  public void onBasicBlockBegin(final CfgBlockStatement block) {
    for (Assignment assignment : block.getAssignments()) {
      final NodeVariable variable = assignment.getVariable();
      final HlddNode node = new HlddNode(nodeStack.peek(), NodeType.ASSIGNMENT);
      node.addExpression(variable);
      node.addExpression(assignment.getValue());
      node.setLevel(currentLevel);
      currentHldd.addNode(node);
    }
  }


  @Override
  public void onCaseBegin(final CfgCaseStatement caseNode) {
    final HlddNode node = new HlddNode(nodeStack.peek(), NodeType.VALUE);
    node.setLevel(currentLevel);
    currentLevel++;
    if (caseNode.isDefault()) {
      node.setDefault();
    } else {
      node.addExpression(caseNode.getValues());
    }
    nodeMapping.put(caseNode, node);
    currentHldd.addNode(node);
    nodeStack.push(node);
  }

  @Override
  public void onCaseEnd(final CfgCaseStatement caseNode) {
    nodeStack.pop();
    currentLevel--;
  }

  @Override
  public void onSwitchBegin(final CfgSwitchStatement switchNode) {
    createVariableFromSwitch(switchNode);

    final HlddNode node =
        new HlddNode(nodeStack.peek(), NodeType.CONDITION,
            switchesAsVariables.get(substituteEvents(switchNode)));
    node.setLevel(currentLevel);
    currentLevel++;
    nodeMapping.put(switchNode, node);
    currentHldd.addNode(node);
    nodeStack.push(node);
  }

  @Override
  public void onSwitchEnd(final CfgSwitchStatement switchNode) {
    nodeStack.pop();
    currentLevel--;
  }

  /*
   * Associates a variable with the specified switch node.
   * 
   * @param switchNode - the specified {@link Switch}.
   */
  private void createVariableFromSwitch(final CfgSwitchStatement switchNode) {
    final Node expression = substituteEvents(switchNode);
    if (!switchesAsVariables.keySet().contains(expression)) {
      final NodeVariable variable =
          new NodeVariable("guard" + guardCounter, expression.getDataType());
      switchesAsVariables.put(expression, variable);
      guardCounter++;
    }
  }

  /* 
   * Substitutes events by adding the corresponding node to the switch condition. 
   */
  private Node substituteEvents(final CfgSwitchStatement switchNode) {
    Node condition = switchNode.getCondition();
    if (switchNode.hasEvents()) {
      for (Event event : switchNode.getEventList().getEvents()) {
        final Node eventCondition = getEventCondition(event);
        if (eventCondition != null) {
          condition = new NodeOperation(StandardOperation.AND, condition, eventCondition);
        }
      }
    }
    return condition;
  }

  /*
   * Constructs the corresponding node from the event. 
   */
  private Node getEventCondition(final Event event) {
    final NodeVariable variable = event.getRangedVariable().getVariable();
    final NodeValue posEdgeValue =
        new NodeValue(Data.newBitVector(1, variable.getDataType().getSize()));
    final NodeValue negEdgeValue =
        new NodeValue(Data.newBitVector(0, variable.getDataType().getSize()));
    switch (event.getType()) {
      case NEGATIVE_EDGE:
        return NodeUtils.getEquation(variable, negEdgeValue);
      case POSITIVE_EDGE:
        return NodeUtils.getEquation(variable, posEdgeValue);
      default:
        break;
    }
    return null;
  }

  /*
   * Constructs a one-node HLDD from the specified switch node and associated variable.
   * 
   * @param switchNode - the specified {@link CfgSwitchStatement}.
   * 
   * @param variable - the specified variable.
   * 
   * @return The {@link Hldd} associated with a switch.
   */
  private Hldd createHlddFromSwitch(final Node switchNode,
      final NodeVariable variable) {
    final Hldd result = new Hldd(variable);
    final HlddNode singleNode = new HlddNode(NodeType.ASSIGNMENT);
    singleNode.addExpression(switchesAsVariables.get(switchNode));
    singleNode.addExpression(switchNode);
    singleNode.setLevel(0);
    result.addNode(singleNode);
    result.setRootNode(singleNode);
    result.setImmediateAssignment(true);
    return result;
  }

}
