/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.hldd;

import ru.ispras.fortress.expression.Node;
import ru.ispras.retrascope.model.hldd.Hldd;
import ru.ispras.retrascope.model.hldd.HlddModel;
import ru.ispras.retrascope.model.hldd.HlddModule;
import ru.ispras.retrascope.model.hldd.HlddNode;
import ru.ispras.retrascope.model.hldd.HlddNode.NodeType;
import ru.ispras.retrascope.model.hldd.HlddProcess;
import ru.ispras.retrascope.model.hldd.HlddVisitor;

import java.util.ArrayList;
import java.util.List;

/**
 * This class represents a visitor which is used to optimize a visited HLDD.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class HlddOptimizationVisitor implements HlddVisitor {

  // The current visited HLDD.
  private Hldd currentHldd;

  @Override
  public void onModelBegin(final HlddModel model) {
    /* Empty */
  }

  @Override
  public void onModelEnd(final HlddModel model) {
    /* Empty */
  }

  @Override
  public void onHlddBegin(final Hldd hldd) {
    currentHldd = hldd;
  }

  @Override
  public void onHlddEnd(final Hldd hldd) {
    /* Empty */
  }

  @Override
  public void onRootBegin(final HlddNode rootNode) {
    /* Empty */
  }

  @Override
  public void onRootEnd(final HlddNode rootNode) {
    /* Empty */
  }

  @Override
  public void onNodeBegin(final HlddNode node) {
    /* Empty */
  }

  @Override
  public void onNodeEnd(final HlddNode node) {
    if (node.getChildren().isEmpty()) {
      addDefaultAssignment(node);
    }
  }

  /*
   * Adds the default assignment node to the value node.
   * 
   * @param node - the specified value node.
   */
  private void addDefaultAssignment(final HlddNode node) {
    if (node.getType() == NodeType.VALUE) {
      final HlddNode defaultNode = new HlddNode(NodeType.ASSIGNMENT);
      defaultNode.addExpression(currentHldd.getVariable());
      defaultNode.addExpression(currentHldd.getVariable());
      defaultNode.setParent(node);
      defaultNode.setLevel(node.getLevel() + 1);
      node.addChild(defaultNode);
    }
  }

  @Override
  public void onConditionBegin(final HlddNode node) {
    if (node.getChildren().size() == 1) {
      final HlddNode onlyValue = node.getChildren().get(0);
      final HlddNode secondValue = new HlddNode(node, NodeType.VALUE);
      secondValue.setLevel(onlyValue.getLevel());
      secondValue.setDefault();
      currentHldd.addNode(secondValue);
    }
  }

  @Override
  public void onConditionEnd(final HlddNode node) {

    if (allGrandchildrenAreAssignments(node)) {
      final List<HlddNode> assignments = getAllGrandChildren(node);
      if (containIdenticalNodes(assignments)) {
        final Node sameExpr = assignments.get(0).getExpressions().get(1);
        final HlddNode singleNode = new HlddNode(node, NodeType.ASSIGNMENT);
        singleNode.addExpression(currentHldd.getVariable());
        singleNode.addExpression(sameExpr);
        final HlddNode parent = node.getParent();
        parent.removeFromChildren(node);
        parent.addChild(singleNode);
      }
    }

  }

  /*
   * Checks whether all nodes in the list contain equal expressions.
   * 
   * @param nodes - the specified list of nodes.
   * 
   * @return <code>true</code> if all nodes contain equal expressions, <code>false</code> otherwise.
   */
  private boolean containIdenticalNodes(final List<HlddNode> nodes) {
    final Node sameNode = nodes.get(0).getExpressions().get(1);
    for (HlddNode assignment : nodes) {
      if (!assignment.getExpressions().get(1).equals(sameNode)) {
        return false;
      }
    }
    return true;
  }

  /*
   * Checks whether all grandchildren of the specified node are of assignment type.
   * 
   * @param node - the specified node.
   * 
   * @return <code>true</code> if all grandchildren are assignments, <code>false</code> otherwise.
   */
  private boolean allGrandchildrenAreAssignments(final HlddNode node) {
    for (HlddNode child : node.getChildren()) {
      for (HlddNode grandchild : child.getChildren()) {
        if (grandchild.getType() != NodeType.ASSIGNMENT) {
          return false;
        }
      }
    }
    return true;
  }

  /*
   * Returns all grandchildren of the specified node.
   * 
   * @param node - the specified node.
   * 
   * @return The list of grandchildren.
   */
  private List<HlddNode> getAllGrandChildren(final HlddNode node) {
    final ArrayList<HlddNode> result = new ArrayList<HlddNode>();
    for (HlddNode child : node.getChildren()) {
      for (HlddNode grandchild : child.getChildren()) {
        result.add(grandchild);
      }
    }
    return result;
  }

  @Override
  public void onValueBegin(final HlddNode node) {
    /* Empty */
  }

  @Override
  public void onValueEnd(final HlddNode node) {
    if (node.getChildren().isEmpty()) {
      addDefaultAssignment(node);
    }
  }

  @Override
  public void onAssignmentBegin(final HlddNode node) {
    /* Empty */
  }

  @Override
  public void onAssignmentEnd(final HlddNode node) {
    /* Empty */
  }

  @Override
  public void onModuleBegin(final HlddModule module) {
    /* Empty */
  }

  @Override
  public void onModuleEnd(final HlddModule module) {
    /* Empty */
  }

  @Override
  public void onProcessBegin(final HlddProcess process) {
    /* Empty */
  }

  @Override
  public void onProcessEnd(final HlddProcess process) {
    /* Empty */
  }

}
