/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.efsm.extractor.assertion;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmModelDefaultVisitor;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.model.efsm.EfsmModuleWalker;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmUtils;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.assertion.AssertionContainer;
import ru.ispras.retrascope.result.assertion.AssertionEntity;
import ru.ispras.retrascope.result.assertion.AssertionGroup;
import ru.ispras.retrascope.util.SolverUtils;
import ru.ispras.retrascope.util.TransformerUtils;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Visitor for EfsmTransitionAssertExtractor.
 *
 * @see EfsmTransitionAssertExtractor
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class EfsmTransitionAssertVisitor extends EfsmModelDefaultVisitor {

  private AssertionEntity entity;
  private AssertionContainer container;
  private AssertionGroup group;

  /**
   * Returns object's output.
   *
   * @return output object that is built by visiting\walking
   */
  public AssertionEntity getOutput() {
    return this.entity;
  }

  @Override
  public void onEfsmModelBegin(final EfsmModel efsmModel) {
    this.entity = new AssertionEntity();
  }

  @Override
  public void onEfsmModuleBegin(final EfsmModule efsmModule) {
    if (this.container == null) {
      this.container = new AssertionContainer(efsmModule.getName(), null);

      /* Set clock-like variable for the top-level container. */
      Collection<NodeVariable> clocks = getClocks(efsmModule);

      if (!clocks.isEmpty()) {
        this.container.setClock(clocks.iterator().next());
      }

      this.entity.addContainer(this.container);
    } else {
      final AssertionContainer parent = this.container;
      this.container = new AssertionContainer(efsmModule.getName(), parent);
      parent.addContainer(this.container);
    }
    this.container.declareVariables(efsmModule.getVariables());
  }

  private Collection<NodeVariable> getClocks(EfsmModule efsmModule) {

    class ClockCollector extends EfsmModelDefaultVisitor {
      private Set<NodeVariable> clocks;

      private Set<NodeVariable> resets;

      ClockCollector() {
        this.clocks = new LinkedHashSet<>();
        this.resets = new LinkedHashSet<>();
      }

      Collection<NodeVariable> getClocks() {
        return CollectionUtils.complementSets(this.clocks, this.resets);
      }

      @Override
      public void onEfsmBegin(final Efsm efsm) {
        this.clocks.addAll(efsm.getClocks());

        Set<GuardedAction> resetActions = efsm.getResetGuardedActions();

        if (!resetActions.isEmpty()) {
          for (final GuardedAction resetAction : resetActions) {
            this.resets.addAll(ExprUtils.getVariables(resetAction.getGuard().toNode()));
          }
        }
      }
    }

    final ClockCollector collector = new ClockCollector();
    final EfsmModuleWalker walker = new EfsmModuleWalker(collector);
    walker.start(efsmModule);

    return collector.getClocks();
  }

  @Override
  public void onEfsmModuleEnd(final EfsmModule efsmModule) {
    this.container = this.container.getParent();
  }

  @Override
  public void onEfsmBegin(final Efsm efsm) {
    this.group = new AssertionGroup(efsm.getName(), this.container);
    this.group.declareVariables(efsm.getVariables());
    this.container.addGroup(this.group);

    /* Set reset-like variable. */
    final Set<GuardedAction> resetGuardedActions = efsm.getResetGuardedActions();
    if (!resetGuardedActions.isEmpty()) {
      final GuardedAction resetGuardedAction = resetGuardedActions.iterator().next();
      if (resetGuardedAction != null) {
        final NodeVariable resetSignal = getResetSignal(resetGuardedAction);
        setResetWhereDeclared(this.group, resetSignal);
      }
    }
  }

  /* It is supposed that the guard depends on exactly one variable. */
  private NodeVariable getResetSignal(final GuardedAction guardedAction) {

    final Node guardCondition = guardedAction.getGuard().toNode();
    final Collection<NodeVariable> uses = ExprUtils.getVariables(guardCondition);

    if (uses.size() == 1) {
      return uses.iterator().next();
    } else {
      throw new RetrascopeRuntimeException(
          String.format(
              "Resetting guarded action cannot depend on more than one variable: %s.",
              Retrascope.getConfiguration().getStylePrinter().toString(guardCondition)));
    }
  }

  private void setResetWhereDeclared(
      final AssertionGroup group,
      final NodeVariable resetSignal) {

    if (group.hasDeclared(resetSignal)) {
      group.setResetVariable(resetSignal);
    } else {
      AssertionContainer parentContainer = group.getParent();
      while (parentContainer != null) {

        if (parentContainer.hasDeclared(resetSignal)) {
          parentContainer.setResetVariable(resetSignal);
          break;
        }

        parentContainer = parentContainer.getParent();
      }
    }
  }

  @Override
  public void onTransitionBegin(final EfsmTransition transition) {

    final EfsmState srcState = transition.getSourceState();
    final GuardedAction guardedAction = transition.getGuardedAction();
    final Guard guard = guardedAction.getGuard();
    final EventList clockEvents = guardedAction.getEventList();
    final Node eventsToExpression = EfsmUtils.eventsToNode(clockEvents.getEvents());

    final Node[] assertionPredicates =
        eventsToExpression == null
            ? new Node[] {srcState.getExpression(), guard.toNode()}
            : new Node[] {srcState.getExpression(), guard.toNode(), eventsToExpression};
    final Node assertion =
        TransformerUtils.standardize(ExprUtils.getConjunction(assertionPredicates));

    /* Filter FALSE assertions. */
    if (!assertion.equals(SolverUtils.FALSE)) {
      this.group.addAssertion(assertion);
    }
  }
}
