/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.SolverResult.Status;
import ru.ispras.fortress.solver.SolverResultBuilder;
import ru.ispras.retrascope.engine.basis.AtomicTestGenerator;
import ru.ispras.retrascope.engine.efsm.generator.test.heuristic.direction.NearestUncoveredEfsmTransitionChooser;
import ru.ispras.retrascope.engine.efsm.simulator.ModuleSimulator;
import ru.ispras.retrascope.engine.efsm.simulator.ProcessSimulator;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.Efsm.TransitionWeightBasis;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.Vector;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Directed test generator for single EFSM.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see RandomGenerator
 * @see EfsmTestGenerator
 */
final class DirectedGenerator extends AtomicTestGenerator {
  private static final Logger LOGGER = Log.getLogger();
  private final String logEntryHeader;

  private final Set<List<EfsmTransition>> traversablePaths;
  private final Map<EfsmTransition, Set<EfsmTransition>> testedCoreTransitions;

  private final ProcessSimulator simulator;
  private int loopLimit;

  private final EfsmTransitionCoverage totalCoverage;

  private NearestUncoveredEfsmTransitionChooser directedChooser;
  private final EfsmAnalyser analyser;

  private final Map<EfsmState, Map<EfsmState, List<EfsmTransition>>> preferredBridges;
  private final Set<List<List<EfsmTransition>>> untraversableFragmentedPaths;
  private final Set<List<EfsmTransition>> untraversablePlainPaths;

  /**
   * Constructs a new EFSM atomic test generator for the specified EFSM.
   *
   * @param efsm the EFSM
   * @param traversablePaths a set of paths which are guaranteed to be traversable
   * @param loopLimit max number of single EFSM loop traversing in one path
   */
  DirectedGenerator(Efsm efsm, Set<List<EfsmTransition>> traversablePaths, int loopLimit,
      ModuleSimulator moduleSimulator) {
    this.logEntryHeader = "EFSM.TestGenerator.DirectedGenerator." + efsm.getName();

    this.loopLimit = loopLimit;

    this.simulator = moduleSimulator.getSubsimulator(efsm);

    this.totalCoverage = new EfsmTransitionCoverage(efsm);

    this.analyser = new EfsmAnalyser(efsm);

    for (List<EfsmTransition> traversablePath: traversablePaths) {
      for (EfsmTransition coveredTransition: traversablePath) {
        this.totalCoverage.markTransitionAsCovered(coveredTransition);
      }
    }

    this.traversablePaths = new HashSet<>(traversablePaths);
    this.testedCoreTransitions = new HashMap<>();

    this.directedChooser = null;

    this.preferredBridges = new HashMap<>();
    this.untraversableFragmentedPaths = new HashSet<>();
    this.untraversablePlainPaths = new HashSet<>();
  }

  /**
   * Returns an EFSM which this generator is related to.
   *
   * @return the EFSM
   */
  public Efsm getEfsm() {
    return this.simulator.getSimulatedModel();
  }

  /**
   * Returns an iterator enabling to retrieve next sequence of vectors produced by this generator.
   *
   * @return the iterator. If this generator is already finished generation,
   *         {@code null} is returned
   */
  @Override
  public Iterator<Vector> getNextSequenceIterator() {
    if (!this.totalCoverage.isFull()) {
      if (this.directedChooser == null) {
        this.directedChooser = new NearestUncoveredEfsmTransitionChooser(
            this.simulator.getSimulatedModel(), this.totalCoverage);
      }

      for (EfsmTransition transition = this.directedChooser.chooseNext(); transition != null;
          transition = this.directedChooser.chooseNext()) {
        List<Vector> vectors = this.generateSequence(transition);

        if (vectors != null) {
          this.directedChooser = new NearestUncoveredEfsmTransitionChooser(
              this.simulator.getSimulatedModel(), this.totalCoverage);

          return vectors.iterator();
        }
      }
    }

    return null;
  }

  private List<Vector> generateSequence(EfsmTransition transition) {
    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader
        + ": generating a sequence to cover the transition: " + transition);

    List<Vector> vectors = this.tryDirectPath(transition);

    if (vectors != null) {
      return vectors;
    }

    Map<EfsmTransition, Set<List<EfsmTransition>>> definitionPaths = new HashMap<>();
    Map<EfsmTransition, Set<Set<List<EfsmTransition>>>> loopDefinitionPaths = new HashMap<>();

    for (List<EfsmTransition> definitionPath: this.analyser.getDefinitionPaths(transition)) {
      EfsmTransition coreTransition = definitionPath.get(0);
      Set<List<EfsmTransition>> relatedPaths = definitionPaths.get(coreTransition);

      if (relatedPaths == null) {
        relatedPaths = new HashSet<>();
        definitionPaths.put(coreTransition, relatedPaths);
      }

      relatedPaths.add(definitionPath);
    }

    for (Set<List<EfsmTransition>> definitionLoops:
        this.analyser.getLoopDefinitionPaths(transition)) {
      EfsmTransition coreTransition = definitionLoops.iterator().next().get(0);
      Set<Set<List<EfsmTransition>>> relatedLoops = loopDefinitionPaths.get(coreTransition);

      if (relatedLoops == null) {
        relatedLoops = new HashSet<>();
        loopDefinitionPaths.put(coreTransition, relatedLoops);
      }

      relatedLoops.add(definitionLoops);
    }

    Set<EfsmTransition> coreTransitions = new HashSet<>(definitionPaths.keySet());
    coreTransitions.addAll(loopDefinitionPaths.keySet());

    if (coreTransitions.isEmpty()) {
      LOGGER.log(LogLevel.DEBUG,
          this.logEntryHeader + ": the transition is independent of inner variables");
    }

    for (EfsmTransition coreTransition: coreTransitions) {
      if (this.testedCoreTransitions.containsKey(transition)
          && this.testedCoreTransitions.get(transition).contains(coreTransition)) {
        continue;
      }

      if (this.totalCoverage.isCovered(coreTransition)) {
        LOGGER.log(LogLevel.DEBUG, this.logEntryHeader
            + ": trying full definition paths from the transition: " + coreTransition);

        List<EfsmTransition> shortestHeadPath = this.findShortestTraversablePath(coreTransition);

        if (shortestHeadPath.isEmpty()) {
          LOGGER.log(LogLevel.DEBUG,
              this.logEntryHeader + ": the shortest next head path is empty");
        } else {
          LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": the shortest head path: ");
          this.logPlainPath(shortestHeadPath, LogLevel.DEBUG);
        }

        if (loopDefinitionPaths.containsKey(coreTransition)) {
          for (Set<List<EfsmTransition>> loopsGroup :
              loopDefinitionPaths.get(coreTransition)) {
            vectors = this.traverseLoop(loopsGroup, shortestHeadPath, transition);

            if (vectors != null) {
              break;
            }
          }
        }

        if ((vectors == null) && definitionPaths.containsKey(coreTransition)) {
          for (List<EfsmTransition> definitionPath : definitionPaths.get(coreTransition)) {
            List<List<EfsmTransition>> fragmentedDefinitionPath =
                this.fragmentPath(definitionPath);
            LOGGER.log(LogLevel.DEBUG,
                this.logEntryHeader + ": trying the definition path: ");
            this.logFragmentedPath(fragmentedDefinitionPath, LogLevel.DEBUG);

            List<List<EfsmTransition>> corePath = new ArrayList<>();
            corePath.add(shortestHeadPath);
            corePath.addAll(fragmentedDefinitionPath);
            corePath.add(Collections.singletonList(transition));
            vectors = this.traverseFragmentedPath(corePath);

            if (vectors != null) {
              break;
            }
          }
        }

        if (vectors != null) {
          return vectors;
        }

        Set<EfsmTransition> testedCoreTransitions = this.testedCoreTransitions.get(transition);

        if (testedCoreTransitions == null) {
          testedCoreTransitions = new HashSet<>();
          this.testedCoreTransitions.put(transition, testedCoreTransitions);
        }

        testedCoreTransitions.add(coreTransition);
      }
    }

    return null;
  }

  private List<Vector> tryDirectPath(EfsmTransition transition) {
    LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": trying the shortest direct path to the target transition");

    List<EfsmTransition> directPath = null;

    for (EfsmTransition predecessor:
        this.simulator.getSimulatedModel().getIncomingTransitions(transition.getSourceState())) {
      if (this.totalCoverage.isCovered(predecessor)) {
        List<EfsmTransition> candidatePath =
            new ArrayList<>(this.findShortestTraversablePath(predecessor));

        if ((directPath == null) || (candidatePath.size() + 1 < directPath.size())) {
          candidatePath.add(predecessor);
          directPath = candidatePath;
        }
      }
    }

    directPath.add(transition);

    return this.traversePlainPath(directPath);
  }

  private List<List<EfsmTransition>> fragmentPath(List<EfsmTransition> plainePath) {
    if (LOGGER.isLoggable(LogLevel.DEBUG)) {
      LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": fragmenting the path: ");
      this.logPlainPath(plainePath, LogLevel.DEBUG);
    }

    List<List<EfsmTransition>> fragmentedPath = new ArrayList<>();
    List<EfsmTransition> currentFragment = new ArrayList<>();
    fragmentedPath.add(currentFragment);
    EfsmState currentState = null;

    for (EfsmTransition transition: plainePath) {
      if (!currentFragment.isEmpty()) {
        EfsmState nextState = transition.getSourceState();

        if (!currentState.equals(nextState)) {
          currentFragment = new ArrayList<>();
          fragmentedPath.add(currentFragment);
        }
      }

      currentFragment.add(transition);
      currentState = transition.getDestinationState();
    }

    if (LOGGER.isLoggable(LogLevel.DEBUG)) {
      LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": fragmented path: ");
      this.logFragmentedPath(fragmentedPath, LogLevel.DEBUG);
    }

    return fragmentedPath;
  }

  private List<List<EfsmTransition>> smoothPath(List<List<EfsmTransition>> fragmentedPath) {
    if (LOGGER.isLoggable(LogLevel.DEBUG)) {
      LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": smoothing the fragmented path: ");
      this.logFragmentedPath(fragmentedPath, LogLevel.DEBUG);
    }

    List<List<EfsmTransition>> smoothPath = new ArrayList<>();
    EfsmState currentState = null;

    for (List<EfsmTransition> fragment: fragmentedPath) {
      if (!fragment.isEmpty()) {
        if (currentState != null) {
          EfsmState nextState = fragment.get(0).getSourceState();

          if (!currentState.equals(nextState)) {
            List<EfsmTransition> bridge;

            if (this.preferredBridges.containsKey(currentState)
                && this.preferredBridges.get(currentState).containsKey(nextState)) {
              bridge = this.preferredBridges.get(currentState).get(nextState);
            } else {
              bridge = this.simulator.getSimulatedModel().getShortestPath(currentState, nextState,
                  TransitionWeightBasis.GUARD_REGISTER_AMOUNT);
            }

            smoothPath.add(bridge);
          }
        }

        smoothPath.add(fragment);
        currentState = fragment.get(fragment.size() - 1).getDestinationState();
      }
    }

    if (LOGGER.isLoggable(LogLevel.DEBUG)) {
      LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": smooth path: ");
      this.logFragmentedPath(fragmentedPath, LogLevel.DEBUG);
    }

    return smoothPath;
  }

  private List<EfsmTransition> planePath(List<List<EfsmTransition>> fragmentedPath) {
    List<EfsmTransition> plainPath = new ArrayList<>();

    for (List<EfsmTransition> fragment: fragmentedPath) {
      plainPath.addAll(fragment);
    }

    return plainPath;
  }

  private List<Vector> traverseLoop(Set<List<EfsmTransition>> plainLoopCores,
      List<EfsmTransition> headPath, EfsmTransition targetTransition) {

    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": fragmenting the loop cores");

    Set<List<List<EfsmTransition>>> loopCores = new HashSet<>(plainLoopCores.size());

    for (List<EfsmTransition> plainLoopCore: plainLoopCores) {
      LinkedList<EfsmTransition> normalisedLoopCore = new LinkedList<>(plainLoopCore);
      normalisedLoopCore.addLast(normalisedLoopCore.removeFirst());
      loopCores.add(this.fragmentPath(normalisedLoopCore));
    }

    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": traversing the loop");
    List<EfsmTransition> currentPath = new ArrayList<>(headPath);
    currentPath.add(plainLoopCores.iterator().next().get(0));

    return this.traverseMixedLoop(loopCores, this.fragmentPath(currentPath), targetTransition, 0);
  }

  private List<Vector> traverseMixedLoop(Set<List<List<EfsmTransition>>> loopCores,
      List<List<EfsmTransition>> currentPath, EfsmTransition targetTransition, int limit) {
    if (limit == this.loopLimit) {
      LOGGER.log(LogLevel.DEBUG,
          this.logEntryHeader + ": loop limit is reached: " + this.loopLimit);

      return null;
    }

    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": current path at iteration " + limit + ":");
    this.logFragmentedPath(currentPath, LogLevel.DEBUG);

    for (List<List<EfsmTransition>> loopCore: loopCores) {
      LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": proceeding with the loop:");
      this.logFragmentedPath(loopCore, LogLevel.DEBUG);

      List<List<EfsmTransition>> newCurrentPath = new ArrayList<>(currentPath);
      newCurrentPath.addAll(loopCore);

      if (this.checkFragmentedPathTraversability(newCurrentPath)) {
        LOGGER.log(LogLevel.DEBUG, this.logEntryHeader
            + ": the new loop is traversable. Trying to reach the target transition");

        List<List<EfsmTransition>> completePath = new ArrayList<>(newCurrentPath);
        completePath.add(Collections.singletonList(targetTransition));

        List<Vector> vectors = this.traverseFragmentedPath(completePath);

        if (vectors != null) {
          return vectors;
        } else {
          LOGGER.log(LogLevel.DEBUG,
              this.logEntryHeader + ": the target transition is still unreachable");

          vectors = this.traverseMixedLoop(loopCores, newCurrentPath, targetTransition, limit + 1);

          if (vectors != null) {
            return vectors;
          }
        }
      } else {
        LOGGER.log(LogLevel.DEBUG,
            this.logEntryHeader + ": the new loop is untraversable. Rolling back");
      }
    }

    return null;
  }

  private boolean checkFragmentedPathTraversability(List<List<EfsmTransition>> corePath) {
    LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": checking the fragmented path for traversability: ");
    this.logFragmentedPath(corePath, LogLevel.DEBUG);

    for (List<List<EfsmTransition>> untraversablePath: this.untraversableFragmentedPaths) {
      if ((corePath.size() >= untraversablePath.size())
          && corePath.subList(0, untraversablePath.size()).equals(untraversablePath)) {

        return false;
      }
    }

    List<List<EfsmTransition>> smoothPath = this.smoothPath(corePath);
    boolean traversability = this.checkPlainPathTraversability(this.planePath(smoothPath));

    if (!traversability) {
      List<List<EfsmTransition>> currentCorePath = new ArrayList<>();
      List<List<EfsmTransition>> newSmoothPath = new ArrayList<>();
      LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": looking for untraversable fragment");

      for (List<EfsmTransition> fragment: smoothPath) {
        if (corePath.contains(fragment)) {
          currentCorePath.add(fragment);
        }

        newSmoothPath.add(fragment);
        List<EfsmTransition> plainPath = this.planePath(newSmoothPath);
        traversability = this.checkPlainPathTraversability(plainPath);

        if (!traversability) {
          LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": the untraversable fragment: ");
          this.logPlainPath(fragment, LogLevel.DEBUG);

          if (corePath.contains(fragment)) {
            LOGGER.log(LogLevel.DEBUG,
                this.logEntryHeader + ": the untraversable fragment is in the core path");

            this.untraversableFragmentedPaths.add(currentCorePath);

            return false;
          } else {
            EfsmState sourceState = fragment.get(0).getSourceState();
            EfsmState destinationState = fragment.get(fragment.size() - 1).getDestinationState();
            Set<EfsmTransition> forbiddenTransitions = new HashSet<>();

            /*for (Iterator<List<EfsmTransition>> i = corePath.listIterator(currentCorePath.size());
                i.hasNext();) {
              forbiddenTransitions.addAll(i.next());
            }*/

            List<List<EfsmTransition>> pathCandidates =
                this.findAllPaths(sourceState, destinationState, forbiddenTransitions);
            boolean success = false;

            for (List<EfsmTransition> pathCandidate: pathCandidates) {
              if (!pathCandidate.equals(fragment)) {
                LOGGER.log(LogLevel.DEBUG,
                    this.logEntryHeader + ": replacing the untraversable fragment with: ");
                this.logPlainPath(pathCandidate, LogLevel.DEBUG);

                newSmoothPath.remove(newSmoothPath.size() - 1);
                newSmoothPath.add(pathCandidate);

                if (this.checkPlainPathTraversability(this.planePath(newSmoothPath))) {
                  success = true;

                  LOGGER.log(LogLevel.DEBUG,
                      this.logEntryHeader + ": replacement is successful");

                  Map<EfsmState, List<EfsmTransition>> bridges =
                      this.preferredBridges.get(sourceState);

                  if (bridges == null) {
                    bridges = new HashMap<>();
                    this.preferredBridges.put(sourceState, bridges);
                  }

                  bridges.put(destinationState, pathCandidate);

                  break;
                }
              }
            }

            if (!success) {
              LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": replacement is unsuccessful");

              currentCorePath.add(corePath.get(currentCorePath.size()));
              this.untraversableFragmentedPaths.add(currentCorePath);

              return false;
            }
          }
        }
      }
    }

    return traversability;
  }

  private List<Vector> traverseFragmentedPath(List<List<EfsmTransition>> corePath) {
    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": traversing the fragmented path: ");
    this.logFragmentedPath(corePath, LogLevel.DEBUG);

    for (List<List<EfsmTransition>> untraversablePath: this.untraversableFragmentedPaths) {
      if ((corePath.size() >= untraversablePath.size())
          && corePath.subList(0, untraversablePath.size()).equals(untraversablePath)) {

        return null;
      }
    }

    List<List<EfsmTransition>> smoothPath = this.smoothPath(corePath);
    List<Vector> vectors = this.traversePlainPath(this.planePath(smoothPath));

    if (vectors == null) {
      List<List<EfsmTransition>> currentCorePath = new ArrayList<>();
      List<List<EfsmTransition>> newSmoothPath = new ArrayList<>();
      LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": looking for untraversable fragment");

      for (List<EfsmTransition> fragment: smoothPath) {
        if (corePath.contains(fragment)) {
          currentCorePath.add(fragment);
        }

        newSmoothPath.add(fragment);
        List<EfsmTransition> plainPath = this.planePath(newSmoothPath);

        boolean success;

        if (currentCorePath.size() == corePath.size()) {
          vectors = this.traversePlainPath(plainPath);
          success = vectors != null;
        } else {
          success = this.checkPlainPathTraversability(plainPath);
        }

        if (!success) {
          LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": the untraversable fragment: ");
          this.logPlainPath(fragment, LogLevel.DEBUG);

          if (corePath.contains(fragment)) {
            LOGGER.log(LogLevel.DEBUG,
                this.logEntryHeader + ": the untraversable fragment is in the core path. "
                    + "Finishing traverse of this core path");

            this.untraversableFragmentedPaths.add(currentCorePath);

            return null;
          } else {
            EfsmState sourceState = fragment.get(0).getSourceState();
            EfsmState destinationState = fragment.get(fragment.size() - 1).getDestinationState();
            Set<EfsmTransition> forbiddenTransitions = new HashSet<>();

            /*for (Iterator<List<EfsmTransition>> i = corePath.listIterator(currentCorePath.size());
                i.hasNext();) {
              forbiddenTransitions.addAll(i.next());
            }*/

            List<List<EfsmTransition>> pathCandidates =
                this.findAllPaths(sourceState, destinationState, forbiddenTransitions);
            success = false;

            for (List<EfsmTransition> pathCandidate: pathCandidates) {
              if (!pathCandidate.equals(fragment)) {
                LOGGER.log(LogLevel.DEBUG,
                    this.logEntryHeader + ": replacing the untraversable fragment with: ");
                this.logPlainPath(pathCandidate, LogLevel.DEBUG);

                newSmoothPath.remove(newSmoothPath.size() - 1);
                newSmoothPath.add(pathCandidate);

                if (this.checkPlainPathTraversability(this.planePath(newSmoothPath))) {
                  success = true;

                  LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": replacement is successful");

                  break;
                }
              }
            }

            if (!success) {
              LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": replacement is unsuccessful");

              currentCorePath.add(corePath.get(currentCorePath.size()));
              this.untraversableFragmentedPaths.add(currentCorePath);

              return null;
            }
          }
        }
      }
    }

    return vectors;
  }

  private List<Vector> traversePlainPath(List<EfsmTransition> path) {
    SolverResult solverResult = this.resolvePlainPath(path);

    if (solverResult.getStatus().equals(Status.SAT)) {
      LOGGER.log(LogLevel.DEBUG,
          this.logEntryHeader + ": the path is traversable. Generating input vectors");

      Map<Integer, Map<String, Data>> inputs =
          Substitutor.deannotateInputs(solverResult.getVariables());

      List<Vector> vectors = new ArrayList<>(path.size());

      for (ListIterator<EfsmTransition> i = path.listIterator(); i.hasNext();) {
        int index = i.nextIndex();
        EfsmTransition transition = i.next();

        Map<String, Data> vectorInputs = inputs.get(index);
        Transaction transaction = new Transaction();

        if (vectorInputs != null) {
          for (Entry<String, Data> input: vectorInputs.entrySet()) {
            transaction.addVariable(input.getKey(), new NodeValue(input.getValue()));
          }
        }

        Vector vector = new Vector();
        vector.addTransaction(transaction);
        vector.addEvents(transition.getGuardedAction().getClocks());
        vectors.add(vector);

        LOGGER.log(LogLevel.DEBUG,
            this.logEntryHeader + ": generated vector: " + vector);

        this.totalCoverage.markTransitionAsCovered(transition);
      }

      this.traversablePaths.add(path);

      return vectors;
    } else {
      LOGGER.log(LogLevel.DEBUG,
          this.logEntryHeader + ": the path is untraversable");

      this.untraversablePlainPaths.add(path);

      return null;
    }
  }

  private boolean checkPlainPathTraversability(List<EfsmTransition> path) {
    boolean traversability = this.resolvePlainPath(path).getStatus().equals(Status.SAT);

    if (!traversability) {
      this.untraversablePlainPaths.add(path);
    }

    return traversability;
  }

  private SolverResult resolvePlainPath(List<EfsmTransition> path) {
    for (List<EfsmTransition> untraversablePath: this.untraversablePlainPaths) {
      if ((path.size() >= untraversablePath.size())
          && path.subList(0, untraversablePath.size()).equals(untraversablePath)) {

        return new SolverResultBuilder(Status.UNSAT).build();
      }
    }

    LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": checking the plain path for traversability: ");
    this.logPlainPath(path, LogLevel.DEBUG);

    if (!this.simulator.getCurrentState().equals(path.get(0).getSourceState())) {
      throw new IllegalArgumentException("The path is started not in the current state of the"
          + " simulator");
    }

    Efsm efsm = this.simulator.getSimulatedModel();
    Substitutor substitutor = new Substitutor(efsm);
    Node pathConstraint = substitutor.backtrack(path);

    LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": the traverse constraint: " + pathConstraint);

    pathConstraint =
        this.simulator.fillExpression(pathConstraint, Collections.<String, NodeValue>emptyMap());

    LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": the constraint to resolve: " + pathConstraint);

    return SolverUtils.solve(pathConstraint);
  }

  private List<List<EfsmTransition>> findAllPaths(EfsmState sourceState,
      EfsmState destinationState, Set<EfsmTransition> forbiddenTransitions) {
    List<List<EfsmTransition>> foundPaths;
    Efsm efsm = this.simulator.getSimulatedModel();

    if (efsm.isReachable(sourceState, destinationState)) {
      foundPaths = new ArrayList<>();

      for (EfsmTransition firstTransition: efsm.getOutgoingTransitions(sourceState)) {
        foundPaths.addAll(this.buildPaths(
            Collections.singletonList(firstTransition), destinationState,
            forbiddenTransitions));
      }
    } else {
      foundPaths = Collections.emptyList();
    }

    Collections.sort(foundPaths, new Comparator<List<EfsmTransition>>() {
      @Override
      public int compare(List<EfsmTransition> list1, List<EfsmTransition> list2) {
        return list1.size() - list2.size();
      }});

    return foundPaths;
  }

  private Set<List<EfsmTransition>> buildPaths(List<EfsmTransition> currentPath,
      EfsmState destinationState, Set<EfsmTransition> forbiddenTransitions) {

    Set<List<EfsmTransition>> foundPaths = new HashSet<>();
    EfsmState currentState = currentPath.get(currentPath.size() - 1).getDestinationState();

    if (currentState.equals(destinationState)) {
      foundPaths.add(currentPath);
    } else {
      Efsm efsm = this.simulator.getSimulatedModel();

      for (EfsmTransition nextTransition: efsm.getOutgoingTransitions(currentState)) {
        if (!currentPath.contains(nextTransition)
            && !forbiddenTransitions.contains(nextTransition)) {
          List<EfsmTransition> newCurrentPath = new ArrayList<>(currentPath);
          newCurrentPath.add(nextTransition);
          foundPaths.addAll(this.buildPaths(newCurrentPath, destinationState,
              forbiddenTransitions));
        }
      }
    }

    return foundPaths;
  }

  private void logFragmentedPath(List<List<EfsmTransition>> fragmentedPath, Level logLevel) {
    for (List<EfsmTransition> fragment: fragmentedPath) {
      LOGGER.log(logLevel, this.logEntryHeader + ": path fragment: ");
      this.logPlainPath(fragment, logLevel);
    }
  }

  private void logPlainPath(List<EfsmTransition> plainPath, Level logLevel) {
    for (EfsmTransition transtion: plainPath) {
      LOGGER.log(logLevel, this.logEntryHeader + ": " + transtion.toString());
    }
  }

  private List<EfsmTransition> findShortestTraversablePath(EfsmTransition transition) {
    List<EfsmTransition> shortestPath = null;

    for (List<EfsmTransition> traversablePath: this.traversablePaths) {
      int index = traversablePath.indexOf(transition);

      if (index != -1) {
        if ((shortestPath == null) || (index + 1 < shortestPath.size())) {
          shortestPath = traversablePath.subList(0, index);
        }
      }
    }

    return shortestPath;
  }

  @Override
  public void reset() {
    // TODO Auto-generated method stub
  }
}
