/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test;

import org.apache.commons.cli.ParseException;

import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.engine.basis.TestGeneratorEngine;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.TestModel;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.Map;
import java.util.logging.Logger;

/**
 * The test generator engine for EFSM-based hardware models.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see EfsmModel
 */
public final class EfsmTestGenerator extends TestGeneratorEngine {
  /**
   * The engine ID of all {@code EfsmTestGenerator} instances.
   */
  public static final String ENGINE_ID = "efsm-test-generator";

  public static final Parameter LOOP_LIMIT_PARAMETER =
      new Parameter("loop-limit", true, "Loop iteration limit");
  private static final int DEFAULT_LOOP_LIMIT = 1;

  private static final Logger logger = Log.getLogger(EfsmTestGenerator.class);
  private static final String LOG_ENTRY_HEADER = "EFSM.TestGenerator: ";

  /**
   * Constructs a generator.
   */
  public EfsmTestGenerator() {
    super(EfsmTestGenerator.ENGINE_ID, EfsmModel.TYPE);
    super.addParameter(EfsmTestGenerator.LOOP_LIMIT_PARAMETER);
  }

  /**
   * Generates and returns tests for the specified EFSM modules.
   *
   * @param inputs a map containing the EFSM modules and command line arguments. It must contain a
   *        mapping from the {@link EfsmModel} entity type to an {@link EfsmModel} object.
   *        Optionally, it may include a mapping from the {@link Arguments} entity type to an
   *        {@link Arguments} object.
   *
   * @throws IllegalArgumentException if {@code inputs} is {@code null}
   * @throws IllegalArgumentException if {@code inputs} contains no mapping from the
   *         {@link EfsmModel} entity type to an {@link EfsmModel} object
   *
   * @return the generated tests
   */
  @Override
  public TestModel start(Map<EntityType, Entity> inputs) {
    EfsmModel efsmModel = Engine.getMandatoryInputEntity(inputs, EfsmModel.class, EfsmModel.TYPE);

    int loopLimit = -1;

    Arguments arguments = Engine.getOptionalInputEntity(inputs, Arguments.class, Arguments.TYPE);

    if (arguments != null) {
      try {
        Arguments parsedArguments = super.parseCommandLine(arguments.getCommandLine());

        loopLimit =
            Integer.parseInt(parsedArguments
                .getValue(EfsmTestGenerator.LOOP_LIMIT_PARAMETER, "-1"));
      } catch (ParseException | NumberFormatException e) {
        throw new IllegalArgumentException("The exception has occurred while parsing command line "
            + "arguments", e);
      }
    }

    if (loopLimit == -1) {
      loopLimit = EfsmTestGenerator.DEFAULT_LOOP_LIMIT;

      EfsmTestGenerator.logger.log(LogLevel.INFO, EfsmTestGenerator.LOG_ENTRY_HEADER
          + "loop limit isn't specified. Setting it to the default value: " + loopLimit);
    } else {
      EfsmTestGenerator.logger.log(LogLevel.DEBUG,
          EfsmTestGenerator.LOG_ENTRY_HEADER + "the specified loop limit: " + loopLimit);
    }

    TestModel testModel = new TestModel();

    for (EfsmModule module : efsmModel.getModules()) {
      String moduleLogPrefix =
          EfsmTestGenerator.LOG_ENTRY_HEADER + "module " + module.getName() + ": ";

      EfsmTestGenerator.logger.log(LogLevel.INFO, moduleLogPrefix + "starting test generation");

      Generator generator = new Generator(module, loopLimit);
      Test test = generator.generateTest();
      testModel.addTest(test);

      EfsmTestGenerator.logger.log(LogLevel.INFO,
          moduleLogPrefix + "test generation has been finished successfully");
      EfsmTestGenerator.logger.log(LogLevel.INFO,
          moduleLogPrefix + "test coverage: " + generator.getCoverage());
      EfsmTestGenerator.logger.log(LogLevel.INFO,
          moduleLogPrefix + "amount of sequences: " + test.getSequences().size());

      int sequenceLength = 0;

      for (Sequence sequence: test.getSequences()) {
        sequenceLength = sequenceLength + sequence.getLength();
      }

      EfsmTestGenerator.logger.log(LogLevel.INFO,
          moduleLogPrefix + "amount of vectors: " + sequenceLength);
    }

    return testModel;
  }
}
