/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test;

import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.retrascope.engine.basis.AtomicTestGenerator;
import ru.ispras.retrascope.engine.efsm.generator.test.heuristic.direction.SemiRandomEfsmTransitionChooser;
import ru.ispras.retrascope.engine.efsm.simulator.ProcessSimulator;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.result.test.Vector;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.Collections;
import java.util.Iterator;
import java.util.Map;
import java.util.logging.Logger;

/**
 * Random-based test generator for single EFSM.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see DirectedGenerator
 * @see EfsmTestGenerator
 */
final class RandomGenerator extends AtomicTestGenerator {
  private final Logger logger;
  private final String logEntryHeader;

  private final ProcessSimulator simulator;

  private final EfsmTransitionCoverage totalCoverage;

  private EventList fixedEvents;
  private Map<String, NodeValue> fixedInputs;

  private SemiRandomEfsmTransitionChooser chooser;

  private long seed;

  /**
   * Constructs a new EFSM random test generator on the basis of the specified process simulator.
   *
   * @param simulator the simulator
   */
  RandomGenerator(ProcessSimulator simulator) {
    Efsm efsm = simulator.getSimulatedModel();
    this.logEntryHeader = "EFSM.TestGenerator.RandomGenerator." + efsm.getName();
    this.logger = Log.getLogger(this.getClass());

    this.simulator = simulator;

    this.totalCoverage = new EfsmTransitionCoverage(efsm);

    this.reset();
  }

  /**
   * Fixes the specified vector for the next generation step.
   *
   * @param vector the vector to be fixed
   */
  void fixInputs(Vector vector) {
    this.fixedEvents = vector.getEventList();
    this.fixedInputs = vector.getAssignments();
  }

  private boolean isCurrentSequenceFinished() {
    return this.simulator.getCoverage().isFull()
        || ((this.chooser != null) && !this.chooser.hasNext());
  }

  @Override
  public Iterator<Vector> getNextSequenceIterator() {
    return new RandomSequenceIterator();
  }

  /**
   * Returns EFSM which this generator is related to.
   *
   * @return the EFSM
   */
  public Efsm getEfsm() {
    return this.simulator.getSimulatedModel();
  }

  private Vector generateInputVectorRandomly() {
    this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": generating a new input vector");

    this.chooser = new SemiRandomEfsmTransitionChooser(this.simulator.getSimulatedModel(),
        this.simulator.getCurrentState(), this.totalCoverage, this.seed);

    for (EfsmTransition transition = this.chooser.chooseNext(); transition != null;
        transition = this.chooser.chooseNext()) {
      Vector vector = this.traverseTransition(transition);

      if (vector != null) {
        return vector;
      }
    }

    return null;
  }

  private Vector traverseTransition(EfsmTransition transition) {
    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": traversing the transition " + transition);

    Vector inputVector = this.simulator.enableGuardedAction(transition.getGuardedAction(),
        this.fixedEvents, this.fixedInputs);

    if (inputVector != null) {
      this.logger.log(LogLevel.DEBUG,
          this.logEntryHeader + ": the transition has been traversed successfully");

      this.totalCoverage.markTransitionAsCovered(transition);
      this.chooser =
          new SemiRandomEfsmTransitionChooser(this.simulator.getSimulatedModel(),
              this.simulator.getCurrentState(), this.simulator.getCoverage(), this.seed);
    } else {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": failed to traverse the transition");
    }

    return inputVector;
  }

  private class RandomSequenceIterator implements Iterator<Vector> {
    RandomSequenceIterator() {
      RandomGenerator.this.seed++;
    }

    @Override
    public boolean hasNext() {
      return !RandomGenerator.this.isCurrentSequenceFinished();
    }

    @Override
    public Vector next() {
      if (this.hasNext()) {
        return RandomGenerator.this.generateInputVectorRandomly();
      } else {
        return null;
      }
    }

    @Override
    public void remove() {
      throw new UnsupportedOperationException("remove");
    }
  }

  @Override
  public void reset() {
    this.fixedEvents = new EventList();
    this.fixedInputs = Collections.emptyMap();

    this.totalCoverage.reset();

    this.chooser = null;
    this.seed = 0L;
  }
}
