/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeBinding.BoundVariable;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmUtils;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 *
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public final class Substitutor {
  private static final String RENAMING_SUFFIX_START = "<";
  private static final String RENAMING_SUFFIX_END = ">";

  private static final Pattern NAME_PATTERN = Pattern.compile(".*(?="
      + Substitutor.RENAMING_SUFFIX_START + ".*"
      + Substitutor.RENAMING_SUFFIX_END + ")");
  private static final Pattern SUFFIX_PATTERN = Pattern.compile("(?<="
      + Substitutor.RENAMING_SUFFIX_START + ").*(?="
      + Substitutor.RENAMING_SUFFIX_END + ")");

  private final Efsm efsm;
  private final Map<String, NodeVariable> annotatedVariables;

  /**
   * Analyses cycle number-annotated variables and returns mappings between cycle numbers and
   * corresponding variables' values.
   *
   * @param variables the cycle number-annotated variables
   *
   * @return a map whose key is a cycle number and values is a mappings between variables' names and
   *     values
   */
  public static Map<Integer, Map<String, Data>> deannotateInputs(Iterable<Variable> variables) {
    Map<Integer, Map<String, Data>> mappedVariables = new HashMap<>();

    for (Variable variable: variables) {
      String annotatedName = variable.getName();
      int number = Substitutor.getCycleNumberForAnnotatedInput(annotatedName);

      if (number != -1) {
        String name = Substitutor.deannotateName(annotatedName);
        Integer boxedNumber = number;

        Map<String, Data> cycleVariables = mappedVariables.get(boxedNumber);

        if (cycleVariables == null) {
          cycleVariables = new HashMap<>();
          mappedVariables.put(boxedNumber, cycleVariables);
        }

        cycleVariables.put(name, variable.getData());
      }
    }

    return mappedVariables;
  }

  /**
   * Extracts a cycle number from the specified cycle number-annotated name.
   *
   * @param name the cycle number-annotated name
   *
   * @return the cycle number
   */
  public static int getCycleNumberForAnnotatedInput(String name) {
    Matcher suffixMatcher = Substitutor.SUFFIX_PATTERN.matcher(name);

    if (suffixMatcher.find()) {
      return Integer.valueOf(suffixMatcher.group());
    } else {
      return -1;
    }
  }

  /**
   * Deannotates the specified cycle number-annotated name.
   *
   * @param name the cycle number-annotated name
   *
   * @return a deannotated name
   */
  public static String deannotateName(String name) {
    Matcher nameMatcher = Substitutor.NAME_PATTERN.matcher(name);

    if (nameMatcher.find()) {
      return nameMatcher.group();
    } else {
      return name;
    }
  }

  /**
   * Constructs a substitutor for the specified EFSM.
   *
   * @param efsm the EFSM
   */
  public Substitutor(Efsm efsm) {
    this.efsm = efsm;
    this.annotatedVariables = new HashMap<>();
  }

  /**
   * Backtracks the specified path performing symbolic substitutions and returns an expression with
   * cycle number-annotated variables that must be satisfied in order to traverse the path.
   *
   * @return the expression with that must be satisfied in order to traverse the path
   */
  public Node backtrack(List<EfsmTransition> path) {
    Node substitutedExpression = null;

    for (ListIterator<EfsmTransition> i = path.listIterator(path.size()); i.hasPrevious();) {
      int index = i.previousIndex();
      EfsmTransition transition = i.previous();
      substitutedExpression = this.simulateAction(substitutedExpression, transition, index);
    }

    return substitutedExpression;
  }

  private Node simulateAction(Node targetExpression, EfsmTransition transition,
      int number) {
    Action action = transition.getGuardedAction().getAction();

    if (targetExpression == null) {
      targetExpression = transition.getGuardedAction().getGuard().toNode();
    } else {
      Map<NodeVariable, Node> substitutions = EfsmUtils.getResultingSubstitutions(action);
      List<BoundVariable> bindingList = new ArrayList<>(substitutions.size());

      for (NodeVariable variable: substitutions.keySet()) {
        Node expression = substitutions.get(variable);
        bindingList.add(NodeBinding.bindVariable(variable, expression));
      }

      targetExpression =
          Transformer.substituteAllBindings(new NodeBinding(targetExpression, bindingList));

      targetExpression =
          Node.AND(targetExpression, transition.getGuardedAction().getGuard().toNode());
    }

    if (targetExpression.getKind() == Node.Kind.OPERATION) {
      targetExpression =
          this.annotateInputsWithCycleNumbers((NodeOperation) targetExpression, number);
    } else {
      if (targetExpression.getKind() == Node.Kind.VARIABLE) {
        targetExpression =
            this.annotateInputWithCycleNumber((NodeVariable) targetExpression, number);
      }
    }

    return targetExpression;
  }

  private NodeOperation annotateInputsWithCycleNumbers(NodeOperation expression, int number) {
    NodeOperation annotatedExpression = expression;

    for (NodeVariable variable: ExprUtils.getVariables(expression)) {
      String originalName = variable.getName();
      NodeVariable annotatedVariable = this.annotateInputWithCycleNumber(variable, number);

      if (!annotatedVariable.equals(variable)) {
        annotatedExpression =
            (NodeOperation) Transformer.substitute(annotatedExpression, originalName,
                annotatedVariable);
      }
    }

    return annotatedExpression;
  }

  private NodeVariable annotateInputWithCycleNumber(NodeVariable variable, int number) {
    String originalName = variable.getName();

    if (this.efsm.containsInput(originalName)
        || this.efsm.getParentContainer().containsInput(originalName)) {
      String newName = originalName + Substitutor.RENAMING_SUFFIX_START
          + number + Substitutor.RENAMING_SUFFIX_END;

      if (this.annotatedVariables.containsKey(newName)) {

        return this.annotatedVariables.get(newName);
      } else {
        NodeVariable annotatedVariable =
            new NodeVariable(new Variable(newName, variable.getData()));
        this.annotatedVariables.put(newName, annotatedVariable);

        return annotatedVariable;
      }
    } else {
      return variable;
    }
  }
}
