/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test;

import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

/**
 * The class which performs test minimisation.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public final class TestMinimiser {

  private TestMinimiser() {
  }

  /**
   * Minimises the specified set of sequences by excluding redundant (in terms of coverage)
   * sequences.
   *
   * @param sequencesCoverage a mappings between source sequences
   *                          and sets of transitions covered by them
   */
  public static Test minimise(Map<Sequence, EfsmTransitionCoverage> sequencesCoverage) {
    Test test = new Test();
    Map<Sequence, Set<EfsmTransition>> coveredTransitions = new HashMap<>(sequencesCoverage.size());
    final Set<EfsmTransition> targetTransitions = new HashSet<>();
    Map<EfsmTransition, Sequence> uniqueTransitions = new HashMap<>();

    for (Entry<Sequence, EfsmTransitionCoverage> entry: sequencesCoverage.entrySet()) {
      Sequence sequence = entry.getKey();
      Set<EfsmTransition> transitions = entry.getValue().getCoveredTransitions();
      coveredTransitions.put(sequence, transitions);

      for (EfsmTransition transition: transitions) {
        if (uniqueTransitions.containsKey(transition)) {
          uniqueTransitions.remove(transition);
        } else {
          if (!targetTransitions.contains(transition)) {
            uniqueTransitions.put(transition, sequence);
          }
        }
      }

      targetTransitions.addAll(transitions);
    }

    for (Sequence sequence: new HashSet<>(uniqueTransitions.values())) {
      test.addSequence(sequence);
      targetTransitions.removeAll(coveredTransitions.remove(sequence));
    }

    while (!targetTransitions.isEmpty()) {
      int maxCoverage = Integer.MIN_VALUE;
      Sequence currentSequence = null;

      for (Iterator<Entry<Sequence, Set<EfsmTransition>>> i =
          coveredTransitions.entrySet().iterator(); i.hasNext();) {
        Entry<Sequence, Set<EfsmTransition>> entry = i.next();
        Set<EfsmTransition> effectiveCoverage = new HashSet<>(entry.getValue());
        effectiveCoverage.retainAll(targetTransitions);

        if (effectiveCoverage.isEmpty()) {
          i.remove();
        } else {
          int coverage = effectiveCoverage.size();
          Sequence sequence = entry.getKey();

          if ((coverage > maxCoverage)
              || ((coverage == maxCoverage)
              && (sequence.getLength() < currentSequence.getLength()))) {
            maxCoverage = coverage;
            currentSequence = sequence;
          }

          entry.setValue(effectiveCoverage);
        }
      }

      test.addSequence(currentSequence);
      targetTransitions.removeAll(coveredTransitions.remove(currentSequence));
    }

    return test;
  }
}
