/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */
 
package ru.ispras.retrascope.engine.efsm.generator.test.fate;

import org.apache.commons.cli.ParseException;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.engine.basis.TestGeneratorEngine;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.result.test.TestModel;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.Map;
import java.util.logging.Logger;

/**
 * The test generator for EFSM-based hardware models based on the algorithm described in the article
 * <a href="http://link.springer.com/article/10.1007%2Fs10836-011-5209-8">Efficient
 * Generation of Stimuli for Functional Verification by Backjumping Across Extended FSMs</a>.
 * <br>
 * FATE is the acronym for "functional ATPG (automated test pattern generator) to traverse EFSMs".
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
abstract class AbstractFateEngine extends TestGeneratorEngine {
  public static final Parameter LOOP_LIMIT =
      new Parameter("loop-limit", true, "Loop iteration limit");
  public static final Parameter SEQUENCE_LENGTH_PARAMETER =
      new Parameter("sequence-length", true,
          "Amount of vectors in one randomly generated sequence");
  public static final Parameter SEQUENCES_NUMBER_PARAMETER =
      new Parameter("sequences-number", true,
          "Amount of sequences in a randomly generated fragment of a test");

  private static final Logger LOGGER = Log.getLogger(AbstractFateEngine.class);

  private final String logEntryHeader;

  /**
   * Constructs a generator.
   */
  public AbstractFateEngine(String engineId, String logEntryHeader) {
    super(engineId, EfsmModel.TYPE);
    super.addParameter(AbstractFateEngine.LOOP_LIMIT);
    super.addParameter(AbstractFateEngine.SEQUENCE_LENGTH_PARAMETER);
    super.addParameter(AbstractFateEngine.SEQUENCES_NUMBER_PARAMETER);

    this.logEntryHeader = logEntryHeader;
  }

  /**
   * Generates and returns tests for the specified EFSM modules.
   *
   * @param inputs a map containing the EFSM modules and command line arguments. It must contain a
   *        mapping from the {@link EfsmModel} entity type to an {@link EfsmModel} object.
   *        Optionally, it may include
   *
   * TODO: finish param text
   *
   * @throws IllegalArgumentException if {@code inputs} is {@code null}
   * @throws IllegalArgumentException if {@code inputs} contains no mapping from the
   *         {@link EfsmModel} entity type to an {@link EfsmModel} object
   *
   * @return generated tests
   */
  @Override
  public TestModel start(Map<EntityType, Entity> inputs) {
    EfsmModel efsmModel = Engine.getMandatoryInputEntity(inputs, EfsmModel.class, EfsmModel.TYPE);

    int loopLimit = -1;
    int sequenceLength = -1;
    int sequencesNumber = -1;

    Arguments arguments = Engine.getOptionalInputEntity(inputs, Arguments.class, Arguments.TYPE);

    if (arguments != null) {
      try {
        Arguments parsedArguments = super.parseCommandLine(arguments.getCommandLine());

        loopLimit = Integer.parseInt(parsedArguments.getValue(AbstractFateEngine.LOOP_LIMIT, "-1"));
        sequenceLength = Integer.parseInt(parsedArguments.getValue(AbstractFateEngine
            .SEQUENCE_LENGTH_PARAMETER, "-1"));
        sequencesNumber = Integer.parseInt(parsedArguments.getValue(AbstractFateEngine
            .SEQUENCES_NUMBER_PARAMETER, "-1"));
      } catch (ParseException | NumberFormatException e) {
        throw new IllegalArgumentException("The exception has occurred while parsing command line "
            + "arguments", e);
      }
    }

    TestModel testModel = new TestModel();

    for (EfsmModule module: efsmModel.getModules()) {
      String moduleLogPrefix = this.logEntryHeader + "module " + module.getName() + ": ";

      AbstractFateEngine.LOGGER.log(LogLevel.INFO, moduleLogPrefix + "starting test generation");

      /*AbstractFate generator =
          this.getGenerator(module, loopLimit, sequenceLength, sequencesNumber);
      Test test = generator.generateTest();
      testModel.addTest(test);

      AbstractFateEngine.LOGGER.log(LogLevel.INFO,
          moduleLogPrefix + "test generation has been finished successfully");
      AbstractFateEngine.LOGGER.log(LogLevel.INFO,
          moduleLogPrefix + "test coverage: " + generator.getCoverage());
      AbstractFateEngine.LOGGER.log(LogLevel.INFO,
          moduleLogPrefix + "test length: " + test.getVectorsAmount());*/
    }

    return testModel;
  }

  /*protected abstract AbstractFate getGenerator(EfsmModule module, int loopLimit, int sequenceLength,
      int sequencesNumber);*/
}
