/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test.fate;

import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.retrascope.engine.basis.AtomicTestGenerator;
import ru.ispras.retrascope.engine.efsm.generator.test.heuristic.direction.EfsmTransitionChooser;
import ru.ispras.retrascope.engine.efsm.simulator.ProcessSimulator;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.result.test.Vector;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.Collections;
import java.util.Iterator;
import java.util.logging.Logger;
import java.util.Map;

/**
 * The FATE test generator which deals with single EFSM and operates at the vector-level.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see EfsmFateTestGenerator
 */
abstract class AbstractRandomFateGenerator extends AtomicTestGenerator {
  private static final Logger logger = Log.getLogger(AbstractRandomFateGenerator.class);

  private String logEntryHeader;

  protected final ProcessSimulator simulator;

  private EventList fixedEvents;
  private Map<String, NodeValue> fixedInputs;

  private boolean wasForcedToChangeDirection;

  /**
   * Constructs a test generator for the specified EFSM.
   *
   * TODO: finish Javadoc comment
   *
   * @param simulator
   * @param logEntryHeader
   */
  AbstractRandomFateGenerator(ProcessSimulator simulator, String logEntryHeader) {
    this.simulator = simulator;

    this.logEntryHeader = logEntryHeader;

    this.fixedEvents = new EventList();
    this.fixedInputs = Collections.emptyMap();

    this.wasForcedToChangeDirection = false;
  }

  /**
   * Returns an EFSM transition chooser to be used for traversal direction.
   *
   * @return the EFSM transition chooser
   */
  protected abstract EfsmTransitionChooser getEfsmTransitionChooser();

  @Override
  public void reset() {
    this.fixedEvents = new EventList();
    this.fixedInputs = Collections.emptyMap();
  }

  /**
   * Returns an iterator enabling to retrieve next sequence of vectors produced by this generator.
   *
   * @return the iterator. If this generator is already finished generation, {@code null} is
   *         returned
   */
  @Override
  public Iterator<Vector> getNextSequenceIterator() {
    return new RandomFateSequenceIterator();
  }

  Vector generateInputVectorRandomly() {
    AbstractRandomFateGenerator.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": generating a new input vector");

    Vector vector = null;
    this.wasForcedToChangeDirection = false;

    EfsmTransitionChooser chooser = this.getEfsmTransitionChooser();

    for (EfsmTransition transition = chooser.chooseNext(); transition != null;
        transition = chooser.chooseNext()) {
      vector = this.traverseTransition(transition);

      if (vector == null) {
        this.wasForcedToChangeDirection = true;
      } else {
        return vector;
      }
    }

    return new Vector();
  }

  private Vector traverseTransition(EfsmTransition transition) {
    AbstractRandomFateGenerator.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": traversing the transition: " + transition);

    Vector inputVector = this.simulator.enableGuardedAction(transition.getGuardedAction(),
        this.fixedEvents, this.fixedInputs);

    if (inputVector != null) {
      AbstractRandomFateGenerator.logger.log(LogLevel.DEBUG,
          this.logEntryHeader + ": the transition has been traversed successfully");
      AbstractRandomFateGenerator.logger.log(LogLevel.DEBUG, this.logEntryHeader
          + ": input vector: " + inputVector);
    } else {
      AbstractRandomFateGenerator.logger.log(LogLevel.DEBUG,
          this.logEntryHeader + ": failed to traverse the transition");
    }

    return inputVector;
  }

  /**
   * Checks if during last vector generation this generator was forced to change target transition
   * because of an incompatibility of its GA with a fixed vector.
   *
   * @return {@code true} if this generator was forced to change direction, otherwise returns
   * {@code false}
   */
  boolean wasForcedToChangeDirection() {
    return this.wasForcedToChangeDirection;
  }

  private class RandomFateSequenceIterator implements Iterator<Vector> {
    @Override
    public boolean hasNext() {
      return true;
    }

    @Override
    public Vector next() {
      if (this.hasNext()) {
        return AbstractRandomFateGenerator.this.generateInputVectorRandomly();
      } else {
        return null;
      }
    }

    @Override
    public void remove() {
      throw new UnsupportedOperationException("remove");
    }
  }
}
