/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test.fate;

import edu.uci.ics.jung.graph.DirectedSparseMultigraph;
import edu.uci.ics.jung.graph.util.EdgeType;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmUtils;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.Map.Entry;
import java.util.logging.Logger;

/**
 * The class which performs the learning phase of the FATE algorithm.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
final class FateEfsmAnalyser {
  private final Logger logger;
  private final String logEntryHeader;

  private final DirectedSparseMultigraph<EfsmTransition, VariableDependency> controlDependencyGraph;
  private final DirectedSparseMultigraph<EfsmTransition, VariableDependency> dataDependencyGraph;
  private final Map<NodeVariable, Set<List<EfsmTransition>>> counters;

  FateEfsmAnalyser(Efsm efsm) {
    this.logEntryHeader = "EFSM.FATE.Analyser." + efsm.getName();
    this.logger = Log.getLogger(this.getClass());

    this.controlDependencyGraph = this.detectControlDependencies(efsm);
    this.dataDependencyGraph = this.detectDataDependencies(efsm);
    this.counters = new HashMap<>();
    this.detectCounters(efsm);
  }

  DirectedSparseMultigraph<EfsmTransition, VariableDependency> getControlDependencyGraph() {
    return this.controlDependencyGraph;
  }

  DirectedSparseMultigraph<EfsmTransition, VariableDependency> getDataDependencyGraph() {
    return this.dataDependencyGraph;
  }

  Map<NodeVariable, Set<List<EfsmTransition>>> getCounters() {
    return this.counters;
  }

  DirectedSparseMultigraph<EfsmTransition, VariableDependency>
    detectControlDependencies(Efsm efsm) {

    this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": detecting control dependencies");

    DirectedSparseMultigraph<EfsmTransition, VariableDependency> controlDependencyGraph =
        new DirectedSparseMultigraph<>();

    for (EfsmTransition transition: efsm.getTransitions()) {
      if (!controlDependencyGraph.containsVertex(transition)) {
        controlDependencyGraph.addVertex(transition);
      }

      for (NodeVariable usedVariable: transition.getGuardedAction().getUsesByGuard()) {
        String usedVariableName = usedVariable.getName();

        Object varData = usedVariable.getUserData();

        if (varData instanceof VariableDescriptor) {
          if (((VariableDescriptor)varData).getVariableType() == VariableType.REG) {
            for (EfsmTransition definingTransition:
                FateEfsmAnalyser.findVariableDefinition(efsm, transition, usedVariableName)) {
              if (!controlDependencyGraph.containsVertex(definingTransition)) {
                controlDependencyGraph.addVertex(definingTransition);
              }

              VariableDependency edge =
                  new VariableDependency(transition, definingTransition, usedVariable);
              controlDependencyGraph.addEdge(edge, definingTransition, transition,
                  EdgeType.DIRECTED);
            }
          }
        } else {
          throw new IllegalArgumentException(
            String.format("Incorrect data type: %s.", varData.getClass()));
        }
      }
    }

    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": control dependencies have been detected successfully");

    return controlDependencyGraph;
  }
 
  DirectedSparseMultigraph<EfsmTransition, VariableDependency> detectDataDependencies(Efsm efsm) {
    this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": detecting data dependencies");

    DirectedSparseMultigraph<EfsmTransition, VariableDependency> dataDependencyGraph =
        new DirectedSparseMultigraph<>();

    for (EfsmTransition transition : efsm.getTransitions()) {
      if (!dataDependencyGraph.containsVertex(transition)) {
        dataDependencyGraph.addVertex(transition);
      }

      for (NodeVariable usedVariable: transition.getGuardedAction().getUsesByAction()) {
        String usedVariableName = usedVariable.getName();

        Object varData = usedVariable.getUserData();

        if (varData instanceof VariableDescriptor) {
          if ((((VariableDescriptor)varData).getVariableType() == VariableType.REG)) {
            for (EfsmTransition definingTransition:
                FateEfsmAnalyser.findVariableDefinition(efsm, transition, usedVariableName)) {
              if (!dataDependencyGraph.containsVertex(definingTransition)) {
                dataDependencyGraph.addVertex(definingTransition);
              }
 
              VariableDependency edge =
                  new VariableDependency(transition, definingTransition, usedVariable);
              dataDependencyGraph.addEdge(edge, definingTransition, transition, EdgeType.DIRECTED);
            }
          }
        } else {
          throw new IllegalArgumentException(
            String.format("Incorrect data type: %s.", varData.getClass()));
        }
      }
    }

    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": data dependencies have been detected successfully");

    return dataDependencyGraph;
  }

  private static List<EfsmTransition> findVariableDefinition(Efsm efsm, EfsmTransition transition,
      String usedVariableName) {
    Set<EfsmTransition> visitedTransitions = new HashSet<>();
    Queue<EfsmTransition> queue = new LinkedList<>();
    queue.add(transition);
    List<EfsmTransition> definingTransitions = new ArrayList<>();

    while (!queue.isEmpty()) {
      EfsmTransition currentTransition = queue.poll();

      for (EfsmTransition previousTransition :
          efsm.getIncomingTransitions(currentTransition.getSourceState())) {

        if (!visitedTransitions.contains(previousTransition)) {
          boolean useDefMatching = false;

          for (NodeVariable definedVariable : previousTransition.getGuardedAction().getDefines()) {
            if (definedVariable.getName().equals(usedVariableName)) {
              useDefMatching = true;

              break;
            }
          }

          if (useDefMatching) {
            definingTransitions.add(previousTransition);
          } else {
            queue.add(previousTransition);
            visitedTransitions.add(previousTransition);
          }
        }
      }
    }

    return definingTransitions;
  }

  private void detectCounters(Efsm efsm) {
    this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": detecting counters");

    Map<NodeVariable, Set<EfsmTransition>> processedCandidates = new HashMap<>();

    for (VariableDependency dependency: this.controlDependencyGraph.getEdges()) {
      EfsmTransition definingTransition = this.controlDependencyGraph.getSource(dependency);

      if (processedCandidates.containsKey(dependency.getVariable())) {
        Set<EfsmTransition> processedTransitions =
            processedCandidates.get(dependency.getVariable());

        if (processedTransitions.contains(definingTransition)) {
          continue;
        } else {
          processedTransitions.add(definingTransition);
        }
      } else {
        Set<EfsmTransition> transitions = new HashSet<>();
        transitions.add(definingTransition);
        processedCandidates.put(dependency.getVariable(), transitions);
      }

      EfsmState sourceState = definingTransition.getSourceState();
      EfsmState destinationState =
          this.controlDependencyGraph.getDest(dependency).getDestinationState();
      Set<NodeVariable> definingVariables = EfsmUtils
          .getResultingDependencies(definingTransition.getGuardedAction().getAction())
              .get(dependency.getVariable());

      for (Iterator<NodeVariable> iterator = definingVariables.iterator(); iterator.hasNext();) {
        String variableName = iterator.next().getName();
        if (efsm.containsInput(variableName)
            || efsm.getParentContainer().containsInput(variableName)) {
          iterator.remove();
        }
      }

      if (!definingVariables.isEmpty()) {
        if (efsm.isReachable(destinationState, sourceState)) {
          Set<NodeVariable> variables = new HashSet<>(1);
          variables.add(dependency.getVariable());

          for (VariableDependency dataDependency:
              this.dataDependencyGraph.getOutEdges(definingTransition)) {
            if (dataDependency.getVariable().equals(dependency.getVariable())) {
              EfsmTransition transition = this.dataDependencyGraph.getDest(dataDependency);
              this.findCounterLoop(dataDependency.getVariable(), transition,
                  definingVariables, variables, Collections.singletonList(definingTransition));
            }
          }
        }
      }
    }

    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": counters have been detected successfully");
  }

  //target transition is the first in the returned list
  private void findCounterLoop(NodeVariable potentialCounter, EfsmTransition currentTransition,
      Set<NodeVariable> targetVariables, Set<NodeVariable> currentVariables,
      List<EfsmTransition> currentPath) {
    if (currentPath.contains(currentTransition)) {
      if (currentTransition.equals(currentPath.get(0))) {
        Set<NodeVariable> intersection = new HashSet<>(targetVariables);
        intersection.retainAll(currentVariables);

        if (!intersection.isEmpty()) {
          Set<List<EfsmTransition>> counterData =
              this.counters.get(potentialCounter);

          if (counterData == null) {
            counterData = new HashSet<>(1);
            this.counters.put(potentialCounter, counterData);
          }

          counterData.add(currentPath);

          this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": the variable "
              + potentialCounter.getName() + " defined by the transition " + currentPath.get(0)
              + " is a counter in the loop " + currentPath);
        }
      }
    } else {
      Set<NodeVariable> newCurrentVariables = new HashSet<>();

      for (Entry<NodeVariable, Set<NodeVariable>> dependency:
          EfsmUtils.getResultingDependencies(currentTransition.getGuardedAction().getAction())
              .entrySet()) {
        NodeVariable variable = dependency.getKey();
        Set<NodeVariable> intersection = new HashSet<>(currentVariables);
        intersection.retainAll(dependency.getValue());

        if (!intersection.isEmpty()) {
          newCurrentVariables.add(variable);
        }
      }

      if (!newCurrentVariables.isEmpty()) {
        List<EfsmTransition> newCurrentPath = new ArrayList<>(currentPath);
        newCurrentPath.add(currentTransition);

        for (VariableDependency dataDependency:
            this.dataDependencyGraph.getOutEdges(currentTransition)) {
          if (newCurrentVariables.contains(dataDependency.getVariable())) {
            EfsmTransition transition = this.dataDependencyGraph.getDest(dataDependency);
            this.findCounterLoop(potentialCounter, transition, targetVariables, newCurrentVariables,
                newCurrentPath);
          }
        }
      }
    }
  }
}
