/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.simulator;

import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.basis.Simulator;
import ru.ispras.retrascope.engine.efsm.simulator.ProcessSimulator.ProcessSimulationSnapshot;
import ru.ispras.retrascope.model.basis.State;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.model.efsm.EfsmModuleState;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.Vector;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;
import java.util.logging.Logger;

/**
 * A simulator for {@code EfsmModule}.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see EfsmModule
 */
public class ModuleSimulator extends Simulator<EfsmModule> {
  private static final Logger LOGGER = Log.getLogger(ModuleSimulator.class);

  private final Set<ProcessSimulator> efsmSimulators;
  private final Map<String, Map<ProcessSimulator, NodeValue>> currentVariableModifications;

  // TODO: field can be converted to local variable
  private final String name;
  private final String logEntryHeader;

  private final Random randomNumberGenerator;

  /**
   * Constructs a simulator for the specified EFSM module.
   *
   * @param efsmModule the EFSM module which the simulator is constructed for
   *
   * @throws IllegalArgumentException if {@code efsmModule} is {@code null}
   */
  public ModuleSimulator(EfsmModule efsmModule) {
    super(efsmModule);

    this.efsmSimulators = new HashSet<>(this.simulatedModel.getNonInitialEfsms().size());

    this.name = this.simulatedModel.getName();
    this.logEntryHeader = "EFSM.Simulator." + this.name + ": ";

    for (Efsm efsm: this.simulatedModel.getNonInitialEfsms()) {
      this.efsmSimulators.add(new ProcessSimulator(this, efsm));
    }

    Set<String> registerNames = this.simulatedModel.getRegisterNames();
    Set<String> outputNames = this.simulatedModel.getOutputNames();

    this.currentVariableModifications = new HashMap<>(registerNames.size() + outputNames.size());

    for (String variable: registerNames) {
      this.currentVariableModifications.put(variable, new HashMap<ProcessSimulator, NodeValue>());
    }

    for (String variable: outputNames) {
      this.currentVariableModifications.put(variable, new HashMap<ProcessSimulator, NodeValue>());
    }

    this.reset();

    this.randomNumberGenerator = new Random(0L);
  }

  @Override
  public EfsmModule getSimulatedModel() {
    return this.simulatedModel;
  }

  @Override
  public Vector simulateVector(Vector inputVector) {
    super.readVector(inputVector);

    ModuleSimulator.LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + "processing vector: " + inputVector);

    // TODO: 'subvectors' collection is updated but never queried
    Map<ProcessSimulator, Vector> subvectors = new HashMap<>(this.efsmSimulators.size());
    Transaction outputTransation = new Transaction();

    for (ProcessSimulator efsmSimulator: this.efsmSimulators) {
      subvectors.put(efsmSimulator, efsmSimulator.simulateVector(inputVector));
    }

    for (Entry<String, NodeValue> modification: this.resolveAssignmentConflicts().entrySet()) {
      String variableName = modification.getKey();
      NodeValue variableValue = modification.getValue();

      if (this.innerVariablesMemory.containsVariable(variableName)) {
        this.innerVariablesMemory.setValue(variableName, variableValue);
      } else {
        if (this.outputsMemory.containsVariable(variableName)) {
          this.outputsMemory.setValue(variableName, variableValue);
          outputTransation.addVariable(variableName, variableValue);
        } else {
          throw new RetrascopeRuntimeException("The variable " + variableName
              + " isn't found neither among inner variables nor among outputs");
        }
      }
    }

    this.checkInnerConsistency();
    Vector outputVector = new Vector(outputTransation);

    ModuleSimulator.LOGGER.log(LogLevel.DEBUG, this.logEntryHeader
        + "vector has been processed successfully. Output vector: " + outputVector);

    return outputVector;
  }

  @Override
  public Sequence simulateSequence(Sequence inputSequence) {
    InvariantChecks.checkNotNull(inputSequence);

    ModuleSimulator.LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + "processing sequence: " + inputSequence);

    Sequence outputSequence = new Sequence(inputSequence.getLength());

    for (Vector inputVector: inputSequence.getVectors()) {
      outputSequence.addVector(this.simulateVector(inputVector));
    }

    ModuleSimulator.LOGGER.log(LogLevel.DEBUG, this.logEntryHeader
        + "sequence has been processed successfully. Output sequence: " + outputSequence);

    return outputSequence;
  }

  @Override
  public void reset() {
    ModuleSimulator.LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + "resetting");

    super.reset();

    for (ProcessSimulator efsmSimulator: this.efsmSimulators) {
      efsmSimulator.reset();
    }

    ModuleSimulator.LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + "reset has been finished successfully");
  }

  @Override
  public EfsmTransitionCoverage getCoverage() {
    EfsmTransitionCoverage coverage = new EfsmTransitionCoverage();

    for (ProcessSimulator efsmSimulator: this.efsmSimulators) {
      coverage.include(efsmSimulator.getCoverage());
    }

    return coverage;
  }

  /**
   * Returns a coverage of the specified EFSM.
   *
   * @param efsm the target EFSM
   *
   * @return the coverage of the specified EFSM. If there is no this EFSM in a simulated module,
   * {@code null} is returned
   */
  public EfsmTransitionCoverage getCoverage(Efsm efsm) {
    for (ProcessSimulator efsmSimulator: this.efsmSimulators) {
      if (efsmSimulator.getSimulatedModel().equals(efsm)) {
        return efsmSimulator.getCoverage();
      }
    }

    return null;
  }

  /**
   * Returns a simulator for the specified EFSM. The returned simulator represents a current state
   * of EFSM simulation.
   *
   * @param efsm the target EFSM
   *
   * @return the simulator for the specified EFSM. If there is no this EFSM in a simulated module,
   * {@code null} is returned
   */
  public ProcessSimulator getSubsimulator(Efsm efsm) {
    for (ProcessSimulator efsmSimulator: this.efsmSimulators) {
      if (efsmSimulator.getSimulatedModel().equals(efsm)) {
        return efsmSimulator;
      }
    }

    return null;
  }

  /**
   * Returns all process simulators of this simulator. The returned simulators represent a current
   * state of simulation of the processes.
   *
   * @return the simulator for the specified EFSM. If there is no this EFSM in a simulated module,
   * {@code null} is returned
   */
  public Set<ProcessSimulator> getSubsimulators() {
    return Collections.unmodifiableSet(this.efsmSimulators);
  }

  /**
   * Submits a variable modification request to this simulator. If there is no inner/output variable
   * with the specified name in a simulated module, this method has no any effects.
   * <br>
   * The request is valid only for the current simulation cycle. At the end of the cycle the
   * simulator performs all requested modifications.
   *
   * @param simulator a process simulator which caused variable modification
   * @param variableName the name of a variable to be modified
   * @param value a new requested value of the variable
   *
   * @throws IllegalArgumentException if {@code efsmModule} is {@code null}
   */
  void modifyVariable(ProcessSimulator simulator, String variableName, NodeValue value) {
    Map<ProcessSimulator, NodeValue> modifications =
        this.currentVariableModifications.get(variableName);

    if (modifications == null) {
      ModuleSimulator.LOGGER.log(LogLevel.WARNING, this.logEntryHeader
          + "a request to modify unknown variable " + variableName + " has been received");
    } else {
      modifications.put(simulator, value);
    }
  }

  private Map<String, NodeValue> resolveAssignmentConflicts() {
    Map<String, NodeValue> effectiveModifications =
        new HashMap<>(this.currentVariableModifications.size());

    for (Entry<String, Map<ProcessSimulator, NodeValue>> modifications:
        this.currentVariableModifications.entrySet()) {
      String variableName = modifications.getKey();
      Map<ProcessSimulator, NodeValue> assignedValues = modifications.getValue();

      switch (assignedValues.size()) {
        case 0: {
          break;
        }

        case 1: {
          effectiveModifications.put(variableName, assignedValues.values().iterator().next());

          break;
        }

        default: {
          ModuleSimulator.LOGGER.log(LogLevel.WARNING, this.logEntryHeader
              + "there are multiple simultaneous assignments to the variable " + variableName
              + ": " + assignedValues);

          int chosenAssignment = this.randomNumberGenerator.nextInt(assignedValues.size());
          NodeValue chosenValue = new ArrayList<>(assignedValues.values()).get(chosenAssignment);

          ModuleSimulator.LOGGER.log(LogLevel.DEBUG,
              this.logEntryHeader + "randomly chosen assigned value: " + chosenValue);

          effectiveModifications.put(variableName, chosenValue);
        }
      }
    }

    return effectiveModifications;
  }

  @Override
  public EfsmModuleState getCurrentState() {
    Collection<State> substates = new ArrayList<>();

    for (ProcessSimulator subsimulator: this.efsmSimulators) {
      substates.add(subsimulator.getCurrentState());
    }

    return new EfsmModuleState(substates);
  }

  @Override
  public void resetCoverage() {
    for (ProcessSimulator subsimulator: this.efsmSimulators) {
      subsimulator.resetCoverage();
    }
  }

  @Override
  protected void checkInnerConsistency() {
    super.checkInnerConsistency();

    for (ProcessSimulator subsimulator: this.efsmSimulators) {
      subsimulator.checkInnerConsistency();
    }
  }

  public class ModuleSimulationSnapshot extends SimulationSnapshot {
    private final Map<ProcessSimulator, ProcessSimulationSnapshot> snapshots;

    ModuleSimulationSnapshot() {
      super();

      this.snapshots = new HashMap<>(ModuleSimulator.this.efsmSimulators.size());

      for (ProcessSimulator subsimulator: ModuleSimulator.this.efsmSimulators) {
        this.snapshots.put(subsimulator, subsimulator.makeSimulationSnapshot());
      }
    }
  }

  @Override
  public ModuleSimulationSnapshot makeSimulationSnapshot() {
    return new ModuleSimulationSnapshot();
  }

  /**
   * TODO: add javadoc here
   *
   * @param snapshot
   */
  public void loadSnapshot(ModuleSimulationSnapshot snapshot) {
    ProcessSimulator.LOGGER.log(LogLevel.DEBUG, this.logEntryHeader
        + ": loading a simulation snapshot");

    super.loadSnapshot(snapshot);

    for (ProcessSimulator subsimulator: ModuleSimulator.this.efsmSimulators) {
      subsimulator.loadSnapshot(snapshot.snapshots.get(subsimulator));
    }
  }
}
