/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.simulator;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeBinding.BoundVariable;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.SolverResult.Status;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.basis.Simulator;
import ru.ispras.retrascope.engine.efsm.generator.test.Substitutor;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.Vector;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.util.LogLevel;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;

/**
 * The EFSM-based representation of a hardware model which encapsulates a runtime state along with a
 * static structure of an original device.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public final class ProcessSimulator extends Simulator<Efsm> {
  private final String logEntryHeader;

  private final ModuleSimulator parentModuleSimulator;

  private final String name;

  private EfsmState currentState;

  private final EfsmTransitionCoverage coverage;

  private EfsmTransition lastTraversedTransition;

  /**
   * Constructs a simulator above the specified EFSM.
   * <br>
   * Immediately after construction the simulator is in the uninitialised state. To drive it to the
   * initial state of the EFSM it is required to feed such inputs which enable so-called reset
   * guarded action of the EFSM.
   *
   * @param parentModuleSimulator a simulator of a module which the specified EFSM belongs to
   * @param efsm an EFSM which the simulator is based on
   *
   * @throws IllegalArgumentException if at least one argument is {@code null}
   *
   * @see Efsm#getInitialState()
   * @see Efsm#getResetGuardedActions()
   */
  public ProcessSimulator(ModuleSimulator parentModuleSimulator, Efsm efsm) {
    super(efsm);
    this.parentModuleSimulator = parentModuleSimulator;
    InvariantChecks.checkNotNull(parentModuleSimulator);

    this.logEntryHeader = "EFSM.Simulator." + this.simulatedModel.getName() + ": ";

    this.name = this.simulatedModel.getName();

    this.coverage = new EfsmTransitionCoverage(this.simulatedModel);

    this.reset();
  }

  /**
   * Returns a name of this simulator.
   *
   * @return the name
   */
  public String getName() {
    return this.name;
  }

  /**
   * Returns a state of an underlying EFSM which this simulator is currently in.
   *
   * @return the current state of this simulator. Returns the same state as
   *         {@link Efsm#getInitialState()} do if called immediately
   *         after construction/reset of this simulator
   *
   * @see ProcessSimulator#reset()
   */
  @Override
  public EfsmState getCurrentState() {
    return this.currentState;
  }

  /**
   * Returns a transition which was traversed by this simulator on last cycle.
   *
   * @return the last traversed transition
   */
  public EfsmTransition getLastTraversedTransition() {
    return this.lastTraversedTransition;
  }

  /**
   * Returns a test coverage reached during simulation since last reset.
   *
   * @return the test coverage
   */
  @Override
  public EfsmTransitionCoverage getCoverage() {
    return new EfsmTransitionCoverage(this.coverage);
  }

  /**
   * Resets this simulator, that is, returns it to an uninitialised state.
   */
  @Override
  public void reset() {
    Simulator.LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + "resetting");

    super.reset();

    this.coverage.reset();

    this.currentState = this.simulatedModel.getInitialState();

    if (this.currentState.equals(Efsm.UNINIT_STATE)) {
      //TODO
      throw new RetrascopeRuntimeException("Efsm.UNINITIALISED_STATE isn't supported yet");
    }

    this.lastTraversedTransition = null;

    super.checkInnerConsistency();
  }

  private void traverse(EfsmTransition transition) {
    Simulator.LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": traversing the transition " + transition);

    this.executeAction(transition.getGuardedAction().getAction());
    this.currentState = transition.getDestinationState();
    this.coverage.markTransitionAsCovered(transition);
    this.lastTraversedTransition = transition;

    Simulator.LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": traversing has been finished successfully");
  }

  private List<EfsmTransition> getTraversableTransitions(EventList events) {
    Simulator.LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": performing transitions traversability test");

    List<EfsmTransition> possibleTransitions =
        new ArrayList<>(this.simulatedModel.getOutgoingTransitions(this.currentState));

    for (Iterator<EfsmTransition> possibleTransitionsIterator = possibleTransitions.iterator();
        possibleTransitionsIterator.hasNext();) {
      EfsmTransition testedTransition = possibleTransitionsIterator.next();

      Simulator.LOGGER.log(LogLevel.DEBUG,
          this.logEntryHeader + ": testing the transition " + testedTransition);

      GuardedAction testedGuardedAction = testedTransition.getGuardedAction();

      if (!testedGuardedAction.isSensible(events)) {
        Simulator.LOGGER.log(LogLevel.DEBUG,
            this.logEntryHeader + ": the transition isn't sensible to the occurred events");

        possibleTransitionsIterator.remove();
      } else {
        final Node testedConstraint = this.guardToConstraint(testedGuardedAction.getGuard());
        if (SolverUtils.isSat(testedConstraint)) {
          Simulator.LOGGER.log(LogLevel.DEBUG,
              this.logEntryHeader + ": the transition is traversable");
        } else {
          Simulator.LOGGER.log(LogLevel.DEBUG,
              this.logEntryHeader + ": the transition isn't traversable now");

          possibleTransitionsIterator.remove();
        }
      }
    }

    return possibleTransitions;
  }

  private EfsmTransition chooseTransitionToTraverse(List<EfsmTransition> traversableTransitions) {
    EfsmTransition traversedTransition;

    if (traversableTransitions.size() == 1) {
      traversedTransition = traversableTransitions.get(0);
    } else {
      Simulator.LOGGER.log(LogLevel.WARNING, this.logEntryHeader
          + ": there are multiple traversable transitions now. Choosing one of them at "
          + "random");
      for (EfsmTransition transition: traversableTransitions) {
        Simulator.LOGGER.log(LogLevel.WARNING,
            this.logEntryHeader + ": possible transition: " + transition);
      }

      Random randomGenerator = new Random(traversableTransitions.size());

      traversedTransition =
          traversableTransitions.get(randomGenerator.nextInt(traversableTransitions.size()));
    }

    return traversedTransition;
  }

  private Node guardToConstraint(Guard guard) {
    return this.substituteVariables(guard.toNode(), Collections.<String, NodeValue>emptyMap());
  }

  @Override
  public void checkInnerConsistency() {
    super.checkInnerConsistency();

    Node consistencyConstraint = this.substituteVariables(this.currentState.getExpression(),
        Collections.<String, NodeValue>emptyMap());

    if (!SolverUtils.isSat(consistencyConstraint)) {
      throw new RetrascopeRuntimeException(
          String.format(
              "State consistency constraint violation: %s; original was %s",
              consistencyConstraint, this.currentState.getExpression()));
    }
  }

  private void executeAction(Action action) {
    InvariantChecks.checkNotNull(action);

    Simulator.LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": executing the action: " + action);

    List<Assignment> assignments = action.getAssignments();
    Map<String, NodeValue> bufferedVariables = new HashMap<>(assignments.size());

    for (Assignment assignment: assignments) {
      bufferedVariables = this.executeAssignment(assignment, bufferedVariables);
    }

    for (Entry<String, NodeValue> bufferedVariable: bufferedVariables.entrySet()) {
      String variableName = bufferedVariable.getKey();
      NodeValue variableValue = bufferedVariable.getValue();

      if (this.simulatedModel.getRegisterNames().contains(variableName)) {
        this.innerVariablesMemory.setValue(variableName, variableValue);
      } else {
        if (this.simulatedModel.getOutputNames().contains(variableName)) {
          this.outputsMemory.setValue(variableName, variableValue);
        } else {
          this.parentModuleSimulator.modifyVariable(this, variableName, variableValue);
        }
      }
    }

    Simulator.LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": the action has been executed successfully: " + action);
  }

  private Map<String, NodeValue> executeAssignment(Assignment assignment,
      Map<String, NodeValue> bufferedVariables) {
    Simulator.LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": executing the assignment: " + assignment);

    RangedVariable rangedVariable = assignment.getTarget();
    Node expression = assignment.getValue();

    if ((rangedVariable == null) && (expression == null)) {
      return Collections.emptyMap();
    }

    InvariantChecks.checkNotNull(rangedVariable);
    InvariantChecks.checkNotNull(expression);

    NodeVariable nodeVariable = rangedVariable.getVariable();
    DataType dataType = nodeVariable.getDataType();

    if (rangedVariable.isRanged()) {
      if (!dataType.getTypeId().equals(DataTypeId.BIT_VECTOR)
          && !dataType.getTypeId().equals(DataTypeId.MAP)) {
        throw new IllegalArgumentException("Unsupported data type of ranged variable "
            + rangedVariable.getVariableName() +  ": " + dataType);
      }

      Range range = rangedVariable.getRange();

      Node leftBoundExpression = this.substituteVariables(range.getYoung(), bufferedVariables);
      Node rightBoundExpression = this.substituteVariables(range.getOld(), bufferedVariables);

      NodeValue leftBound = this.computeExpression(leftBoundExpression, DataType.INTEGER);
      NodeValue rightBound = this.computeExpression(rightBoundExpression, DataType.INTEGER);

      int leftBoundInteger = leftBound.getInteger().intValue();
      int rightBoundInteger = rightBound.getInteger().intValue();

      int variableSize = dataType.getSize();

      if (leftBoundInteger < rightBoundInteger) {
        throw new IllegalArgumentException("This type of subvectors is currently unsupported");
      }

      NodeOperation mostSignificantPart = null;
      NodeOperation lessSignificantPart = null;

      if (leftBoundInteger < variableSize - 1) {
        NodeValue leftIndex =
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(variableSize - 1), 10));
        NodeValue rightIndex =
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(leftBoundInteger + 1), 10));

        mostSignificantPart = new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex,
            nodeVariable);
      }

      if (rightBoundInteger > 0) {
        NodeValue leftIndex =
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(rightBoundInteger - 1), 10));
        NodeValue rightIndex = new NodeValue(DataType.INTEGER.valueOf("0", 10));

        lessSignificantPart = new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex,
            nodeVariable);
      }

      if ((mostSignificantPart != null) && (lessSignificantPart != null)) {
        expression = new NodeOperation(StandardOperation.BVCONCAT, mostSignificantPart, expression,
            lessSignificantPart);
      } else {
        if (mostSignificantPart != null) {
          expression =
              new NodeOperation(StandardOperation.BVCONCAT, mostSignificantPart, expression);
        } else {
          if (lessSignificantPart != null) {
            expression =
                new NodeOperation(StandardOperation.BVCONCAT, expression, lessSignificantPart);
          }
        }
      }
    }

    expression = this.substituteVariables(expression, bufferedVariables);
    NodeValue assignedValue = this.computeExpression(expression, dataType);
    bufferedVariables.put(rangedVariable.getVariableName(), assignedValue);

    Simulator.LOGGER.log(LogLevel.DEBUG, this.logEntryHeader
        + ": the assigned value: " + assignedValue);

    return bufferedVariables;
  }

  private NodeValue computeExpression(Node expression, DataType resultDataType) {
    if (expression instanceof NodeValue) {
      NodeValue value = (NodeValue) expression;

      if (!value.getDataType().equals(resultDataType)) {
        throw new IllegalArgumentException("Incompatible data types: " + value.getDataType()
            + " vs " + resultDataType);
      }

      return value;
    } else {
      if (!ExprUtils.getVariables(expression).isEmpty()) {
        throw new IllegalArgumentException("The expression to be computed (" + expression
            + ") contains unevaluated variables: " + ExprUtils.getVariables(expression));
      }

      Simulator.LOGGER.log(LogLevel.DEBUG, this.logEntryHeader
          + ": computing the expression " + expression);

      expression =
          NodeUtils.getEquation(new NodeVariable("assignmentResult", resultDataType), expression);

      SolverResult result = SolverUtils.solve(expression);

      if (result.getStatus().equals(Status.SAT)) {
        NodeValue resultValue = new NodeValue(result.getVariables().iterator().next().getData());

        Simulator.LOGGER.log(LogLevel.DEBUG,
            this.logEntryHeader + ": the computation result: " + resultValue);

        return resultValue;
      } else {
        throw new RetrascopeRuntimeException(
            "Failed to resolve the assignment constraint due to errors: " + result.getErrors());
      }
    }
  }

  private void logInputDump() {
    if (Simulator.LOGGER.isLoggable(LogLevel.DEBUG)) {
      Simulator.LOGGER.log(LogLevel.DEBUG,
          this.logEntryHeader + "input dump: " + this.inputsMemory);
    }
  }

  private void logInnerDump() {
    if (Simulator.LOGGER.isLoggable(LogLevel.DEBUG)) {
      Simulator.LOGGER.log(LogLevel.DEBUG,
          this.logEntryHeader + "current state: " + this.currentState);
      Simulator.LOGGER.log(LogLevel.DEBUG,
          this.logEntryHeader + "inner variables' dump: " + this.innerVariablesMemory);
      Simulator.LOGGER.log(LogLevel.DEBUG,
          this.logEntryHeader + "output dump: " + this.outputsMemory);
      Simulator.LOGGER.log(LogLevel.DEBUG,
          this.logEntryHeader + "current coverage: " + this.coverage);
    }
  }

  @Override
  public Vector simulateVector(Vector inputVector) {
    Simulator.LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + "processing vector: " + inputVector);
    super.readVector(inputVector);

    List<EfsmTransition> possibleTransitions =
        this.getTraversableTransitions(inputVector.getEventList());

    if (possibleTransitions.isEmpty()) {
      Simulator.LOGGER.log(LogLevel.WARNING, this.logEntryHeader
          + "there are no traversable transitions now. Finishing event handling");

      this.lastTraversedTransition = null;
    } else {
      this.traverse(this.chooseTransitionToTraverse(possibleTransitions));
    }

    this.logInnerDump();

    Simulator.LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + "model has finished to process the events");

    return super.getOutputVector();
  }

  @Override
  public Sequence simulateSequence(Sequence inputSequence) {
    throw new UnsupportedOperationException("simulateSequence(Sequence) isn't implemented yet");
  }

  @Override
  public Efsm getSimulatedModel() {
    return this.simulatedModel;
  }

  @Override
  public Set<String> getInnerVariableNames() {
    Set<String> variableNames = new HashSet<>(this.innerVariablesMemory.getValues().keySet());
    variableNames.addAll(this.parentModuleSimulator.getInnerVariableNames());

    return variableNames;
  }

  @Override
  public Set<String> getOutputVariableNames() {
    Set<String> variableNames = new HashSet<>(this.outputsMemory.getValues().keySet());
    variableNames.addAll(this.parentModuleSimulator.getOutputVariableNames());

    return variableNames;
  }

  @Override
  public boolean hasInput(String variableName) {
    return this.simulatedModel.isInputDeclared(variableName);
  }

  @Override
  public boolean hasInnerVariable(String variableName) {
    return this.simulatedModel.isRegisterDeclared(variableName);
  }

  @Override
  public boolean hasOutput(String variableName) {
    return this.simulatedModel.isOutputDeclared(variableName);
  }

  @Override
  public NodeValue getVariableValue(String variableName) {
    NodeValue value = super.getVariableValue(variableName);

    if (value == null) {
      value = this.parentModuleSimulator.getVariableValue(variableName);
    }

    return value;
  }

  /**
   * Generates an input vector to enable the specified guarded action considering the specified
   * fixed events and inputs.
   *
   * @param guardedAction the guarded action to be enabled
   * @param fixedEvents the fixed events
   * @param fixedInputs the fixed inputs
   *
   * @return the generated vector; if one failed to enable the guarded action, {@code null} is
   *     returned
   */
  public Vector enableGuardedAction(GuardedAction guardedAction, EventList fixedEvents,
      Map<String, NodeValue> fixedInputs) {
    if (!fixedEvents.isEmpty()) {
      EventList gaEventList = new EventList();
      gaEventList.addEvents(guardedAction.getClocks());

      if (!gaEventList.compatibleWith(fixedEvents)) {
        return null;
      }
    }

    Node guardExpression = this.fillExpression(guardedAction.getGuard().toNode(), fixedInputs);

    SolverResult solverResult = SolverUtils.solve(guardExpression);

    if (solverResult.getStatus().equals(Status.SAT)) {
      Vector inputVector = new Vector();
      Transaction transaction = new Transaction();

      for (Variable variable : solverResult.getVariables()) {
        String variableName = variable.getName();

        if (this.hasInput(variableName)) {
          transaction.addVariable(variableName, new NodeValue(variable.getData()));
        }
      }

      inputVector.addEvents(guardedAction.getClocks());

      if (!transaction.isEmpty()) {
        inputVector.addTransaction(transaction);
      }

      inputVector.addEvents(fixedEvents.getEvents());

      if (!fixedInputs.isEmpty()) {
        inputVector.addTransaction(new Transaction(fixedInputs));
      }

      return inputVector;
    } else {
      return null;
    }
  }

  /**
   * Fills the specified expression with the specified inputs and current values of inner variables.
   *
   * @param expression the expression to be filled
   * @param fixedInputs the fixed inputs
   *
   * @return an expression where variables replaced by their values
   */
  public Node fillExpression(Node expression, Map<String, NodeValue> fixedInputs) {
    Set<NodeVariable> usedVariables = new HashSet<>(ExprUtils.getVariables(expression));
    List<BoundVariable> bindingList = new ArrayList<>(usedVariables.size());
    Set<Node> inputConstraints = new HashSet<>();

    for (NodeVariable usedVariable: usedVariables) {
      String variableName = Substitutor.deannotateName(usedVariable.getName());

      if (this.hasInput(variableName)) {
        Node variableConstraint = this.simulatedModel.getInvariant(variableName);

        if (variableConstraint != null) {
          inputConstraints.add(Transformer.substitute(variableConstraint, variableName,
              usedVariable));
        }

        if (fixedInputs.containsKey(variableName)) {
          bindingList.add(NodeBinding.bindVariable(usedVariable,
              fixedInputs.get(variableName)));
        }
      } else {
        if (this.hasInnerVariable(variableName)) {
          bindingList.add(NodeBinding.bindVariable(usedVariable,
              this.getVariableValue(variableName)));
        }
      }
    }

    expression =
        Transformer.substituteAllBindings(new NodeBinding(expression, bindingList));

    if (!inputConstraints.isEmpty()) {
      Node inputConstraint =
          ExprUtils.getConjunction(inputConstraints.toArray(new Node[inputConstraints.size()]));
      expression = Node.AND(expression, inputConstraint);
    }

    return expression;
  }

  @Override
  public void resetCoverage() {
    this.coverage.reset();
  }

  /**
   * TODO: add Javadoc here
   */
  public class ProcessSimulationSnapshot extends SimulationSnapshot {
    private final EfsmState currentState;
    private final EfsmTransition lastTraversedTransition;
    private final EfsmTransitionCoverage coverage;

    ProcessSimulationSnapshot() {
      super();

      this.currentState = ProcessSimulator.this.currentState;
      this.lastTraversedTransition = ProcessSimulator.this.lastTraversedTransition;
      this.coverage = new EfsmTransitionCoverage(ProcessSimulator.this.coverage);
    }
  }

  @Override
  public ProcessSimulationSnapshot makeSimulationSnapshot() {
    return new ProcessSimulationSnapshot();
  }

  /**
   * TODO: add Javadoc here
   *
   * @param snapshot
   */
  public void loadSnapshot(ProcessSimulationSnapshot snapshot) {
    ProcessSimulator.LOGGER.log(LogLevel.DEBUG, this.logEntryHeader
        + ": loading a simulation snapshot");

    this.currentState = snapshot.currentState;
    this.lastTraversedTransition = snapshot.lastTraversedTransition;
    this.coverage.reset();
    this.coverage.include(snapshot.coverage);

    super.loadSnapshot(snapshot);
  }
}
