/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv;

import org.apache.commons.cli.ParseException;

import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.FileNames;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.model.hldd.HlddModel;
import ru.ispras.retrascope.model.hldd.HlddModule;
import ru.ispras.retrascope.model.hldd.HlddVisitor;
import ru.ispras.retrascope.model.hldd.HlddWalker;
import ru.ispras.retrascope.result.assertion.AssertionEntity;
import ru.ispras.retrascope.util.Log;

import java.util.Map;

/**
 * This engine prints the HLDD model as a NuXMV file.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class HlddSmvPrinter extends Engine {

  /** The engine ID. */
  public static final String ENGINE_ID = "hldd-smv-printer";

  /** The engine output entity type. */
  public static final EntityType OUT_TYPE = EntityType.get("smv");

  /** The command-line parameter. */
  public static final Parameter SMV_FILE_NAME = new Parameter("output-file", true,
      "Output file name");
  
  public static final Parameter CHECK_METHOD = new Parameter("check-method", new String[] {"bdd",
  "bmc"}, "Model checking method");

  /**
   * Constructs a new printer object.
   */
  public HlddSmvPrinter() {
    super(ENGINE_ID);

    addInputType(HlddModel.TYPE);
    addInputType(AssertionEntity.TYPE);
    setOutputType(OUT_TYPE);

    addParameter(SMV_FILE_NAME);
    addParameter(CHECK_METHOD);

  }

  @Override
  public Entity start(Map<EntityType, Entity> inputs) {

    final FileNames result = new FileNames(OUT_TYPE.getId());

    try {

      final HlddModel model = (HlddModel) inputs.get(HlddModel.TYPE);

      if (!containsEmptyModules(model)) {
        AssertionEntity assertions = null;
        if (inputs.containsKey(AssertionEntity.TYPE)) {
          assertions = (AssertionEntity) inputs.get(AssertionEntity.TYPE);
        }

        final Arguments arguments =
            this.parseCommandLine(((Arguments) inputs.get(Arguments.TYPE)).getCommandLine());
        final String fileName = arguments.getValue(SMV_FILE_NAME, "model.smv");
        result.add(fileName);
        
        final String checkType = arguments.getValue(CHECK_METHOD, "bmc").toLowerCase();

        final HlddVisitor visitor = new HlddSmvVisitor(fileName);
        ((HlddSmvVisitor) visitor).setSpecType(checkType);
        
        if (assertions != null) {
          ((HlddSmvVisitor) visitor).addAssertions(assertions.getContainerMap());
        }

        final HlddWalker walker = new HlddWalker(visitor);
        walker.start(model);
      }

    } catch (ParseException e) {
      e.printStackTrace();
    }

    return result;
  }

  /* 
   * Checks whether the model contains empty modules. 
   */
  private boolean containsEmptyModules(final HlddModel model) {
    if (model == null) {
      return true;
    }

    for (HlddModule module : model.getModules()) {
      if (module.getModule().getName().isEmpty()) {
        Log.printWarn("Empty module found!");
        return true;
      }
    }

    return false;
  }

}
