/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.Node.Kind;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.retrascope.basis.FileCreator;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.MapUtils;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.SmvExprPrinter;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.SmvPrinterUtils;
import ru.ispras.retrascope.model.basis.VariableDeclaration;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.hldd.Hldd;
import ru.ispras.retrascope.model.hldd.HlddModel;
import ru.ispras.retrascope.model.hldd.HlddModule;
import ru.ispras.retrascope.model.hldd.HlddNode;
import ru.ispras.retrascope.model.hldd.HlddProcess;
import ru.ispras.retrascope.model.hldd.HlddVisitor;
import ru.ispras.retrascope.result.assertion.AssertionContainer;
import ru.ispras.retrascope.result.assertion.AssertionGroup;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.NodeUtils;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * This class represents a visitor which contains SMV printing methods.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class HlddSmvVisitor implements HlddVisitor {

  // Prints expressions into file.
  private PrintWriter printWriter;
  // Module being visited.
  private CfgModule currentModule;
  // Process being visited.
  private HlddProcess currentProcess;
  // XMV expression printer.
  private SmvExprPrinter exprPrinter;
  // Indentation.
  private String indentation;
  // Model assertions.
  private Map<String, AssertionContainer> assertions;
  // Current assertion container (module).
  private AssertionContainer currentAssertionContainer;
  // Current assertion group (process).
  private AssertionGroup currentAssertionGroup;
  // Variables which are used in the process nodes.
  private List<NodeVariable> currentUsedVariables;
  // Current HLDD variable.
  private NodeVariable currentHlddVariable;
  // Enables printing.
  private boolean printingEnabled;
  // The node which children are not printed.
  private HlddNode disablingNode;
  // Enables initial reset printing.
  // private boolean addReset;

  private boolean printCTL;

  /**
   * Constructs a new visitor and creates a new *.smv file.
   * 
   * @param fileName - the specified *.smv file name.
   */
  public HlddSmvVisitor(final String fileName) {
    final File smvFile = FileCreator.newFile(fileName);
    exprPrinter = new SmvExprPrinter();
    indentation = "      ";
    assertions = new HashMap<>();
    currentUsedVariables = new ArrayList<>();
    printingEnabled = true;
    disablingNode = null;
    printCTL = false;
    // addReset = false;

    try {
      printWriter = new PrintWriter(new BufferedWriter(new FileWriter(smvFile)));
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  /**
   * Adds model assertions to the visitor.
   * 
   * @param assertions - the specified assertions.
   */
  public void addAssertions(final Map<String, AssertionContainer> assertions) {
    this.assertions = assertions;
  }

  // /**
  // * Enables initial reset printing.
  // *
  // * @param addReset - the specified flag.
  // */
  // public void addReset(final boolean addReset) {
  // this.addReset = addReset;
  // }

  public void setSpecType(final String method) {
    if (method.equals("bdd")) {
      printCTL = true;
    } else if (method.equals("bmc")) {
      printCTL = false;
    } else {
      throw new RetrascopeRuntimeException("Unknown model checking method: " + method);
    }

  }

  private void append(final String string) {
    if (printingEnabled) {
      printWriter.print(string);
    }
  }

  private void appendLine(final String line) {
    if (printingEnabled) {
      printWriter.println(line);
    }
  }

  private void appendLine() {
    if (printingEnabled) {
      printWriter.println();
    }
  }

  @Override
  public void onModelBegin(final HlddModel model) {

    appendLine();
    appendLine(SmvPrinterUtils.getModuleDeclaration("main"));

    final Map<NodeVariable, VariableDeclaration> inputVariables = new HashMap<>();

    // append(SmvExprPrinter.VARIABLE_DECLARATION + "\n");
    for (HlddModule hlddModule : model.getModules()) {
      final CfgModule module = hlddModule.getModule();
      final String moduleName = SmvPrinterUtils.getFormattedName(module.getName());
      final String instanceName = "m_" + moduleName;

      for (NodeVariable input : module.getInputs()) {
        inputVariables.put(input, module.getDeclaration(input));
      }
      appendLine(SmvPrinterUtils.getModuleInstantiation(instanceName, moduleName,
          module.getInputs()));
    }

    appendLine(SmvPrinterUtils.getVarDeclareSection(inputVariables, exprPrinter));
  }

  @Override
  public void onModelEnd(final HlddModel model) {
    appendLine();
    printWriter.close();
  }

  @Override
  public void onHlddBegin(final Hldd hldd) {
    if (!hldd.getVariable().isType(DataTypeId.MAP)) {
      final String name = SmvPrinterUtils.getName(hldd.getVariable());
      if (hldd.isImmediateAssignment()) {
        append("  " + name + " := ");
      } else {
        append("  next(" + name + ") := ");
      }
    }
    currentHlddVariable = hldd.getVariable();
  }

  @Override
  public void onHlddEnd(final Hldd hldd) {
    appendLine();
  }

  @Override
  public void onRootBegin(final HlddNode rootNode) {
    /* Empty */
  }

  @Override
  public void onRootEnd(final HlddNode rootNode) {
    /* Empty */
  }

  @Override
  public void onNodeBegin(final HlddNode node) {
    /* Empty */
  }

  @Override
  public void onNodeEnd(final HlddNode node) {
    /* Empty */
  }

  @Override
  public void onConditionBegin(final HlddNode node) {
    if (!isValue(node)) {
      indentation = SmvPrinterUtils.addWhitespace(indentation);
      appendLine("\n" + indentation + "case");
    }
  }

  /*
   * Checks whether the specified HlddNode contains a NodeValue expression.
   * 
   * @param node - the specified node.
   * 
   * @return <code>true</code> if the specified node contains a NodeValue, <code>false</code>
   * otherwise.
   */
  private boolean isValue(final HlddNode node) {
    final Node guard = node.getOnlyExpression();
    for (Hldd hldd : currentProcess.getHldds()) {
      if (hldd.getVariable().equals(guard)) {
        final Node expr = hldd.getTerminalNodes().iterator().next().getExpressions().get(1);
        if (expr instanceof NodeValue) {
          return true;
        }
      }
    }
    return false;
  }

  @Override
  public void onConditionEnd(final HlddNode node) {
    if (!isValue(node)) {
      appendLine(indentation + "esac;");
      indentation = SmvPrinterUtils.removeWhitespace(indentation);
    }
  }

  @Override
  public void onValueBegin(final HlddNode node) {

    final HlddNode parent = node.getParent();

    if (!isValue(parent)) {
      if (node.isDefault()) {
        indentation = SmvPrinterUtils.addWhitespace(indentation);
        append(indentation + "TRUE" + " : ");
      } else {
        final ArrayList<Node> expressions = new ArrayList<>();

        for (Node value : node.getExpressions()) {
          expressions.add(NodeUtils.getEquation(parent.getOnlyExpression(), value));
        }

        Node finalExpression = ExprUtils.getDisjunction(expressions.toArray(new Node[0]));
        finalExpression = SmvPrinterUtils.makeSubstitutions(finalExpression);
        indentation = SmvPrinterUtils.addWhitespace(indentation);
        final String exprDescription = getExpressionDescription(finalExpression);
        append(indentation + exprDescription + " : ");
      }
    } else {
      if (!((NodeValue) node.getOnlyExpression()).getBoolean()) {
        printingEnabled = false;
        disablingNode = parent;
        Log.printWarn("Unreachable path detected. Variable: "
            + parent.getOnlyExpression().toString());
      }
    }
  }

  @Override
  public void onValueEnd(final HlddNode node) {
    final HlddNode parent = node.getParent();
    if (!isValue(parent)) {
      indentation = SmvPrinterUtils.removeWhitespace(indentation);
    } else {
      if (parent.equals(disablingNode)) {
        printingEnabled = true;
      }
    }
  }

  @Override
  public void onAssignmentBegin(final HlddNode node) {
    Node expression = node.getExpressions().get(1);
    if (expression.getKind() == Kind.VALUE) {
      DataType curVarDataType = currentHlddVariable.getDataType();
      if (!expression.getDataType().equals(curVarDataType)
          && expression.getDataTypeId() != DataTypeId.MAP) {
        final Node coercedExpr = TypeConversion.coerce(expression, curVarDataType);
        if (coercedExpr != null) {
          expression = coercedExpr;
        }
      }
      appendLine(getExpressionDescription(expression) + ";");

    } else {
      expression = SmvPrinterUtils.makeSubstitutions(expression);
      final String exprDescription = getExpressionDescription(expression);
      final DataType exprType = getExpressionType(expression);
      convertAndPrint(exprType, exprDescription);
    }
  }

  /*
   * Returns the expression description string.
   */
  private String getExpressionDescription(final Node expression) {
    String exprDescription = exprPrinter.toString(expression)/* .toLowerCase() */;
    exprDescription = exprDescription.replaceAll("true", "TRUE");
    exprDescription = exprDescription.replaceAll("false", "FALSE");
    return exprDescription;
  }

  /*
   * Returns the specified expressions data type if possible.
   */
  private DataType getExpressionType(final Node expression) {
    DataType exprType = DataType.UNKNOWN;
    try {
      exprType = expression.getDataType();
    } catch (IllegalStateException | NullPointerException e) {
      exprType = DataType.UNKNOWN;
    } finally {
      if (exprType == null) {
        exprType = DataType.UNKNOWN;
      }
    }
    return exprType;
  }

  /*
   * Prints the specified expression description and adds the type conversion expression if needed.
   */
  private void convertAndPrint(final DataType exprType, final String exprDescription) {
    if (exprType.getTypeId() != currentHlddVariable.getDataTypeId()
        && exprType.getTypeId() != DataTypeId.UNKNOWN) {
      append(SmvPrinterUtils.getTypeConvertion(exprType.getTypeId(),
          currentHlddVariable.getDataTypeId())
          + "(");
    }
    append(exprDescription);
    if (exprType.getTypeId() != currentHlddVariable.getDataTypeId()
        && exprType.getTypeId() != DataTypeId.UNKNOWN) {
      append(")");
    }
    appendLine(";");
  }

  @Override
  public void onAssignmentEnd(final HlddNode node) {
    final Node expression = node.getExpressions().get(1);
    searchMissingVariables(expression);
  }

  @Override
  public void onModuleBegin(final HlddModule hlddModule) {
    final CfgModule module = hlddModule.getModule();
    final ArrayList<NodeVariable> inputVariables = new ArrayList<>(module.getInputs());

    currentModule = module;

    final String moduleName = " " + SmvPrinterUtils.getFormattedName(module.getName());

    setCurrentAssertionContainer(module.getName());

    appendLine();
    appendLine(SmvPrinterUtils.getModuleDeclaration(moduleName, inputVariables));

    if (currentAssertionContainer != null && currentAssertionContainer.getClock() != null) {
      append(SmvPrinterUtils.getClockAssignment(currentAssertionContainer.getClock()));
      appendLine();
      // printReset(currentAssertionContainer.getResetVariable());
    }

    for (CfgProcess process : module.getProcesses()) {
      String processName = SmvPrinterUtils.getFormattedName(process.getId());
      final String instanceName = "  m_" + processName;
      // appendLine(SmvExprPrinter.VARIABLE_DECLARATION);
      append(SmvPrinterUtils.getModuleInstantiation(instanceName, processName, inputVariables));
    }
  }

  /*
   * Searches for AssertionContainer using the specified key name. If found, assigns the value to
   * the corresponding variable, assigns null otherwise.
   */
  private void setCurrentAssertionContainer(final String name) {
    if (assertions.containsKey(name)) {
      currentAssertionContainer = assertions.get(name);
    } else {
      currentAssertionContainer = null;
    }
  }



  @Override
  public void onModuleEnd(final HlddModule module) {
    appendLine();
  }

  @Override
  public void onProcessBegin(final HlddProcess process) {
    currentProcess = process;
    final ArrayList<NodeVariable> moduleVariables = new ArrayList<>(currentModule.getInputs());

    final String moduleName = SmvPrinterUtils.getFormattedName(process.getProcessId());
    setCurrentAssertionGroup(process.getProcessId());

    final Map<NodeVariable, VariableDeclaration> variables = new HashMap<>();

    for (NodeVariable var : process.getVariables()) {
      if (process.getCfgProcess().getDeclaration(var) != null) {
        variables.put(var, process.getCfgProcess().getDeclaration(var));
      } else if (currentModule.getDeclaration(var) != null) {
        variables.put(var, currentModule.getDeclaration(var));
      } else {
        variables.put(var, null);
      }
    }

    appendLine();
    appendLine(SmvPrinterUtils.getModuleDeclaration(moduleName, moduleVariables));
    appendLine();
    appendLine(SmvPrinterUtils.getVarDeclareSection(variables, exprPrinter));
    appendLine(SmvExprPrinter.ASSIGNMENT_DECLARATION);

    currentUsedVariables = new ArrayList<>();
  }

  /*
   * Sets the current assertion group.
   */
  private void setCurrentAssertionGroup(final String name) {
    if (currentAssertionContainer != null) {
      final Map<String, AssertionGroup> groups = currentAssertionContainer.getGroupMap();
      if (groups.containsKey(name)) {
        currentAssertionGroup = groups.get(name);
      } else {
        currentAssertionGroup = null;
      }
    }
  }

  @Override
  public void onProcessEnd(final HlddProcess process) {
    final List<NodeVariable> currentVariables = new ArrayList<>();
    currentVariables.addAll(currentModule.getInputs());
    currentVariables.addAll(process.getVariables());
    if (currentAssertionGroup != null) {
      printAssertions();
      // printReset(currentAssertionGroup.getResetVariable());
    }
    printMissingVariables(currentVariables);
    appendLine(MapUtils.printMaps(exprPrinter));
    MapUtils.resetMaps();
  }

  /*
   * Prints the current assertion group.
   */
  private void printAssertions() {
    final String assertionDecl =
        printCTL ? SmvExprPrinter.CTL_SPEC_DECLARATION : SmvExprPrinter.LTL_SPEC_DECLARATION;
    final String operator = printCTL ? "EF" : "F";
    for (Node node : currentAssertionGroup.getAssertions()) {
      append(assertionDecl);
      final Node subNode = SmvPrinterUtils.makeSubstitutions(node);
      searchMissingVariables(subNode);
      final String exprDescription = getExpressionDescription(subNode);
      appendLine("  ! " + operator + " " + exprDescription + ";");
    }
    appendLine();
  }

  /*
   * Prints reset initial value.
   */
  // private void printReset(final NodeVariable reset) {
  // if (addReset && reset != null) {
  // appendLine(SmvExprPrinter.ASSIGNMENT_DECLARATION);
  // final String varName = SmvPrinterUtils.getName(reset);
  // final DataType varType = reset.getDataType();
  // final Data value = varType.valueOf("1", 2);
  // append("  init(" + varName +") := " + exprPrinter.getConstant(value/*, varType*/) + ";\n");
  // }
  // }

  /*
   * Searches for the variables which haven't been declared in the current process.
   */
  private void searchMissingVariables(final Node node) {
    final List<NodeVariable> usedVariables = SmvPrinterUtils.getUsedVariables(node);
    for (NodeVariable var : usedVariables) {
      if (!currentUsedVariables.contains(var) && !MapUtils.getMaps().containsValue(var)) {
        currentUsedVariables.add(var);
      }
    }
  }

  /*
   * Prints the variables which haven't been declared in the current process.
   */
  private void printMissingVariables(final List<NodeVariable> variables) {
    if (!variables.containsAll(currentUsedVariables)) {
      appendLine(SmvExprPrinter.VARIABLE_DECLARATION);
      for (NodeVariable var : currentUsedVariables) {
        if (!variables.contains(var)) {
          append(SmvPrinterUtils.getVariableDeclaration(var, null, exprPrinter, ""));
        }
      }
      appendLine();
    }
  }

}
