/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv.utils;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.datamap.DataMap;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.NodeTransformer;
import ru.ispras.fortress.transformer.TransformerRule;

/**
 * This class contains static methods for proper MAP data type printing.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class MapUtils {

  // Represents MAP data type node-name pairs.
  private static Map<Node, Node> arrays;
  // Counts MAP data type values in the program.
  private static int mapCounter;

  private MapUtils() {}
  
  static {
    resetMapCounter();
    resetMaps();
  }
  
  /**
   * Resets MAP data type container.
   */
  public static void resetMaps() {
    arrays = new HashMap<Node, Node>();
  }

  /**
   * Returns the MAP data type values with their SMV variable name.
   * 
   * @return The map which keys are MAP data type nodes and values are corresponding variable names.
   */
  public static Map<Node, Node> getMaps() {
    return arrays;
  }

  /**
   * Resets data maps counter.
   */
  public static void resetMapCounter() {
    mapCounter = 0;
  }
  
  /**
   * Searches for DataTypeId.MAP values in the specified node and replaces them by generated
   * array variables.
   * 
   * @param node - the specified {@link Node}.
   * @return The node with substitution performed.
   */
  public static Node substituteMaps(final Node node) {
    final NodeTransformer transformer = new NodeTransformer();

    final TransformerRule mapRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node arg0) {
        return arg0 instanceof NodeOperation
            && ((NodeOperation) arg0).getOperationId() == StandardOperation.SELECT
            && ((NodeOperation) arg0).getOperand(0).isType(DataTypeId.MAP);
        
      }

      @Override
      public Node apply(final Node arg0) {
        final NodeOperation operation = (NodeOperation) arg0;
        final Node map = operation.getOperand(0);
        final Node index = operation.getOperand(1);
        final DataMap dataMap = (DataMap) ((NodeValue) map).getValue();
        Node mapVar;        
        if (!arrays.containsKey(map)) {
          final DataType dataType = dataMap.getValueType();
          mapVar = new NodeVariable("map" + mapCounter, dataType);
          arrays.put(map, mapVar);
          mapCounter++;
        } else {
          mapVar = arrays.get(map);
        }
        
        final NodeValue size = new NodeValue(Data.newInteger(dataMap.size()));
        final Node modIndex = new NodeOperation(StandardOperation.MOD, index, size);
        
        return new NodeOperation(StandardOperation.SELECT, mapVar, modIndex);
      }

    };

    transformer.addRule(StandardOperation.SELECT, mapRule);
    transformer.walk(node);
    return transformer.getResult().iterator().next();
  }
  
  /**
   * Returns the description of data arrays in SMV format.
   * 
   * @return String representation of data arrays.
   */
  public static String printMaps(final SmvExprPrinter printer) {
    final StringBuilder result = new StringBuilder();

    if (!arrays.isEmpty()) {
      result.append(SmvExprPrinter.DEFINE_DECLARATION + "\n");
    }

    for (Entry<Node, Node> entry : arrays.entrySet()) {
      final Node node = entry.getKey();
      final DataMap dataMap = (DataMap) ((NodeValue) node).getValue();
      final String mapName = ((NodeVariable) entry.getValue()).getName();

      result.append("  " + mapName + " := [");

      int ind = 0;
      for (Entry<Data, Data> mapEntry : dataMap.entrySet()) {
        if (ind != 0) {
          result.append(", ");
        }
        result.append(printer.getConstant(mapEntry.getValue()/*, valueType*/));
        ind++;
      }
      result.append("];\n");
    }
    return result.toString();
  }

}
