/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv.utils;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.expression.printer.MapBasedPrinter;
import ru.ispras.fortress.expression.printer.OperationDescription.Type;
import ru.ispras.fortress.transformer.TypeConversion;

import java.math.BigInteger;
import java.util.Collection;
import java.util.Stack;

/**
 * This class represents the printer into *.smv format.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SmvExprPrinter extends MapBasedPrinter {

  /** Announces module declaration. */
  public static final String MODULE_DECLARATION = "MODULE";
  /** Variable declaration section header. */
  public static final String VARIABLE_DECLARATION = "VAR";
  /** Input variable declaration section header. */
  public static final String INPUT_VARIABLE_DECLARATION = "IVAR";
  /** Assignment section header. */
  public static final String ASSIGNMENT_DECLARATION = "ASSIGN";
  /** Definition section header. */
  public static final String DEFINE_DECLARATION = "DEFINE";
  /** CTL specification header. */
  public static final String CTL_SPEC_DECLARATION = "CTLSPEC";
  /** LTL specification header. */
  public static final String LTL_SPEC_DECLARATION = "LTLSPEC";
  /** The size of integer type.*/
  public static final int INTEGER_SIZE = 32;
  /** Substituted variable suffix.*/
  public static final String SUB_ENDING = "_SUBSTITUTED";
  /** SMV language reserved words. */
  public static final String[] FORBIDDEN_NAMES = {"A", "ABF", "ABG", "AF", "AG", "array", "ASSIGN",
      "AX", "bool", "boolean", "BU", "case", "COMPASSION", "COMPID", "COMPUTE", "COMPWFF",
      "CONSTANTS", "CONSTRAINT", "count", "CTLSPEC", "CTLWFF", "DEFINE", "E", "EBF", "EBG", "EF",
      "EG", "esac", "EX", "extend", "F", "FAIRNESS", "FALSE", "floor", "FROZENVAR", "FUN", "G",
      "H", "IN", "in", "INIT", "init", "Integer", "integer", "INVAR", "INVARSPEC", "ISA", "ITYPE",
      "IVAR", "JUSTICE", "LTLSPEC", "LTLWFF", "MAX", "max", "MDEFINE", "MIN", "MIRROR", "mod",
      "MODULE", "NAME", "next", "NEXTWFF", "O", "of", "PRED", "PREDICATES", "PSLSPEC", "READ",
      "Real", "real", "resize", "S", "self", "signed", "SIMPWFF", "sizeof", "SPEC", "swconst", "T",
      "toint", "TRANS", "TRUE", "U", "union", "unsigned", "uwconst", "V", "VAR", "Word", "word",
      "word1", "WRITE", "X", "xnor", "xor", "Y", "Z"};

  // Integer printing mode flag.
  private boolean intAsBv;
  // Hexadecimal bit vector printing flag.
  private boolean bvHex;
  // Binary bit vector printing flag.
  private boolean bvBin;
  // Decimal bit vector printing flag.
  private boolean bvDec;
  // Operation stack.
  private Stack<Enum<?>> operationIdStack;
  // Current operand number.
  private int currentOperandIndex;
  // SELECT operation flag.
  private boolean insideSelect;

  /** Types of NuXMV constraints. */
  public enum ConstraintType {
    FAIRNESS, INVAR, CTLSPEC, LTLSPEC
  }

  /**
   * Default constructor. Sets operation mappings.
   */
  public SmvExprPrinter() {
    // Supported operations.
    addMapping(StandardOperation.MINUS, "-", Type.PREFIX);
    addMapping(StandardOperation.PLUS, "+", Type.PREFIX);
    addMapping(StandardOperation.NOT, "!", Type.PREFIX, false);
    addMapping(StandardOperation.BVNOT, "!", Type.PREFIX, false);
    addMapping(StandardOperation.BVANDR, "&", Type.PREFIX);
    addMapping(StandardOperation.BVORR, "|", Type.PREFIX);
    addMapping(StandardOperation.BVXORR, "xor", Type.PREFIX);
    addMapping(StandardOperation.BVXNORR, "xnor", Type.PREFIX);
    addMapping(StandardOperation.BVUREM, "mod", Type.INFIX);
    addMapping(StandardOperation.MUL, "*", Type.INFIX);
    addMapping(StandardOperation.BVMUL, "*", Type.INFIX);
    addMapping(StandardOperation.DIV, "/", Type.INFIX);
    addMapping(StandardOperation.REM, "mod", Type.INFIX);
    addMapping(StandardOperation.MOD, "mod", Type.INFIX);
    addMapping(StandardOperation.BVSMOD, "mod", Type.INFIX);
    addMapping(StandardOperation.ADD, "+", Type.INFIX);
    addMapping(StandardOperation.BVADD, "+", Type.INFIX);
    addMapping(StandardOperation.SUB, "-", Type.INFIX);
    addMapping(StandardOperation.BVSUB, "-", Type.INFIX);
    addMapping(StandardOperation.BVLSHR, ">>", Type.INFIX);
    addMapping(StandardOperation.BVLSHL, "<<", Type.INFIX);
    addMapping(StandardOperation.BVASHR, ">>", Type.INFIX);
    addMapping(StandardOperation.BVASHL, "<<", Type.INFIX);
    addMapping(StandardOperation.LESS, "<", Type.INFIX);
    addMapping(StandardOperation.BVULT, "<", Type.INFIX);
    addMapping(StandardOperation.BVSLT, "<", Type.INFIX);
    addMapping(StandardOperation.LESSEQ, "<=", Type.INFIX);
    addMapping(StandardOperation.BVULE, "<=", Type.INFIX);
    addMapping(StandardOperation.BVSLE, "<=", Type.INFIX);
    addMapping(StandardOperation.GREATER, ">", Type.INFIX);
    addMapping(StandardOperation.BVUGT, ">", Type.INFIX);
    addMapping(StandardOperation.BVSGT, ">", Type.INFIX);
    addMapping(StandardOperation.GREATEREQ, ">=", Type.INFIX);
    addMapping(StandardOperation.BVUGE, ">=", Type.INFIX);
    addMapping(StandardOperation.BVSGE, ">=", Type.INFIX);
    addMapping(StandardOperation.EQ, "=", Type.INFIX);
    addMapping(StandardOperation.NOTEQ, "!=", Type.INFIX);
    addMapping(StandardOperation.EQCASE, "=", Type.INFIX);
    addMapping(StandardOperation.NOTEQCASE, "!=", Type.INFIX);
    addMapping(StandardOperation.BVAND, "&", Type.INFIX);
    addMapping(StandardOperation.BVXOR, "xor", Type.INFIX);
    addMapping(StandardOperation.BVXNOR, "xnor", Type.INFIX);
    addMapping(StandardOperation.BVOR, "|", Type.INFIX);
    addMapping(StandardOperation.AND, "&", Type.INFIX);
    addMapping(StandardOperation.OR, "|", Type.INFIX);
    addMapping(StandardOperation.IMPL, "->", Type.INFIX);
    addMapping(StandardOperation.ITE, new String[] {"?", ":"});
    addMapping(StandardOperation.BVCONCAT, "(", " :: ", ")");
    addMapping(StandardOperation.BVREPEAT, "(", "(", "))");
    addMapping(StandardOperation.BVEXTRACT, "", new String[] {"[", ":"}, "]", new int[] {2, 0, 1});
    addMapping(StandardOperation.SELECT, "", "[", "]");
    addMapping(StandardOperation.BV2BOOL, new String[] {"bool(", ")"});
    addMapping(StandardOperation.BOOL2BV, new String[] {"word1(", ")"});
    addMapping(StandardOperation.STORE, new String[] {"[", "] :="});

    // Unsupported operations.
    addMapping(StandardOperation.BVROL, "BVROL(", ", ", ")");
    addMapping(StandardOperation.BVROR, "BVROR(", ", ", ")");
    addMapping(StandardOperation.BVUREM, "BVUREM(", ", ", ")");
    addMapping(StandardOperation.BVSREM, "BVSREM(", ", ", ")");
    addMapping(StandardOperation.BVZEROEXT, "BVZEROEXT(", ", ", ")");
    addMapping(StandardOperation.BVSIGNEXT, "BVSIGNEXT(", ", ", ")");
    addMapping(StandardOperation.ABS, "ABS(", ", ", ")");
    addMapping(StandardOperation.MIN, "MIN(", ", ", ")");
    addMapping(StandardOperation.MAX, "MAX(", ", ", ")");

    addMapping(StandardOperation.BVNANDR, "~&", Type.PREFIX);
    addMapping(StandardOperation.BVNORR, "~|", Type.PREFIX);
    addMapping(StandardOperation.POWER, "**", Type.INFIX);
    addMapping(StandardOperation.BVNAND, "~&", Type.INFIX);
    addMapping(StandardOperation.BVNOR, "~|", Type.INFIX);


    setVisitor(new Visitor());
    printIntsAsBv();
    setDec();

    operationIdStack = new Stack<Enum<?>>();
    insideSelect = false;
  }

  /**
   * Enables integer printing as 32-bit vector.
   */
  public void printIntsAsBv() {
    intAsBv = true;
  }

  /**
   * Disables integer printing as 32-bit vector.
   */
  public void printIntsNormal() {
    intAsBv = false;
  }

  /**
   * Sets binary bit vector printing mode.
   */
  public void setBin() {
    bvBin = true;
    bvDec = false;
    bvHex = false;
  }

  /**
   * Sets decimal bit vector printing mode.
   */
  public void setDec() {
    bvBin = false;
    bvDec = true;
    bvHex = false;
  }

  /**
   * Sets hexadecimal bit vector printing mode.
   */
  public void setHex() {
    bvBin = false;
    bvDec = false;
    bvHex = true;
  }

  /**
   * Returns the description of a constraint in SMV format using the given {@link ConstraintType}
   * and {@link Node}.
   * 
   * @param type - the specified constraint type.
   * @param node - the specified expression node.
   * @return The description string.
   */
  public String getConstraintDescription(final ConstraintType type, final Node node) {
    return type.name() + "\n  " + this.toString(node);
  }

  /**
   * Returns the description of a constraint in SMV format using the given {@link ConstraintType}
   * and the collection of nodes.
   * 
   * @param type - the specified constraint type.
   * @param nodes - the specified node collection.
   * @return The description string.
   */
  public String getConstraintDescription(final ConstraintType type, final Collection<Node> nodes) {
    final StringBuilder builder = new StringBuilder();
    for (Node node : nodes) {
      builder.append(getConstraintDescription(type, node));
    }
    return builder.toString();
  }

  /* The visitor which is used to print expressions. */
  private final class Visitor extends ExprTreeVisitor {

    @Override
    public void onValue(final NodeValue value) {
      NodeValue coercedValue = null;
      boolean intAsBvOld = intAsBv;
      final Enum<?> currentOperation = operationIdStack.isEmpty() ? null : operationIdStack.peek();

      if (!operationIdStack.isEmpty() && currentOperation == StandardOperation.BVEXTRACT) {
        intAsBv = false;
        if (currentOperandIndex != 0) {
          coercedValue = (NodeValue) TypeConversion.coerce(value, DataType.INTEGER);
        }
      }
      
      if (!operationIdStack.isEmpty() && (currentOperation == StandardOperation.BVASHL 
          || (currentOperation == StandardOperation.MOD && insideSelect)) 
          && value.isType(DataType.INTEGER) && intAsBv) {
        coercedValue = (NodeValue) TypeConversion.coerce(value, DataType.BIT_VECTOR(INTEGER_SIZE));
      }
      
      if (null != coercedValue) {
        appendText(getConstant(coercedValue.getData()/*, coercedValue.getDataType()*/));
      } else {
        appendText(getConstant(value.getData()/*, value.getDataType()*/));
      }
      intAsBv = intAsBvOld;
    }

    @Override
    public void onOperationBegin(final NodeOperation operation) {
      final Enum<?> operationId = operation.getOperationId(); 
      operationIdStack.push(operationId);
      if (intAsBv && operationId == StandardOperation.BVASHL) {
        appendText("signed");
      }
      if (operationId == StandardOperation.SELECT) {
        insideSelect = true;
      }
      super.onOperationBegin(operation);
    }

    @Override
    public void onOperationEnd(final NodeOperation operation) {
      operationIdStack.pop();
      if (operation.getOperationId() == StandardOperation.SELECT) {
        insideSelect = false;
      }
      super.onOperationEnd(operation);
    }

    @Override
    public void onOperandBegin(final NodeOperation operation, final Node operand, final int index) {
      currentOperandIndex = index;
      if (operation.getOperationId() == StandardOperation.ITE) {
        if (index == 0 && operand.getDataTypeId() != DataTypeId.LOGIC_BOOLEAN) {
          appendText("bool(");
        }
      }
      super.onOperandBegin(operation, operand, index);
    }

    @Override
    public void onOperandEnd(final NodeOperation operation, final Node operand, final int index) {
      super.onOperandEnd(operation, operand, index);
      if (operation.getOperationId() == StandardOperation.ITE) {
        if (index == 0 && operand.getDataTypeId() != DataTypeId.LOGIC_BOOLEAN) {
          appendText(")");
        }
      }
    }

    @Override
    public void onVariable(final NodeVariable variable) {
      final String varName = SmvPrinterUtils.getName(variable);
      final Enum<?> currentOperation = operationIdStack.isEmpty() ? null : operationIdStack.peek();
      if (!operationIdStack.isEmpty() && (currentOperation == StandardOperation.BVASHL
          || (currentOperation == StandardOperation.MOD && insideSelect)) 
          && variable.isType(DataType.INTEGER) && intAsBv) {
        appendText("unsigned(" + varName + ")");
      } else {
        appendText(varName);
      }
    }

  }

  /**
   * Returns {@link Data} description in SMV format.
   * 
   * @param node - the specified data node.
   * @return Data string description.
   */
  public String getConstant(final Data node) {
    String result = "";
    String signed = "";
    final DataType dataType = node.getType();
    switch (dataType.getTypeId()) {
      case LOGIC_BOOLEAN:
        if (node.getBoolean() == true) {
          result = "TRUE";
        } else {
          result = "FALSE";
        }
        break;
      default:
        BigInteger value = BigInteger.ZERO;
        int size = 0;
        if (dataType.getTypeId() == DataTypeId.BIT_VECTOR) {
          value = node.getBitVector().bigIntegerValue(false);
          size = node.getBitVector().getBitSize();
        }
        if (dataType.getTypeId() == DataTypeId.LOGIC_INTEGER) {
          value = node.getInteger();
          size = INTEGER_SIZE;
          signed = "s";
        }

        final String sign = (value.signum() == -1) ? "-" : "";

        if (intAsBv || dataType.getTypeId() != DataTypeId.LOGIC_INTEGER) {
          if (bvBin) {
            result = sign + "0" + signed + "b" + size + "_" + value.abs().toString(2);
          }

          if (bvDec) {
            result = sign + "0" + signed + "d" + size + "_" + value.abs().toString(10);
          }

          if (bvHex) {
            result = sign + "0" + signed + "h" + size + "_" + value.abs().toString(16);
          }
        } else {
          result = value.toString();
        }

    }
    return result;
  }
  
  /**
   * Returns a data type in SMV format.
   * 
   * @param dataType - the specified {@link DataType}.
   * 
   * @return The data type string.
   */
  public String getType(final DataType dataType) {
    final StringBuilder result = new StringBuilder();
    switch (dataType.getTypeId()) {
      case BIT_VECTOR:
        result.append("word[" + dataType.getSize() + "]");
        break;
      case LOGIC_BOOLEAN:
        result.append("boolean");
        break;
      case LOGIC_INTEGER:
        if (intAsBv) {
          result.append("signed word[" + INTEGER_SIZE + "]");
        } else {
          result.append("integer");
        }
        break;
      case MAP:
        result.append("array 0..");
        final Object[] types = dataType.getParameters();
        final DataType keyType = (DataType) types[0];
        final DataType valueType = (DataType) types[1];
        if (keyType.getTypeId() == DataTypeId.LOGIC_INTEGER) {
          result.append(Integer.MAX_VALUE);
        }
        if (keyType.getTypeId() == DataTypeId.BIT_VECTOR) {
          final BigInteger upperRange =
              BigInteger.valueOf(2).pow(keyType.getSize()).subtract(BigInteger.ONE);
          result.append(upperRange.toString());
        }
        result.append(" of " + getType(valueType));
        break;
      default:
        result.append("boolean");
        break;
    }
    return result.toString();
  }


}
