/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv.utils;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprTreeVisitorDefault;
import ru.ispras.fortress.expression.ExprTreeWalker;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.NodeTransformer;
import ru.ispras.fortress.transformer.TransformerRule;
import ru.ispras.retrascope.model.basis.VariableDeclaration;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

/**
 * This class contains static methods for *.smv file printing.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SmvPrinterUtils {



 
  private SmvPrinterUtils() {}

  /**
   * Generates a module declaration in SMV format.
   * 
   * @param name - the specified module name.
   * @param inputs - the specified collection of parameters.
   * @return The string containing the module declaration.
   */
  public static String getModuleDeclaration(final String name, 
      final Collection<NodeVariable> inputs) {

    final StringBuilder result = new StringBuilder();

    result.append(SmvExprPrinter.MODULE_DECLARATION);
    result.append(" " + name);

    if (inputs.isEmpty()) {
      result.append("\n");
    } else {
      result.append("(");
      result.append(getModuleParametersDescription(inputs));
      result.append(")\n");
    }
    return result.toString();

  }

  /**
   * Generates a module declaration in SMV format with empty parameter list.
   * 
   * @param name - the specified module name.
   * @return The string containing the module declaration.
   */
  public static String getModuleDeclaration(final String name) {
    return getModuleDeclaration(name, new ArrayList<NodeVariable>());
  }

  /**
   * Generates a module instantiation in SMV format.
   * 
   * @param instanceName - the specified module instance name.
   * @param moduleName - the specified module name.
   * @param parameters - the specified collection of parameters.
   * @return The string containing the module instantiation.
   */
  public static String getModuleInstantiation(final String instanceName, final String moduleName,
      final Collection<NodeVariable> parameters) {
    final StringBuilder result = new StringBuilder();
    result.append(SmvExprPrinter.VARIABLE_DECLARATION + "\n");
    result.append(instanceName);
    result.append(" : ");
    result.append(moduleName);
    result.append("(");
    result.append(getModuleParametersDescription(parameters));
    result.append(");\n");
    return result.toString();
  }

  /*
   * Generates a list of module parameters from a collection of variables.
   * 
   * @param parameters - the specified collection of variables.
   * 
   * @return The string containing the list of module parameters.
   */
  private static String getModuleParametersDescription(final Collection<NodeVariable> parameters) {
    final StringBuilder result = new StringBuilder();
    final ArrayList<NodeVariable> variables = new ArrayList<NodeVariable>(parameters);
    for (int j = 0; j < variables.size(); j++) {
      if (j != 0) {
        result.append(", ");
      }
      result.append(getName(variables.get(j)));
    }
    return result.toString();
  }

  /**
   * Checks whether the variable name is acceptable by SMV syntax.
   * 
   * @param variable - the specified variable.
   * @return The modified variable name if it is forbidden by SMV, lower case variable name
   *         otherwise.
   */
  public static String getName(final NodeVariable variable) {
    final String varName = variable.getName()/* .toLowerCase() */;
    if (Arrays.asList(SmvExprPrinter.FORBIDDEN_NAMES).contains(varName)) {
      return varName + SmvExprPrinter.SUB_ENDING;
    }
    return varName;
  }

  /**
   * Generates a IVAR declaration section in SMV format from a collection of input variables.
   * 
   * @param inputs - the input variable-declaration mapping.
   * @param printer - the specified {@link SmvExprPrinter}.
   * @return The string containing the IVAR section.
   */
  public static String getInputVarDeclareSection(
      final Map<NodeVariable, VariableDeclaration> inputs, final SmvExprPrinter printer) {
    return getVarsDescription(inputs, SmvExprPrinter.INPUT_VARIABLE_DECLARATION, printer);
  }

  /**
   * Generates a VAR declaration section in SMV format from a collection of variables.
   * 
   * @param variables - the variable-declaration mapping
   * @param printer - the specified {@link SmvExprPrinter}.
   * @return The string containing the VAR section.
   */
  public static String getVarDeclareSection(final Map<NodeVariable, VariableDeclaration> variables,
      final SmvExprPrinter printer) {
    return getVarsDescription(variables, SmvExprPrinter.VARIABLE_DECLARATION, printer);
  }

  /**
   * Generates a variable declaration in SMV format.
   * 
   * @param variable - the specified variable.
   * @param declaration - the variable declaration.
   * @param printer - the specified {@link SmvExprPrinter}.
   * @return The string containing the variable declaration.
   */
  public static String getVariableDeclaration(final NodeVariable variable,
      final VariableDeclaration declaration, final SmvExprPrinter printer, final String keyword) {
    final StringBuilder result = new StringBuilder();
    final String varName = getName(variable);
    result.append(keyword + "\n");
    result.append("  " + varName + " : " + printer.getType(variable.getDataType()) + ";\n\n");

    if (declaration != null && declaration.containsInvariant()) {
      Node invariant = declaration.getInvariant();

      result.append("INVAR\n");
      invariant = makeSubstitutions(invariant);
      String exprDescription = printer.toString(invariant)/* .toLowerCase() */;
      exprDescription = exprDescription.replaceAll("true", "TRUE");
      exprDescription = exprDescription.replaceAll("false", "FALSE");
      result.append("  " + exprDescription + ";\n\n");
      
    }
    
    if (declaration != null && declaration.containsInitialValue()) {
      Node initialValue = declaration.getInitialValue();
      result.append(SmvExprPrinter.ASSIGNMENT_DECLARATION + "\n");
      result.append("  init(" + varName + ") := " + printer.toString(initialValue)+ ";\n\n");
      
    }
    
    
    return result.toString();
  }

  /*
   * Generates a declaration section in SMV format from variable collection.
   * 
   * @param variables
   * 
   * @param keyword - the specified declaration section keyword.
   * 
   * @param printer
   * 
   * @return The string containing the declaration section.
   */
  private static String getVarsDescription(final Map<NodeVariable, VariableDeclaration> variables,
      final String keyword, final SmvExprPrinter printer) {
    final StringBuilder result = new StringBuilder();
//    result.append(keyword);
    result.append("\n");
    for (Entry<NodeVariable, VariableDeclaration> entry : variables.entrySet()) {
      result.append(getVariableDeclaration(entry.getKey(), entry.getValue(), printer, keyword));
    }
    return result.toString();
  }

  

  /**
   * Adds two whitespaces to the end of the string.
   * 
   * @param string - the specified string.
   * @return The modified string.
   */
  public static String addWhitespace(final String string) {
    return string.concat("  ");
  }

  /**
   * Removes two symbols from the end of the string. The method is used to remove whitespaces from
   * indentation string.
   * 
   * @param string - the specified string.
   * @return The modified string.
   */
  public static String removeWhitespace(final String string) {
    if (string.length() > 1) {
      return string.substring(0, string.length() - 2);
    } else {
      return "";
    }
  }

  /**
   * Searches for power operations in the node and replaces them by arithmetic left shift operation.
   * 
   * @param node - the specified {@link Node}.
   * @return The node with substitution performed.
   */
  public static Node substitutePower(final Node node) {
    final NodeTransformer transformer = new NodeTransformer();

    final TransformerRule powerRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node arg0) {
        if (arg0.getKind() == Node.Kind.OPERATION) {
          return ((NodeOperation) arg0).getOperationId() == StandardOperation.POWER;
        }
        return false;
      }

      @Override
      public Node apply(final Node arg0) {
        return new NodeOperation(StandardOperation.BVASHL, new NodeValue(Data.newInteger(1)),
            ((NodeOperation) arg0).getOperand(1));
      }
    };

    transformer.addRule(StandardOperation.POWER, powerRule);
    transformer.walk(node);
    return transformer.getResult().iterator().next();
  }


  /**
   * Performs the following substitutions on the specified Node:
   * <ul>
   * <li>Substitutes a power operation by an arithmetic left shift operation.
   * <li>If a map value is found, a corresponding variable is created in the declarations section
   * and the value is replaced by this variable.
   * <li>If a BVEXTRACT operation extracts only one bit it is replaced by a SELECT operation.
   * </ul>
   * 
   * @param node - the specified node.
   * @return The node with the substitutions made.
   */
  public static Node makeSubstitutions(final Node node) {
    Node subNode = substitutePower(node);
    subNode = MapUtils.substituteMaps(subNode);
    return subNode;
  }

  private static final class Visitor extends ExprTreeVisitorDefault {

    private final List<NodeVariable> variables;

    public Visitor() {
      super();
      variables = new ArrayList<NodeVariable>();
    }

    @Override
    public void onVariable(final NodeVariable variable) {
      variables.add(variable);
    }

    public List<NodeVariable> getVariables() {
      return variables;
    }
  }

  /**
   * Returns the variables which are used in the specified {@link Node}.
   * 
   * @param node - the specified node.
   * @return The list of {@link NodeVariable}.
   */
  public static List<NodeVariable> getUsedVariables(final Node node) {
    final Visitor visitor = new Visitor();
    final ExprTreeWalker walker = new ExprTreeWalker(visitor);
    walker.visit(node);
    return visitor.getVariables();
  }

  /**
   * Returns the type conversion string from the source type into the target type if possible.
   * 
   * @param sourceType - the specified source type.
   * @param targetType - the specified target type.
   * @return Type conversion expression in SMV format if conversion is possible, empty string
   *         otherwise.
   */
  public static String getTypeConvertion(final DataTypeId sourceType, final DataTypeId targetType) {
    String result = "";
    if (targetType == DataTypeId.BIT_VECTOR && sourceType == DataTypeId.LOGIC_BOOLEAN) {
      result = "word1";
    }
    if (targetType == DataTypeId.LOGIC_BOOLEAN && sourceType == DataTypeId.BIT_VECTOR) {
      result = "bool";
    }
    if (targetType == DataTypeId.LOGIC_INTEGER
        && (sourceType == DataTypeId.LOGIC_BOOLEAN || sourceType == DataTypeId.BIT_VECTOR)) {
      result = "toint";
    }
    return result;
  }

  /**
   * Generates clock assignment.
   * 
   * @param clock - the specified clock variable.
   * @return Clock assignment in SMV format.
   */
  public static String getClockAssignment(final NodeVariable clock) {
    final StringBuilder result = new StringBuilder();
    final String clockName = getName(clock);
    result.append(SmvExprPrinter.ASSIGNMENT_DECLARATION + "\n");
    result.append("  next(" + clockName + ") := ! " + clockName + ";\n");
    return result.toString();
  }
  
  /*
   * Converts a string of process identifiers into a single identifier.
   */
  public static String getFormattedName(final String name) {
    String result = name.replaceAll("[\\s\\.\\,\\(\\)\\{\\}\\:\\=\\-\\[\\]\\@]", "_");
    result = result.replaceAll("\\_+", "_");
    return result;
  }

}
