/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.smv.launcher;

import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.FileNames;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.hldd.printer.smv.HlddSmvPrinter;
import ru.ispras.retrascope.engine.smv.test.parser.SmvLogParser;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.SmvChecker;
import ru.ispras.retrascope.util.SmvChecker.Result;

import java.io.File;
import java.util.Map;

/**
 * This engine checks the generated *.smv files with the model checker.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SmvModelCheckerLauncher extends Engine {

  /** The engine ID. */
  public static final String ENGINE_ID = "smv-modelchecker-launcher";

  /** The engine output entity type. */
  public static final EntityType OUT_TYPE = EntityType.get("smv-launch");

  /** NuXMV log file name. */
  public static final String NUXMV_LOG_FILE = "nuxmv_run.log";

  // Command file for NuXMV.
  private static final String BDD_COMMANDS = "nuxmv/nuxmv_commands_bdd.txt";
  private static final String BMC_COMMANDS = "nuxmv/nuxmv_commands_bmc.txt";

  public static final Parameter CHECK_METHOD = new Parameter("check-method", new String[] {"bdd",
      "bmc"}, "Model checking method");

  /**
   * Constructs a new engine object.
   */
  public SmvModelCheckerLauncher() {
    super(ENGINE_ID);
    addParameter(CHECK_METHOD);
    addInputType(HlddSmvPrinter.OUT_TYPE);
    setOutputType(OUT_TYPE);
  }

  @Override
  public Entity start(Map<EntityType, Entity> inputs) {

    final FileNames result = new FileNames(OUT_TYPE.getId());
    final SmvChecker checker = SmvChecker.getChecker();
    checker.setNuXmv();

    try {
      final FileNames smvFiles = (FileNames) inputs.get(HlddSmvPrinter.OUT_TYPE);
      final Arguments arguments =
          this.parseCommandLine(((Arguments) inputs.get(Arguments.TYPE)).getCommandLine());
      final String checkType = arguments.getValue(CHECK_METHOD, "bmc").toLowerCase();
      String commands = "";

      if (checkType.equals("bdd")) {
        commands = BDD_COMMANDS;
      } else if (checkType.equals("bmc")) {
        commands = BMC_COMMANDS;
      } else {
        throw new RetrascopeRuntimeException("Unknown model checking method: " + checkType);
      }

      final String commandFile =
          new File(this.getClass().getClassLoader().getResource(commands).getFile())
              .getAbsolutePath();

      for (String file : smvFiles.getArray()) {
        checker.runChecker(file, "-source", commandFile);
        if (checker.getResult() == Result.OK && checker.getLogFile() != null) {
          result.add(NUXMV_LOG_FILE);
        }
        Log.printInfo("Checker result: " + checker.getResult().name());
        if (checker.getResult() == Result.OK) {
          printLaunchInformation(file);
        }
      }

    } catch (Exception e) {
      e.printStackTrace();
    }

    return result;
  }

  /*
   * Prints model checking information.
   */
  private void printLaunchInformation(final String fileName) {
    final SmvLogParser parser = new SmvLogParser();
    parser.parseCheckerLog(fileName + "log");
    parser.parseSpecifications(fileName);
    Log.printInfo("Specifications total: " + parser.getOverallSpecs());
    Log.printInfo("Counterexamples generated: " + parser.getCounterexamples());
    Log.printInfo("Specifications proven TRUE: " + parser.getTrueSpecs());
  }

}
