/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.smv.test.parser;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.SmvExprPrinter;
import ru.ispras.retrascope.result.cfginterface.CfgInterface;
import ru.ispras.retrascope.result.cfginterface.CfgSignalInterface;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.Vector;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This class is used to parse a SMV log file and construct a {@link Test} object.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SmvLogParser {

  private static final String COMMENT_PATTERN = "^\\*{3}";
  private static final String COUNTEREXAMPLE_PATTERN = "Trace Type: Counterexample";
  private static final String TRUE_SPEC_PATTERN = "(specification).+(is true)";
  private static final String SPEC_PATTERN = "[CL]TLSPEC";
  private static final String STATE_PATTERN = "(State: )\\d+\\.\\d+";
  private static final String VALUE_PATTERN = "(^\\s+[\\w_]+(\\.[\\w_]+)*) = (\\-)*([\\w_]+)";

  private final Pattern comment;
  private final Pattern counterexample;
  private final Pattern state;
  private final Pattern value;
  private final Pattern trueSpec;
  private final Pattern specification;

  private Test test;
  private Sequence sequence;
  private Vector vector;
  private Transaction transaction;

  private int trueSpecsCount;
  private int counterexamplesCount;
  private int overallSpecs;

  private Map<String, DataType> inputs;

  private enum Method {
    CE, SPECS, STATS
  }

  /**
   * Constructs a new parser object.
   */
  public SmvLogParser() {
    comment = Pattern.compile(COMMENT_PATTERN);
    counterexample = Pattern.compile(COUNTEREXAMPLE_PATTERN, Pattern.CASE_INSENSITIVE);
    state = Pattern.compile(STATE_PATTERN);
    value = Pattern.compile(VALUE_PATTERN, Pattern.CASE_INSENSITIVE);
    trueSpec = Pattern.compile(TRUE_SPEC_PATTERN);
    specification = Pattern.compile(SPEC_PATTERN);
    trueSpecsCount = 0;
    counterexamplesCount = 0;
    overallSpecs = 0;
    inputs = new HashMap<>();
  }

  /**
   * Returns the number of true specifications after model checking.
   * 
   * @return The number of true specifications.
   */
  public int getTrueSpecs() {
    return trueSpecsCount;
  }

  /**
   * Returns the number of counterexamples generated during model checking.
   * 
   * @return The number of counterexamples.
   */
  public int getCounterexamples() {
    return counterexamplesCount;
  }

  /**
   * Returns the total number of specifications.
   * 
   * @return The total number of specifications.
   */
  public int getOverallSpecs() {
    return overallSpecs;
  }

  // TODO: Replace this method after refactoring a SMV model.
  public void setSignalInterface(final CfgInterface iface) {
    InvariantChecks.checkNotNull(iface);
    inputs = new HashMap<>();
    for (Map.Entry<String, CfgSignalInterface> entry : iface.getMapping().entrySet()) {
      for (NodeVariable input : entry.getValue().getInputs()) {
        inputs.put(input.getName(), input.getDataType());
      }
    }
  }

  /**
   * Parses a file for counterexamples.
   * 
   * @param fileName - the specified log file name.
   * @return The {@link Test} object extracted from the file.
   */
  public Test parseCounterexamples(final String fileName) {
    test = new Test();
    parse(fileName, Method.CE);
    addVector();
    addSequence();
    return test;
  }

  /**
   * Parses checker output for run statistics.
   * 
   * @param fileName - the specified *.smvlog file name.
   */
  public void parseCheckerLog(final String fileName) {
    trueSpecsCount = 0;
    counterexamplesCount = 0;
    parse(fileName, Method.STATS);
  }

  /**
   * Parses a file for specification statistics.
   * 
   * @param fileName - the specified file name.
   */
  public void parseSpecifications(final String fileName) {
    overallSpecs = 0;
    parse(fileName, Method.SPECS);
  }

  /*
   * Performs file parsing depending on the specified target method. 
   */
  private void parse(final String fileName, final Method method) {
    try {
      final BufferedReader fileReader = new BufferedReader(new FileReader(fileName));
      while (fileReader.ready()) {
        final String line = fileReader.readLine();
        if (line != null) {
          switch (method) {
            case CE:
              parseLine(line);
              break;
            case STATS:
              parseStats(line);
              break;
            case SPECS:
              parseSpecs(line);
              break;
          }
        }
      }
      fileReader.close();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  /*
   * Searches for specifications. 
   */
  private void parseSpecs(final String line) {
    final Matcher specMatcher = specification.matcher(line);
    if (specMatcher.find()) {
      overallSpecs++;
    }
  }


  /*
   * Searches for a specification in the line.
   */
  private void parseStats(final String line) {
    final Matcher trueMatcher = trueSpec.matcher(line);
    final Matcher specMatcher = counterexample.matcher(line);

    if (trueMatcher.find()) {
      trueSpecsCount++;
    }
    if (specMatcher.find()) {
      counterexamplesCount++;
    }
  }

  /*
   * Parses a line.
   */
  private void parseLine(final String line) {
    final Matcher commentMatcher = comment.matcher(line);
    final Matcher specMatcher = counterexample.matcher(line);
    final Matcher stateMatcher = state.matcher(line);
    final Matcher valueMatcher = value.matcher(line);

    if (!line.isEmpty() && !commentMatcher.find()) {
      if (specMatcher.find()) {
        addVector();
        addSequence();
        sequence = new Sequence();
        vector = null;
      }

      if (stateMatcher.find()) {
        addVector();
        vector = new Vector();
      }

      if (valueMatcher.find()) {
        addTransaction(line);
      }
    }
  }

  /*
   * Adds a new transaction to the test.
   */
  private void addTransaction(final String line) {
    transaction = createTransaction(line);
    if (!transaction.isEmpty()) {
      vector.addTransaction(transaction);
    }
  }

  /*
   * Adds a new vector to the test.
   */
  private void addVector() {
    if (vector != null && !vector.isEmpty()) {
      sequence.addVector(vector);
    }
  }

  /*
   * Adds a new sequence to the test.
   */
  private void addSequence() {
    if (sequence != null && sequence.getLength() != 0 && !test.getSequences().contains(sequence)) {
      test.addSequence(sequence);
    }
  }

  /*
   * Creates a new transaction from the specified string.
   */
  private Transaction createTransaction(final String line) {
    final String[] elements = line.split(" = ");
    String varName = elements[0];
    final String value = elements[1];
    final Transaction result = new Transaction();
    varName = parseName(varName);
    if (varName != null) {
      result.addVariable(varName, parseValue(value, inputs.get(varName)));
    }
    return result;
  }

  /*
   * Parses a variable name.
   */
  private String parseName(final String varName) {
    String result = "";
    String[] values = varName.split("\\.");
    if (values.length == 1) {
      result = values[0];
      result = result.replaceAll("\\s", "");
      if (result.endsWith(SmvExprPrinter.SUB_ENDING)) {
        result = result.substring(0, result.length() - SmvExprPrinter.SUB_ENDING.length());
      }
      if (!inputs.containsKey(result)) {
        return null;
      }
    } else {
      return null;
    }
    return result;
  }

  /*
   * Parses a variable value.
   */
  private NodeValue parseValue(String string, final DataType dataType) {
    NodeValue result = null;
    String sign = "";
    if (string.matches("TRUE|FALSE")) {
      result = new NodeValue(Data.newBoolean(Boolean.valueOf(string)));
    }
    if (string.matches("\\d+")) {
      result = new NodeValue(Data.newInteger(Integer.valueOf(string)));
    }
    if (string.matches("\\-\\w+")) {
      string = string.substring(1);
      sign = "-";
    }
    if (string.matches("0ud\\d+_\\d+") || string.matches("0sd\\d+_\\d+")) {
      final String[] values = string.split("_");
      final String size = values[0].substring(3, values[0].length());
      final String value = sign + values[1];
      final int intSize = Integer.parseInt(size);
      final long intValue = Long.parseLong(value);
      if (dataType.getTypeId() == DataTypeId.LOGIC_INTEGER) {
        result = new NodeValue(Data.newInteger(intValue));
      } else {
        result = new NodeValue(Data.newBitVector(intValue, intSize));
      }
    }
    return result;
  }

}
