/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.test.printer.testbench;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Vector;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.FileAlreadyExistsException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

/**
 * The class for printing tests to files in the DECIDER-like manner.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class DeciderLikeTestPatternPrinter {
  private static final String PATTERN_HEADER = "#Pattern: ";
  private static final String SEQUENCE_HEADER = "#New sequence";
  private static final String SIGNAL_DELIMETER = " ";
  private static final String VALUE_FOR_UNSPECIFIED_SIGNALS = "X";

  private DeciderLikeTestPatternPrinter() {}

  /**
   * Prints the specified sequence to the specified file.
   *
   * @param sequence the test sequence to be printed
   * @param file the file where the test will be printed to
   * @param inputNames a set of all input signals of a tested module
   * @param overwrite indicates whether or not the target file should be overwritten if it already
   *     exists
   *
   * @throws IOException if the specified file already exists or if an I/O error occurred
   */
  public static void print(Sequence sequence, Path file, Set<String> inputNames, boolean overwrite)
      throws IOException {
    if (!overwrite && Files.exists(file)) {
      throw new FileAlreadyExistsException("The specified file already exists: " + file);
    }

    try (BufferedWriter testWriter = Files.newBufferedWriter(file, StandardCharsets.UTF_8)) {
      testWriter.append(DeciderLikeTestPatternPrinter.PATTERN_HEADER);

      for (Iterator<String> i = inputNames.iterator(); i.hasNext();) {
        testWriter.append(i.next());

        if (i.hasNext()) {
          testWriter.append(DeciderLikeTestPatternPrinter.SIGNAL_DELIMETER);
        }
      }

      testWriter.newLine();

      testWriter.append(DeciderLikeTestPatternPrinter.SEQUENCE_HEADER);
      testWriter.newLine();

      for (Vector vector: sequence.getVectors()) {
        Map<String, NodeValue> eventValues = DeciderLikeTestPatternPrinter.getEventValues(vector);
        Map<String, NodeValue> invertedEventValues =
            DeciderLikeTestPatternPrinter.invertEventValues(eventValues);

        StringBuilder firstHalfBuilder = new StringBuilder();
        StringBuilder secondHalfBuilder = new StringBuilder();

        for (Iterator<String> i = inputNames.iterator(); i.hasNext();) {
          String inputName = i.next();

          if (eventValues.containsKey(inputName)) {
            firstHalfBuilder.append(invertedEventValues.get(inputName).getData().getValue());
            secondHalfBuilder.append(eventValues.get(inputName).getData().getValue());
          } else {
            if (vector.containsVariable(inputName)) {
              Node value = vector.getValue(inputName);

              if (value instanceof NodeValue) {
                firstHalfBuilder.append(((NodeValue) value).getData().getValue());
                secondHalfBuilder.append(((NodeValue) value).getData().getValue());
              } else {
                throw new IllegalArgumentException("Unsupported value type: "
                    + value.getClass().getSimpleName());
              }
            } else {
              firstHalfBuilder.append(VALUE_FOR_UNSPECIFIED_SIGNALS);
              secondHalfBuilder.append(VALUE_FOR_UNSPECIFIED_SIGNALS);
            }
          }

          if (i.hasNext()) {
            firstHalfBuilder.append(DeciderLikeTestPatternPrinter.SIGNAL_DELIMETER);
            secondHalfBuilder.append(DeciderLikeTestPatternPrinter.SIGNAL_DELIMETER);
          }
        }

        testWriter.append(firstHalfBuilder);
        testWriter.newLine();
        testWriter.append(secondHalfBuilder);
        testWriter.newLine();
      }

      testWriter.flush();
    }
  }

  private static Map<String, NodeValue> getEventValues(Vector vector) {
    Map<String, NodeValue> eventValues = new HashMap<>();

    for (Event event: vector.getEventList().getEvents()) {
      NodeVariable variable = event.getRangedVariable().getVariable();
      String variableName = variable.getName();
      DataType dataType = variable.getDataType();

      switch (event.getType()) {
        case NEGATIVE_EDGE:
          if (dataType.getTypeId().equals(DataTypeId.LOGIC_BOOLEAN)) {
            eventValues.put(variableName, new NodeValue(Data.newBoolean(false)));
          } else {
            eventValues.put(variableName, new NodeValue(dataType.valueOf("0", 10)));
          }

          break;
        case POSITIVE_EDGE:
          if (dataType.getTypeId().equals(DataTypeId.LOGIC_BOOLEAN)) {
            eventValues.put(variableName, new NodeValue(Data.newBoolean(true)));
          } else {
            eventValues.put(variableName, new NodeValue(dataType.valueOf("1", 10)));
          }

          break;
        case ANY_EDGE:
          if (vector.containsVariable(variableName)) {
            eventValues.put(variableName, (NodeValue) vector.getValue(variableName));
          } else {
            if (dataType.getTypeId().equals(DataTypeId.LOGIC_BOOLEAN)) {
              eventValues.put(variableName, new NodeValue(Data.newBoolean(true)));
            } else {
              eventValues.put(variableName, new NodeValue(dataType.valueOf("1", 10)));
            }
          }

          break;
        default:
          throw new IllegalArgumentException("Unsupported event type: " + event.getType());
      }
    }

    return eventValues;
  }

  private static Map<String, NodeValue> invertEventValues(Map<String, NodeValue> eventValues) {
    Map<String, NodeValue> invertedEventValues = new HashMap<>();

    for (Entry<String, NodeValue> entry: eventValues.entrySet()) {
      NodeValue value = entry.getValue();
      DataType dataType = value.getDataType();

      if (dataType.getTypeId().equals(DataTypeId.LOGIC_BOOLEAN)) {
        if (value.getData().equals(Data.newBoolean(false))) {
          invertedEventValues.put(entry.getKey(), new NodeValue(Data.newBoolean(true)));
        } else {
          if (value.getData().equals(Data.newBoolean(true))) {
            invertedEventValues.put(entry.getKey(), new NodeValue(Data.newBoolean(false)));
          } else {
            throw new IllegalArgumentException("Unexpected boolean event value: " + value);
          }
        }
      } else {
        if (value.getData().equals(dataType.valueOf("0", 10))) {
          invertedEventValues.put(entry.getKey(), new NodeValue(dataType.valueOf("1", 10)));
        } else {
          if (value.getData().equals(dataType.valueOf("1", 10))) {
            invertedEventValues.put(entry.getKey(), new NodeValue(dataType.valueOf("0", 10)));
          } else {
            throw new IllegalArgumentException("Unexpected numeric event value: " + value);
          }
        }
      }
    }

    return invertedEventValues;
  }
}
