/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.test.printer.testbench.vhdl;

import org.apache.commons.cli.ParseException;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.test.printer.testbench.DeciderLikeTestPatternPrinter;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.result.cfginterface.CfgInterface;
import ru.ispras.retrascope.result.cfginterface.CfgSignalInterface;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.TestModel;
import ru.ispras.retrascope.result.testbench.Testbench;
import ru.ispras.retrascope.result.testbench.VhdlTestbenchModel;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.Charset;
import java.nio.file.FileAlreadyExistsException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;

/**
 * The engine that makes VHDL testbenches from tests.
 * 
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 * @see Test
 * @see VhdlTestbenchModel
 */
public class TestVhdlTestbenchPrinter extends Engine {

  /**
   * The engine ID of all {@code TestVhdlTestbenchPrinter} instances.
   */
  public static final String ENGINE_ID = "test-vhdl-testbench-printer";

  public static final Parameter TESTBENCH_DIRECTORY_PARAMETER = new Parameter("testbench-dir",
      true, "A directory where testbenches are saved");
  public static final Parameter OVERWRITE_EXISTING_FILES_PARAMETER = new Parameter(
      "overwrite-existing", false, "Overwrite target files if they already exist");

  private static final String MODULE_NAME_PLACEHOLDER = "%MODULE_NAME%";

  private static final String DEFAULT_TESTBENCH_DIRECTORY = "testbenches/"
      + MODULE_NAME_PLACEHOLDER;

  private static final String TESTBENCH_TEMPLATE_RESOURCE_NAME = "templates/vhdl/testbench";
  private static final String TST_PARSER_RESOURCE_NAME = "templates/vhdl/tst_parser";

  private static final String DEFAULT_CLOCK_PERIOD_VALUE = "10";
  private static final String DEFAULT_CLOCK_PERIOD_UNIT = "ns";

  private static final String CLOCK_PERIOD_VALUE_PLACEHOLDER = "%PERIOD_VALUE%";
  private static final String CLOCK_PERIOD_UNIT_PLACEHOLDER = "%PERIOD_UNIT%";
  private static final String TEST_FILE_NAME_PLACEHOLDER = "%TEST_FILE_NAME%";
  private static final String PORT_DECLARATION_PLACEHOLDER = "%PORT_DECLARATION%";
  private static final String SIGNAL_DECLARATION_PLACEHOLDER = "%SIGNAL_DECLARATION%";
  private static final String BUFFER_VARIABLE_DECLARATION_PLACEHOLDER =
      "%BUFFER_VARIABLE_DECLARATION%";
  private static final String PORT_MAP_DECLARATION_PLACEHOLDER = "%PORT_MAP_DECLARATION%";
  private static final String TABULATION_PLACEHOLDER = "%TABULATION%";
  private static final String PORT_NAME_PLACEHOLDER = "%PORT_NAME%";
  private static final String PORT_KIND_PLACEHOLDER = "%PORT_KIND%";
  private static final String PORT_TYPE_PLACEHOLDER = "%PORT_TYPE%";
  private static final String SIGNAL_NAME_PLACEHOLDER = "%SIGNAL_NAME%";
  private static final String SIGNAL_TYPE_PLACEHOLDER = "%SIGNAL_TYPE%";
  private static final String MAX_LINE_LENGTH_PLACEHOLDER = "%MAX_LINE_LENGTH%";
  private static final String MAX_VALUE_LENGTH_PLACEHOLDER = "%MAX_VALUE_LENGTH%";
  private static final String SIGNAL_ASSIGNMENT_PLACEHOLDER = "%SIGNAL_ASSIGNMENT%";
  private static final String BUFFER_VARIABLE_ASSIGNMENT_PLACEHOLDER =
      "%BUFFER_VARIABLE_ASSIGNMENT%";
  private static final String CONVERSION_FUNCTION_PLACEHOLDER = "%CONVERSION_FUNCTION%";
  private static final String BUFFER_VARIABLE_NAME_POSTFIX = "_buffer";

  private static final String PORT_DECLARATION = TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER
      + TestVhdlTestbenchPrinter.PORT_NAME_PLACEHOLDER + ": "
      + TestVhdlTestbenchPrinter.PORT_KIND_PLACEHOLDER + " "
      + TestVhdlTestbenchPrinter.PORT_TYPE_PLACEHOLDER;
  private static final String SIGNAL_DECLARATION = TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER
      + "signal " + TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER + ": "
      + TestVhdlTestbenchPrinter.SIGNAL_TYPE_PLACEHOLDER + ";";
  private static final String BUFFER_VARIABLE_DECLARATION =
      TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER + "variable "
          + TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER
          + TestVhdlTestbenchPrinter.BUFFER_VARIABLE_NAME_POSTFIX + ": "
          + TestVhdlTestbenchPrinter.SIGNAL_TYPE_PLACEHOLDER + ";";
  private static final String PORT_MAP_DECLARATION =
      TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER
          + TestVhdlTestbenchPrinter.PORT_NAME_PLACEHOLDER + " => "
          + TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER;

  private static final String SIGNAL_ASSIGNMENT = TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER
      + TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER + " <= "
      + TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER
      + TestVhdlTestbenchPrinter.BUFFER_VARIABLE_NAME_POSTFIX + ";";

  private static final String BUFFER_VARIABLE_ASSIGNMENT =
      TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER
          + TestVhdlTestbenchPrinter.CONVERSION_FUNCTION_PLACEHOLDER + "(test_line_string, "
          + TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER
          + TestVhdlTestbenchPrinter.BUFFER_VARIABLE_NAME_POSTFIX + ");";

  private static final String INPUT_PORT_KIND = "in";
  private static final String OUTPUT_PORT_KIND = "out";

  private static final String STD_LOGIC_VECTOR_PATTERN =
      "std_logic_vector (%LOWER_BOUND% to %UPPER_BOUND%)";
  private static final String UPPER_BOUND_PLACEHOLDER = "%UPPER_BOUND%";
  private static final String LOWER_BOUND_PLACEHOLDER = "%LOWER_BOUND%";

  private static final Map<DataType, String> DATA_TYPES_MAP = new HashMap<>();

  static {
    DATA_TYPES_MAP.put(DataType.BIT_VECTOR(1), "std_logic");
    DATA_TYPES_MAP.put(DataType.INTEGER, "integer");
    DATA_TYPES_MAP.put(DataType.BOOLEAN, "bit");
    DATA_TYPES_MAP.put(DataType.REAL, "real");
  }

  private static final Map<String, String> CONVERSION_FUNCTIONS_MAP = new TreeMap<>(
      String.CASE_INSENSITIVE_ORDER);

  static {
    CONVERSION_FUNCTIONS_MAP.put("std_logic", "string_to_std_logic");
    CONVERSION_FUNCTIONS_MAP.put("std_logic_vector", "string_to_std_logic_vector");
    CONVERSION_FUNCTIONS_MAP.put("bit", "string_to_bit");
    CONVERSION_FUNCTIONS_MAP.put("bit_vector", "string_to_bit_vector");
    CONVERSION_FUNCTIONS_MAP.put("integer", "string_to_integer");
    CONVERSION_FUNCTIONS_MAP.put("real", "string_to_real");
  }

  private static String getTestFileName(String moduleName, int testNumber) {
    StringBuilder testFileNameBuilder = new StringBuilder(moduleName).append("_test");

    if (testNumber > 0) {
      testFileNameBuilder.append('_').append(testNumber);
    }

    return testFileNameBuilder.append(".tst").toString();
  }

  private static String getTestbenchFileName(String moduleName, int testNumber) {
    StringBuilder testFileNameBuilder = new StringBuilder(moduleName).append("_testbench");

    if (testNumber > 0) {
      testFileNameBuilder.append('_').append(testNumber);
    }

    return testFileNameBuilder.append(".vhd").toString();
  }

  private static String getTstParserFileName(String moduleName, int testNumber) {
    StringBuilder testFileNameBuilder = new StringBuilder(moduleName).append("_tst_parser");

    if (testNumber > 0) {
      testFileNameBuilder.append('_').append(testNumber);
    }

    return testFileNameBuilder.append(".vhd").toString();
  }

  /**
   * Creates a new {@code TestVhdlTestbenchPrinter} instance.
   */
  public TestVhdlTestbenchPrinter() {
    super(ENGINE_ID, TestModel.TYPE, VhdlTestbenchModel.TYPE);
    super.addInputType(CfgInterface.TYPE);
    super.addParameter(TESTBENCH_DIRECTORY_PARAMETER);
    super.addParameter(OVERWRITE_EXISTING_FILES_PARAMETER);
  }

  @Override
  public VhdlTestbenchModel start(Map<EntityType, Entity> inputs) {
    TestModel testModel = getMandatoryInputEntity(inputs, TestModel.class, TestModel.TYPE);
    CfgInterface cfgInterface =
        getMandatoryInputEntity(inputs, CfgInterface.class, CfgInterface.TYPE);

    final List<Test> tests = testModel.getTests();
    final Map<String, CfgSignalInterface> interfaceMapping = cfgInterface.getMapping();

    if (tests.size() != interfaceMapping.size()) {
      throw new IllegalArgumentException("Amount of tests (" + tests.size() + ") isn't equal to "
          + "amount of interfaces (" + interfaceMapping.size() + ")");
    }

    String testbenchDirectory = this.getTestbenchDirectory(inputs);

    List<Set<Testbench>> testbenches = new ArrayList<>(tests.size());

    Iterator<Test> testIterator = tests.iterator();
    Iterator<Entry<String, CfgSignalInterface>> entryIterator =
        interfaceMapping.entrySet().iterator();

    Arguments arguments = Engine.getOptionalInputEntity(inputs, Arguments.class, Arguments.TYPE);
    boolean overwrite = false;

    if (arguments != null) {
      try {
        Arguments parsedArguments = super.parseCommandLine(arguments.getCommandLine());
        overwrite = parsedArguments.contains(OVERWRITE_EXISTING_FILES_PARAMETER);
      } catch (ParseException e) {
        throw new RetrascopeRuntimeException("The exception occurred while parsing command line"
            , e);
      }
    }

    while (testIterator.hasNext() && entryIterator.hasNext()) {
      Entry<String, CfgSignalInterface> interfaceEntry = entryIterator.next();
      Test test = testIterator.next();

      Set<Testbench> testbenchSet = new HashSet<>();
      testbenchSet.add(this.printTestbench(test, interfaceEntry.getKey(),
          interfaceEntry.getValue(), testbenchDirectory, overwrite));
      testbenches.add(testbenchSet);
    }

    return new VhdlTestbenchModel(testbenches);
  }

  private String getTestbenchDirectory(Map<EntityType, Entity> inputs) {
    String testbenchDirectory = DEFAULT_TESTBENCH_DIRECTORY;
    Arguments arguments = Engine.getOptionalInputEntity(inputs, Arguments.class, Arguments.TYPE);

    try {
      if (arguments != null) {
        Arguments parsedArguments = super.parseCommandLine(arguments.getCommandLine());
        testbenchDirectory =
            parsedArguments.getValue(TESTBENCH_DIRECTORY_PARAMETER, testbenchDirectory);
      }
    } catch (ParseException e) {
      throw new RetrascopeRuntimeException("The exception occurred while parsing command line", e);
    }

    return testbenchDirectory;
  }

  private Testbench printTestbench(Test test, String moduleName,
      CfgSignalInterface externalSignals, String testbenchDirectoryString, boolean overwrite) {
    String normalisedModuleName = this.normaliseModuleName(moduleName);

    Path testbenchDirectory =
        Paths.get(this.processFileName(testbenchDirectoryString, normalisedModuleName));

    try {
      Files.createDirectories(testbenchDirectory);
    } catch (IOException exception) {
      throw new RetrascopeRuntimeException(
          "The exception has occurred while creating the target directory", exception);
    }

    Collection<NodeVariable> inputSignals = externalSignals.getInputs();
    Set<String> inputSignalNames = new LinkedHashSet<>(inputSignals.size());

    for (NodeVariable input : inputSignals) {
      inputSignalNames.add(input.getName());
    }

    Set<Path> additionalFiles = new HashSet<>();

    int valueLength = 0;
    int lineLength = 0;
    int sequenceNumber = 0;

    Path testPatternFile =
        testbenchDirectory.resolve(getTestFileName(normalisedModuleName, sequenceNumber));

    for (Sequence sequence : test.getSequences()) {
      Path deciderFile =
          testbenchDirectory.resolve(getTestFileName(normalisedModuleName, sequenceNumber));
      try {
        DeciderLikeTestPatternPrinter.print(sequence, deciderFile, inputSignalNames, overwrite);
      } catch (IOException e) {
        throw new RetrascopeRuntimeException(
            "The exception has occurred while printing test pattern file", e);
      }
      sequenceNumber++;

      additionalFiles.add(deciderFile);

      try (BufferedReader testReader =
          Files.newBufferedReader(deciderFile, Charset.defaultCharset())) {

        int maxLineLength = 0;
        int maxValueLength = 0;

        for (String line = testReader.readLine(); line != null; line = testReader.readLine()) {
          if (line.length() > maxLineLength) {
            maxLineLength = line.length();
          }

          String[] values = line.split(" ");

          for (String value : values) {
            if (value.length() > maxValueLength) {
              maxValueLength = value.length();
            }
          }

          if (maxLineLength > lineLength) {
            lineLength = maxLineLength;
          }
          if (maxValueLength > valueLength) {
            valueLength = maxValueLength;
          }
        }
      } catch (IOException exception) {
        String exceptionMessage = "The exception has occurred while reading test pattern file";

        throw new RetrascopeRuntimeException(exceptionMessage, exception);
      }

    }

    try {
      Path tstParserFile =
          this.processTstParser(testbenchDirectory, normalisedModuleName, valueLength, 0
              , overwrite);
      Path testbenchFile =
          this.processTestbenchFile(testbenchDirectory, normalisedModuleName, testPatternFile,
              lineLength, valueLength, externalSignals, inputSignalNames, 0, overwrite);

      additionalFiles.add(tstParserFile);

      return new Testbench(testbenchFile, additionalFiles);
    } catch (IOException e) {
      throw new RetrascopeRuntimeException("The exception has occurred while printing a file", e);
    }
  }

  private Path processTstParser(Path testbenchDirectory, String moduleName, int maxValueLength,
      int testNumber, boolean overwrite) throws FileAlreadyExistsException {
    String fileName = getTstParserFileName(moduleName, testNumber);
    Path tstParserFile = testbenchDirectory.resolve(fileName);

    if (!overwrite && Files.exists(tstParserFile)) {
      throw new FileAlreadyExistsException("The specified file already exists: " + tstParserFile);
    }

    ClassLoader classLoader = this.getClass().getClassLoader();
    InputStream tstParserResourceStream = classLoader.getResourceAsStream(TST_PARSER_RESOURCE_NAME);

    try (BufferedReader tstParserReader =
        new BufferedReader(new InputStreamReader(tstParserResourceStream));
        BufferedWriter tstParserWriter =
            Files.newBufferedWriter(tstParserFile, Charset.defaultCharset())) {
      for (String line = tstParserReader.readLine(); line != null; line =
          tstParserReader.readLine()) {
        line = line.replace(MAX_VALUE_LENGTH_PLACEHOLDER, String.valueOf(maxValueLength));
        tstParserWriter.write(line);
        tstParserWriter.newLine();
      }

      tstParserWriter.flush();
    } catch (IOException exception) {
      String exceptionMessage = "The exception occurred while copying writing parser library";

      throw new RetrascopeRuntimeException(exceptionMessage, exception);
    }

    return tstParserFile;
  }

  private Path processTestbenchFile(Path testbenchDirectory, String moduleName,
      Path testPatternFile, int maxLineLength, int maxValueLength,
      CfgSignalInterface externalSignals, Set<String> inputSignalNames, int testNumber,
      boolean overwrite) throws FileAlreadyExistsException {
    String fileName = getTestbenchFileName(moduleName, testNumber);
    Path testbenchFile = testbenchDirectory.resolve(fileName);

    if (!overwrite && Files.exists(testbenchFile)) {
      throw new FileAlreadyExistsException("The specified file already exists: " + testbenchFile);
    }

    InputStream testbenchTemplateResourceStream =
        this.getClass().getClassLoader().getResourceAsStream(TESTBENCH_TEMPLATE_RESOURCE_NAME);

    try (BufferedReader templateReader =
        new BufferedReader(new InputStreamReader(testbenchTemplateResourceStream));
        BufferedWriter testbenchWriter =
            Files.newBufferedWriter(testbenchFile, Charset.defaultCharset())) {
      Map<String, VhdlSignal> signalMapping = this.mapExternalSignals(externalSignals);

      for (String templateLine = templateReader.readLine(); templateLine != null; templateLine =
          templateReader.readLine()) {

        String testbenchLine =
            templateLine.replace(MODULE_NAME_PLACEHOLDER, moduleName)
                .replace(TEST_FILE_NAME_PLACEHOLDER, testPatternFile.getFileName().toString())
                .replace(CLOCK_PERIOD_UNIT_PLACEHOLDER, DEFAULT_CLOCK_PERIOD_UNIT)
                .replace(CLOCK_PERIOD_VALUE_PLACEHOLDER, DEFAULT_CLOCK_PERIOD_VALUE)
                .replace(MAX_LINE_LENGTH_PLACEHOLDER, String.valueOf(maxLineLength))
                .replace(MAX_VALUE_LENGTH_PLACEHOLDER, String.valueOf(maxValueLength));

        if (testbenchLine.contains(TestVhdlTestbenchPrinter.PORT_DECLARATION_PLACEHOLDER)) {
          String tabulation =
              testbenchLine.substring(0, testbenchLine.indexOf(PORT_DECLARATION_PLACEHOLDER));

          for (String portDeclaration : this.buildPortDeclaration(tabulation, externalSignals)) {
            testbenchWriter.append(portDeclaration);
            testbenchWriter.newLine();
          }

          continue;
        }

        if (testbenchLine.contains(TestVhdlTestbenchPrinter.SIGNAL_DECLARATION_PLACEHOLDER)) {
          String tabulation =
              testbenchLine.substring(0, testbenchLine.indexOf(SIGNAL_DECLARATION_PLACEHOLDER));

          for (String signalDeclaration : this.buildSignalDeclaration(tabulation,
              signalMapping.values())) {
            testbenchWriter.append(signalDeclaration);
            testbenchWriter.newLine();
          }

          continue;
        }

        if (testbenchLine.contains(BUFFER_VARIABLE_DECLARATION_PLACEHOLDER)) {
          String tabulation =
              testbenchLine.substring(0,
                  testbenchLine.indexOf(BUFFER_VARIABLE_DECLARATION_PLACEHOLDER));

          for (String signalDeclaration : this.buildBufferVariableDeclaration(tabulation,
              signalMapping.values())) {
            testbenchWriter.append(signalDeclaration);
            testbenchWriter.newLine();
          }

          continue;
        }

        if (testbenchLine.contains(PORT_MAP_DECLARATION_PLACEHOLDER)) {
          String tabulation =
              testbenchLine.substring(0, testbenchLine.indexOf(PORT_MAP_DECLARATION_PLACEHOLDER));

          for (String signalDeclaration : this.buildPortMapDeclaration(tabulation, signalMapping)) {
            testbenchWriter.append(signalDeclaration);
            testbenchWriter.newLine();
          }

          continue;
        }

        if (testbenchLine.contains(SIGNAL_ASSIGNMENT_PLACEHOLDER)) {
          String tabulation =
              testbenchLine.substring(0, testbenchLine.indexOf(SIGNAL_ASSIGNMENT_PLACEHOLDER));

          for (String inputSignalName : inputSignalNames) {
            testbenchWriter.append(this.buildSignalAssignemnt(tabulation,
                signalMapping.get(inputSignalName)));
            testbenchWriter.newLine();
          }

          continue;
        }

        if (testbenchLine.contains(BUFFER_VARIABLE_ASSIGNMENT_PLACEHOLDER)) {
          String tabulation =
              testbenchLine.substring(0,
                  testbenchLine.indexOf(BUFFER_VARIABLE_ASSIGNMENT_PLACEHOLDER));

          for (String inputSignalName : inputSignalNames) {
            testbenchWriter.append(this.buildBufferVariableAssignemnt(tabulation,
                signalMapping.get(inputSignalName)));
            testbenchWriter.newLine();
          }

          continue;
        }

        testbenchWriter.append(testbenchLine);
        testbenchWriter.newLine();
      }

      testbenchWriter.flush();
    } catch (IOException exception) {
      String exceptionMessage = "The exception has occurred while processing testbench file";

      throw new RetrascopeRuntimeException(exceptionMessage, exception);
    }

    return testbenchFile;
  }

  private List<String> buildPortDeclaration(String tabulation, CfgSignalInterface externalSignals) {
    List<String> portDeclaration = new ArrayList<>();

    for (NodeVariable input : externalSignals.getInputs()) {
      if (!portDeclaration.isEmpty()) {
        int lastIndex = portDeclaration.size() - 1;
        portDeclaration.set(lastIndex, portDeclaration.get(lastIndex) + ";");
      }

      portDeclaration.add(PORT_DECLARATION.replace(TABULATION_PLACEHOLDER, tabulation)
          .replace(PORT_NAME_PLACEHOLDER, input.getName())
          .replace(PORT_KIND_PLACEHOLDER, INPUT_PORT_KIND)
          .replace(PORT_TYPE_PLACEHOLDER, this.mapDataType(input)));
    }

    for (NodeVariable output : externalSignals.getOutputs()) {
      if (!portDeclaration.isEmpty()) {
        int lastIndex = portDeclaration.size() - 1;
        portDeclaration.set(lastIndex, portDeclaration.get(lastIndex) + ";");
      }

      portDeclaration.add(PORT_DECLARATION.replace(TABULATION_PLACEHOLDER, tabulation)
          .replace(PORT_NAME_PLACEHOLDER, output.getName())
          .replace(PORT_KIND_PLACEHOLDER, OUTPUT_PORT_KIND)
          .replace(PORT_TYPE_PLACEHOLDER, this.mapDataType(output)));
    }

    return portDeclaration;
  }

  private List<String> buildSignalDeclaration(String tabulation, Collection<VhdlSignal> signals) {
    List<String> signalDeclaration = new ArrayList<>();

    for (VhdlSignal signal : signals) {
      signalDeclaration.add(SIGNAL_DECLARATION.replace(TABULATION_PLACEHOLDER, tabulation)
          .replace(SIGNAL_NAME_PLACEHOLDER, signal.name)
          .replace(SIGNAL_TYPE_PLACEHOLDER, signal.type));
    }

    return signalDeclaration;
  }

  private List<String> buildBufferVariableDeclaration(String tabulation,
      Collection<VhdlSignal> signals) {
    List<String> signalDeclaration = new ArrayList<>();

    for (VhdlSignal signal : signals) {
      signalDeclaration.add(BUFFER_VARIABLE_DECLARATION.replace(TABULATION_PLACEHOLDER, tabulation)
          .replace(SIGNAL_NAME_PLACEHOLDER, signal.name)
          .replace(SIGNAL_TYPE_PLACEHOLDER, signal.type));
    }

    return signalDeclaration;
  }

  private List<String> buildPortMapDeclaration(String tabulation, Map<String, VhdlSignal> portMap) {
    List<String> portMapDeclaration = new ArrayList<>();

    for (Map.Entry<String, VhdlSignal> mapping : portMap.entrySet()) {
      if (!portMapDeclaration.isEmpty()) {
        int lastIndex = portMapDeclaration.size() - 1;
        portMapDeclaration.set(lastIndex, portMapDeclaration.get(lastIndex) + ",");
      }

      portMapDeclaration.add(PORT_MAP_DECLARATION.replace(TABULATION_PLACEHOLDER, tabulation)
          .replace(PORT_NAME_PLACEHOLDER, mapping.getKey())
          .replace(SIGNAL_NAME_PLACEHOLDER, mapping.getValue().name));
    }

    return portMapDeclaration;
  }

  private String mapDataType(NodeVariable variable) {
    VariableDescriptor variableDescriptor = (VariableDescriptor) variable.getUserData();

    if (variableDescriptor != null) {
      String vhdlDataType = variableDescriptor.getMetaInfo(MetaInfoType.HDL_TYPE_STR.name());

      if (vhdlDataType != null) {
        return vhdlDataType;
      }
    }

    DataType dataType = variable.getDataType();

    if (TestVhdlTestbenchPrinter.DATA_TYPES_MAP.containsKey(dataType)) {
      return TestVhdlTestbenchPrinter.DATA_TYPES_MAP.get(dataType);
    } else {
      if (dataType.getTypeId().equals(DataTypeId.BIT_VECTOR)) {
        return STD_LOGIC_VECTOR_PATTERN.replace(UPPER_BOUND_PLACEHOLDER,
            String.valueOf(dataType.getSize() - 1)).replace(LOWER_BOUND_PLACEHOLDER, "0");
      } else {
        throw new IllegalArgumentException("Unsupported data type: " + dataType);
      }
    }
  }

  private Map<String, VhdlSignal> mapExternalSignals(CfgSignalInterface externalSignals) {
    Collection<NodeVariable> inputSignals = externalSignals.getInputs();
    Collection<NodeVariable> outputSignals = externalSignals.getOutputs();

    Map<String, VhdlSignal> mapping = new HashMap<>(inputSignals.size() + outputSignals.size());

    for (NodeVariable input : inputSignals) {
      mapping.put(input.getName(), new VhdlSignal(input.getName(), this.mapDataType(input)));
    }

    for (NodeVariable output : outputSignals) {
      mapping.put(output.getName(), new VhdlSignal(output.getName(), this.mapDataType(output)));
    }

    return mapping;
  }

  private String buildBufferVariableAssignemnt(String tabulation, VhdlSignal signal) {
    String assignment =
        BUFFER_VARIABLE_ASSIGNMENT.replace(TABULATION_PLACEHOLDER, tabulation).replace(
            SIGNAL_NAME_PLACEHOLDER, signal.name);

    String signalCoreType = signal.type.split(" |\\(")[0];

    if (CONVERSION_FUNCTIONS_MAP.containsKey(signalCoreType)) {
      assignment =
          assignment.replace(CONVERSION_FUNCTION_PLACEHOLDER,
              CONVERSION_FUNCTIONS_MAP.get(signalCoreType));
    } else {
      throw new RetrascopeRuntimeException("Unsupported signal type: " + signal.type);
    }

    return assignment;
  }

  private String buildSignalAssignemnt(String tabulation, VhdlSignal signal) {
    return SIGNAL_ASSIGNMENT.replace(TABULATION_PLACEHOLDER, tabulation).replace(
        SIGNAL_NAME_PLACEHOLDER, signal.name);
  }

  private class VhdlSignal {
    public final String name;
    public final String type;

    public VhdlSignal(String name, String type) {
      this.name = name;
      this.type = type;
    }
  }

  private String normaliseModuleName(String moduleName) {
    InvariantChecks.checkNotNull(moduleName);

    String[] splittedName = moduleName.split("\\.");

    if (splittedName.length == 1) {
      return moduleName;
    } else {
      return splittedName[splittedName.length - 1].replaceAll("\\(.*\\)", "");
    }
  }

  private String processFileName(String name, String moduleName) {
    return name.replace(MODULE_NAME_PLACEHOLDER, moduleName);
  }
}
