/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.jaxb.JaxbEventAdapter;

import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;

/**
 * Variable (signal) occurrence.
 *
 * <p>Data that characterizes changing of some variable (signal) level.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 *
 */
@XmlJavaTypeAdapter(JaxbEventAdapter.class)
public class Event extends MetaInfo {

  private final EventType type;

  private final RangedVariable variable;

  /**
   * Constructs an object with the specified variable and event type.
   *
   * @param variable Ranged variable.
   * @param type Event type.
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public Event(final RangedVariable variable, final EventType type) {
    super();
    InvariantChecks.checkNotNull(variable);
    InvariantChecks.checkNotNull(type);

    this.type = type;
    this.variable = variable;
  }

  /**
   * Constructs an object with the specified variable and of any edge event type.
   *
   * @param variable Ranged variable.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public Event(final RangedVariable variable) {
    super();
    InvariantChecks.checkNotNull(variable);

    this.variable = variable;
    this.type = EventType.ANY_EDGE;
  }

  /**
   * Returns type.
   *
   * @return Event type.
   */
  public EventType getType() {
    return this.type;
  }

  /**
   * Returns variable.
   *
   * @return Ranged variable.
   */
  public RangedVariable getRangedVariable() {
    return this.variable;
  }

  /**
   * Checks whether this object is triggered by the specified one.
   *
   * <p>Checks if this event is always triggered along with the specified one (for example, if this
   * event is any edge of some variable while other is positive edge of the same variable).</p>
   *
   * @param event Event to be tested.
   *
   * @return {@code true} if this event is always triggered along with the specified one, otherwise
   *         returns {@code false}.
   */
  public boolean isTriggered(final Event event) {
    return this.variable.equals(event.variable)
        && (this.type == event.type || this.type == EventType.ANY_EDGE);
  }

  /**
   * Checks whether this object is compatible with the specified one.
   *
   * <p>Checks if this event is compatible with the specified one, that is, they aren't mutually
   * exclusive.</p>
   *
   * @param event Event to be tested for compatibility with this one.
   *
   * @return {@code true} if the events are compatible, otherwise returns {@code false}.
   */
  public boolean isCompatible(final Event event) {
    return !this.variable.equals(event.variable)
        || this.type == event.type
        || this.type == EventType.ANY_EDGE
        || event.type == EventType.ANY_EDGE;
  }

  /**
   * Returns a deep copy of this object.
   *
   * @return Deep copy of this object.
   */
  public Event deepCopy() {
    return new Event(this.variable.deepCopy(), this.type);
  }

  @Override
  public String toString() {
    String result;
    if (this.getType() == EventType.ANY_EDGE) {
      result = this.getRangedVariable().toString();
    } else {
      result = this.getType().toCompactString() + " of " + this.getRangedVariable().toString();
    }
    return result;
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }

    final Event event = (Event) obj;

    return this.type == event.type && this.variable.equals(event.variable);
  }

  @Override
  public int hashCode() {
    int result = this.type.hashCode();
    result = 31 * result + this.variable.hashCode();
    return result;
  }
}
