/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.data.DataType;

import java.util.Collection;
import java.util.Map;

/**
 * The abstract representation of a device model encapsulating only a static structure of an
 * original device.
 * 
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public abstract class Model extends MetaInfo {
  /**
   * Adds a new input with the specified name and type to this model.
   * 
   * @param name    the name of the input to be added
   * @param type    the type of the input to be added
   */
  public abstract void addInput(String name, DataType type);

  /**
   * Adds a new inner variable with the specified name and type to this model.
   * 
   * @param name    the name of the variable to be added
   * @param type    the type of the variable to be added
   */
  public abstract void addInnerVariable(String name, DataType type);

  /**
   * Adds a new output with the specified name and type to this model.
   * 
   * @param name    the name of the output to be added
   * @param type    the type of the output to be added
   */
  public abstract void addOutput(String name, DataType type);

  /**
   * Checks if this model has an input with the specified name.
   * 
   * @param name    the name to be checked
   * 
   * @return {@code true} if this model has the input, otherwise returns {@code false}
   */
  public abstract boolean containsInput(String name);

  /**
   * Checks if this model has an inner variable with the specified name.
   * 
   * @param name    the name to be checked
   * 
   * @return {@code true} if this model has the inner variable, otherwise returns {@code false}
   */
  public abstract boolean containsInnerVariable(String name);

  /**
   * Checks if this model has an output with the specified name.
   * 
   * @param name    the name to be checked
   * 
   * @return {@code true} if this model has the output, otherwise returns {@code false}
   */
  public abstract boolean containsOutput(String name);

  /**
   * Returns a data type of the specified input variable of this model.
   * 
   * @param name    a name of the input
   * 
   * @return the data type of the specified input
   * 
   * @throws IllegalArgumentException if the specified variable is not found among the model inputs
   */
  public abstract DataType getInputDataType(String name);

  /**
   * Returns data types of the specified input variables of this model.
   * 
   * @param names   a collection containing names of the inputs
   * 
   * @return a map containing mappings between names and data types of the specified inputs
   *         
   * @throws IllegalArgumentException if at least on of the specified variables is not found
   *         among the model inputs
   */
  public abstract Map<String, DataType> getInputDataTypes(Collection<String> names);

  /**
   * Returns data types of all input variables of this model.
   * 
   * @return a map containing mappings between names and data types of the inputs
   */
  public abstract Map<String, DataType> getInputDataTypes();

  /**
   * Returns a data type of the specified inner variable of this model.
   * 
   * @param name    a name of the variable
   * 
   * @return the data type of the specified variable
   * 
   * @throws IllegalArgumentException if the specified variable is not found among the model inner
   *         variables
   */
  public abstract DataType getInnerDataType(String name);

  /**
   * Returns data types of the specified inner variables of this model.
   * 
   * @param names   a collection containing names of the variables
   * 
   * @return a map containing mappings between names and data types of the specified variables
   *         
   * @throws IllegalArgumentException if at least on of the specified variables is not found among
   *         the model inner variables
   */
  public abstract Map<String, DataType> getInnerDataTypes(Collection<String> names);

  /**
   * Returns data types of all inner variables of this model.
   * 
   * @return a map containing mappings between names and data types of the inner variables
   */
  public abstract Map<String, DataType> getInnerDataTypes();

  /**
   * Returns a data type of the specified output of this model.
   * 
   * @param name    a name of the output
   * 
   * @return the data type of the specified output
   * 
   * @throws IllegalArgumentException if the specified output is not found among the model outputs
   */
  public abstract DataType getOutputDataType(String name);

  /**
   * Returns data types of the specified output variables of this model.
   * 
   * @param names   a collection containing names of the outputs
   * 
   * @return a map containing mappings between names and data types of the specified outputs
   *         
   * @throws IllegalArgumentException if at least on of the specified signals is not found among the
   *         model outputs
   */
  public abstract Map<String, DataType> getOutputDataTypes(Collection<String> names);

  /**
   * Returns data types of all output signals of this model.
   * 
   * @return a map containing mappings between names and data types of the outputs
   */
  public abstract Map<String, DataType> getOutputDataTypes();

  /**
   * Returns an initial state of this model.
   * 
   * @return the initial state of this model
   */
  public abstract State getInitialState();
}
