/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis;

import org.apache.commons.collections15.Closure;
import org.apache.commons.collections15.CollectionUtils;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Variable declarations container.
 *
 * <p>Abstract class that contains {@link VariableDeclaration} variable declarations
 * and bindings between internal variables and external expressions.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class VariableContainer extends MetaInfo {

  /**
   * Prefix for new variables that are created by the tool.
   */
  public static final String NEW_VAR_PREFIX = "NEW_VAR_";

  /* Variable declarations. */
  private final Map<NodeVariable, VariableDeclaration> variables;

  /* Bindings between variables and external expressions. */
  private final Map<String, Node> bindings;

  /**
   * Default constructor.
   */
  public VariableContainer() {
    super();

    this.variables = new LinkedHashMap<>();
    this.bindings = new LinkedHashMap<>();
  }

  /**
   * Returns variables declarations.
   *
   * @return declarations of node variables
   */
  public Collection<VariableDeclaration> getDeclarations() {
    return this.variables.values();
  }

  /**
   * Returns declaration for the variable with the specified name.
   *
   * @param variableName Name of variable which declaration is to be returned.
   *
   * @return Declaration of the variable with the specified name
   *         or {@code null} if there is no such.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public VariableDeclaration getDeclaration(final String variableName) {
    InvariantChecks.checkNotNull(variableName);

    for (final NodeVariable variable : this.variables.keySet()) {
      if (variable.getName().equals(variableName)) {
        return this.variables.get(variable);
      }
    }

    return null;
  }

  /**
   * Returns declared variable with specified name.
   *
   * @param name name of variable to be returned
   *
   * @return declared variable with the specified name,
   *         or {@code null} if there is no such variable.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public NodeVariable getVariable(final String name) {
    InvariantChecks.checkNotNull(name);

    NodeVariable result = null;

    for (final NodeVariable variable : this.variables.keySet()) {
      if (variable.getName().equals(name)) {
        result = variable;
        break;
      }
    }

    return result;
  }

  /**
   * Returns declared variables.
   *
   * @return variables collection
   */
  public List<NodeVariable> getVariables() {
    return new LinkedList<>(this.variables.keySet());
  }

  /**
   * Returns a collection of names of the declared input signal variables.
   *
   * @return input variables names
   */
  public Set<String> getInputNames() {
    final Set<String> names = new LinkedHashSet<>();

    final Closure<NodeVariable> inputClosure = new Closure<NodeVariable>() {
      @Override
      public void execute(final NodeVariable variable) {

        final Object userData = variable.getUserData();
        if (userData instanceof VariableDescriptor
            && VariableType.isInputSignal(((VariableDescriptor) userData).getVariableType())) {
          names.add(variable.getName());
        }
      }
    };

    CollectionUtils.forAllDo(this.variables.keySet(), inputClosure);
    return names;
  }

  /**
   * Returns a collection of names of the declared output signal variables.
   *
   * @return output variables names
   */
  public Set<String> getOutputNames() {
    final Set<String> names = new LinkedHashSet<>();

    final Closure<NodeVariable> outputClosure = new Closure<NodeVariable>() {
      @Override
      public void execute(final NodeVariable variable) {

        final Object userData = variable.getUserData();
        if (userData instanceof VariableDescriptor
            && VariableType.isOutputSignal(((VariableDescriptor) userData).getVariableType())) {
          names.add(variable.getName());
        }
      }
    };

    CollectionUtils.forAllDo(this.variables.keySet(), outputClosure);
    return names;
  }

  /**
   * Returns a collection of names of the declared register variables.
   *
   * @return register variables names
   */
  public Set<String> getRegisterNames() {
    final Set<String> names = new LinkedHashSet<>();

    final Closure<NodeVariable> regClosure = new Closure<NodeVariable>() {
      @Override
      public void execute(final NodeVariable variable) {

        final Object userData = variable.getUserData();
        if (userData instanceof VariableDescriptor
            && VariableType.isRegister(((VariableDescriptor) userData).getVariableType())) {
          names.add(variable.getName());
        }
      }
    };

    CollectionUtils.forAllDo(this.variables.keySet(), regClosure);
    return names;
  }

  /**
   * Returns bindings.
   *
   * @return bindings
   */
  public Map<String, Node> getBindings() {
    return this.bindings;
  }

  /**
   * Returns data type of the declared variable with the specified name.
   *
   * @param name variable name
   *
   * @return data type if exists the declared variable with the specified name,
   *         {@code null} otherwise.
   */
  public DataType getDataType(final String name) {
    DataType dataType = null;

    for (final NodeVariable variable : this.variables.keySet()) {
      if (variable.getName().equals(name)) {
        dataType = variable.getDataType();
        break;
      }
    }

    return dataType;
  }

  /**
   * Returns initial value of the declared variable with the specified name.
   *
   * @param name variable name
   *
   * @return initial value if it exist for the declared variable with the specified name,
   *         {@code null} otherwise.
   */
  public Node getInitialValue(final String name) {
    Node initial = null;

    for (final Map.Entry<NodeVariable, VariableDeclaration> entry : this.variables.entrySet()) {
      if (entry.getKey().getName().equals(name)) {
        initial = entry.getValue().getInitialValue();
        break;
      }
    }

    return initial;
  }

  /**
   * Returns invariant of the declared variable with the specified name.
   *
   * @param name variable name
   *
   * @return invariant if it exist for the declared variable with the specified name,
   *         {@code null} otherwise.
   */
  public Node getInvariant(final String name) {
    Node invariant = null;

    for (final Map.Entry<NodeVariable, VariableDeclaration> entry : this.variables.entrySet()) {
      if (entry.getKey().getName().equals(name)) {
        invariant = entry.getValue().getInvariant();
        break;
      }
    }

    return invariant;
  }

  /**
   * Checks does this object contain the specified variable.
   *
   * @param variable variable to be checked
   *
   * @return {@code true} if module already has such variable declared, {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public boolean containsVariable(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    return this.variables.containsKey(variable);
  }

  /**
   * Checks whether object contains a variable with the specified name.
   *
   * @param name variable name
   *
   * @return {@code true} if object contains a variable with the specified name,
   *         {@code false} otherwise.
   */
  public boolean containsVariable(final String name) {

    boolean result = false;

    for (final NodeVariable variable : this.variables.keySet()) {
      if (variable.getName().equals(name)) {
        result = true;
        break;
      }
    }

    return result;
  }

  /**
   * Checks whether this object contain input signal variable with the specified name.
   *
   * @param name variable name
   *
   * @return {@code true} if this object contains an input signal variable with the specified name,
   *         {@code false} otherwise.
   */
  public boolean containsInput(final String name) {
    boolean result = false;

    for (final NodeVariable variable : this.variables.keySet()) {

      final Object userData = variable.getUserData();
      if (variable.getName().equals(name)
          && userData instanceof VariableDescriptor
          && VariableType.isInputSignal(((VariableDescriptor) userData).getVariableType())) {
        result = true;
        break;
      }
    }

    return result;
  }

  /**
   * Checks whether this object contain output signal variable with the specified name.
   *
   * @param name variable name
   *
   * @return {@code true} if this object contains an output signal variable with the specified name,
   *         {@code false} otherwise.
   */
  public boolean containsOutput(final String name) {
    boolean result = false;

    for (final NodeVariable variable : this.variables.keySet()) {

      final Object userData = variable.getUserData();
      if (variable.getName().equals(name)
          && userData instanceof VariableDescriptor
          && VariableType.isOutputSignal(((VariableDescriptor) userData).getVariableType())) {
        result = true;
        break;
      }
    }

    return result;
  }

  /**
   * Checks whether this object contain register variable with the specified name.
   *
   * @param name variable name
   *
   * @return {@code true} if this object contains a register variable with the specified name,
   *         {@code false} otherwise.
   */
  public boolean containsRegister(final String name) {
    boolean result = false;

    for (final NodeVariable variable : this.variables.keySet()) {

      final Object userData = variable.getUserData();
      if (variable.getName().equals(name)
          && userData instanceof VariableDescriptor
          && VariableType.isRegister(((VariableDescriptor) userData).getVariableType())) {
        result = true;
        break;
      }
    }

    return result;
  }

  /**
   * Declares the specified collection of variables.
   *
   * @param declarations variables declarations
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void declareVariables(final Map<NodeVariable, VariableDeclaration> declarations) {
    InvariantChecks.checkNotNull(declarations);

    for (final Map.Entry<NodeVariable, VariableDeclaration> entry : declarations.entrySet()) {
      final VariableDeclaration declaration = entry.getValue();
      declareVariable(entry.getKey(), declaration.getInitialValue(), declaration.getInvariant());
    }
  }

  /**
   * Declares non-existing variable.
   *
   * <p>Object MUST NOT contain the specified variable.
   * Check this fact by corresponding check-method first.</p>
   *
   * @param variable variable to be added
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws IllegalStateException when this object
   *         already contains such variable.
   */
  public void declareVariable(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    declareVariable(variable, null, null);
  }

  /**
   * Declares the specified variable with the specified invariant.
   *
   * <p>mObject MUST NOT contain the specified variable.
   * Check this fact by corresponding method first.</p>
   *
   * @param variable variable to be added
   * @param invariant variable invariant (can be {@code null})
   *
   * @throws IllegalArgumentException when variable argument is {@code null}.
   * @throws IllegalStateException when this module object
   *         already contains such variable.
   */
  public void declareVariable(final NodeVariable variable, final Node invariant) {
    InvariantChecks.checkNotNull(variable);

    if (invariant == null) {
      declareVariable(variable);
    } else {
      declareVariable(variable, null, invariant);
    }
  }

  /**
   * Declares the specified variable with the specified initial value and the specified invariant.
   *
   * <p>Invariant keeps a constraint on variable value.</p>
   *
   * @param variable variable
   * @param initialValue initial value of variable
   * @param invariant variable invariant
   */
  public void declareVariable(
      final NodeVariable variable,
      final Node initialValue,
      final Node invariant) {

    if (this.variables.keySet().contains(variable)) {
      this.variables.remove(variable);
    }
    this.variables.put(variable, new VariableDeclaration(initialValue, invariant));
  }

  /**
   * Sets bindings.
   *
   * @param bindings new bindings
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void setBindings(final Map<String, Node> bindings) {
    InvariantChecks.checkNotNull(bindings);

    this.bindings.clear();
    this.bindings.putAll(bindings);
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }
    if (!super.equals(obj)) {
      return false;
    }

    final VariableContainer that = (VariableContainer) obj;

    return variables.equals(that.variables) && bindings.equals(that.bindings);
  }

  @Override
  public int hashCode() {
    int result = super.hashCode();
    result = 31 * result + variables.hashCode();
    result = 31 * result + bindings.hashCode();
    return result;
  }
}
