/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;

import java.util.Objects;

/**
 * Variable data descriptor.
 *
 * <p>Should be stored in the user data field of node variable object.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class VariableDescriptor extends MetaInfo {

  /* The variable type. */
  private final VariableType variableType;

  /* VHDL-related attributes (VHDL2008, 16.2) */

  // LEFT
  private Node left;

  // RIGHT
  private Node right;

  // LOW
  private Node low;

  // HIGH
  private Node high;

  // ASCENDING
  private Node ascending;

  // BASE
  private DataType base;

  // RANGE
  private Range range;

  // REVERSE_RANGE
  private Range reverseRange;

  // LENGTH
  private Node length;

  /**
   * Constructs object with the specified variable type.
   *
   * @param variableType Variable type.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public VariableDescriptor(final VariableType variableType) {
    super();
    InvariantChecks.checkNotNull(variableType);

    this.variableType = variableType;
  }

  /**
   * Returns variable type.
   *
   * @return Variable type of this object.
   */
  public VariableType getVariableType() {
    return this.variableType;
  }

  /**
   * Returns "'left" attribute value.
   *
   * @return Left attribute value of this object.
   */
  public Node getLeft() {
    return this.left;
  }

  /**
   * Sets the specified value to the non-initialized left field.
   *
   * @param left Value to bet set to left field.
   *
   * @throws RetrascopeRuntimeException when left field is already set to another value.
   */
  public void setLeft(Node left) {
    if (this.left != null && !this.left.equals(left)) {
      throwException(left.toString());
    }
    this.left = left;
  }

  /**
   * Returns "'right" attribute value.
   *
   * @return Right attribute value of this object.
   */
  public Node getRight() {
    return this.right;
  }

  /**
   * Sets the specified value to the non-initialized right field.
   *
   * @param right Value to bet set to right field.
   *
   * @throws RetrascopeRuntimeException when right field is already set to another value.
   */
  public void setRight(Node right) {
    if (this.right != null && ! this.right.equals(right)) {
      throwException(right.toString());
    }
    this.right = right;
  }

  /**
   * Returns "'low" attribute value.
   *
   * @return Low attribute value.
   */
  public Node getLow() {
    return this.low;
  }

  /**
   * Sets the specified value to the non-initialized low field.
   *
   * @param low Value to bet set to low field.
   *
   * @throws RetrascopeRuntimeException when low field is already set to another value.
   */
  public void setLow(Node low) {
    if (this.low != null && !this.low.equals(low)) {
      throwException(low.toString());
    }
    this.low = low;
  }

  /**
   * Returns "'high" attribute value.
   *
   * @return High attribute value.
   */
  public Node getHigh() {
    return this.high;
  }

  /**
   * Sets the specified value to the non-initialized high field.
   *
   * @param high Value to bet set to high field.
   *
   * @throws RetrascopeRuntimeException when high field is already set to another value.
   */
  public void setHigh(Node high) {
    if (this.high != null && !this.high.equals(high)) {
      throwException(high.toString());
    }
    this.high = high;
  }

  /**
   * Returns "'ascending" field value.
   *
   * @return Ascending attribute value.
   */
  public Node getAscending() {
    return this.ascending;
  }

  /**
   * Sets the specified value to the non-initialized ascending field.
   *
   * @param ascending Value to bet set to ascending field.
   *
   * @throws RetrascopeRuntimeException when ascending field is already set to another value.
   */
  public void setAscending(Node ascending) {
    if (this.ascending != null && !this.ascending.equals(ascending)) {
      throwException(ascending.toString());
    }
    this.ascending = ascending;
  }

  /**
   * Returns "'base" field value.
   *
   * @return Base attribute value.
   */
  public DataType getBase() {
    return this.base;
  }

  /**
   * Sets the specified value to the non-initialized base field.
   *
   * @param base Value to bet set to base field.
   *
   * @throws RetrascopeRuntimeException when base field is already set to another value.
   */
  public void setBase(DataType base) {
    if (this.base != null && !this.base.equals(base)) {
      throwException(base.toString());
    }

    this.base = base;
  }

  /**
   * Returns "'range" field value.
   *
   * @return Range attribute value.
   */
  public Range getRange() {
    return this.range;
  }

  /**
   * Sets the specified value to the non-initialized range field.
   *
   * @param range Value to bet set to range field.
   *
   * @throws RetrascopeRuntimeException when range field is already set to another value.
   */
  public void setRange(Range range) {
    if (this.range != null && !this.range.equals(range)) {
      throwException(range.toString());
    }

    this.range = range;
  }

  /**
   * Returns "'reverse_range" field value.
   *
   * @return Reverse_range attribute value.
   */
  public Range getReverseRange() {
    return this.reverseRange;
  }

  /**
   * Sets the specified value to the non-initialized reverse_range field.
   *
   * @param reverseRange Value to bet set to reverse_range field.
   *
   * @throws RetrascopeRuntimeException when reverse_range field is already set to another value.
   */
  public void setReverseRange(Range reverseRange) {
    if (this.reverseRange != null && !this.reverseRange.equals(reverseRange)) {
      throwException(reverseRange.toString());
    }

    this.reverseRange = reverseRange;
  }

  /**
   * Returns "'length" field value.
   *
   * @return Length attribute value.
   */
  public Node getLength() {
    return this.length;
  }

  /**
   * Sets the specified value to the non-initialized length field.
   *
   * @param length Value to bet set to length field.
   *
   * @throws RetrascopeRuntimeException when length field is already set to another value.
   */
  public void setLength(Node length) {
    if (this.length != null && ! this.length.equals(length)) {
      throwException(length.toString());
    }

    this.length = length;
  }

  private void throwException(final String field) {
    throw new RetrascopeRuntimeException(
        String.format("Cannot set: %s; it's already initialized.", field));
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }
    if (!super.equals(obj)) {
      return false;
    }
    VariableDescriptor that = (VariableDescriptor) obj;
    return Objects.equals(getAscending(), that.getAscending())
        && getVariableType() == that.getVariableType()
        && Objects.equals(getLeft(), that.getLeft())
        && Objects.equals(getRight(), that.getRight())
        && Objects.equals(getLow(), that.getLow())
        && Objects.equals(getHigh(), that.getHigh())
        && Objects.equals(getBase(), that.getBase())
        && Objects.equals(getRange(), that.getRange())
        && Objects.equals(getReverseRange(), that.getReverseRange())
        && Objects.equals(getLength(), that.getLength());
  }

  @Override
  public int hashCode() {
    return Objects.hash(
        super.hashCode(),
        getVariableType(),
        getLeft(),
        getRight(),
        getLow(),
        getHigh(),
        getAscending(),
        getBase(),
        getRange(),
        getReverseRange(),
        getLength());
  }
}
