/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.AssignmentContainer;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Basic block node of process' control flow graph.
 *
 * <p>Basic blocks can be either of concurrent (non-blocking) type
 * or of sequential (blocking) type. The concurrent type means
 * that all the assignments should be executed simultaneously
 * in such a concurrent mode. The sequential type means that
 * all the assignments should be executed sequentially in the same order
 * they have in it's collection.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 *
 */
public class CfgBlockStatement extends CfgAtomicStatement implements AssignmentContainer {

  /* The assignments. */
  private final List<Assignment> assignments;

  /* The concurrent type flag. */
  private boolean concurrent;

  /**
   * Constructs object of the specified type flag.
   *
   * @param concurrent block type flag
   */
  public CfgBlockStatement(final boolean concurrent) {
    super();
    this.assignments = new LinkedList<>();
    this.concurrent = concurrent;
  }

  /**
   * Constructs object with the specified assignments and the type flag.
   *
   * @param assignments assignments
   * @param concurrent block type flag
   *
   * @throws IllegalArgumentException when first argument is {@code null}.
   */
  public CfgBlockStatement(final List<Assignment> assignments, final boolean concurrent) {
    super();
    InvariantChecks.checkNotNull(assignments);

    this.assignments = assignments;
    this.concurrent = concurrent;
  }

  /**
   * Constructs object with the specified assignment and the flag type.
   *
   * @param assignment assignment
   * @param concurrent block type flag
   *
   * @throws IllegalArgumentException when assignment argument is {@code null}.
   */
  public CfgBlockStatement(final Assignment assignment, final boolean concurrent) {
    super();
    InvariantChecks.checkNotNull(assignment);

    this.assignments = new LinkedList<>();
    this.assignments.add(assignment);
    this.concurrent = concurrent;
  }

  /**
   * Constructs object with the specified collection of assignments.
   *
   * <p>The default value of concurrent flag is {@code false}</p>
   *
   * @param assignments assignments
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public CfgBlockStatement(final List<Assignment> assignments) {
    this(assignments, false);
  }

  /**
   * Constructs object with the specified assignment and of non-concurrent type.
   *
   * @param assignment assignment
   */
  public CfgBlockStatement(final Assignment assignment) {
    this(assignment, false);
  }

  /**
   * Constructs object that contains one assignment with the specified target and value.
   *
   * @see Assignment
   *
   * @param target target variable
   * @param value expression value
   */
  public CfgBlockStatement(final RangedVariable target, final Node value) {
    this(new Assignment(target, value));
  }

  /**
   * Constructs object that contains one assignment with the specified target and value.
   *
   * @see Assignment
   *
   * @param target target variable
   * @param value expression value
   */
  public CfgBlockStatement(final NodeVariable target, final Node value) {
    this(new Assignment(target, value));
  }

  private CfgBlockStatement(final CfgBlockStatement blockStatement) {
    super(blockStatement);
    this.concurrent = blockStatement.isConcurrent();
    this.assignments = new LinkedList<>(blockStatement.getAssignments());
  }

  @Override
  public CfgStatementType getType() {
    return CfgStatementType.BASIC_BLOCK;
  }

  @Override
  public CfgStatement deepCopy() {
    return new CfgBlockStatement(this);
  }

  /**
   * Adds assignment statement.
   *
   * <p>Assignment will be added with on 1 greater number
   * than other assignments have.</p>
   *
   * @param stmt assignment to be added
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  @Override
  public void addAssignment(final Assignment stmt) {
    InvariantChecks.checkNotNull(stmt);

    this.assignments.add(stmt);
  }

  /**
   * Appends assignments list to the last assignment of basic block.
   *
   * @param statements list of concurrent assignments
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  @Override
  public void addAssignments(final List<Assignment> statements) {
    InvariantChecks.checkNotNull(statements);

    this.assignments.addAll(statements);
  }

  /**
   * Returns all block assignments.
   *
   * @return assignments list
   */
  @Override
  public List<Assignment> getAssignments() {
    return this.assignments;
  }

  /**
   * Removes all block assignments.
   */
  @Override
  public void removeAssignments() {
    this.assignments.clear();
  }

  /**
   * Checks whether object contains only concurrent assignments.
   *
   * @return {@code true} if contains only concurrent assignments,
   *         {@code false} otherwise.
   */
  public boolean isConcurrent() {
    return this.concurrent;
  }

  /**
   * Sets the concurrent flag to the specified value.
   *
   * @param concurrent concurrent flag
   */
  public void setConcurrent(final boolean concurrent) {
    this.concurrent = concurrent;
  }

  @Override
  public void apply(Map<String, Node> bindings) {

    for (final Assignment stmt : this.assignments) {

      final RangedVariable rangedVariable = stmt.getTarget();
      final String name = rangedVariable.getVariableName();

      /* If defined variable is used-in-bindings. */
      if (bindings.containsKey(name)) {

        final Node binding = bindings.get(name);

        if (bindings.get(name).getKind() == Node.Kind.VARIABLE) {

          stmt.setVariable(new RangedVariable((NodeVariable) binding));
        } else {
          throw RetrascopeRuntimeException.makeException(LogMessage.WRONG_TYPE, binding.getKind());
        }
      }

      final Set<String> usedVarNames = stmt.getUsedVariableNames();

      for (final String usedVarName : usedVarNames) {

        /* If used-in-expression variable is used-in-bindings. */
        if (bindings.containsKey(usedVarName)) {
          stmt.setValue(
              Transformer.substitute(stmt.getValue(), usedVarName, bindings.get(usedVarName)));
        }
      }
    }
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {
    StringBuilder buffer = new StringBuilder();

    final Iterator<Assignment> iterator = this.assignments.iterator();
    while (iterator.hasNext()) {
      final Assignment stmt = iterator.next();

      buffer.append(
          stmt.getSeparatedDescription(printer, stmt.getSeparator(printer, isConcurrent())));

      if (iterator.hasNext()) {
        buffer.append("\n");
      }
    }
    return buffer.toString();
  }


  @Override
  public Set<NodeVariable> getUses() {
    final Set<NodeVariable> uses = new LinkedHashSet<>();

    for (final Assignment assignment : this.assignments) {
      final Collection<NodeVariable> assignmentUses = assignment.getUses();

      for (final NodeVariable assignmentUse : assignmentUses) {
        if (!uses.contains(assignmentUse)) {
          uses.add(assignmentUse);
        }
      }
    }
    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() {
    final Set<NodeVariable> defines = new LinkedHashSet<>();

    for (final Assignment assignment : this.assignments) {
      final Collection<NodeVariable> assignmentDefines = assignment.getDefines();

      for (final NodeVariable assignmentDefine : assignmentDefines) {
        if (!defines.contains(assignmentDefine)) {
          defines.add(assignmentDefine);
        }
      }
    }
    return defines;
  }
}
