/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Case statement of control flow graph model.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgCaseStatement extends CfgSequentialStatement {

  private final Set<NodeValue> values = new LinkedHashSet<>();
  private boolean isDefault;

  /**
   * Constructs case statement with the specified value.
   *
   * @param value case value
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public CfgCaseStatement(final NodeValue value) {
    InvariantChecks.checkNotNull(value);

    this.values.add(value);
    this.isDefault = false;
  }

  /**
   * Constructs case statement with the specified values.
   *
   * @param values case values
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeRuntimeException when values are not of the same data type.
   */
  public CfgCaseStatement(final Collection<NodeValue> values) {
    InvariantChecks.checkNotNull(values);
    InvariantChecks.checkNotEmpty(values);
    haveSameDataType(values);

    this.values.addAll(values);
    this.isDefault = false;
  }

  private static void haveSameDataType(final Collection<NodeValue> values) {

    DataType dataType = null;
    for (final NodeValue value : values) {
      if (dataType == null) {
        dataType = value.getDataType();
      } else if (!dataType.equals(value.getDataType())) {
        throw new RetrascopeRuntimeException(
            String.format(
                "Argument should contain same data type but it contains: %s and %s.",
                dataType,
                value.getDataType()));
      }
    }
  }

  /**
   * Constructs case statement of the specified default type.
   *
   * @param defaultNode default type
   */
  public CfgCaseStatement(final boolean defaultNode) {
    this.isDefault = defaultNode;
  }

  private CfgCaseStatement(final CfgCaseStatement statement) {
    super(statement);
    this.values.addAll(statement.getValues());
    this.isDefault = statement.isDefault();
  }

  /**
   * Returns case statement values.
   *
   * @return values
   */
  public Set<NodeValue> getValues() {
    return Collections.unmodifiableSet(this.values);
  }

  /**
   * Returns case statement default type.
   *
   * @return default type
   */
  public boolean isDefault() {
    return this.isDefault;
  }

  /**
   * Sets default type of this case statement object.
   * @param isDefault default type
   *
   * @throws RetrascopeRuntimeException when argument is {@code true}
   *         and this statement already contains related {@link NodeValue} values
   */
  public void setDefault(final boolean isDefault) {
    if (!this.values.isEmpty() && isDefault) {
      throw new RetrascopeRuntimeException(
          String.format("Cannot be both default and have values: %s.", getType()));
    }
    this.isDefault = isDefault;
  }

  /**
   * Adds value to this case statement object.
   *
   * @param value case value
   *
   * @throws RetrascopeRuntimeException when this object is of default kind.
   */
  public void add(final NodeValue value) {
    InvariantChecks.checkNotNull(value);

    if (isDefault()) {
      throw new RetrascopeRuntimeException(
          String.format("Cannot be both default and have values: %s.", getType()));
    }
    this.values.add(value);
  }

  /**
   * Removes the specified value from this case statement.
   *
   * @param value case value
   */
  public void remove(final NodeValue value) {
    InvariantChecks.checkNotNull(value);

    this.values.remove(value);
  }

  /**
   * Removes all values of this case statement.
   */
  public void removeValues() {
    this.values.clear();
  }

  @Override
  public String getDescription(ExprTreePrinter printer) {

    final Iterator<NodeValue> iterator = this.values.iterator();
    final StringBuilder builder = new StringBuilder(iterator.hasNext() ? "" : "default  ");

    while (iterator.hasNext()) {
      builder.append(printer.toString(iterator.next()));
      if (iterator.hasNext()) {
        builder.append(',');
      }
    }
    return builder.toString();
  }

  @Override
  public CfgStatementType getType() {
    return CfgStatementType.CASE;
  }

  @Override
  public CfgStatement deepCopy() {
    return new CfgCaseStatement(this);
  }

  @Override
  public void setParent(CfgStatement parent) {
    checkClass(parent);
    super.setParent(parent);
  }

  @Override
  public void removeParent(CfgStatement parent) {
    checkClass(parent);
    super.removeParent(parent);
  }

  private static void checkClass(CfgStatement parent) {
    if (!parent.isType(CfgStatementType.SWITCH)) {
      throw RetrascopeRuntimeException.makeException(LogMessage.WRONG_TYPE, parent);
    }
  }
}
