/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

/**
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
abstract class CfgConditionStatement extends CfgSequentialStatement {

  private EventList list;
  private Node condition;

  protected CfgConditionStatement(final EventList list, final Node condition) {
    super();
    this.list = list;
    this.condition = condition;
  }

  protected CfgConditionStatement(final EventList list) {
    this(list, SolverUtils.TRUE);
  }

  protected CfgConditionStatement(final Event event) {
    this(new EventList(event));
  }

  protected CfgConditionStatement(final Node condition) {
    this(new EventList(), condition);
  }

  /**
   * Returns {@link Node} condition.
   *
   * @return condition
   */
  public Node getCondition() {
    return this.condition;
  }

  /**
   * Sets condition to this statement.
   *
   * @param condition condition node
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void setCondition(final Node condition) {
    InvariantChecks.checkNotNull(condition);

    this.condition = condition;
  }

  /**
   * Returns event list.
   *
   * @return event list
   */
  public EventList getEventList() {
    return this.list;
  }

  /**
   * Checks whether this object contains an event.
   *
   * @return {@code true} when this object containt an event,
   *         {@code false} otherwise
   */
  public boolean hasEvents() {
    return this.list != null && !this.list.isEmpty();
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {
    InvariantChecks.checkNotNull(printer);

    return String.format(
        "%s%s",
        getEventList().isEmpty() ? "" : getEventList().toString() + " ",
        printer.toString(getCondition()));
  }

  @Override
  public Set<NodeVariable> getUses() {
    final Set<NodeVariable> uses = new LinkedHashSet<>();

    /* Get variables that are used in expression. */
    final Node expr = getCondition();
    final Collection<NodeVariable> usedExprVariables = ExprUtils.getVariables(expr);
    uses.addAll(usedExprVariables);

    /* Get variables that are used in event. */
    if (hasEvents()) {
      final Collection<Event> events = getEventList().getEvents();

      for (final Event event : events) {
        final RangedVariable rVariable = event.getRangedVariable();
        final NodeVariable eventVariable = rVariable.getVariable();

        uses.add(eventVariable);

        if (rVariable.isRanged()) {
          final Node youngNum = rVariable.getRange().getYoung();
          final Node oldNum = rVariable.getRange().getOld();

          uses.addAll(ExprUtils.getVariables(youngNum));
          uses.addAll(ExprUtils.getVariables(oldNum));
        }
      }
    }
    uses.addAll(super.getUses());
    return uses;
  }

  @Override
  public void apply(final Map<String, Node> bindings) {
    getEventList().applyToEventList(bindings);

    if (getCondition() != null) {
      for (final Map.Entry<String, Node> entry : bindings.entrySet()) {
        setCondition(Transformer.substitute(getCondition(), entry.getKey(), entry.getValue()));
      }
    }
  }
}
