/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;

/**
 * Loop statement for control flow graph node.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgLoopStatement extends CfgSequentialStatement {

  /* Loop type. */
  private Type type;

  /* Loop counter initialization (only for {@code FOR} loops). */
  private Assignment init;

  /* Loop condition (or iteration number). */
  private Node condition;

  /* Loop counter increment (only for {@code FOR} loops). */
  private Assignment increment;

  /**
   * Loop statement types.
   */
  public enum Type {
    /* Forever loop: {@code forever ...}. */
    FOREVER,
    /* Repeat loop: {@code repeat(n) ...}. */
    REPEAT,
    /* While loop: {@code while(c) ...}. */
    WHILE,
    /* For loop: {@code for(a; b; c) ...}. */
    FOR,
  }

  /**
   * Constructs loop statement with the specified parameters.
   * <p>Constructs loop statement of the specified type, with the specified condition,
   * initialization and incrementation assignments on loop counter</p>
   * @param type loop type
   * @param condition loop counter condition
   * @param init loop counter initialization assignment
   * @param increment loop counter incrementation assignment
   *
   * @throws IllegalArgumentException when type argument is {@code null}.
   */
  public CfgLoopStatement(
      final Type type,
      final Node condition,
      final Assignment init,
      final Assignment increment) {
    super();
    InvariantChecks.checkNotNull(type);

    setType(type);
    setCondition(condition);
    setInit(init);
    setIncrement(increment);
  }

  private CfgLoopStatement(final CfgLoopStatement statement) {
    super(statement);
    this.type = statement.getLoopType();
    this.init = statement.getInit();
    this.increment = statement.getIncrement();
    this.condition = statement.getCondition();
  }

  /**
   * Returns loop type.
   * @return loop type
   */
  public Type getLoopType() {
    return type;
  }

  /**
   * Sets loop type.
   * @param type loop type
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void setType(Type type) {
    InvariantChecks.checkNotNull(type);

    this.type = type;
  }

  /**
   * Returns loop counter initialization assignment.
   * @return loop counter initialization assignment
   */
  public Assignment getInit() {
    return this.init;
  }

  /**
   * Sets loop counter initialization assignment
   * @param init loop counter initialization assignment
   *
   * @throws IllegalArgumentException when argument is {@code null} for {@code REPEAT}
   *         and {@code FOR} loops.
   */
  public void setInit(Assignment init) {

    if (getLoopType().equals(Type.FOR) || getLoopType().equals(Type.REPEAT)) {
      InvariantChecks.checkNotNull(init);
    }

    this.init = init;
  }

  /**
   * Returns loop counter condition.
   * @return loop counter condition
   */
  public Node getCondition() {
    return this.condition;
  }

  /**
   * Sets loop counter condition.
   * @param condition loop counter condition
   *
   * @throws IllegalArgumentException when argument is {@code null} and loop type
   *         is not equal to {@code FOREVER}.
   */
  public void setCondition(Node condition) {

    if (!getLoopType().equals(Type.FOREVER)) {
      InvariantChecks.checkNotNull(condition);
    }

    this.condition = condition;
  }

  /**
   * Returns loop counter incrementation assignment.
   * @return loop counter incrementation assignment
   */
  public Assignment getIncrement() {
    return increment;
  }

  /**
   * Sets loop counter incrementation assignment.
   * @param increment loop counter incrementation assignment
   *
   * @throws IllegalArgumentException when argument is {@code null} for {@code REPEAT}
   *         and {@code FOR} loops.
   */
  public void setIncrement(Assignment increment) {
    if (getLoopType().equals(Type.FOR) || getLoopType().equals(Type.REPEAT)) {
      InvariantChecks.checkNotNull(increment);
    }

    this.increment = increment;
  }

  @Override
  public CfgStatementType getType() {
    return CfgStatementType.LOOP;
  }

  @Override
  public CfgStatement deepCopy() {
    return new CfgLoopStatement(this);
  }

  @Override
  public String getDescription(ExprTreePrinter printer) {

    final String description =
        getCondition() == null
            ? ""
            : String.format(
                "(%s%s%s)",
                getInit() == null ? "" : getInit().getDescription(printer) + ';',
                printer.toString(getCondition()),
                getIncrement() == null ? "" : ';' + getIncrement().getDescription(printer));
    return String.format("%s %s", getLoopType().name(), description);
  }
}
