/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Module of control flow graph model.
 *
 * <p>Object contains child {@link CfgProcess} processes and {@link CfgModule} sub-modules.
 * When object is an instance of another module, it MUST have the instance name
 * with module name.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgModule extends CfgDeclarationStatement {

  private final String name;

  private String instanceName;

  /**
   * Constructs object with the specified name.
   *
   * @param name module name
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public CfgModule(final String name) {
    super();
    InvariantChecks.checkNotNull(name);

    this.name = name;
  }

  private CfgModule(final CfgModule module) {
    super(module);
    this.name = module.getName();
    this.instanceName = module.getInstanceName();
  }

  /**
   * Sets instance name.
   *
   * @param instanceName name of current instance
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void setInstanceName(final String instanceName) {
    InvariantChecks.checkNotNull(instanceName);

    this.instanceName = instanceName;
  }

  /**
   * Returns processes.
   *
   * @return processes
   */
  public Collection<CfgProcess> getProcesses() {
    final Set<CfgProcess> processes = new LinkedHashSet<>();
    for (int i = 0; i < getStatementNum(); i++) {
      final CfgStatement statement = getStatement(i);
      if (statement.getType() == CfgStatementType.PROCESS) {
        processes.add((CfgProcess) statement);
      }
    }
    return processes;
  }

  /**
   * Returns sub-modules.
   *
   * @return sub-modules
   */
  public Collection<CfgModule> getModules() {
    final Set<CfgModule> modules = new LinkedHashSet<>();
    for (int i = 0; i < getStatementNum(); i++) {
      final CfgStatement statement = getStatement(i);
      if (statement.getType() == CfgStatementType.MODULE) {
        modules.add((CfgModule) statement);
      }
    }
    return modules;
  }

  /**
   * Instantiates this object with the specified name and bindings.
   *
   * @param instanceName instance name
   * @param bindings instance bindings
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public void instantiate(final String instanceName, final Map<String, Node> bindings) {
    InvariantChecks.checkNotNull(instanceName);
    InvariantChecks.checkNotNull(bindings);

    setInstanceName(instanceName);
    apply(bindings);
  }

  /**
   * Returns a full name.
   * <p>A full name is a combination of name and instance name.</p>
   * @return full name
   */
  public String getFullName() {
    final StringBuilder builder = new StringBuilder(this.name);

    if (this.instanceName != null) {
      builder.append('.');
      builder.append(this.instanceName);
    }
    return builder.toString();
  }

  /**
   * Returns name.
   *
   * @return module name
   */
  public String getName() {
    return this.name;
  }

  /**
   * Returns instance name.
   *
   * @return instance name
   */
  public String getInstanceName() {
    return this.instanceName;
  }


  /**
   * Returns all the variables, that are of input signal type,
   * or empty collection, if there are no such.
   *
   * @return input signals
   */
  public Collection<NodeVariable> getInputs() {
    return getVariables(VariableType.IN, VariableType.INOUT);
  }

  /**
   * Returns all variables that are of output signal type,
   * or empty collection, if there are no such.
   *
   * @return output signals
   */
  public Collection<NodeVariable> getOutputs() {
    return getVariables(VariableType.INOUT, VariableType.OUT);
  }

  /**
   * Remove internal processes.
   */
  public void removeProcesses() {

    final List<CfgProcess> processes = new LinkedList<>();

    for (int i = 0; i < getStatementNum(); i++) {
      final CfgStatement statement = getStatement(i);
      if (statement.isType(CfgStatementType.PROCESS)) {
        processes.add((CfgProcess) statement);
      }
    }

    for (final CfgProcess process : processes) {
      remove(process);
    }
  }

  @Override
  public void add(final CfgStatement statement) {
    if (!(statement instanceof CfgDeclarationStatement)) {
      throw RetrascopeRuntimeException.makeException(LogMessage.WRONG_TYPE, statement);
    }
    super.add(statement);
  }

  @Override
  public void setParent(final CfgStatement statement) {
    if (!(statement.getType() == CfgStatementType.MODULE)) {
      throw RetrascopeRuntimeException.makeException(LogMessage.WRONG_TYPE, statement);
    }
    super.setParent(statement);
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {
    return String.format("%s %s", getType().name(), getFullName());
  }

  @Override
  public CfgStatementType getType() {
    return CfgStatementType.MODULE;
  }

  @Override
  public CfgStatement deepCopy() {
    return new CfgModule(this);
  }

  @Override
  public void apply(final Map<String, Node> bindings) {

    /* Change declarations. */
    for (final String varName : bindings.keySet()) {
      if (containsVariable(varName)) {

        /* Remove old declaration for the variable that is substituted by bindings. */
        removeDeclaration(varName);
      }
    }
    super.apply(bindings);
  }

  private Collection<NodeVariable> getVariables(final VariableType ... types) {

    final Set<NodeVariable> variables = new LinkedHashSet<>();

    final Set<NodeVariable> moduleVariables = getDeclarationMap().keySet();
    for (final NodeVariable var : moduleVariables) {

      // extract and analyse variable data
      final Object varData = var.getUserData();
      if (varData instanceof VariableDescriptor) {
        final VariableType varType = ((VariableDescriptor) varData).getVariableType();
        for (final VariableType type : types) {
          if (varType.equals(type)) {
            variables.add(var);
            break;
          }
        }
      }
    }
    return variables;
  }
}
