/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.util.LogMessage;

import java.util.List;
import java.util.Map;

/**
 * Process of control flow graph model.
 *
 * <p>The object contains the embedded control flow graph nodes
 * and {@link EventList} sensitivity list. Also it can be either
 * of initial or of non-initial type.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgProcess extends CfgDeclarationStatement {

  /* The sensitivity list. */
  private EventList eventList = new EventList();
  private final boolean initial;

  /**
   * Default constructor.
   *
   * <p>Constructs non-initial process.</p>
   */
  public CfgProcess() {
    this(false);
  }

  /**
   * Constructs new process with the specified initial flag and sensitivity list.
   *
   * @param initial initial flag
   * @param list event list
   *
   * @throws IllegalArgumentException when sensitivity list argument is {@code null}.
   */
  public CfgProcess(final boolean initial, final EventList list) {
    super();
    InvariantChecks.checkNotNull(list);

    this.initial = initial;
    if (!list.isEmpty()) {
      this.eventList.addEvents(list.getEvents());
    }
  }
  /**
   * Constructs new process with the specified value of initial flag.
   *
   * @param initial initial flag
   */
  public CfgProcess(final boolean initial) {
    super();

    this.initial = initial;
  }

  private CfgProcess(final CfgProcess process) {
    super(process);
    this.eventList = process.getSensitivityList();
    this.initial = process.isInitial();
  }

  /**
   * Checks whether process is of initial kind.
   *
   * @return {@code true} if process is initial, {@code false} otherwise.
   */
  public boolean isInitial() {
    return this.initial;
  }

  /**
   * Returns {@link ru.ispras.retrascope.model.basis.EventList} sensitivity list.
   *
   * @return sensitivity list
   */
  public EventList getSensitivityList() {
    return this.eventList.deepCopy();
  }

  /**
   * Sets new {@link ru.ispras.retrascope.model.basis.EventList} sensitivity list.
   *
   * @param eventList event list
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeRuntimeException when object is initial and argument is non-empty.
   */
  public void setSensitivityList(final EventList eventList) {
    InvariantChecks.checkNotNull(eventList);

    if (isInitial() && !eventList.isEmpty()) {
      throw new RetrascopeRuntimeException(
          "Cannot set to initial process a non-empty event collection.");
    }

    this.eventList.clear();
    this.eventList.addEvents(eventList.getEvents());
  }

  /**
   * Add the specified statement which this process is sensible to.
   *
   * @param event event to be added
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeRuntimeException when this process is initial.
   */
  public void addEvent(final Event event) {
    InvariantChecks.checkNotNull(event);

    if (isInitial()) {
      throw new RetrascopeRuntimeException("Cannot add event to initial process");
    }
    this.eventList.add(event);
  }

  /** Applies bindings to process control flow graph and sensitivity list.
   *
   * @param bindings "variable name - expression" bindings
   */
  @Override
  public void apply(final Map<String, Node> bindings) {
    this.eventList.applyToEventList(bindings);
    super.apply(bindings);
  }

  @Override
  public CfgStatement deepCopy() {
    return new CfgProcess(this);
  }

  @Override
  public void add(final CfgStatement statement) {

    if (!(statement instanceof CfgSequentialStatement)) {
      throw RetrascopeRuntimeException.makeException(LogMessage.WRONG_TYPE, statement);
    }
    super.add(statement);
  }

  @Override
  public void addAll(final List<CfgStatement> list) {
    InvariantChecks.checkNotNull(list);

    for (final CfgStatement statement : list) {
      add(statement);
    }
  }

  @Override
  public void setParent(final CfgStatement parent) {

    if (parent.getType() != CfgStatementType.MODULE) {
      throw RetrascopeRuntimeException.makeException(LogMessage.WRONG_TYPE, parent);
    }
    super.setParent(parent);
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {
    return String.format(
        "%s %s%s (%s)",
        getPrefix(),
        getType().name().toLowerCase(),
        this.eventList.isEmpty() ? "" : " " + this.eventList.toString(),
        getSuffix());
  }

  private String getPrefix() {
    return ((CfgModule)getParent()).getFullName();
  }

  private String getSuffix() {
    return Integer.toHexString(hashCode());
  }

  @Override
  public CfgStatementType getType() {
    return CfgStatementType.PROCESS;
  }
}
