/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.model.basis.Identifiable;
import ru.ispras.retrascope.model.basis.Instantiated;
import ru.ispras.retrascope.model.basis.MetaInfo;
import ru.ispras.retrascope.model.basis.UseDef;
import ru.ispras.retrascope.model.cfg.walker.CfgWalker;
import ru.ispras.retrascope.model.cfg.walker.CfgWalkerFactory;
import ru.ispras.retrascope.model.cfg.walker.CfgWalkerMode;
import ru.ispras.retrascope.util.LogMessage;

import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;

/**
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class CfgStatement extends MetaInfo implements Instantiated,Identifiable,UseDef {

  private CfgStatement parent;
  private final List<CfgStatement> statements = new LinkedList<>();

  /**
   * Returns statement type.
   * @return statement type
   */
  public abstract CfgStatementType getType();

  protected CfgStatement() {
    /* Empty. */
  }

  protected CfgStatement(final CfgStatement statement) {
    for (int i = 0; i < statement.getStatementNum(); i++) {
      add(statement.getStatement(i).deepCopy());
    }
  }

  /**
   * Checks whether this statement is of specified type.
   * @param type Statement type.
   * @return {@code true} if this statement is of the specified type,
   *         {@code false} otherwise.
   */
  public boolean isType(final CfgStatementType type) {
    return getType().equals(type);
  }

  /**
   * Checks whether this statement contains any other embedded statements.
   *
   * @return {@code true} if this statement contains statements, {@code false} otherwise.
   */
  public boolean containStatements() {
    return getStatementNum() > 0;
  }

  /**
   * Returns a copy of this statement and all it's sub-statements.
   *
   * @return Statement copy.
   */
  public abstract CfgStatement deepCopy();

  /**
   * Returns a number of child statements.
   * @return Child statement number.
   */
  public int getStatementNum() {
    return this.statements.size();
  }

  /**
   * Returns a child statement of the specified number.
   *
   * @param num Child statement number.
   *
   * @return child statement that is stored at the specified number
   *         or {@code null} if there is no such.
   */
  public CfgStatement getStatement(final int num) {
    return this.statements.get(num);
  }

  /**
   * Returns sub-statement with the specified identifier.
   *
   * @param id Statement identifier.
   * @return Sub-statement that has the specified identifier.
   *
   * @throws NoSuchElementException when there is no such sub-statement with the specified
   *         identifier.
   */
  public CfgStatement getStatement(final String id) {

    class CfgGetByIdVisitor extends CfgDefaultVisitor {

      private String id;
      private CfgStatement result;


      public CfgGetByIdVisitor(final String id) {
        this.id = id;
        this.result = null;
      }

      public CfgStatement getStatementWithId() {
        return this.result;
      }

      @Override
      public void onNodeBegin(final CfgStatement statement) {
        if (statement.getId().equals(this.id)) {
          this.result = statement;
        }
      }
    }

    final CfgGetByIdVisitor visitor = new CfgGetByIdVisitor(id);
    final CfgWalker walker = CfgWalkerFactory.newWalker(visitor, CfgWalkerMode.DFS_NO_RPT);
    walker.visitNode(this);

    final CfgStatement statementWithId = visitor.getStatementWithId();

    if (statementWithId == null) {
      throw new NoSuchElementException(
          String.format("%s node with '%s' id.", LogMessage.CANT_FIND, id));
    }
    return statementWithId;
  }

  /**
   * Returns unique node identifier in string form.
   *
   * @return Statement identifier.
   */
  public String getId() {
    return this.toString();
  }

  /**
   * Returns index of the specified statement.
   *
   * @param statement Statement which index is returned.
   * @return index of the specified child statement, or {@code -1} otherwise.
   */
  public int indexOf(final CfgStatement statement) {
    return this.statements.indexOf(statement);
  }

  /**
   * Checks whether this statement contains the specified as the child.
   *
   * @param statement Statement which existence in this object is checked.
   * @return {@code true} if this object contains the specified statement, {@code false} otherwise.
   */
  public boolean contains(final CfgStatement statement) {
    return this.statements.contains(statement);
  }

  /**
   * Adds the specified statement as child to this object.
   *
   * @param statement Statement to be added.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws IllegalArgumentException when this object already contains the specified statement.
   * @throws IllegalArgumentException when the specified statement already has parent statement
   *         and this parent is not equal to this statement.
   */
  public void add(final CfgStatement statement) {
    InvariantChecks.checkNotNull(statement);

    if (this.statements.contains(statement)) {
      throw new IllegalArgumentException(
          LogMessage.makeMessage(LogMessage.ALREADY_CONTAIN, statement));
    }
    this.statements.add(statement);
    if (statement.hasParent() && !statement.getParent().equals(this)) {
      throw new IllegalArgumentException(
        LogMessage.makeMessage(LogMessage.CANT_REASSIGN, statement.getParent()));
    }
    statement.setParent(this);
  }

  /**
   * Adds the specified child statement to this statement at the specified position.
   *
   * @param statement Statement to be added as child.
   * @param num New child statement position number.
   *
   * @throws IndexOutOfBoundsException if position argument is greater or equal to this statement
   *         children number, or less that {@code 0}.
   */
  public void add(final CfgStatement statement, final int num) {
    InvariantChecks.checkBounds(num, getStatementNum() + 1);

    /* Keep statements that should go after specified */
    List<CfgStatement> afterList = new LinkedList<>();
    for (int i = num; i < getStatementNum(); i++) {
      afterList.add(getStatement(i));
    }
    for (final CfgStatement removed : afterList) {
      remove(removed);
    }
    add(statement);
    addAll(afterList);
  }

  /**
   * Removes the specified statement from children.
   *
   * @param statement Statement to be removed from child statements.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws IllegalArgumentException when this statement already contain the specified
   *         statement as child.
   * @throws IllegalArgumentException when the specified statement already has parent statement
   *         and the parent statement is not equal to this statement.
   */
  public void remove(final CfgStatement statement) {
    InvariantChecks.checkNotNull(statement);

    if (!this.statements.contains(statement)) {
      throw new IllegalArgumentException(
          LogMessage.makeMessage(LogMessage.CANT_FIND, statement));
    }
    this.statements.remove(statement);

    if (statement.hasParent()) {
      if (!statement.getParent().equals(this)) {
        throw new IllegalArgumentException(
            LogMessage.makeMessage(LogMessage.CANT_REASSIGN, statement.getParent()));
      }
      statement.removeParent(this);
    }
  }

  /**
   * Remove nested statement with the specified identifier.
   * <p>Removes the statement with the specified identifier whether it is included into this
   * statement's children or some of it's children at any degree of inclusion.</p>
   * @param id Identifier of the statement to be removed.
   * @throws NoSuchElementException when there is no nested statement with the specified identifier.
   */
  public void removeStatement(final String id) {
    class CfgGetByIdVisitor extends CfgDefaultVisitor {

      private String id;
      private boolean removed = false;

      public CfgGetByIdVisitor(final String id) {
        this.id = id;
      }

      boolean foundAndRemoved() {
        return this.removed;
      }

      @Override
      public void onNodeBegin(final CfgStatement statement) {

        CfgStatement childWithId = null;
        for (int i = 0; i < statement.getStatementNum(); i++) {
          final CfgStatement child = statement.getStatement(i);
          if (child.getId().equals(this.id)) {
            childWithId = child;
            break;
          }
        }

        if (childWithId != null) {
          statement.remove(childWithId);
          this.removed = true;
        }
      }
    }

    final CfgGetByIdVisitor visitor = new CfgGetByIdVisitor(id);
    final CfgWalker walker = CfgWalkerFactory.newWalker(visitor, CfgWalkerMode.DFS_NO_RPT);
    walker.visitNode(this);

    if (!visitor.foundAndRemoved()) {
      throw new NoSuchElementException(
        String.format("%s node with '%s' id.", LogMessage.CANT_FIND, id));
    }
  }

  /**
   * Adds all the statements from the specified list as children.
   *
   * @param list Children statements to be added.
   */
  public void addAll(final List<CfgStatement> list) {
    InvariantChecks.checkNotNull(list);

    for (final CfgStatement statement : list) {
      add(statement);
    }
  }

  /**
   * Removes all the child statements.
   */
  public void removeAll() {
    final List<CfgStatement> statementsCopy = new LinkedList<>(this.statements);
    for (CfgStatement statementCopy : statementsCopy) {
      remove(statementCopy);
    }
  }

  /**
   * Checks whether this statement has parent statement.
   * @return {@code true} if this statement has non-null parent statement, {@code false} otherwise.
   */
  public boolean hasParent() {
    return this.parent != null;
  }

  /**
   * Returns parent statement.
   * @return Parent statement.
   */
  public CfgStatement getParent() {
    return this.parent;
  }

  /**
   * Sets the specified statement as parent.
   *
   * @param parent Statement to be set as parent.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws IllegalStateException when this statement already has non-null parent statement
   *         and the parent statement does not equal to the specified statement.
   */
  public void setParent(final CfgStatement parent) {
    InvariantChecks.checkNotNull(parent);

    if (hasParent() && !getParent().equals(parent)) {
      throw new IllegalStateException(
          LogMessage.makeMessage(LogMessage.CANT_REASSIGN, getParent()));
    }
    this.parent = parent;

    if (!parent.contains(this)) {
      parent.add(this);
    }
  }

  /**
   * Removes link with the specified statement.
   * <p>The specified statement should be a parent statement for this before this method.</p>
   * @param parent parent statement
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws IllegalArgumentException when either this statement does not have parent
   *         or it's parent is not equal to the specified statement.
   */
  public void removeParent(final CfgStatement parent) {
    InvariantChecks.checkNotNull(parent);

    if (!hasParent() || !getParent().equals(parent)) {
      throw new IllegalArgumentException(
          LogMessage.makeMessage(LogMessage.CANT_FIND, parent));
    }
    this.parent = null;

    if (parent.contains(this)) {
      parent.remove(this);
    }
  }

  /**
   * Applies bindings to this statement.
   * @param bindings Pairs of variable names and bounded expressions.
   */
  public void apply(final Map<String, Node> bindings) {
    for (final CfgStatement statement : this.statements) {
      statement.apply(bindings);
    }
  }

  @Override
  public Set<NodeVariable> getUses() {
    final Set<NodeVariable> uses = new LinkedHashSet<>();
    for (int i = 0; i < getStatementNum(); i++) {
      uses.addAll(getStatement(i).getUses());
    }
    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() {
    final Set<NodeVariable> defines = new LinkedHashSet<>();
    for (int i = 0; i < getStatementNum(); i++) {
      defines.addAll(getStatement(i).getDefines());
    }
    return defines;
  }

  @Override
  public boolean equals(final Object obj) {
    return this == obj;
  }

  @Override
  public int hashCode() {
    return System.identityHashCode(this);
  }

  @Override
  public String getDescription() {
    return getDescription(Retrascope.getConfiguration().getStylePrinter());
  }
}
