/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.TransformerUtils;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Utility methods for control flow graph model and it's components.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgUtils {

  private CfgUtils() {
    /* Empty. */
  }

  /**
   * Substitutes usages of variable with the specified name
   * by the specified expression in all the elements of the specified collection.
   * <p>
   * Substitutes usages of variable with the specified {@link java.lang.String} name
   * by the specified {@link ru.ispras.fortress.expression.Node} expression
   * in all the elements of the specified collection. The resulting collection
   * consists of node copies of the specified collection. In which node the every
   * usage of the variable with the specified name will be substituted by the
   * specified expression.
   * </p>
   *
   * @param list collection which nodes will be copied and transformed
   * @param name variable name
   * @param node expression
   *
   * @return collection of modified nodes
   */
  public static List<CfgStatement> substitute(
      final List<CfgStatement> list,
      final String name,
      final Node node) {

    final int listSize = list.size();
    final List<CfgStatement> result = new LinkedList<>();
    final Map<CfgStatement, CfgStatement> map = new LinkedHashMap<>(listSize);

    /* Substitute variable with the specified name by the specified node. */
    for (final CfgStatement elem : list) {
      CfgStatement copy = substituteStatement(elem, name, node);

      map.put(elem, copy);
      result.add(copy);
    }

    /* Establish parent-child links between nodes if they have them. */
    for (int i = 0; i < listSize; i++) {
      final CfgStatement one = list.get(i);
      for (CfgStatement another : list) {
        if (one.contains(another)) {
          map.get(one).add(map.get(another));
        }
      }
    }

    return result;
  }

  private static CfgStatement substituteStatement(
      final CfgStatement elem,
      final String name,
      final Node node) {

    switch (elem.getType()) {
      case SWITCH:
        return new CfgSwitchStatement(
            Transformer.substitute(((CfgSwitchStatement) elem).getCondition(), name, node));
      case BASIC_BLOCK:
        final List<Assignment> elemAssignments = ((CfgBlockStatement) elem).getAssignments();
        final List<Assignment> assignments = new LinkedList<>();

        for (Assignment elemAssign : elemAssignments) {
          final Node value = elemAssign.getValue();
          final RangedVariable target = elemAssign.getTarget();
          Assignment copyAssignment;
          final Node copyValue = Transformer.substitute(value, name, node);
          if (target.isRanged()) {
            final Node old = target.getRange().getOld();
            final Node young = target.getRange().getYoung();

            final Range copyRange =
                new Range(
                    Transformer.reduce(Transformer.substitute(young, name, node)),
                    Transformer.reduce(Transformer.substitute(old, name, node)));
            copyAssignment =
                new Assignment(new RangedVariable(target.getVariable(), copyRange), copyValue);
          } else {
            copyAssignment = new Assignment(target.getVariable(), copyValue);
          }
          assignments.add(copyAssignment);
        }

        return new CfgBlockStatement(assignments, ((CfgBlockStatement) elem).isConcurrent());
      case ASSERT:
        return new CfgAssertStatement(
            Transformer.substitute(
                ((CfgAssertStatement)elem).getCondition(), name, node),
            ((CfgAssertStatement)elem).getSeverity(),
            ((CfgAssertStatement)elem).getReport());
      case WAIT:
        return new CfgWaitStatement(
            ((CfgWaitStatement)elem).getEventList(),
            Transformer.substitute(((CfgConditionStatement)elem).getCondition(), name, node));
      default:
        throw new RetrascopeRuntimeException(
            LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, elem.getType()));
    }
  }

  /**
   * Checks whether the specified nodes are of the specified type.
   *
   * @param type node type
   * @param nodes nodes to be checked
   *
   * @return {@code true} if all the specified nodes are of the specified type,
   *         {@code false} otherwise.
   *
   * @throws IllegalArgumentException when either the specified type is {@code null}
   *         or the specified array of nodes is empty.
   */
  public static boolean isType(
      final CfgStatementType type,
      final CfgStatement... nodes) {
    InvariantChecks.checkNotNull(type);
    InvariantChecks.checkNotEmpty(nodes);

    boolean result = true;

    for (CfgStatement node : nodes) {

      if (!node.isType(type)) {
        result = false;
        break;
      }
    }

    return result;
  }

  /**
   * Returns node that contains a declaration of variable with the specified name.
   *
   * @param node root node which and which child nodes are checked
   * @param name variable name
   *
   * @return node that contains the declaration of variable with the specified name
   *         or {@code null} when the specified node and all the it's children
   *         do not contain such declaration
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  public static CfgDeclarationStatement getVariableContainer(
      final CfgStatement node,
      final String name) {
    InvariantChecks.checkNotNull(node);
    InvariantChecks.checkNotNull(name);

    if (node instanceof CfgDeclarationStatement
        && ((CfgDeclarationStatement)node).containsVariable(name)) {

      return (CfgDeclarationStatement) node;
    } else if (node.getStatementNum() > 0) {

      for (int i = 0; i < node.getStatementNum(); i++) {
        CfgDeclarationStatement result = getVariableContainer(node.getStatement(i), name);
        if (result != null) {
          return result;
        }
      }
      return null;
    }
    return null;
  }

  /**
   * Returns variable declaration container that is closest to the specified statement.
   *
   * @param statement statement that is included in some degree to the resulting container
   * @return declaration container that is closest to the specified statement in terms of inclusion
   */
  public static CfgDeclarationStatement getNearVariableContainer(final CfgStatement statement) {
    InvariantChecks.checkNotNull(statement);

    if (statement instanceof CfgDeclarationStatement) {
      return (CfgDeclarationStatement) statement;
    } else if (statement.getParent() != null) {
      return getNearVariableContainer(statement.getParent());
    }
    return null;
  }

  /**
   * Returns closest declaration container of the specified type to the specified statement.
   * <p>Closest container is in terms of statement inclusion. For example, we have a
   * {@link CfgModule} object {@code x} that includes {@link CfgProcess} object {@code y} that
   * includes some statement. For that statement the closest declaration container of the
   * {@code CfgStatementType.MODULE} type is an {@code x} object.</p>
   *
   * @param statement statement that is included in some degree to the resulting container
   * @param containerType type of the returned object
   *
   * @return the declaration container that is closest to the specified statement in terms
   *         of inclusion and has the specified type or {@code null} if there is no such.
   *
   * @throws RetrascopeRuntimeException when there is no variable declaration container
   *         of the specified type among parents of the specified statement.
   */
  public static CfgDeclarationStatement getNearVariableContainer(
      final CfgStatement statement,
      final CfgStatementType containerType) {

    InvariantChecks.checkNotNull(statement);
    if (statement instanceof CfgDeclarationStatement && statement.isType(containerType)) {
      return (CfgDeclarationStatement) statement;
    } else if (statement.getParent() != null) {
      return getNearVariableContainer(statement.getParent(), containerType);
    }
    throw new RetrascopeRuntimeException(
        String.format(
            "%s container of %s type for %s.",
            LogMessage.CANT_FIND,
            containerType,
            statement.getDescription()));
  }

  /**
   * Returns statements that go after the specified statement.
   *
   * @param statement statement which successors are returned
   * @return statements that go after the specified
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static Collection<CfgStatement> getSuccessors(final CfgStatement statement) {
    InvariantChecks.checkNotNull(statement);

    Collection<CfgStatement> successors = new LinkedHashSet<>();
    final CfgStatement statementParent = statement.getParent();

    if (!statement.containStatements()) {
      CfgStatement current = statement;
      CfgStatement parent = statementParent;
      while (parent != null && !parent.isType(CfgStatementType.MODULE)) {

        /* If parent node is SWITCH, ignore it's other children since they're not successors. */
        if (!parent.isType(CfgStatementType.SWITCH)) {
          final int currentIndex = parent.indexOf(current);
          if (currentIndex < parent.getStatementNum() - 1) {
            successors.add(parent.getStatement(currentIndex + 1));
            break;
          }
        }
        current = parent;
        parent = current.getParent();
      }
    } else if (statement.isType(CfgStatementType.SWITCH)
        || statement.isType(CfgStatementType.MODULE)) {
      for (int i = 0; i < statement.getStatementNum(); i++) {
        successors.add(statement.getStatement(i));
      }
    } else {
      successors.add(statement.getStatement(0));
    }

    return successors;
  }

  /**
   * Checks whether the specified statement has successors or not.
   *
   * @param statement statement that is checked
   * @return {@code true} when the specified statement has successors, {@code false} otherwise
   */
  public static boolean hasSuccessors(final CfgStatement statement) {
    return !CfgUtils.getSuccessors(statement).isEmpty();
  }

  /**
   * Replaces the first statement by the second statement.
   * <p>Finds parent statement for the specified first statement and substitutes it
   * by the second statement.</p>
   * @param one statement to be replaced
   * @param another statement that substitutes the first argument
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}
   * @throws IllegalArgumentException when first argument does not have parent statement.
   */
  public static void replace(
      final CfgStatement one,
      final CfgStatement another) {
    InvariantChecks.checkNotNull(one);
    InvariantChecks.checkNotNull(another);

    final CfgStatement parent = one.getParent();
    if (parent == null) {
      throw new IllegalArgumentException("Node to be replaced does not belong to any statement.");
    }

    final int oneIndex = parent.indexOf(one);
    parent.remove(one);
    parent.add(another, oneIndex);
  }

  /**
   * Constructs an expression from the expression of the specified switch
   * {@link CfgSwitchStatement} node and it's possible values
   * in the specified case {@link CfgCaseStatement} node.
   *
   * @param caseNode case node
   * @param switchNode switch node
   *
   * @return expression
   */
  public static Node getExpression(
      final CfgCaseStatement caseNode,
      final CfgSwitchStatement switchNode) {

    Node expr;

    if (caseNode.isDefault()) {

      /* The 'default' case processing. */
      final Collection<CfgStatement> children = CfgUtils.getSuccessors(switchNode);
      final List<NodeOperation> list = new LinkedList<>();
      for (final CfgStatement child : children) {
        if (!child.equals(caseNode)) {
          list.add(constructExpr((CfgCaseStatement)child, switchNode));
        }
      }
      expr = ExprUtils.getComplement(list.toArray(new Node[list.size()]));
    } else {

      /* The 'normal' case processing. */
      expr = constructExpr(caseNode, switchNode);
    }
    return TransformerUtils.standardize(expr);
  }

  private static NodeOperation constructExpr(
      final CfgCaseStatement child,
      final CfgSwitchStatement parent) {

    Node expr = null;

    for (final NodeValue value: child.getValues()) {
      if (expr == null) {
        expr = NodeUtils.getEquation(parent.getCondition(), value);
      } else {
        expr = ExprUtils.getDisjunction(
            expr,
            NodeUtils.getEquation(parent.getCondition(), value));
      }
    }
    return (NodeOperation) expr;
  }
}
