/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

/**
 * Control flow graph model visitor.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public interface CfgVisitor {

  /*
   * The enumeration describes possible statuses of the visitor. Status serves as a directive
   * for the walker to alter its behavior depending on events that may occur in the visitor.
   */
  enum Status {
    /** Continue traversing. */
    OK,

    /** Skip child nodes. */
    SKIP,

    /** Stop traversing. */
    ABORT
  }

  /**
   * Returns the current status of the visitor. The status guides further actions of the walker.
   *
   * @return Current visitor status.
   */
  Status getStatus();

  /**
   * Notifies that processing has been started.
   */
  void onRootBegin();

  /**
   * Notifies that processing has been finished.
   */
  void onRootEnd();

  /**
   * Starts visiting {@link CfgModel} object.
   */
  void onCfgModelBegin();

  /**
   * Finishes visiting {@link CfgModel} object.
   */
  void onCfgModelEnd();

  /**
   * Starts visiting {@link CfgStatement} node.
   *
   * @param statement Statement to be visited.
   */
  void onNodeBegin(CfgStatement statement);

  /**
   * Finishes visiting {@link CfgStatement} node.
   *
   * @param statement Statement to be visited.
   */
  void onNodeEnd(CfgStatement statement);

  /**
   * Starts visiting {@link CfgModule} node.
   *
   * @param module Module node to be visited.
   */
  void onModuleBegin(CfgModule module);

  /**
   * Finishes visiting {@link CfgModule} node.
   *
   * @param module Module node to be visited.
   */
  void onModuleEnd(CfgModule module);

  /**
   * Starts visiting {@link CfgProcess} node.
   *
   * @param process Process node to be visited.
   */
  void onProcessBegin(CfgProcess process);

  /**
   * Finishes visiting {@link CfgProcess} node.
   *
   * @param process Process node to be visited.
   */
  void onProcessEnd(CfgProcess process);

  /**
   * Starts visiting {@link CfgAssertStatement} node.
   *
   * @param assertStatement Assert statement to be visited.
   */
  void onAssertBegin(CfgAssertStatement assertStatement);

  /**
   * Finishes visiting {@link CfgAssertStatement} node.
   *
   * @param assertStatement Assert statement to be visited.
   */
  void onAssertEnd(CfgAssertStatement assertStatement);

  /**
   * Starts visiting {@link CfgBlockStatement} node.
   *
   * @param block Block statement to be visited.
   */
  void onBasicBlockBegin(CfgBlockStatement block);

  /**
   * Finishes visiting {@link CfgBlockStatement} node.
   *
   * @param block Block statement to be visited.
   */
  void onBasicBlockEnd(CfgBlockStatement block);

  /**
   * Starts visiting {@link CfgCaseStatement} node.
   *
   * @param caseNode Case statement to be visited.
   */
  void onCaseBegin(CfgCaseStatement caseNode);

  /**
   * Finishes visiting {@link CfgCaseStatement} node.
   *
   * @param caseNode Case statement to be visited.
   */
  void onCaseEnd(CfgCaseStatement caseNode);

  /**
   * Starts visiting {@link CfgSwitchStatement} node.
   *
   * @param switchNode Switch statement to be visited.
   */
  void onSwitchBegin(CfgSwitchStatement switchNode);

  /**
   * Finishes visiting {@link CfgSwitchStatement} node.
   *
   * @param switchNode Switch statement to be visited.
   */
  void onSwitchEnd(CfgSwitchStatement switchNode);

  /**
   * Starts visiting {@link CfgLoopStatement} node.
   *
   * @param loopStatement Loop statement to be visited.
   */
  void onLoopBegin(CfgLoopStatement loopStatement);

  /**
   * Finishes visiting {@link CfgLoopStatement} node.
   *
   * @param loopStatement Loop statement to be visited.
   */
  void onLoopEnd(CfgLoopStatement loopStatement);

  /**
   * Starts visiting {@link CfgWaitStatement} node.
   *
   * @param wait Wait statement to be visited.
   */
  void onWaitBegin(CfgWaitStatement wait);

  /**
   * Finishes visiting {@link CfgWaitStatement} node.
   *
   * @param wait Wait statement to be visited.
   */
  void onWaitEnd(CfgWaitStatement wait);
}
