/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg.walker;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.cfg.CfgVisitor;

import java.util.Collection;

/**
 * Control flow graph model walker running in a depth-first search mode.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgDfsWalker extends CfgWalker {

  /**
   * Constructs walker object with the specified parameters.
   *
   * @param visitor visitor to be applied to control flow graph model
   * @throws IllegalArgumentException when any argument is {@code null}.
   * @throws IllegalArgumentException when {@code mode} argument equals neither
   * {@link CfgWalkerMode#DFS} nor {@link CfgWalkerMode#DFS_NO_RPT}
   */
  public CfgDfsWalker(
      final CfgVisitor visitor,
      final CfgWalkerMode mode) {
    super(visitor, mode);
    InvariantChecks.checkTrue(mode == CfgWalkerMode.DFS || mode == CfgWalkerMode.DFS_NO_RPT);
  }

  @Override
  public void goToNextStatement() {
    /* Do nothing. */
  }

  @Override
  public void processChildren(CfgStatement statement) {
    visitChildren(statement);
  }

  private void visitChildren(CfgStatement node) {

    Collection<CfgStatement> successors = CfgUtils.getSuccessors(node);

    for (final CfgStatement statement : successors) {

      if (getMode().equals(CfgWalkerMode.DFS_NO_RPT)) {
        if (getVisited().isEmpty() || !getVisited().contains(statement.getId())) {
          addVisited(statement.getId());
          visitNode(statement);
          if (isStatus(CfgVisitor.Status.ABORT)) {
            return;
          }
        }
      } else {
        visitNode(statement);

        if (isStatus(CfgVisitor.Status.ABORT)) {
          return;
        }
      }
    }
  }
}
