/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg.walker;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.cfg.CfgAssertStatement;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgLoopStatement;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgVisitor;
import ru.ispras.retrascope.model.cfg.CfgWaitStatement;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Control flow graph model walker.
 *
 * <p>Provides methods for traversing nodes of {@link CfgModel} control flow graph model.</p>
 *
 * <p>
 * Actions taken in the traversal process depend on the current status of the visitor (see
 * {@link ru.ispras.retrascope.model.cfg.CfgVisitor.Status}). Possible status values are:</p>
 * <ol>
 * <li>OK - continue traversal
 * <li>SKIP - skip child nodes
 * <li>ABORT - stop traversal
 * </ol>
 *
 * <p>The status is checked after calling any visitor method. Once {@code Status.ABORT} is set,
 * all traversal methods cease the execution. If after a call to a method having the Begin suffix
 * (e.g. onModuleBegin), the {@code Status.SKIP} status is set, nested elements of the visited node
 * (child nodes or subtrees) are not traversed and a corresponding terminating method
 * (that has the End suffix eg. onModuleEnd) is called.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class CfgWalker {

  private final CfgVisitor visitor;
  private final CfgWalkerMode mode;

  private final Set<String> visited = new LinkedHashSet<>();

  /**
   * Constructs object with the specified parameters.
   *
   * @param visitor visitor to be applied to control flow graph model
   * @param mode walking mode
   *
   * @throws IllegalArgumentException when any argument is {@code null}.
   */
  public CfgWalker(
      final CfgVisitor visitor,
      final CfgWalkerMode mode) {
    InvariantChecks.checkNotNull(visitor);
    InvariantChecks.checkNotNull(mode);

    this.visitor = visitor;
    this.mode = mode;
  }

  /**
   * Returns visitor.
   *
   * @see CfgVisitor
   *
   * @return visitor
   */
  public CfgVisitor getVisitor() {
    return this.visitor;
  }

  public CfgWalkerMode getMode() {
    return this.mode;
  }

  /**
   * Returns a collection of identifiers for statements that are already visited.
   *
   * @return identifiers of already visited statements
   */
  public Collection<String> getVisited() {
    return this.visited;
  }

  /**
   * Add the specified statement identifier to the storage of identifiers
   * for statements that are already visited by this walker.
   *
   * @param visitedId statement identifier that should be treated by this walker as already visited
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addVisited(final String visitedId) {
    InvariantChecks.checkNotNull(visitedId);

    this.visited.add(visitedId);
  }

  /**
   * Add statement identifiers from the specified collection to the storage of identifiers
   * for statements that are already visited by this walker.
   *
   * @param visitedIds identifiers for statements that should be treated as already visited
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addVisited(final Collection<String> visitedIds) {
    InvariantChecks.checkNotNull(visitedIds);

    for (final String visitedId : visitedIds) {
      addVisited(visitedId);
    }
  }

  /**
   * Starts control flow graph model walking.
   *
   * @param model control flow graph model
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void start(final CfgModel model) {
    InvariantChecks.checkNotNull(model);

    visitor.onRootBegin();

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visitCfgModel(model);
      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    visitor.onRootEnd();
  }

  /**
   * Visits the specified node.
   *
   * @param node node to be visited
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void visitNode(final CfgStatement node) {
    InvariantChecks.checkNotNull(node);

    visitor.onNodeBegin(node);
    switch (node.getType()) {
      case MODULE:
        visitModule((CfgModule) node);
        break;
      case PROCESS:
        visitProcess((CfgProcess) node);
        break;
      case ASSERT:
        visitAssert((CfgAssertStatement) node);
        break;
      case BASIC_BLOCK:
        visitBasicBlock((CfgBlockStatement) node);
        break;
      case CASE:
        visitCase((CfgCaseStatement) node);
        break;
      case SWITCH:
        visitSwitch((CfgSwitchStatement) node);
        break;
      case LOOP:
        visitLoop((CfgLoopStatement) node);
        break;
      case WAIT:
        visitWait((CfgWaitStatement) node);
        break;
      default:
        throw RetrascopeRuntimeException.makeException(LogMessage.UNSUPPORTED_TYPE, node.getType());
    }
    visitor.onNodeEnd(node);

    goToNextStatement();
  }

  /**
   * Process child statements of the specified one.
   *
   * @param statement Statement those children are to be processed.
   */
  public abstract void processChildren(final CfgStatement statement);

  /**
   * Next statement for walking choosing method.
   */
  public abstract void goToNextStatement();

  private void visitLoop(CfgLoopStatement node) {
    this.visitor.onLoopBegin(node);

    if (processStatement(node)) {
      return;
    }

    this.visitor.onLoopEnd(node);
  }

  /**
   * Checks rather this walker object has the specified status value.
   *
   * @param status Status value
   * @return {@code true} whether this object has the specified status value,
   *         {@code false} otherwise.
   */
  protected boolean isStatus(final CfgVisitor.Status status) {
    return this.visitor.getStatus() == status;
  }

  private void visitCfgModel(final CfgModel model) {

    this.visitor.onCfgModelBegin();

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {

      final Collection<CfgModule> modules = model.getModules();

      for (final CfgModule module : modules) {

        visitModule(module);

        if (isStatus(CfgVisitor.Status.ABORT)) {
          return;
        }
      }
    }
    this.visitor.onCfgModelEnd();
  }

  private void visitModule(final CfgModule node) {

    this.visitor.onModuleBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {

      final Collection<CfgProcess> processes = node.getProcesses();

      for (final CfgProcess process : processes) {

        visitProcess(process);

        if (isStatus(CfgVisitor.Status.ABORT)) {
          return;
        }
      }

      final Collection<CfgModule> modules = node.getModules();

      for (final CfgModule module : modules) {

        visitModule(module);

        if (isStatus(CfgVisitor.Status.ABORT)) {
          return;
        }
      }
    }

    this.visitor.onModuleEnd(node);
  }

  private void visitProcess(final CfgProcess node) {

    this.visitor.onProcessBegin(node);

    if (processStatement(node)) {
      return;
    }

    this.visitor.onProcessEnd(node);
  }

  private boolean processStatement(final CfgStatement node) {
    if (isStatus(CfgVisitor.Status.ABORT)) {
      return true;
    }

    if (isStatus(CfgVisitor.Status.OK)) {

      processChildren(node);

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return true;
      }
    }
    return false;
  }

  private void visitAssert(final CfgAssertStatement assertStatement) {

    this.visitor.onAssertBegin(assertStatement);

    if (processStatement(assertStatement)) {
      return;
    }

    this.visitor.onAssertEnd(assertStatement);
  }

  private void visitSwitch(final CfgSwitchStatement node) {

    this.visitor.onSwitchBegin(node);

    if (processStatement(node)) {
      return;
    }
    this.visitor.onSwitchEnd(node);
  }

  private void visitCase(final CfgCaseStatement node) {

    this.visitor.onCaseBegin(node);

    if (processStatement(node)) {
      return;
    }
    this.visitor.onCaseEnd(node);
  }

  private void visitBasicBlock(final CfgBlockStatement node) {

    this.visitor.onBasicBlockBegin(node);

    if (processStatement(node)) {
      return;
    }
    this.visitor.onBasicBlockEnd(node);
  }

  private void visitWait(final CfgWaitStatement node) {

    this.visitor.onWaitBegin(node);

    if (processStatement(node)) {
      return;
    }
    this.visitor.onWaitEnd(node);
  }
}
