/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.VariableContainer;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * The representation of an extended finite state machine.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class Efsm extends VariableContainer {

  /**
   * An object representing a state based on uninitialised values.
   */
  public static final EfsmState UNINIT_STATE = new EfsmState(SolverUtils.FALSE);

  private static final String EFSM_STATE_NOT_FOUND =
      "The specified state is not found in this EFSM: ";

  private final EfsmModule parent;
  private final String name;

  /* Clock-like variables for this object. */
  private final Set<NodeVariable> clocks;

  private final Set<EfsmState> states;
  private final Set<EfsmTransition> transitions;

  private EfsmState initialState;
  private final Set<GuardedAction> resetGuardedActions;

  /**
   * Constructs an empty EFSM with the specified name and parent module.
   *
   * @param parent the module which contains the constructed EFSM
   * @param name the EFSM name
   *
   * @throws IllegalArgumentException if either {@code parent}
   *         or {@code name} is {@code null}
   */
  public Efsm(EfsmModule parent, String name) {
    super();

    InvariantChecks.checkNotNull(parent);
    InvariantChecks.checkNotNull(name);

    this.parent = parent;
    this.name = name;

    this.clocks = new LinkedHashSet<>();
    this.states = new HashSet<>();
    this.transitions = new HashSet<>();

    this.initialState = null;

    this.resetGuardedActions = new HashSet<>(1);
  }

  /**
   * Returns parent container of this EFSM.
   *
   * @return the parent EFSM container
   */
  public EfsmModule getParentContainer() {
    return this.parent;
  }

  /**
   * Checks if the specified input declared either in this EFSM or in its parent module.
   *
   * @param variableName a name of the input
   *
   * @return {@code true} if the input is declared. Otherwise returns {@code false}.
   */
  public boolean isInputDeclared(String variableName) {
    return this.containsInput(variableName) || this.parent.containsInput(variableName);
  }

  /**
   * Checks if the specified output declared either in this EFSM or in its parent module.
   *
   * @param variableName a name of the output
   *
   * @return {@code true} if the output is declared. Otherwise returns {@code false}.
   */
  public boolean isOutputDeclared(String variableName) {
    return this.containsOutput(variableName) || this.parent.containsOutput(variableName);
  }

  /**
   * Checks if the specified register declared either in this EFSM or in its parent module.
   *
   * @param variableName a name of the register
   *
   * @return {@code true} if the register is declared. Otherwise returns {@code false}.
   */
  public boolean isRegisterDeclared(String variableName) {
    return this.containsRegister(variableName)
        || this.parent.containsRegister(variableName);
  }

  @Override
  public DataType getDataType(final String variableName) {
    DataType dataType = super.getDataType(variableName);

    if (dataType == null) {
      dataType = this.parent.getDataType(variableName);
    }

    return dataType;
  }

  @Override
  public Node getInitialValue(final String variableName) {
    Node initialValue = super.getInitialValue(variableName);

    if (initialValue == null) {
      initialValue = this.parent.getInitialValue(variableName);
    }

    return initialValue;
  }

  @Override
  public Node getInvariant(final String variableName) {
    Node condition = super.getInvariant(variableName);

    if (condition == null) {
      condition = this.parent.getInvariant(variableName);
    }

    return condition;
  }

  /**
   * Creates a new state based on the specified logical expression and adds it into this EFSM. All
   * variables which are in the expression must be declared either in this EFSM or in its parent
   * before the creation.
   *
   * @param expression the expression representing the state
   *
   * @return the created state
   *
   * @throws IllegalArgumentException if at least one variable which is in the expression is not
   *         declared neither in this EFSM nor in its parent
   */
  public EfsmState createState(Node expression) {
    EfsmState state = new EfsmState(expression);
    this.addState(state);

    return state;
  }

  /**
   * Adds the specified state into this EFSM.
   *
   * @param state the state to be added
   *
   * @throws NullPointerException if {@code state} is {@code null}
   * @throws IllegalArgumentException if at least one variable which is in the expression is not
   *         declared neither in this EFSM nor in its parent
   * @throws IllegalArgumentException when argument is {@code null}
   */
  public void addState(EfsmState state) {
    InvariantChecks.checkNotNull(state);
    this.processAddedExpression(state.getExpression());
    this.states.add(state);
  }

  /**
   * Creates a new transition with the specified guarded action between the specified states and
   * adds it into this EFSM. All variables which are in the guarded action must be declared either
   * in this EFSM or in its parent before the creation.
   *
   * @param sourceState a source state
   * @param destinationState a destination state
   * @param guardedAction the guarded action
   *
   * @return the created transition
   *
   * @throws IllegalArgumentException if the specified source and/or destination state is not found
   *         in this EFSM or at least one variable which is in {@code guardedAction} is not
   *         declared neither in this EFSM nor in its parent
   */
  public EfsmTransition createTransition(EfsmState sourceState, EfsmState destinationState,
      GuardedAction guardedAction) {
    if (!this.states.contains(sourceState)) {
      throw new IllegalArgumentException(
          "The specified source state of a new transition is not found in this EFSM: "
              + sourceState);
    }

    if (!this.states.contains(destinationState)) {
      throw new IllegalArgumentException(
          "The specified destination state of a new transition is not found in this EFSM: "
              + destinationState);
    }

    EfsmTransition transition = new EfsmTransition(sourceState, destinationState, guardedAction);
    this.transitions.add(transition);
    this.processAddedExpression(guardedAction.getGuard().toNode());

    for (Assignment assignment: guardedAction.getAction().getAssignments()) {
      this.processAddedExpression(assignment.getValue());

      if (assignment.getTarget() != null) {
        this.processAddedExpression(assignment.getVariable());
      }
    }

    return transition;
  }

  /**
   * Returns a name of this EFSM.
   *
   * @return the EFSM name
   */
  public String getName() {
    return this.name;
  }

  /**
   * Checks if the specified state is in this EFSM.
   *
   * @param state the state whose presence in the EFSM to be tested
   *
   * @return {@code true} if this EFSM contains the specified state. Otherwise returns {@code false}
   */
  public boolean containsState(EfsmState state) {
    return this.states.contains(state);
  }

  /**
   * Marks the specified state of this EFSM as an initial state.
   *
   * @param state the new initial state of this EFSM
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  public void setInitialState(EfsmState state) {
    if (!this.states.contains(state)) {
      throw new IllegalArgumentException(EFSM_STATE_NOT_FOUND + state);
    }

    this.initialState = state;
  }

  /**
   * Returns an initial state of this EFSM.
   *
   * @return the initial state of this EFSM
   */
  public EfsmState getInitialState() {
    return this.initialState;
  }

  /**
   * Adds the specified guarded action to this EFSM. This guarded action represents a precondition
   * of the initial state and an action which is performed during initialisation of this EFSM.
   *
   * @param resetGuardedAction the reset guarded action
   *
   * @throws IllegalArgumentException if the {@code resetGuardedAction} is {@code null}
   */
  public void addResetGuardedAction(GuardedAction resetGuardedAction) {
    InvariantChecks.checkNotNull(resetGuardedAction);

    /* If variable is used at reset guarded action guard, it cannot be in model clocks. */
    final Collection<NodeVariable> resetVariables =
        ExprUtils.getVariables(resetGuardedAction.getGuard().toNode());

    final Set<NodeVariable> filtered = new LinkedHashSet<>();
    for (final NodeVariable reset : resetVariables) {
      if (getClocks().contains(reset)) {
        filtered.add(reset);
      }
    }
    this.clocks.removeAll(filtered);

    this.resetGuardedActions.add(resetGuardedAction);
  }

  /**
   * Returns reset guarded actions of this EFSM, that is, such guarded actions which represent an
   * initialisation of this EFSM.
   *
   * @return the reset guarded actions
   */
  public Set<GuardedAction> getResetGuardedActions() {
    return Collections.unmodifiableSet(this.resetGuardedActions);
  }

  /**
   * Add the specified object as clock-like variable for this model.
   * <p>When the specified variable is used in a guard of some resetting transition,
   * this method does nothing, because a variable cannot be placed both at clocks and resets.</p>
   * @param clockVariable clock-like variable
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addClock(final NodeVariable clockVariable) {
    InvariantChecks.checkNotNull(clockVariable);

    if (!usedAsReset(clockVariable)) {
      this.clocks.add(clockVariable);
    }
  }

  private boolean usedAsReset(final NodeVariable variable) {

    for (final GuardedAction reset : getResetGuardedActions()) {
      if (ExprUtils.getVariables(reset.getGuard().toNode()).contains(variable)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Returns clock-like variables related to this model.
   * @return clock-like variables
   */
  public Collection<NodeVariable> getClocks() {
    return Collections.unmodifiableSet(this.clocks);
  }

  /**
   * Returns all states of this EFSM.
   *
   * @return a set containing the states of this EFSM
   */
  public Set<EfsmState> getStates() {
    return Collections.unmodifiableSet(this.states);
  }

  /**
   * Returns all transitions of this EFSM.
   *
   * @return a set containing the transitions of this EFSM
   */
  public Set<EfsmTransition> getTransitions() {
    return Collections.unmodifiableSet(this.transitions);
  }

  /**
   * Checks if this EFSM is synchronous, that is, contains at least one event-sensible guarded
   * action.
   *
   * @return {@code true} if this EFSM is synchronous. Otherwise returns {@code false}
   */
  public boolean isSynchronous() {
    for (EfsmTransition transition: this.transitions) {
      if (!transition.getGuardedAction().getClocks().isEmpty()) {
        return true;
      }
    }

    return false;
  }

  /**
   * Checks if there is a path from the specified source state to the specified destination state.
   *
   * @param sourceState the source state
   * @param destinationState the destination state
   *
   * @return {@code true} if there is a path from the source state to the destination state.
   *         Otherwise returns {@code false}
   *
   * @throws IllegalArgumentException if the specified source and/or destination state is not found
   *         in this EFSM
   */
  public boolean isReachable(EfsmState sourceState, EfsmState destinationState) {
    if (!this.states.contains(sourceState)) {
      throw new IllegalArgumentException("The specified source state is not found in this EFSM");
    }

    if (!this.states.contains(destinationState)) {
      throw new IllegalArgumentException(
          "The specified destination state is not found in this EFSM");
    }

    if (sourceState.equals(destinationState)) {
      return true;
    }

    LinkedHashSet<EfsmState> visitedStates = new LinkedHashSet<>(this.states.size());
    LinkedList<EfsmState> queue = new LinkedList<>();
    visitedStates.add(sourceState);
    queue.add(sourceState);

    while (!queue.isEmpty()) {
      for (EfsmState nextState : this.getDestinationStates(queue.poll())) {
        if (nextState.equals(destinationState)) {
          return true;
        }

        if (!visitedStates.contains(nextState)) {
          queue.add(nextState);
          visitedStates.add(nextState);
        }
      }
    }

    return false;
  }

  /**
   * Looks for the shortest path from the specified source state to the specified destination state
   * with the specified principle of transition weight defining. This method is based on the
   * Dijkstra's algorithm.
   * <br>
   * Currently CONST and GUARD_REGISTER_AMOUNT TransitionWeightBasis types are supported.
   *
   * @param sourceState the source state
   * @param destinationState the destination state
   * @param weightBasis the principle used for transition weight computing
   *
   * @return a list containing the transitions of found path.
   *         It's empty if the source state matches
   *         to the destination state. If there is no path
   *         from the source state to the destination
   *         state {@code null} is returned
   *
   * @throws IllegalArgumentException if the specified source and/or destination state is not found
   *         in this EFSM or if unsupported {@link TransitionWeightBasis} is passed
   */
  public List<EfsmTransition> getShortestPath(EfsmState sourceState, EfsmState destinationState,
      TransitionWeightBasis weightBasis) {
    if (!this.states.contains(sourceState)) {
      throw new IllegalArgumentException("The specified source state is not found in this EFSM");
    }

    if (!this.states.contains(destinationState)) {
      throw new IllegalArgumentException(
          "The specified destination state is not found in this EFSM");
    }

    if (sourceState.equals(destinationState)) {
      return Collections.emptyList();
    }

    if (!this.isReachable(sourceState, destinationState)) {
      return null;
    }

    final LinkedHashMap<EfsmState, Integer> distances = new LinkedHashMap<>(this.states.size());
    LinkedHashMap<EfsmState, EfsmTransition> previousNodes =
        new LinkedHashMap<>(this.states.size());

    for (EfsmState state : this.states) {
      distances.put(state, Integer.MAX_VALUE);
      previousNodes.put(state, null);
    }

    distances.put(sourceState, 0);
    final LinkedHashMap<EfsmTransition, Integer> transitionWeights = new LinkedHashMap<>();

    for (EfsmTransition transition : this.transitions) {
      switch (weightBasis) {
        case CONSTANT:
          transitionWeights.put(transition, 1);

          break;
        case GUARD_REGISTER_AMOUNT:
          int weight = 0;

          for (NodeVariable variable : transition.getGuardedAction().getGuard().getUses()) {
            String variableName = variable.getName();
            if (super.containsRegister(variableName)
                || this.parent.containsRegister(variableName)) {
              weight++;
            }
          }

          transitionWeights.put(transition, weight);

          break;
        default:
          throw new IllegalArgumentException("Unsupported weight defining principle: "
              + weightBasis);
      }
    }

    Comparator<EfsmTransition> transitionComparator = new Comparator<EfsmTransition>() {
      @Override
      public int compare(EfsmTransition transition1, EfsmTransition transition2) {
        return transitionWeights.get(transition1) - transitionWeights.get(transition2);
      }
    };

    Comparator<EfsmState> stateComparator = new Comparator<EfsmState>() {
      @Override
      public int compare(EfsmState state1, EfsmState state2) {
        return distances.get(state1) - distances.get(state2);
      }
    };
    LinkedHashSet<EfsmState> visitedStates = new LinkedHashSet<>(this.states.size());
    visitedStates.add(sourceState);
    LinkedList<EfsmState> queue = new LinkedList<>();
    queue.add(sourceState);

    while (!queue.isEmpty()) {
      EfsmState currentState = queue.poll();
      int currentDistance = distances.get(currentState);
      LinkedList<EfsmTransition> nextTransitions =
          new LinkedList<>(this.getOutgoingTransitions(currentState));
      Collections.sort(nextTransitions, transitionComparator);

      for (EfsmTransition nextTransition : nextTransitions) {
        EfsmState nextState = nextTransition.getDestinationState();
        int newDistance = currentDistance + transitionWeights.get(nextTransition);

        if (newDistance < distances.get(nextState)) {
          previousNodes.put(nextState, nextTransition);
          distances.put(nextState, newDistance);

          if (nextState.equals(destinationState)) {
            break;
          }

          if (!visitedStates.contains(nextState)) {
            visitedStates.add(nextState);
            queue.add(nextState);
          }
        }
      }

      Collections.sort(queue, stateComparator);
    }

    LinkedList<EfsmTransition> result = new LinkedList<>();
    EfsmState currentState = destinationState;

    while (previousNodes.get(currentState) != null) {
      result.addFirst(previousNodes.get(currentState));
      currentState = previousNodes.get(currentState).getSourceState();
    }

    return result;
  }

  /**
   * Returns transitions which ends at the specified state.
   *
   * @param state the state to be tested
   *
   * @return a set containing incoming transitions for the specified state
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  public Set<EfsmTransition> getIncomingTransitions(EfsmState state) {
    if (!this.states.contains(state)) {
      throw new IllegalArgumentException(EFSM_STATE_NOT_FOUND + state);
    }

    Set<EfsmTransition> incomingTransitions = new HashSet<>();

    for (EfsmTransition transition: this.transitions) {
      if (transition.getDestinationState().equals(state)) {
        incomingTransitions.add(transition);
      }
    }

    return incomingTransitions;
  }

  /**
   * Returns transitions which starts at the specified state.
   *
   * @param state the state to be tested
   *
   * @return a set containing outgoing transitions for the specified state
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  public Set<EfsmTransition> getOutgoingTransitions(EfsmState state) {
    if (!this.states.contains(state)) {
      throw new IllegalArgumentException(EFSM_STATE_NOT_FOUND + state);
    }

    Set<EfsmTransition> outgoingTransitions = new HashSet<>();

    for (EfsmTransition transition: this.transitions) {
      if (transition.getSourceState().equals(state)) {
        outgoingTransitions.add(transition);
      }
    }

    return outgoingTransitions;
  }

  /**
   * Returns states which are start points of incoming transitions of the specified state.
   *
   * @param state the state to be tested
   *
   * @return a set containing the source states for the specified state
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  public Set<EfsmState> getSourceStates(EfsmState state) {
    if (!this.states.contains(state)) {
      throw new IllegalArgumentException(EFSM_STATE_NOT_FOUND + state);
    }

    Set<EfsmTransition> incomingTransitions = this.getIncomingTransitions(state);

    if (incomingTransitions.isEmpty()) {
      return Collections.emptySet();
    }

    Set<EfsmState> sourceStates = new HashSet<>(incomingTransitions.size());

    for (EfsmTransition transition: incomingTransitions) {
      sourceStates.add(transition.getSourceState());
    }

    return sourceStates;
  }

  /**
   * Returns states which are end points of outgoing transitions of the specified state.
   *
   * @param state the state to be tested
   *
   * @return a set containing the destination states for the specified state
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  public Set<EfsmState> getDestinationStates(EfsmState state) {
    if (!this.states.contains(state)) {
      throw new IllegalArgumentException(EFSM_STATE_NOT_FOUND + state);
    }

    Set<EfsmTransition> outgoingTransitions = this.getOutgoingTransitions(state);

    if (outgoingTransitions.isEmpty()) {
      return Collections.emptySet();
    }

    Set<EfsmState> destinationStates = new HashSet<>(outgoingTransitions.size());

    for (EfsmTransition transition : outgoingTransitions) {
      destinationStates.add(transition.getDestinationState());
    }

    return destinationStates;
  }

  /**
   * Checks if this EFSM contains states or not.
   *
   * @return {@code true} if this EFSM contains states, {@code false} otherwise
   */
  public boolean containsStates() {
    return !this.states.isEmpty();
  }

  private void processAddedExpression(final Node expression) {
    if (expression != null) {
      for (final NodeVariable variable: ExprUtils.getVariables(expression)) {
        final String variableName = variable.getName();

        if (!EfsmUtils.hasDeclaringAncestor(this, variable)) {
          throw new IllegalArgumentException("There is no declaration of variable " + variableName
              + " neither in this EFSM nor in its ancestors.");
        }
      }
    }
  }

  public enum TransitionWeightBasis {
    /**
     * All transitions are considered as having one and the same weight.
     */
    CONSTANT,
    /**
     * A transition's weight is equal to amount of registers used in the transition's guard.
     */
    GUARD_REGISTER_AMOUNT
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = super.hashCode();
    result = prime * result + ((this.initialState == null) ? 0 : this.initialState.hashCode());
    result = prime * result + this.name.hashCode();
    result = prime * result + this.resetGuardedActions.hashCode();
    result = prime * result + this.states.hashCode();

    return prime * result + this.transitions.hashCode();
  }

  @Override
  public boolean equals(Object object) {
    if (this == object) {
      return true;
    }

    if (!super.equals(object) || !(object instanceof Efsm)) {
      return false;
    }

    Efsm otherEfsm = (Efsm) object;

    if (this.initialState == null) {
      if (otherEfsm.initialState != null) {
        return false;
      }
    } else if (!this.initialState.equals(otherEfsm.initialState)) {
      return false;
    }

    return this.name.equals(otherEfsm.name) && this.states.equals(otherEfsm.states)
        && this.resetGuardedActions.equals(otherEfsm.resetGuardedActions)
        && this.transitions.equals(otherEfsm.transitions);
  }

  @Override
  public String toString() {
    return this.parent.getName() + "." + this.name;
  }
}
