/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

/**
 * The ordered container for EFSM modules.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see Entity
 */
public class EfsmModel extends Entity {
  /**
   * The entity type of all {@code EfsmModel} entities.
   */
  public static final EntityType TYPE = EntityType.get("efsm");

  private final List<EfsmModule> efsmModules;

  /**
   * Creates an empty EFSM model.
   */
  public EfsmModel() {
    super(EfsmModel.TYPE);

    this.efsmModules = new ArrayList<>();
  }

  /**
   * Creates an EFSM model containing the specified EFSM module.
   *
   * @param efsmModule the EFSM module to be placed into the EFSM model
   *
   * @throws IllegalArgumentException if {@code efsmModule} is {@code null}
   */
  public EfsmModel(EfsmModule efsmModule) {
    this();

    InvariantChecks.checkNotNull(efsmModule);
    this.efsmModules.add(efsmModule);
  }

  /**
   * Creates an EFSM model containing the specified EFSM modules.
   *
   * @param efsmModules the EFSM modules to be placed into the EFSM model
   *
   * @throws IllegalArgumentException if {@code efsmModules} is {@code null}
   */
  public EfsmModel(Collection<EfsmModule> efsmModules) {
    this();

    InvariantChecks.checkNotNull(efsmModules);
    this.efsmModules.addAll(efsmModules);
  }

  /**
   * Adds the specified EFSM module to this EFSM model.
   *
   * @param efsmModule the EFSM module to be added
   *
   * @throws IllegalArgumentException if {@code efsmModule} is {@code null}
   */
  public void addModule(EfsmModule efsmModule) {
    InvariantChecks.checkNotNull(efsmModule);
    this.efsmModules.add(efsmModule);
  }

  /**
   * Adds the specified EFSM modules to this EFSM model.
   *
   * @param efsmModules the EFSM modules to be added
   *
   * @throws IllegalArgumentException if {@code efsmModules} is {@code null}
   */
  public void addModules(Collection<EfsmModule> efsmModules) {
    InvariantChecks.checkNotNull(efsmModules);
    this.efsmModules.addAll(efsmModules);
  }

  /**
   * Returns EFSM modules of this EFSM model.
   *
   * @return a set of the EFSM modules
   */
  public List<EfsmModule> getModules() {
    return Collections.unmodifiableList(this.efsmModules);
  }

  @Override
  public void add(Entity entity) {
    if (entity.getEntityType() == EfsmModel.TYPE) {
      this.efsmModules.addAll(((EfsmModel) entity).getModules());
    } else {
      throw new IllegalArgumentException("Illegal entity type: " + entity.getEntityType());
    }
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = super.hashCode();
    result = prime * result + this.efsmModules.hashCode();

    return result;
  }

  @Override
  public boolean equals(Object object) {
    if (this == object) {
      return true;
    }

    if (!super.equals(object) || !(object instanceof EfsmModel)) {
      return false;
    }

    EfsmModel otherEfsmModel = (EfsmModel) object;

    return this.efsmModules.equals(otherEfsmModel.efsmModules);
  }
}
