/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.VariableContainer;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

/**
 * The EFSM representation of an HDL module. It consists of EFSMs which share all variables declared
 * in {@code EfsmModule} and nested modules having their own variable space.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class EfsmModule extends VariableContainer {
  private final String name;

  private EfsmModule parent;
  private final Set<Efsm> initialEfsms;
  private final Set<Efsm> nonInitialEfsms;

  private final Set<EfsmModule> nestedModules;

  /**
   * Constructs an empty EFSM module.
   *
   * @param name a name of the module
   *
   * @throws IllegalArgumentException if {@code name} is {@code null}
   */
  public EfsmModule(String name) {
    super();

    InvariantChecks.checkNotNull(name);

    this.name = name;

    this.initialEfsms = new HashSet<>();
    this.nonInitialEfsms = new HashSet<>();

    this.nestedModules = new HashSet<>();
  }

  /**
   * Sets the specified module as parent for this one.
   * <p>This method also checks whether parent module contains this one
   * as a child, and if not - adds it to the nested modules collection.</p>
   *
   * @param parent module to be set as parent
   */
  public void setParent(final EfsmModule parent) {
    InvariantChecks.checkNotNull(parent);

    this.parent = parent;
    if (!parent.getNestedModules().contains(this)) {
      parent.addNestedModule(this);
    }
  }

  /**
   * Returns parent module for this one.
   *
   * @return parent module
   */
  public EfsmModule getParent() {
    return this.parent;
  }

  /**
   * Returns this module's name.
   *
   * @return this module's name
   */
  public String getName() {
    return this.name;
  }

  /**
   * Adds the specified non-initial EFSM into this module.
   *
   * @param efsm the efsm to be added
   *
   * @throws IllegalArgumentException if {@code efsm} is {@code null}
   */
  public void addNonInitialEfsm(Efsm efsm) {
    InvariantChecks.checkNotNull(efsm);

    this.nonInitialEfsms.add(efsm);
  }

  /**
   * Adds the specified EFSM into this module as so-called "initial EFSM" which must be
   * simulated in the very beginning of this module simulation. If there are more than one initial
   * EFSMs in this module, they are executed in parallel.
   *
   * @param efsm the efsm to be added
   *
   * @throws IllegalArgumentException if {@code efsm} is {@code null}
   */
  public void addInitialEfsm(Efsm efsm) {
    InvariantChecks.checkNotNull(efsm);

    this.initialEfsms.add(efsm);
  }

  /**
   * Returns top-level non-initial EFSMs of this EFSM module.
   *
   * @return the EFSMs
   */
  public Set<Efsm> getNonInitialEfsms() {
    return Collections.unmodifiableSet(this.nonInitialEfsms);
  }

  /**
   * Returns top-level initial EFSMs of this EFSM module.
   *
   * @return the EFSMs
   */
  public Set<Efsm> getInitialEfsms() {
    return Collections.unmodifiableSet(this.initialEfsms);
  }

  /**
   * Adds the specified EFSM module into this module.
   *
   * @param efsmModule the EFSM module to be added
   *
   * @throws IllegalArgumentException if {@code efsmModule} is {@code null}
   */
  public void addNestedModule(EfsmModule efsmModule) {
    InvariantChecks.checkNotNull(efsmModule);

    this.nestedModules.add(efsmModule);
    if (efsmModule.getParent() != this) {
      efsmModule.setParent(this);
    }
  }

  /**
   * Returns nested EFSM modules of this EFSM container.
   *
   * @return the EFSM modules
   */
  public Set<EfsmModule> getNestedModules() {
    return Collections.unmodifiableSet(this.nestedModules);
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = super.hashCode();
    result = prime * result + this.initialEfsms.hashCode();
    result = prime * result + this.nonInitialEfsms.hashCode();
    result = prime * result + this.name.hashCode();

    return prime * result + this.nestedModules.hashCode();
  }

  @Override
  public boolean equals(Object object) {
    if (this == object) {
      return true;
    }

    if (!super.equals(object) || !(object instanceof EfsmModule)) {
      return false;
    }

    EfsmModule otherModule = (EfsmModule) object;

    return this.nonInitialEfsms.equals(otherModule.nonInitialEfsms)
        && this.name.equals(otherModule.name) && this.initialEfsms.equals(otherModule.initialEfsms)
        && this.nestedModules.equals(otherModule.nestedModules);
  }

  @Override
  public String toString() {
    return this.name;
  }
}
