/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.State;

/**
 * A state of an extended finite state machine. The state representation is based on a logical
 * expression in form of a {@link ru.ispras.fortress.expression.Node} instance.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class EfsmState extends State {
  private final Node expression;

  /**
   * Constructs a state represented by the specified expression.
   *
   * @param expression the expression representing the state
   *
   * @throws IllegalArgumentException if {@code expression} is {@code null}
   */
  public EfsmState(Node expression) {
    InvariantChecks.checkNotNull(expression);

    this.expression = expression;
  }

  /**
   * Returns an expression which this state is represented by.
   *
   * @return the expression representing this state
   */
  public Node getExpression() {
    return this.expression;
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + this.expression.hashCode();

    return result;
  }

  @Override
  public String getId() {
    return this.expression.toString();
  }

  /**
   * Compares the specified object with this state for equality. The object is considered as equal
   * to this state if and only if it is also an EFSM state and its underlying logical
   * expression is the same name as this state's one, or more formally:
   * <ul>
   *    <li>it is an instance of the {@code EfsmState} class or its subclass;
   *    <li>its underlying logical expression is equal to this state's one.
   * </ul>
   *
   * @param obj the object to be compared for equality with this state
   *
   * @return {@code true} if the specified object is equal to this state. Otherwise returns
   * {@code false}
   */
  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }

    if (!(obj instanceof EfsmState) || (this.hashCode() != obj.hashCode())) {
      return false;
    }

    EfsmState stateForComparison = (EfsmState) obj;

    return this.expression.equals(stateForComparison.expression);
  }

  @Override
  public String toString() {
    return getId();
  }
}
