/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.RangedVariableUtils;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Utility methods for extended finite state machine model and it's components.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class EfsmUtils {

  private EfsmUtils() {
    /* Empty. */
  }

  /**
   * Returns mapping between modified variables and affecting variables.
   * <p>
   * Returns mappings between modified variables and variables which affects corresponding
   * modifications.
   * </p>
   * <br>
   * <p>
   * For example, if this action consists of two subsequent assignments {@code x = y} and
   * {@code z = x + i}, the resulting map contains mappings from {@code x} to {@code y} and from
   * {@code z} to {@code y} and {@code i}.
   * </p>
   *
   * @return a map containing modified variables and variables which affect corresponding
   *         modifications
   */
  public static Map<NodeVariable, Set<NodeVariable>> getResultingDependencies(final Action action) {

    final List<Assignment> assignments = action.getAssignments();
    final Map<NodeVariable, Set<NodeVariable>> substitutions = new HashMap<>(assignments.size());

    for (final Assignment assignment : assignments) {
      final Collection<NodeVariable> usedVariables = ExprUtils.getVariables(assignment.getValue());

      Set<NodeVariable> affecting;

      if (usedVariables.isEmpty()) {
        affecting = Collections.emptySet();
      } else {
        affecting = new HashSet<>(usedVariables.size());

        for (final NodeVariable variable : usedVariables) {
          if (substitutions.containsKey(variable)) {
            affecting.addAll(substitutions.get(variable));
          } else {
            affecting.add(variable);
          }
        }
      }
      substitutions.put(assignment.getVariable(), affecting);
    }
    return substitutions;
  }

  /**
   * Returns mapping between variables and corresponding substitutions.
   * <p>
   * Returns mappings between variables and corresponding substitutions representing total variable
   * modifications during execution of this action. That is, execution of this action is equal to
   * parallel evaluation of the substitutions and subsequent assignment of derived values to the
   * corresponding variables.
   * </p>
   *
   * @return a map containing modified variables and the substitutions representing these
   *         modifications
   */
  public static Map<NodeVariable, Node> getResultingSubstitutions(final Action action) {

    final List<Assignment> assignments = action.getAssignments();
    final Map<NodeVariable, Node> substitutions = new LinkedHashMap<>();

    for (final Assignment assignment : assignments) {
      Node expression = assignment.getValue();

      for (final NodeVariable variable : ExprUtils.getVariables(expression)) {
        if (substitutions.containsKey(variable)) {
          expression =
              Transformer.substitute(expression, variable.getName(), substitutions.get(variable));
        }
      }

      final RangedVariable rangedVariable = assignment.getTarget();
      final NodeVariable nodeVariable = rangedVariable.getVariable();

      if (rangedVariable.isRanged()) {
        final DataType dataType = nodeVariable.getDataType();

        if (!dataType.getTypeId().equals(DataTypeId.BIT_VECTOR)) {
          throw new IllegalArgumentException(
              LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, dataType));
        }

        final Range range = rangedVariable.getRange();

        if (!(range.getOld() instanceof NodeValue) || !(range.getYoung() instanceof NodeValue)) {
          throw new IllegalArgumentException(LogMessage.UNSUPPORTED_TYPE + "varying range bounds");
        }

        final NodeValue leftValue = (NodeValue) range.getYoung();
        final NodeValue rightValue = (NodeValue) range.getOld();

        final int leftBound = leftValue.getInteger().intValue();
        final int rightBound = rightValue.getInteger().intValue();

        if (leftBound < rightBound) {
          throw new IllegalArgumentException(LogMessage.UNSUPPORTED_TYPE.toString());
        }

        NodeOperation mostSignPart = null;
        NodeOperation lessSignPart = null;
        final int variableSize = dataType.getSize();
        final int decRadix = 10;

        if (leftBound < variableSize - 1) {
          final NodeValue leftIndex =
              new NodeValue(DataType.INTEGER.valueOf(String.valueOf(variableSize - 1), decRadix));
          final NodeValue rightIndex =
              new NodeValue(DataType.INTEGER.valueOf(String.valueOf(leftBound + 1), decRadix));

          mostSignPart =
              new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex, nodeVariable);
        }

        if (rightBound > 0) {
          final NodeValue leftIndex =
              new NodeValue(DataType.INTEGER.valueOf(String.valueOf(rightBound - 1), decRadix));
          final NodeValue rightIndex = new NodeValue(DataType.INTEGER.valueOf("0", decRadix));

          lessSignPart =
              new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex, nodeVariable);
        }

        if (mostSignPart == null || lessSignPart == null) {
          if (mostSignPart == null && lessSignPart != null) {
            expression = new NodeOperation(StandardOperation.BVCONCAT, expression, lessSignPart);
          } else {
            expression = new NodeOperation(StandardOperation.BVCONCAT, mostSignPart, expression);
          }
        } else {
          expression =
              new NodeOperation(StandardOperation.BVCONCAT, mostSignPart, expression, lessSignPart);
        }
      }

      if (substitutions.containsKey(nodeVariable)) {
        substitutions.remove(nodeVariable);
      }

      substitutions.put(nodeVariable, expression);
    }

    return substitutions;
  }

  /**
   * Constructs a synchronous composition from a list of EFSMs.
   *
   * @param efsmList the list of EFSMs
   * @return the synchronous composition of the EFSMs
   * @throws IllegalArgumentException if the specified list of EFSMs is {@code null}
   *         or contains less than two models
   */
  public static Efsm product(Efsm... efsmList) {

    if ((efsmList == null) || (efsmList.length < 2)) {
      throw new IllegalArgumentException("The number of EFSMs should be greater than 1!");
    }

    // This data structure represents a state with its outgoing transitions.

    class StateWithTransitions {
      public EfsmState state;
      public Collection<EfsmTransition> transitions;

      public StateWithTransitions(EfsmState state, Collection<EfsmTransition> transitions) {
        this.state = state;
        this.transitions = transitions;
      }
    }

    // Constructing a new EFSM name.

    String productName = efsmList[0].getName();
    for (int i = 1; i < efsmList.length; i++) {
      productName += (" || " + efsmList[i].getName());
    }
    final Efsm product = new Efsm(new EfsmModule("product"), productName);

    // Each EFSM is represented as a list of states with transitions
    List<List<StateWithTransitions>> oldStates = new ArrayList<>(efsmList.length);
    List<EfsmState> initialStates = new ArrayList<>();

    for (Efsm efsm : efsmList) {

      // Collecting initial states.
      initialStates.add(efsm.getInitialState());

      // Collecting input, output and inner signals.
      final List<EfsmModule> parentModules = getAllParentModules(efsm.getParentContainer());
      for (EfsmModule module : parentModules) {
        for (NodeVariable variable : module.getVariables()) {
          if (!product.containsVariable(variable)) {
            product.declareVariable((NodeVariable) variable.deepCopy());
          }
        }
      }
      for (NodeVariable variable : efsm.getVariables()) {
        if (!product.containsVariable(variable)) {
          product.declareVariable((NodeVariable) variable.deepCopy());
        }
      }

      ArrayList<StateWithTransitions> swt = new ArrayList<>(efsm.getStates().size());

      for (EfsmState state : efsm.getStates()) {
        Collection<EfsmTransition> trans = efsm.getOutgoingTransitions(state);
        if (!trans.isEmpty()) {
          List<Node> guards = new ArrayList<>();

          // If the guards of the outgoing transitions are not complete (a situation is
          // possible when no transition is enabled), a negation of disjunction of guards
          // is constructed.

          for (EfsmTransition item : trans) {
            guards.add(item.getGuardedAction().getGuard().toNode());
          }
          if (!areComplete(trans)) {
            trans.add(new EfsmTransition(state, state, new GuardedAction(new Guard(ExprUtils
                .getComplement(guards.toArray(new Node[guards.size()]))), new Action(
                new Assignment()))));
          }
        } else {

          // If a state has no outgoing transitions, a new transition is constructed
          // (guard = true, action = null). This is needed by the construction algorithm.

          trans.add(new EfsmTransition(state, state, new GuardedAction(new Guard(NodeValue
              .newBoolean(true)), new Action(new Assignment()))));
        }
        swt.add(new StateWithTransitions(state, new ArrayList<>(trans)));

      }
      oldStates.add(swt);

    }

    // Constructing a new initial state.

    Node initialStateNode = null;
    for (EfsmState initState : initialStates) {
      if (initialStateNode == null) {
        initialStateNode = initState.getExpression();
      } else {
        initialStateNode = Node.AND(initialStateNode, initState.getExpression());
      }
    }

    EfsmState newInitialState = new EfsmState(initialStateNode);

    // Combining old EFSM states to construct new EFSM states.

    List<List<StateWithTransitions>> stateCombinations = combineLists(oldStates);

    LinkedHashMap<EfsmState, List<List<EfsmTransition>>> transitionsMapping = new LinkedHashMap<>();

    // Combining transitions of new state combinations.

    for (List<StateWithTransitions> combinedStateList : stateCombinations) {
      Node stateNode = null;
      List<List<EfsmTransition>> newStateTransitions = new ArrayList<>();
      List<List<EfsmTransition>> combinedTransitions;

      for (StateWithTransitions item : combinedStateList) {
        if (stateNode == null) {
          stateNode = item.state.getExpression();
        } else {
          stateNode = Node.AND(stateNode, item.state.getExpression());
        }
        newStateTransitions.add((ArrayList<EfsmTransition>) item.transitions);
      }

      combinedTransitions = combineLists(newStateTransitions);

      // Creating a new state.

      EfsmState newState = product.createState(stateNode);
      transitionsMapping.put(newState, combinedTransitions);

      if (newState.equals(newInitialState)) {
        product.setInitialState(newState);
      }

    }

    // Constructing new transitions.

    for (Map.Entry<EfsmState, List<List<EfsmTransition>>> entry : transitionsMapping.entrySet()) {
      for (List<EfsmTransition> item : entry.getValue()) {
        EfsmState destState = getNewDestinationState(item);
        if (transitionsMapping.keySet().contains(destState)) {
          GuardedAction ga = constructGuardedAction(item);
          if (ga != null) {
            product.createTransition(entry.getKey(), destState, ga);
          }
        }
      }
    }

    return product;
  }

  private static List<EfsmModule> getAllParentModules(EfsmModule module) {
    final List<EfsmModule> modules = new ArrayList<>();

    modules.add(module);

    final EfsmModule parentModule = module.getParent();

    if (parentModule != null) {
      final List<EfsmModule> parentModules = getAllParentModules(parentModule);
      if (!parentModules.isEmpty()) {
        modules.addAll(parentModules);
      }
    }
    return modules;
  }

  /**
   * Checks whether the guards of the specified transitions are complete.
   *
   * @param transitions the specified {@link Collection} of {@link EfsmTransition}
   * @return {@code true} if the guards are complete, {@code false} otherwise
   */
  private static boolean areComplete(Collection<EfsmTransition> transitions) {

    Node constraint = null;
    List<Node> guards = new ArrayList<>();

    for (EfsmTransition item : transitions) {
      guards.addAll(item.getGuardedAction().getGuard().getPredicates());
      if (constraint == null) {
        constraint = item.getGuardedAction().getGuard().toNode();
      } else {
        constraint = Node.AND(constraint, item.getGuardedAction().getGuard().toNode());
      }
    }

    return SolverUtils.areComplete(guards.toArray(new Node[guards.size()]));
  }

  /**
   * Performs a Cartesian product of several lists (sets) of elements.
   *
   * @param lists {@link List} containing the lists to product
   * @return the new {@link List} containing a Cartesian product of the input lists
   * @throws IllegalArgumentException when argument is {@code null}
   */
  private static <T> List<List<T>> combineLists(List<List<T>> lists) {
    InvariantChecks.checkNotNull(lists);

    int combinations = 1;
    for (List<T> item : lists) {
      combinations = combinations * item.size();
    }

    List<List<T>> product = new ArrayList<>();

    for (int i = 0; i < combinations; i++) {
      List<T> combination = new ArrayList<>();
      int index = 1;
      for (List<T> item : lists) {
        combination.add(item.get((i / index) % item.size()));
        index = index * item.size();
      }
      product.add(combination);
    }

    return product;

  }

  /**
   * Constructs a new EfsmState according to the destinations of transitions, contained in the
   * specified collection.
   *
   * @param transitions the specified {@link Collection} of {@link EfsmTransition}
   * @return a new {@link EfsmState}
   */
  private static EfsmState getNewDestinationState(Collection<EfsmTransition> transitions) {
    List<EfsmState> result = new ArrayList<>();
    Node constraint = null;

    for (EfsmTransition item : transitions) {
      result.add(item.getDestinationState());
    }

    for (EfsmState item : result) {
      if (constraint == null) {
        constraint = item.getExpression();
      } else {
        constraint = Node.AND(constraint, item.getExpression());
      }
    }

    return new EfsmState(constraint);

  }

  /**
   * Constructs a single {@link GuardedAction} from the guarded actions of the specified
   * transitions.
   *
   * @param transitions a collection of the transitions
   * @return the new guarded action
   */
  private static GuardedAction constructGuardedAction(Collection<EfsmTransition> transitions) {
    List<Node> guardPredicates = new ArrayList<>();
    List<Assignment> assignments = new ArrayList<>();
    List<Event> events = new ArrayList<>();

    for (EfsmTransition item : transitions) {
      guardPredicates.addAll(item.getGuardedAction().getGuard().getPredicates());
      events.addAll(item.getGuardedAction().getClocks());
      if (item.getGuardedAction().getAction() != null) {
        Action action = item.getGuardedAction().getAction();
        for (Assignment atomic : action.getAssignments()) {
          if (atomic.getValue() != null) {
            assignments.add(atomic);
          }
        }
      }
    }

    if (!SolverUtils.areCompatible(guardPredicates.toArray(new Node[guardPredicates.size()]))) {
      return null;
    }

    GuardedAction ga = new GuardedAction(new Guard(guardPredicates), new Action(assignments));
    ga.addClocks(events);

    return ga;
  }

  /**
   * Checks whether the specified EFSM (or it's ancestor) declares the specified variable.
   *
   * @param efsm EFSM to be checked if it (or it's ancestor) declares variable
   * @param variable variable to be checked for declaration
   * @return {@code true} if the specified EFSM (or some it's ancestor) declares the specified
   *         variable, {@code false} otherwise
   */
  public static boolean hasDeclaringAncestor(final Efsm efsm, final NodeVariable variable) {
    if (efsm.containsVariable(variable)) {
      return true;
    } else {
      EfsmModule module = efsm.getParentContainer();
      while (module != null) {
        if (module.containsVariable(variable)) {
          return true;
        }
        module = module.getParent();
      }
      return false;
    }
  }

  /**
   * Transforms the specified events container to the node.
   *
   * @param events Container of events.
   * @return Node that is a conjunction of "signal == signal_level" equations that are extracted
   *         from events.
   */
  public static Node eventsToNode(final Collection<Event> events) {

    final List<Node> nodes = new LinkedList<>();

    for (final Event event : events) {
      final RangedVariable rangedVariable = event.getRangedVariable();
      final NodeVariable variable = rangedVariable.getVariable();

      Node value = null;
      switch (event.getType()) {
        case ANY_EDGE:
          /* Do nothing. */
          break;
        case NEGATIVE_EDGE:
          value = createValue(variable.getDataType(), 0);
          break;
        case POSITIVE_EDGE:
          value = createValue(variable.getDataType(), 1);
          break;
      }

      if (value != null) {
        nodes.add(NodeUtils.getEquation(
          variable,
          rangedVariable.isRanged()
            ? RangedVariableUtils.transformNodeWithRanged(rangedVariable, value) : value));
      }
    }
    return nodes.isEmpty() ? null : ExprUtils.getConjunction(nodes.toArray(new Node[nodes.size()]));
  }

  private static Node createValue(DataType dataType, int num) {

    switch (dataType.getTypeId()) {
      case BIT_VECTOR:
        return NodeValue.newBitVector(BitVector.valueOf(num, dataType.getSize()));
      case LOGIC_BOOLEAN:
        return NodeValue.newBoolean(num != 0);
      case LOGIC_INTEGER:
        return NodeValue.newInteger(num);
      default:
        return null;
    }
  }

}
