/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;

/**
 * The EFSM walker.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class EfsmWalker {
  private final EfsmVisitor visitor;

  /**
   * Constructs an EFSM walker for the specified visitor.
   * 
   * @param visitor the visitor
   * 
   * @throws IllegalArgumentException if {@code visitor} is {@code null}
   */
  public EfsmWalker(EfsmVisitor visitor) {
    InvariantChecks.checkNotNull(visitor);

    this.visitor = visitor;
  }

  /**
   * Performs walking across the specified EFSM.
   *
   * @param efsm the EFSM to be walked across
   *
   * @throws IllegalArgumentException if {@code efsm} is {@code null}
   */
  public void start(Efsm efsm) {
    InvariantChecks.checkNotNull(efsm);

    this.visitor.onEfsmBegin(efsm);
    if (efsm.containsStates()) {
      EfsmState currentState = efsm.getInitialState();
      this.visitState(currentState);
      Set<EfsmState> visitedStates = new HashSet<>(efsm.getStates().size());
      visitedStates.add(currentState);
      LinkedList<EfsmTransition> transitionsToVisit =
          new LinkedList<>(efsm.getOutgoingTransitions(currentState));

      while (!transitionsToVisit.isEmpty()) {
        EfsmTransition currentTransition = transitionsToVisit.poll();
        this.visitTransition(currentTransition);
        currentState = currentTransition.getDestinationState();

        if (!visitedStates.contains(currentState)) {
          this.visitState(currentState);
          visitedStates.add(currentState);
          transitionsToVisit.addAll(efsm.getOutgoingTransitions(currentState));
        }
      }

      if (!visitedStates.containsAll(efsm.getStates())) {
        Set<EfsmState> unvisitedStates = new HashSet<>(efsm.getStates());
        unvisitedStates.removeAll(visitedStates);
        LinkedList<EfsmState> statesToVisit = new LinkedList<>(unvisitedStates);

        while (!statesToVisit.isEmpty()) {
          currentState = statesToVisit.poll();

          if (!visitedStates.contains(currentState)) {
            this.visitState(currentState);
            visitedStates.add(currentState);

            for (EfsmTransition transition : efsm.getOutgoingTransitions(currentState)) {
              this.visitTransition(transition);
            }
          }
        }
      }
    }

    this.visitor.onEfsmEnd(efsm);
  }

  /**
   * Visits the specified state.
   *
   * @param state the state to be visited
   *
   * @throws IllegalArgumentException if {@code state} is {@code null}
   */
  public void visitState(EfsmState state) {
    InvariantChecks.checkNotNull(state);

    this.visitor.onStateBegin(state);
    this.visitExpression(state.getExpression());
    this.visitor.onStateEnd(state);
  }

  /**
   * Visits the specified expression.
   *
   * @param expression the expression to be visited
   */
  public void visitExpression(Node expression) {
    this.visitor.onExpression(expression);
  }

  /**
   * Visits the specified transition.
   *
   * @param transition the transition to be visited.
   *
   * @throws IllegalArgumentException if {@code transition} is {@code null}.
   */
  public void visitTransition(EfsmTransition transition) {
    InvariantChecks.checkNotNull(transition);

    this.visitor.onTransitionBegin(transition);
    this.visitGuardedAction(transition.getGuardedAction());
    this.visitor.onTransitionEnd(transition);
  }

  /**
   * Visits the specified guarded action.
   *
   * @param guardedAction the guarded action to be visited
   *
   * @throws IllegalArgumentException if {@code guardedAction} is {@code null}
   */
  public void visitGuardedAction(GuardedAction guardedAction) {
    InvariantChecks.checkNotNull(guardedAction);

    this.visitor.onGuardedActionBegin(guardedAction);

    for (Event event: guardedAction.getClocks()) {
      this.visitEvent(event);
    }

    this.visitGuard(guardedAction.getGuard());
    Action action = guardedAction.getAction();
    
    if (action != null) {
      this.visitAssignmentAction(action);
    } else {
      throw new IllegalStateException(
          "Null action is found with guard: " + guardedAction.getGuard() + ".");
    }
    this.visitor.onGuardedActionEnd(guardedAction);
  }

  /**
   * Visits the specified guard.
   *
   * @param guard the guard to be visited
   *
   * @throws IllegalArgumentException if {@code guard} is {@code null}
   */
  public void visitGuard(Guard guard) {
    InvariantChecks.checkNotNull(guard);

    this.visitor.onGuardBegin(guard);

    for (Node expression: guard.getPredicates()) {
      this.visitExpression(expression);
    }

    this.visitor.onGuardEnd(guard);
  }

  /**
   * Visits the specified event.
   *
   * @param event the event to be visited
   *
   * @throws IllegalArgumentException if {@code event} is {@code null}
   */
  public void visitEvent(Event event) {
    InvariantChecks.checkNotNull(event);

    this.visitor.onEvent(event);
  }

  /**
   * Visits the specified assignment action.
   *
   * @param action the assignment action to be visited
   *
   * @throws IllegalArgumentException if {@code action} is {@code null}
   */
  public void visitAssignmentAction(Action action) {
    InvariantChecks.checkNotNull(action);

    this.visitor.onAssignmentActionBegin(action);

    for (Assignment assignment: action.getAssignments()) {
      this.visitAssignment(assignment);
    }

    this.visitor.onAssignmentActionEnd(action);
  }

  /**
   * Visits the specified assignment.
   *
   * @param assignment the assignment to be visited
   *
   * @throws IllegalArgumentException if {@code assignment} is {@code null}
   */
  public void visitAssignment(Assignment assignment) {
    InvariantChecks.checkNotNull(assignment);

    this.visitor.onAssignmentBegin(assignment);
    this.visitExpression(assignment.getValue());
    this.visitor.onAssignmentEnd(assignment);
  }
}
