/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.MetaInfo;
import ru.ispras.retrascope.model.basis.UseDef;

import java.util.Collection;
import java.util.Set;

/**
 * Clocked guarded action.
 *
 * <p>Clocked guarded action is a triple: a collection of clock
 * {@link ru.ispras.retrascope.model.basis.Event} events, a guard
 * {@link ru.ispras.retrascope.model.efsm.Guard} condition,
 * and an {@link ru.ispras.retrascope.model.efsm.Action} object.</p>
 * <br>
 * <p>When some clock event happens and guard is {@code true}
 * then the action proceeds.</p>
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 * 
 */
public class GuardedAction extends MetaInfo implements UseDef {

  /* The delimiter for object's string representation. */
  private static final String GA_DELIMITER = "->";

  private final Guard guard;
  private final Action action;
  private final EventList clocks;

  /**
   * Constructs object with the specified guard condition and action.
   *
   * @param guard guard
   * @param action action
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  public GuardedAction(final Guard guard, final Action action) {
    super();
    InvariantChecks.checkNotNull(guard);
    InvariantChecks.checkNotNull(action);

    this.guard = guard;
    this.action = action;
    this.clocks = new EventList();
  }

  /**
   * Returns action.
   *
   * @return action
   */
  public Action getAction() {
    return this.action;
  }

  /**
   * Returns guard.
   *
   * @return guard
   */
  public Guard getGuard() {
    return this.guard;
  }

  /**
   * Returns list of object's clock events.
   *
   * @return clocks list
   */
  public EventList getEventList() {
    return this.clocks;
  }

  /**
   * Returns clock events.
   *
   * @return clock events collection
   */
  public Collection<Event> getClocks() {
    return this.clocks.getEvents();
  }

  /**
   * Adds the specified clock event to this object's collection.
   *
   * @param clock clock event
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addClock(final Event clock) {
    InvariantChecks.checkNotNull(clock);

    this.clocks.add(clock);
  }

  /**
   * Adds the specified clock events collection to the object's collection.
   *
   * @param clocks clocks collection
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addClocks(final Collection<Event> clocks) {
    InvariantChecks.checkNotNull(clocks);

    this.clocks.addEvents(clocks);
  }

  /**
   * Returns variables that are used in {@link Guard} condition.
   *
   * @return sequence of variables
   */
  public Iterable<NodeVariable> getUsesByGuard() {
    return this.getGuard().getUses();
  }

  /**
   * Returns variables that are used in {@link Action} action.
   *
   * @return sequence of variables
   */
  public Iterable<NodeVariable> getUsesByAction() {
    return this.getAction().getUses();
  }

  /**
   * Checks that the specified event is in the object's clock collection.
   *
   * @param event event
   *
   * @return {@code true} if the specified event is in the object's clock collection
   *         or the list is empty, {@code false} otherwise.
   */
  public boolean isSensible(final Event event) {
    return this.clocks.isTriggered(event);
  }

  /**
   * Checks if this guarded action is sensible to the specified events.
   *
   * @param eventList an event list to be tested
   *
   * @return {@code true} if at least one of the specified events
   *         is in the object's clock collection or it is empty, {@code false} otherwise.
   */
  public boolean isSensible(final EventList eventList) {
    for (final Event event: eventList.getEvents()) {
      if (this.isSensible(event)) {
        return true;
      }
    }

    return false;
  }

  @Override
  public Set<NodeVariable> getUses() {
    return CollectionUtils.uniteSets(this.getGuard().getUses(), this.getAction().getUses());
  }

  @Override
  public Set<NodeVariable> getDefines() {
    return this.getAction().getDefines();
  }

  @Override
  public String toString() {
    final StringBuilder builder = new StringBuilder();
    builder.append(String.format("{%s: %s%s{", this.clocks.toString(), getGuard(), GA_DELIMITER));

    if (this.getAction() != null) {
      builder.append(this.getAction());
    }
    builder.append("}}");
    return builder.toString();
  }

  @Override
  public boolean equals(final Object obj) {
    return this == obj;
  }

  @Override
  public int hashCode() {
    return System.identityHashCode(this);
  }
}
