/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.hldd;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.hldd.HlddNode.NodeType;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;

/**
 * This class represents a High-Level Decision Diagram (HLDD) of a variable assignment function.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class Hldd {

  // Diagram nodes.
  private Collection<HlddNode> nodes;
  // Terminal nodes.
  private Collection<HlddNode> terminalNodes;
  // Assigned variable.
  private NodeVariable variable;
  // Root node.
  private HlddNode rootNode;
  // Immediate assignment flag.
  private boolean isImmAssignment;

  /**
   * Constructs a new HLDD using the specified variable.
   * 
   * @param variable - the specified variable.
   */
  public Hldd(final NodeVariable variable) {
    setVariable(variable);
    nodes = new LinkedList<HlddNode>();
    terminalNodes = new LinkedList<HlddNode>();
  }

  /**
   * Constructs a new HLDD using the specified values.
   * 
   * @param variable - the specified variable.
   * @param nodes - the specified collection of nodes.
   */
  public Hldd(final NodeVariable variable, final Collection<HlddNode> nodes) {
    setVariable(variable);
    this.nodes = new LinkedList<HlddNode>();
    terminalNodes = new LinkedList<HlddNode>();
    addNodes(nodes);
  }

  /**
   * Constructs a new HLDD.
   */
  public Hldd() {
    nodes = new LinkedList<HlddNode>();
    terminalNodes = new LinkedList<HlddNode>();
  }

  /**
   * Sets the root node.
   * 
   * @param root - the specified node.
   */
  public void setRootNode(final HlddNode root) {
    InvariantChecks.checkNotNull(root);
    root.setRoot(true);
    rootNode = root;
  }

  /**
   * Returns the root node.
   * 
   * @return The root node.
   */
  public HlddNode getRootNode() {
    return rootNode;
  }

  /**
   * Adds the specified node to the diagram.
   * 
   * @param node - the specified node.
   */
  public void addNode(final HlddNode node) {
    InvariantChecks.checkNotNull(node);

    nodes.add(node);
    if (node.getType() == NodeType.ASSIGNMENT) {
      terminalNodes.add(node);
    }
  }

  /**
   * Adds the collection of nodes to the diagram.
   * 
   * @param nodes - the specified collection of nodes.
   */
  public void addNodes(final Collection<HlddNode> nodes) {
    for (HlddNode node : nodes) {
      addNode(node);
    }
  }

  /**
   * Returns the diagram's nodes.
   * 
   * @return The collection of nodes.
   */
  public Collection<HlddNode> getNodes() {
    return nodes;
  }

  /**
   * Returns the terminal nodes.
   * 
   * @return The collection of terminal nodes.
   */
  public Collection<HlddNode> getTerminalNodes() {
    return terminalNodes;
  }

  /**
   * Returns a deep copy of the diagram using the specified associated variable.
   * 
   * @return A deep copy.
   */
  public Hldd getCopy() {
    final Hldd result = new Hldd();
    final Map<HlddNode, HlddNode> nodeCopies = new HashMap<HlddNode, HlddNode>();

    for (HlddNode node : nodes) {
      nodeCopies.put(node, new HlddNode(node.getType()));
    }

    for (Entry<HlddNode, HlddNode> entry : nodeCopies.entrySet()) {
      final HlddNode node = entry.getKey();
      final HlddNode nodeCopy = entry.getValue();
      final HlddNode parent = node.getParent();
      final Collection<HlddNode> children = node.getChildren();

      nodeCopy.addExpression(node.getExpressions());
      nodeCopy.setLevel(node.getLevel());

      final HlddNode parentCopy = nodeCopies.get(parent);
      nodeCopy.setParent(parentCopy);

      for (HlddNode child : children) {
        final HlddNode childCopy = nodeCopies.get(child);
        nodeCopy.addChild(childCopy);
      }
      
      if (node.isDefault()) {
        nodeCopy.setDefault();
      }

      if (node.isRoot()) {
        result.setRootNode(nodeCopy);
      }
      
    }

    result.addNodes(nodeCopies.values());
    result.setVariable(variable);
    return result;
  }

  /**
   * Returns the associated variable.
   * 
   * @return The associated variable.
   */
  public NodeVariable getVariable() {
    return variable;
  }

  /**
   * Sets the associated variable.
   * 
   * @param variable - the specified variable.
   */
  public void setVariable(final NodeVariable variable) {
    this.variable = variable;
  }

  /**
   * Shows whether the HLDD implements an immediate assignment. 
   * 
   * @return {@code true} if the HLDD implements condition, {@code false} otherwise.
   */
  public boolean isImmediateAssignment() {
    return isImmAssignment;
  }

  /**
   * Sets an immediate assignment property.
   * 
   * @param isGuard - the specified value. 
   */
  public void setImmediateAssignment(final boolean isGuard) {
    this.isImmAssignment = isGuard;
  }

}
