/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.hldd;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.cfg.CfgModule;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * This class represents a module of a HLDD model. Contains a collection of HlddProcesses and a
 * corresponding CFG {@link CfgModule}.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class HlddModule {

  // Corresponding CFG module.
  private CfgModule module;
  // Nested modules.
  private List<HlddModule> nestedModules;
  // Nested processes.
  private List<HlddProcess> processes;

  /**
   * Constructs a new module.
   */
  public HlddModule() {
    nestedModules = new ArrayList<>();
    processes = new ArrayList<>();
  }

  /**
   * Constructs a new module and sets the specified data.
   * 
   * @param module - the specified CFG {@link CfgModule}.
   * @param processes - the specified collection of processes.
   */
  public HlddModule(final CfgModule module, final Collection<HlddProcess> processes) {
    setModule(module);
    nestedModules = new ArrayList<>();
    this.processes = new ArrayList<>();
    add(processes);
  }

  /**
   * Sets the corresponding CFG {@link CfgModule}.
   * 
   * @param module - the specified module.
   */
  public void setModule(final CfgModule module) {
    InvariantChecks.checkNotNull(module);
    this.module = module;
  }

  /**
   * Returns the corresponding CFG {@link CfgModule}.
   * 
   * @return The corresponding module.
   */
  public CfgModule getModule() {
    return module;
  }

  /**
   * Adds a process to this module.
   * 
   * @param process - the specified process.
   */
  public void add(final HlddProcess process) {
    InvariantChecks.checkNotNull(process);
    processes.add(process);
  }

  /**
   * Adds a collection of processes to this module.
   * 
   * @param processes - the specified collection.
   */
  public void add(final Collection<HlddProcess> processes) {
    for (HlddProcess object : processes) {
      add(object);
    }
  }

  /**
   * Returns the processes of this module.
   * 
   * @return The collection of processes.
   */
  public List<HlddProcess> getProcesses() {
    return new ArrayList<>(processes);
  }

  /**
   * Returns the nested modules of this module.
   * 
   * @return The collection of modules.
   */
  public List<HlddModule> getNestedModules() {
    return new ArrayList<>(nestedModules);
  }

  /**
   * Adds a collection of nested modules to this module.
   * 
   * @param nestedModules - the specified collection of modules.
   */
  public void addNestedModules(final Collection<HlddModule> nestedModules) {
    this.nestedModules.addAll(nestedModules);
  }

  /**
   * Adds a nested module to this module.
   * 
   * @param nestedModule - the specified module.
   */
  public void addNestedModule(final HlddModule nestedModule) {
    nestedModules.add(nestedModule);
  }

  @Override
  public String toString() {
    return "HlddModule [module=" + module.getName() + ", nestedModules=" + nestedModules
        + ", processes=" + processes + "]";
  }

}
