/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.hldd;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

/**
 * This class represents a node of a high-level decision diagram (HLDD).
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class HlddNode {

  // Parent nodes.
  // private List<HlddNode> parents;
  private HlddNode parent;
  // Type of this node.
  private NodeType type;
  // Children nodes.
  private List<HlddNode> children;
  // Expressions which are stored in this node.
  private List<Node> expression;
  // Indicates that this node is a root node of HLDD.
  private boolean root;
  // Indicates that this node represents a default value.
  private boolean isDefault;
  // Node level in the HLDD.
  private int level;

  /**
   * Type of a node.
   */
  public enum NodeType {
    CONDITION, VALUE, ASSIGNMENT, OTHER
  }

  /**
   * Constructs a new node.
   * 
   * @param parent - the specified parent of a node;
   * @param type - the specified type of a node;
   * @param expression - the specified expression.
   */
  public HlddNode(final HlddNode parent, final NodeType type, final Node expression) {
    children = new ArrayList<>();
    this.expression = new LinkedList<>();
    root = false;
    isDefault = false;
    setParent(parent);
    parent.addChild(this);
    setType(type);
    addExpression(expression);

  }

  /**
   * Constructs a new node.
   * 
   * @param parent - the specified parent of a node;
   * @param type - the specified type of a node;
   * @param expressions - the specified expressions.
   */
  public HlddNode(final HlddNode parent, final NodeType type,
      final Collection<? extends Node> expressions) {
    children = new ArrayList<>();
    expression = new LinkedList<>(expressions);
    root = false;
    isDefault = false;
    setParent(parent);
    parent.addChild(this);
    setType(type);
  }

  /**
   * Constructs a new node.
   * 
   * @param parent - the specified parent of a node;
   * @param type - the specified type of a node.
   */
  public HlddNode(final HlddNode parent, final NodeType type) {
    children = new ArrayList<>();
    expression = new LinkedList<Node>();
    root = false;
    isDefault = false;
    setParent(parent);
    parent.addChild(this);
    setType(type);
  }

  /**
   * Constructs a new node.
   * 
   * @param type - the specified type of a node.
   */
  public HlddNode(final NodeType type) {
    children = new ArrayList<>();
    expression = new LinkedList<Node>();
    root = false;
    isDefault = false;
    setType(type);
  }

  /**
   * Returns the level of this node.
   * 
   * @return The node level.
   */
  public int getLevel() {
    return level;
  }

  /**
   * Sets the level of this node.
   * 
   * @param level - the specified level in the diagram.
   */
  public void setLevel(final int level) {
    this.level = level;
  }

  /**
   * Shows whether the node is a root node.
   * 
   * @return <code>true</code> if the node is a root node, <code>false</code> otherwise.
   */
  public boolean isRoot() {
    return root;
  }

  /**
   * Sets the value of a root flag.
   * 
   * @param root - the specified flag value.
   */
  public void setRoot(final boolean root) {
    this.root = root;
  }

  /**
   * Sets the node as a default node.
   */
  public void setDefault() {
    if (type == NodeType.VALUE) {
      isDefault = true;
    }
  }

  /**
   * Shows whether the node is a default node.
   * 
   * @return <code>true</code> if the node is a default node, <code>false</code> otherwise.
   */
  public boolean isDefault() {
    return isDefault;
  }

  /**
   * Returns the type of the node.
   * 
   * @return {@link NodeType}.
   */
  public NodeType getType() {
    return type;
  }

  /**
   * Sets the type of the node.
   * 
   * @param type - the specified {@link NodeType}.
   */
  public void setType(final NodeType type) {
    this.type = type;
  }

  /**
   * Returns the only expression stored in the node.
   * 
   * @return The only expression if the node contains only one expression, the first expression if
   *         the node contains multiple expressions, <code>null</code> if no expressions are stored
   *         in the node.
   */
  public Node getOnlyExpression() {
    if (!expression.isEmpty()) {
      return expression.iterator().next();
    } else {
      return null;
    }
  }

  /**
   * Returns all expressions stored in the node.
   * 
   * @return {@link List} containing all expressions.
   */
  public List<Node> getExpressions() {
    return expression;
  }

  /**
   * Adds the specified {@link Node} to the stored expressions.
   * 
   * @param expression - the specified {@link Node}.
   */
  public void addExpression(final Node expression) {
    InvariantChecks.checkNotNull(expression);
    this.expression.add(expression);
  }

  /**
   * Adds the specified collection of {@link Node} to the stored expressions.
   * 
   * @param expressions - the specified collection of {@link Node}.
   */
  public void addExpression(final Collection<? extends Node> expressions) {
    InvariantChecks.checkNotNull(expressions);
    expression.addAll(expressions);
  }

  /**
   * Returns the parent of the node.
   * 
   * @return The parent node.
   */
  public HlddNode getParent() {
    return parent;
  }

  /**
   * Sets the specified {@link HlddNode} as the parent node.
   * 
   * @param parent - the specified parent node.
   */
  public void setParent(final HlddNode parent) {
    this.parent = parent;
  }

  /**
   * Returns the children of the node.
   * 
   * @return Collection of children.
   */
  public List<HlddNode> getChildren() {
    return new ArrayList<HlddNode>(children);
  }

  /**
   * Adds the specified {@link HlddNode} to the children of the node.
   * 
   * @param child - the specified child node.
   */
  public void addChild(final HlddNode child) {
    InvariantChecks.checkNotNull(child);
    children.add(child);
  }

  /**
   * Adds the specified collection of {@link HlddNode} to the children of the node.
   * 
   * @param children - the specified collection of child nodes.
   */
  public void addChildren(final Collection<HlddNode> children) {
    for (HlddNode child : children) {
      this.children.add(child);
    }
  }

  /**
   * Removes the specified parent node from children.
   * 
   * @param node - the specified child node.
   */
  public void removeFromChildren(final HlddNode node) {
    InvariantChecks.checkNotNull(node);
    children.remove(node);
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
//    result = prime * result + ((children == null) ? 0 : children.hashCode());
    result = prime * result + ((expression == null) ? 0 : expression.hashCode());
    result = prime * result + (isDefault ? 1231 : 1237);
    result = prime * result + ((parent == null) ? 0 : parent.hashCode());
    result = prime * result + (root ? 1231 : 1237);
    result = prime * result + ((type == null) ? 0 : type.hashCode());
    result = prime * result + level;
    return result;
  }


  @Override
  public boolean equals(final Object object) {
    if (this == object) {
      return true;
    }
    if (object == null) {
      return false;
    }
    if (object.getClass() != this.getClass()) {
      return false;
    } else {
      final HlddNode node = (HlddNode) object;
      if (parent != null && node.getParent() != null) {
        if (!node.getParent().equals(parent)) {
          return false;
        }
      } else {
        if (parent != null || node.getParent() != null) {
          return false;
        }
      }
      if (node.getType() != type) {
        return false;
      }
//       if (!node.getChildren().equals(children)) {
//       return false;
//       }
      if (!node.getExpressions().equals(expression)) {
        return false;
      }
      if (node.isRoot() != root) {
        return false;
      }
      if (node.isDefault() != isDefault) {
        return false;
      }
      if (node.getLevel() != level) {
        return false;
      }
    }
    return true;
  }

  @Override
  public String toString() {
    return "HlddNode [type=" + type + ", expression=" + expression.toString() + ", level=" + level
        + "]";
  }


}
