/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.hldd;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.cfg.CfgProcess;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * This class represents a process of a HLDD model. Contains a collection of HLDDs and the
 * corresponding variables.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class HlddProcess {

  // Corresponding variables.
  private Collection<NodeVariable> processVariables;
  // Process id string.
  private String processId;
  // HLDDs of the process.
  private List<Hldd> hldds;
  // Corresponding CFG process.
  private CfgProcess cfgProcess;

  /**
   * Constructs a new process.
   */
  public HlddProcess() {
    hldds = new ArrayList<>();
    processVariables = new ArrayList<>();
  }

  /**
   * Adds a HLDD to this process.
   * 
   * @param hldd - the specified HLDD.
   */
  public void add(final Hldd hldd) {
    InvariantChecks.checkNotNull(hldd);
    hldds.add(hldd);
    if (hldd.getVariable() != null) {
      processVariables.add(hldd.getVariable());
    }
  }
  
  /**
   * Adds a collection of HLDDs to this process.
   * 
   * @param hldds - the specified collection.
   */
  public void add(Collection<Hldd> hldds) {
    for (Hldd hldd : hldds) {
      add(hldd);
    }
  }

  /**
   * Returns the corresponding variables.
   * 
   * @return The collection of {@link NodeVariable}.
   */
  public Collection<NodeVariable> getVariables() {
    return processVariables;
  }

  /**
   * Returns the process id.
   * 
   * @return The process id string.
   */
  public String getProcessId() {
    return processId;
  }

  /**
   * Sets the process id.
   * 
   * @param processId - the specified string.
   */
  public void setProcessId(String processId) {
    this.processId = processId;
  }
  
  /**
   * Returns the contents of the container.
   * 
   * @return The collection of objects.
   */
  public Collection<Hldd> getHldds() {
    return new ArrayList<>(hldds);
  }

  /**
   * Returns the corresponding CFG process.
   * 
   * @return The {@link CfgProcess}.
   */
  public CfgProcess getCfgProcess() {
    return cfgProcess;
  }

  /**
   * Sets the corresponding CFG process.
   * 
   * @param process - the specified process.
   */
  public void setCfgProcess(CfgProcess process) {
    this.cfgProcess = process;
  }
}
