/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.hldd;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.Iterator;

/**
 * Hldd model walker.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class HlddWalker {

  // Model visitor.
  private HlddVisitor visitor;

  /**
   * Constructs a new walker object.
   * 
   * @param visitor - the specified {@link HlddVisitor}.
   */
  public HlddWalker(final HlddVisitor visitor) {
    InvariantChecks.checkNotNull(visitor);
    this.visitor = visitor;
  }

  /**
   * Returns the current visitor.
   * 
   * @return Current {@link HlddVisitor}.
   */
  public HlddVisitor getVisitor() {
    return visitor;
  }

  /**
   * Sets the current visitor.
   * 
   * @param visitor - the specified {@link HlddVisitor}.
   */
  public void setVisitor(final HlddVisitor visitor) {
    this.visitor = visitor;
  }

  /**
   * Starts walking a model.
   * 
   * @param hlddModel - the specified {@link HlddModel}.
   */
  public void start(final HlddModel hlddModel) {
    InvariantChecks.checkNotNull(hlddModel);
    visitor.onModelBegin(hlddModel);
    for (HlddModule module : hlddModel.getModules()) {
      visitModule(module);
    }
    visitor.onModelEnd(hlddModel);
  }

  /*
   * Visits a module.
   * 
   * @param module - the specified {@link HlddModule}.
   */
  private void visitModule(final HlddModule module) {
    visitor.onModuleBegin(module);
    for (HlddProcess process : module.getProcesses()) {
      visitProcess(process);
    }
    for (HlddModule nested : module.getNestedModules()) {
      visitModule(nested);
    }
    visitor.onModuleEnd(module);
  }

  /*
   * Visits a process.
   * 
   * @param process - the specified {@link HlddProcess}.
   */
  private void visitProcess(final HlddProcess process) {
    visitor.onProcessBegin(process);
    for (Hldd hldd : process.getHldds()) {
      visitHldd(hldd);
    }
    visitor.onProcessEnd(process);
  }

  /*
   * Visits a high-level decision diagram.
   * 
   * @param hldd - the specified {@link Hldd}.
   */
  private void visitHldd(final Hldd hldd) {
    InvariantChecks.checkNotNull(hldd);
    visitor.onHlddBegin(hldd);
    visitRoot(hldd.getRootNode());
    visitor.onHlddEnd(hldd);
  }

  /*
   * Visits a root node.
   * 
   * @param node - the specified root node.
   */
  private void visitRoot(final HlddNode node) {
    InvariantChecks.checkNotNull(node);
    visitor.onRootBegin(node);
    visitNode(node);
    visitor.onRootEnd(node);
  }

  /*
   * Visits a collection of nodes.
   * 
   * @param nodes - the specified collection of nodes.
   */
  private void visit(final Collection<HlddNode> nodes) {
    InvariantChecks.checkNotNull(nodes);
    
    Iterator<HlddNode> iterator = nodes.iterator();
    while (iterator.hasNext()) {
      visitNode(iterator.next());
    }
  }

  /*
   * Visits a node.
   * 
   * @param node - the specified {@link HlddNode}.
   */
  private void visitNode(final HlddNode node) {
    InvariantChecks.checkNotNull(node);

    switch (node.getType()) {

      case CONDITION:
        visitor.onConditionBegin(node);
        visit(node.getChildren());
        visitor.onConditionEnd(node);
        break;

      case VALUE:
        visitor.onValueBegin(node);
        visit(node.getChildren());
        visitor.onValueEnd(node);
        break;

      case ASSIGNMENT:
        visitor.onAssignmentBegin(node);
        visit(node.getChildren());
        visitor.onAssignmentEnd(node);
        break;

      default:
        visitor.onNodeBegin(node);
        visit(node.getChildren());
        visitor.onNodeEnd(node);
        break;

    }

  }
}
