/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgStatementType;
import ru.ispras.retrascope.model.cfg.CfgUtils;

/**
 * The control flow graph model visitor that transforms sequences of basic block nodes.
 *
 * <p>The component transforms sequences of {@link CfgBlockStatement}
 * basic block nodes of the same kind of concurrency
 *  (all nodes of the sequence should be either concurrent
 *  or non-concurrent).</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgBasicBlockSequenceBackend extends CfgDefaultVisitor {

  /**
   * Default constructor.
   */
  public CfgBasicBlockSequenceBackend() {
    super();
  }

  @Override
  public void onBasicBlockBegin(final CfgBlockStatement block) {
    while (ofSameConcurrentKind(block)) {
      mergeNext(block);
    }
  }

  private void mergeNext(final CfgBlockStatement block) {

    final CfgStatement parent = block.getParent();
    final CfgBlockStatement next =
        (CfgBlockStatement) CfgUtils.getSuccessors(block).iterator().next();

    block.addAssignments(next.getAssignments());
    parent.remove(next);
  }

  private boolean ofSameConcurrentKind(final CfgBlockStatement block) {

    final CfgStatement parent = block.getParent();
    final int blockIndex = parent.indexOf(block);

    if (blockIndex < parent.getStatementNum() - 1) {

      final CfgStatement next = parent.getStatement(blockIndex + 1);

      return next.isType(CfgStatementType.BASIC_BLOCK)
        && (((CfgBlockStatement)next).isConcurrent() == block.isConcurrent());
    }
    return false;
  }
}
