/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgLoopStatement;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * The control flow graph model visitor that detects loops or cycles and transforms them.
 *
 * <p>Finite loops are transformed into sequences of statements (i.e. loop unrolling),
 * infinite loops are transformed into separate processes.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgLoopBackend extends CfgDefaultVisitor {

  /* The current module. */
  private CfgModule module;

  /* Loop parents & loops that are already processed. */
  private Map<CfgLoopStatement, CfgStatement> processed = new LinkedHashMap<>();

  /**
   * Default constructor.
   */
  public CfgLoopBackend() {
    super();
  }

  private CfgModule getModule() {
    return this.module;
  }

  @Override
  public void onModuleBegin(final CfgModule module) {
    this.module = module;
  }

  @Override
  public void onModuleEnd(final CfgModule module) {
    this.module = null;

    /* Remove loop statement that are already processed. */
    for (final Map.Entry<CfgLoopStatement, CfgStatement> entry : this.processed.entrySet()) {

      final CfgStatement parent = entry.getValue();
      if (parent.contains(entry.getKey())) {
        parent.remove(entry.getKey());
      }
    }
  }

  @Override
  public void onLoopEnd(final CfgLoopStatement loop) {

    if (!this.processed.containsKey(loop)) {
      switch (loop.getLoopType()) {
        case FOR:
        case REPEAT:
          processFiniteLoop(loop);
          break;
        case FOREVER:
          processInfiniteLoop(loop);
          break;
        case WHILE:
          processWhileLoop(loop);
          break;
        default:
          throw new RetrascopeRuntimeException(
              LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, loop.getLoopType()));
      }
      this.processed.put(loop, loop.getParent());
    }
  }

  private void processWhileLoop(CfgLoopStatement loop) {
    Log.printWarn(LogMessage.CANT_PARSE + loop.getDescription());
  }

  private void processInfiniteLoop(final CfgLoopStatement loop) {

    final Node condition = loop.getCondition();

    if (condition == null || SolverUtils.isSat(condition)) {

      final CfgProcess newProcess = new CfgProcess();

      final NodeVariable loopVariable = getModule().declareNewVariable(DataType.BOOLEAN);

      getModule().declareVariable(loopVariable);
      getModule().add(newProcess);

      newProcess.addEvent(new Event(new RangedVariable(loopVariable)));

      /* Init loop variable in the block node & add it to the parent process. */
      final CfgBlockStatement loopVariableBlock =
          new CfgBlockStatement(loopVariable, SolverUtils.TRUE);

      /* Put trigger block after the loop place. */
      final CfgStatement loopParent = loop.getParent();
      loopParent.add(loopVariableBlock, loopParent.indexOf(loop) + 1);

      /* Construct the sub-process control flow graph. */
      final CfgSwitchStatement switchStatement =
          new CfgSwitchStatement(NodeUtils.getEquation(loopVariable, SolverUtils.TRUE));
      final CfgCaseStatement trueCase = new CfgCaseStatement(SolverUtils.TRUE);
      newProcess.add(switchStatement);
      switchStatement.add(trueCase);

      for (int i = 0; i < loop.getStatementNum(); i++) {
        final CfgStatement loopStatement = loop.getStatement(i);
        loopStatement.removeParent(loop);
        trueCase.add(loopStatement);
      }
    } else {
      /* Dead code has been found. */
      Log.printWarn("Dead code: " + loop.getId());
    }
  }

  private void processFiniteLoop(final CfgLoopStatement loop) {

    final Node loopCondition = loop.getCondition();

    final Assignment init = loop.getInit();
    final String counterName = init.getVariable().getName();
    final Node counterInitValue = init.getValue();

    final Assignment incrementAssignment = loop.getIncrement();
    final Node increment = incrementAssignment.getValue();

    Node whileExpr = Transformer.substitute(loopCondition, counterName, counterInitValue);

    Node currentCounterValue = counterInitValue;

    final List<CfgStatement> processed = new LinkedList<>();

    while (SolverUtils.isSat(whileExpr)) {

      final List<CfgStatement> copies = new LinkedList<>();
      for (int i = 0; i < loop.getStatementNum(); i++) {

        copies.add(loop.getStatement(i).deepCopy());

      }
      final List<CfgStatement> substituted =
          CfgUtils.substitute(copies, counterName, currentCounterValue);
      processed.addAll(substituted);

      currentCounterValue =
          Transformer.reduce(Transformer.substitute(increment, counterName, currentCounterValue));

      whileExpr = Transformer.substitute(loopCondition, counterName, currentCounterValue);
    }

    /* Insert processing result instead of loop statement. */
    final CfgStatement parent = loop.getParent();
    final int loopIndex = parent.indexOf(loop);

    for (int i = loopIndex; i < loopIndex + processed.size(); i++) {
      parent.add(processed.get(i - loopIndex), i + 1);
    }
  }
}
