/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatementType;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Control flow graph model visitor that detects one-basic-block processes and transforms them.
 *
 * <p>This backend merges {@link CfgProcess} processes,
 * which contain only one {@link CfgBlockStatement} basic block node
 * of concurrent type, with only one {@link ru.ispras.retrascope.model.basis.Assignment} assignment,
 * and equal {@link ru.ispras.retrascope.model.basis.EventList} event lists.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgOneAssignmentProcessBackend extends CfgDefaultVisitor {

  /**
   * Default constructor.
   */
  public CfgOneAssignmentProcessBackend() {
    super();
  }

  @Override
  public void onModuleBegin(final CfgModule module) {

    final Set<CfgProcess> merged = new LinkedHashSet<>();
    final Collection<CfgProcess> processes = module.getProcesses();

    if (!processes.isEmpty()) {
      for (final CfgProcess process : processes) {
        if (isOneBlockCfg(process)) {

          if (merged.isEmpty()) {
            merged.add(process);

          } else {

            boolean isMerged = false;
            for (final CfgProcess mergedProcess : merged) {

              /* If the target process can be merged into "merged" process. */
              if (isOneBlockCfg(mergedProcess)
                  && mergedProcess.getSensitivityList().equals(process.getSensitivityList())) {
                mergeProcesses(mergedProcess, process);
                isMerged = true;
                break;
              }
            }
            if (!isMerged) {
              merged.add(process);
            }
          }
        } else {
          merged.add(process);
        }
      }

      /* Remove "old" processes. */
      module.removeProcesses();

      /* Add new "merged" processes. */
      for (final CfgProcess process : merged) {
        module.add(process);
      }
    }
  }

  private void mergeProcesses(final CfgProcess result, final CfgProcess target) {

    /* Merge sensitivity lists. */
    final EventList rList = result.getSensitivityList();
    final Collection<Event> tList = target.getSensitivityList().getEvents();

    for (final Event event : tList) {
      rList.add(event);
    }

    /* Merge control flow graphs. */

    /* Get basic bock node at the result process. */
    final CfgBlockStatement resultBlock = (CfgBlockStatement) result.getStatement(0);

    /* Get basic block at the target process. */
    final CfgBlockStatement targetBlock = (CfgBlockStatement) target.getStatement(0);

    resultBlock.addAssignments(targetBlock.getAssignments());
  }

  private boolean isOneBlockCfg(final CfgProcess process) {
    return process.getStatementNum() == 1
        && isConcurrentAssignmentBlockCfg(process);
  }

  private boolean isConcurrentAssignmentBlockCfg(final CfgProcess node) {
    return node.getStatement(0).isType(CfgStatementType.BASIC_BLOCK)
        && ((CfgBlockStatement)node.getStatement(0)).isConcurrent();
  }
}
