/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgStatementType;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Stack;

/**
 * Control flow graph model visitor that detects and transforms sequences of switch nodes.
 *
 * <p>The HDL parser backend that optimises sequences
 * of control flow graph {@link ru.ispras.retrascope.model.cfg.CfgModel} model
 * {@link CfgSwitchStatement} switch nodes. A sequence can be treated as nested switch statements
 * and as sequential switch statements. Switch nodes must contain conditions of the following form:
 * {@code x == const_1 || ... || x == const_n}, and only the {@code x} must be used
 * in every node of the sequences to be transformed.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgSwitchSequenceBackend extends CfgDefaultVisitor {

  private Stack<String> hammockSequentialSwitchIds;

  /**
   * Default constructor.
   */
  public CfgSwitchSequenceBackend() {
    super();
  }

  @Override
  public void onProcessBegin(final CfgProcess process) {
    this.hammockSequentialSwitchIds = new Stack<>();
  }

  @Override
  public void onProcessEnd(final CfgProcess process) {
    while (!this.hammockSequentialSwitchIds.isEmpty()) {
      final String switchId = this.hammockSequentialSwitchIds.pop();
      final CfgSwitchStatement switchStatement =
          (CfgSwitchStatement) process.getStatement(switchId);
      processNextSwitches(switchStatement);
    }
  }

  @Override
  public void onSwitchBegin(final CfgSwitchStatement switchNode) {

    processNestedSwitches(switchNode);

    if (canBeCombinedWithNext(switchNode)) {
      this.hammockSequentialSwitchIds.push(switchNode.getId());
    }
  }

  private void processNextSwitches(final CfgSwitchStatement switchNode) {

    while (canBeCombinedWithNext(switchNode)) {
      combineWithNext(switchNode);
    }
  }

  private void combineWithNext(final CfgSwitchStatement switchStatement) {

    final CfgStatement parent = switchStatement.getParent();
    final CfgSwitchStatement next =
        (CfgSwitchStatement) parent.getStatement(parent.indexOf(switchStatement) + 1);
    parent.remove(next);

    updateConditionAndValues(next);

    final Node switchCondition = switchStatement.getCondition();
    if (ExprUtils.isKind(Node.Kind.OPERATION, switchCondition)
        && ((NodeOperation)switchCondition).getOperationId().equals(StandardOperation.EQ)) {
      updateConditionAndValues(switchStatement);
    }

    for (int i = 0; i < next.getStatementNum(); i++) {
      final CfgCaseStatement nextCase = (CfgCaseStatement) next.getStatement(i);
      if (nextCase.containStatements()) {
        switchStatement.add(nextCase.deepCopy());
      }
    }
  }

  private void updateConditionAndValues(final CfgSwitchStatement switchStatement) {

    final NodeOperation switchCondition = (NodeOperation) switchStatement.getCondition();
    final Node commonVariableExpression = getExpression(switchCondition);

    assert commonVariableExpression != null
        : "Can't extract variable expression from: " + switchCondition;

    if (!commonVariableExpression.equals(switchCondition)) {

      final NodeValue caseValue = getAssignedValue(switchCondition);
      switchStatement.setCondition(commonVariableExpression);

      for (int i = 0; i < switchStatement.getStatementNum(); i++) {
        final CfgCaseStatement caseStatement = (CfgCaseStatement) switchStatement.getStatement(i);
        caseStatement.removeValues();
        if (caseStatement.containStatements()) {
          caseStatement.add(caseValue);
        } else {
          caseStatement.setDefault(true);
        }
      }
    }
  }

  private Node getExpression(final Node condition) {

    if (ExprUtils.isKind(Node.Kind.OPERATION, condition)) {

      if (((NodeOperation)condition).getOperationId().equals(StandardOperation.EQ)) {

        for (int i = 0; i < ((NodeOperation) condition).getOperandCount(); i++) {
          final Node operand = ((NodeOperation) condition).getOperand(i);

          if (!ExprUtils.isKind(Node.Kind.VALUE, operand)) {
            return operand;
          }
        }
      } else {
        return condition;
      }
    }
    return null;
  }

  private boolean canBeCombinedWithNext(final CfgSwitchStatement switchStatement) {

    final CfgStatement parent = switchStatement.getParent();

    final int switchIndex = parent.indexOf(switchStatement);

    if (switchIndex == parent.getStatementNum() - 1) {

      /* Switch statement is the last. */
      return false;
    }

    final CfgStatement next = parent.getStatement(switchIndex + 1);

    if (!next.isType(CfgStatementType.SWITCH)) {

      /* Next statement is not a switch. */
      return false;
    }

    final Node switchCond = switchStatement.getCondition();
    final Node nextCond = ((CfgSwitchStatement)next).getCondition();

    return canCombineConditions(switchCond, nextCond)
        && canCombineSwitches(switchStatement, (CfgSwitchStatement) next);
  }

  private boolean canCombineConditions(
      final Node one,
      final Node another) {

    return isOneVarEquation(one)
        && isOneVarEquation(another)
        && (getCommonExpression(one, another) != null);
  }

  private Node getCommonExpression(
      final Node one,
      final Node another) {

    final Node oneExpression = getExpression(one);
    final Node anotherExpression = getExpression(another);

    if (oneExpression != null
        && anotherExpression != null
        && oneExpression.equals(anotherExpression)) {
      return oneExpression;
    }
    return null;
  }

  private boolean canCombineSwitches(
      final CfgSwitchStatement one,
      final CfgSwitchStatement another) {

    for (int i = 0; i < one.getStatementNum(); i++) {

      final CfgCaseStatement oneCase = (CfgCaseStatement) one.getStatement(i);
      final Node oneCaseCondition = CfgUtils.getExpression(oneCase, one);

      for (int j = 0; j < another.getStatementNum(); j++) {
        final CfgCaseStatement anotherCase = (CfgCaseStatement) another.getStatement(j);
        final Node anotherCaseCondition = CfgUtils.getExpression(anotherCase, another);

        final Node compatibleBranches =
            ExprUtils.getConjunction(oneCaseCondition, anotherCaseCondition);

        if (SolverUtils.isSat(compatibleBranches)) {
          if (oneCase.containStatements() && anotherCase.containStatements()) {
            return false;
          }
        }
      }
    }
    return true;
  }

  private void processNestedSwitches(final CfgSwitchStatement switchNode) {
    CfgSwitchStatement grandChild;
    while ((grandChild = getSimilarExprGrandChild(switchNode)) != null) {
      collapseSwitchSubTree(switchNode, grandChild);
    }
  }

  private void collapseSwitchSubTree(
      final CfgSwitchStatement switchNode,
      final CfgSwitchStatement grandChild) {

    final CfgCaseStatement child = (CfgCaseStatement) grandChild.getParent();
    switchNode.remove(child);
    child.remove(grandChild);

    final Node switchCondition = switchNode.getCondition();
    final Node grandCondition = grandChild.getCondition();
    final NodeVariable commonUsed = getCommonUsedVariable(switchCondition, grandCondition);

    if (!switchCondition.equals(commonUsed)) {
      switchNode.setCondition(commonUsed);
    }

    final NodeValue switchValue = getAssignedValue(switchCondition);
    final NodeValue grandValue = getAssignedValue(grandCondition);

    /* If switch condition is not collapsed, modify it and it's non-removed case child. */
    if (switchValue != null) {

      final CfgCaseStatement restChild = (CfgCaseStatement) switchNode.getStatement(0);
      if (restChild.isDefault()) {
        restChild.setDefault(false);
      }
      restChild.removeValues();
      restChild.add(switchValue);
    }

    final List<CfgStatement> moved = new LinkedList<>();

    /* Change values for grandChild children */
    for (int i = 0; i < grandChild.getStatementNum(); i++) {
      final CfgCaseStatement toBeMoved = (CfgCaseStatement) grandChild.getStatement(i);
      final Set<NodeValue> caseValues = new LinkedHashSet<>(toBeMoved.getValues());
      toBeMoved.removeValues();
      if (caseValues.contains(SolverUtils.TRUE)) {
        toBeMoved.add(grandValue);
      } else {
        toBeMoved.setDefault(true);
      }
      moved.add(toBeMoved);
    }

    /* Re-link grandchild children case statements to root switch statement */
    for (final CfgStatement toBeMoved : moved) {
      toBeMoved.removeParent(grandChild);
      toBeMoved.setParent(switchNode);
    }
  }

  private static NodeValue getAssignedValue(final Node operation) {

    switch (operation.getKind()) {
      case OPERATION:
        for (int i = 0; i < ((NodeOperation)operation).getOperandCount(); i++) {
          final Node operand = ((NodeOperation)operation).getOperand(i);

          if (operand.getKind() == Node.Kind.VALUE) {
            return (NodeValue) operand;
          }
        }
        break;
      case VALUE:
        return (NodeValue) operation;
      default:
        return null;
    }
    return null;
  }

  private CfgSwitchStatement getSimilarExprGrandChild(final CfgSwitchStatement switchNode) {
    if (switchNode.getStatementNum() > 0) {
      for (int i = 0; i < switchNode.getStatementNum(); i++) {

        final CfgCaseStatement switchNodeChild = (CfgCaseStatement) switchNode.getStatement(i);

        if (hasSimilarExprGrandChild(switchNode, switchNodeChild)) {
          return (CfgSwitchStatement) switchNodeChild.getStatement(0);
        }
      }
    }
    return null;
  }

  private boolean hasSimilarExprGrandChild(
      final CfgSwitchStatement node,
      final CfgCaseStatement nodeChild) {

    Node nodeCondition = node.getCondition();
    if ((ExprUtils.isKind(Node.Kind.VARIABLE, nodeCondition)
            || nodeCondition.isType(DataType.BOOLEAN))
        && isFalseCase(node, nodeChild)
        && nodeChild.getStatementNum() == 1
        && nodeChild.getStatement(0).isType(CfgStatementType.SWITCH)) {

      final CfgSwitchStatement grandChild = (CfgSwitchStatement) nodeChild.getStatement(0);
      final Node grandChildCondition = grandChild.getCondition();

      return similar(nodeCondition, grandChildCondition);
    }
    return false;
  }

  private static boolean isFalseCase(
      final CfgSwitchStatement parent,
      final CfgCaseStatement statement) {
    final Set<NodeValue> values = statement.getValues();

    return statement.isDefault() && hasTrueCaseChildOtherThan(parent, statement)
        || (values.size() == 1 && values.iterator().next().equals(SolverUtils.FALSE));
  }

  private static boolean hasTrueCaseChildOtherThan(
      final CfgSwitchStatement parent,
      final CfgCaseStatement child) {

    for (int i = 0; i < parent.getStatementNum(); i++) {
      final CfgCaseStatement statement = (CfgCaseStatement) parent.getStatement(i);

      if (!statement.equals(child)) {
        return statement.getValues().contains(SolverUtils.TRUE);
      }
    }
    return false;
  }

  private static boolean isOneVarEquation(final Node node) {

    if (node.getKind() == Node.Kind.OPERATION
        && ((NodeOperation) node).getOperationId().equals(StandardOperation.EQ)) {
      int varCount = 0;
      int valCount = 0;

      for (int i = 0; i < ((NodeOperation) node).getOperandCount(); i++) {
        final Node operand = ((NodeOperation) node).getOperand(i);
        switch (operand.getKind()) {
          case VARIABLE:
            varCount++;
            break;
          case VALUE:
            valCount++;
            break;
          case OPERATION:
            varCount += ExprUtils.getVariables(operand).size();
            break;
          default:
            break;
        }
      }

      return (varCount == 1) && (valCount == 1);
    } else {
      return isOneVariable(node);
    }
  }

  private static boolean isOneVariable(final Node node) {
    return ExprUtils.isKind(Node.Kind.VARIABLE, node);
  }

  private boolean similar(
      final Node first,
      final Node second) {
    return isOneVarEquation(second)
        && isOneVarEquation(first)
        && getCommonUsedVariable(first, second) != null;
  }

  private NodeVariable getCommonUsedVariable(
      final Node switchCondition,
      final Node grandCondition) {
    if (switchCondition.getKind().equals(Node.Kind.VARIABLE)) {
      return (NodeVariable) switchCondition;
    } else {
      final Set<NodeVariable> switchUsed = new LinkedHashSet<>();

      for (int i = 0; i < ((NodeOperation) switchCondition).getOperandCount(); i++) {
        final Node operand = ((NodeOperation) switchCondition).getOperand(i);
        if (operand.getKind().equals(Node.Kind.VARIABLE)) {
          switchUsed.add((NodeVariable) operand);
        }
      }

      for (int i = 0; i < ((NodeOperation)grandCondition).getOperandCount(); i++) {
        final Node grandChildOperand = ((NodeOperation) grandCondition).getOperand(i);
        if (grandChildOperand instanceof NodeVariable && switchUsed.contains(grandChildOperand)) {
          return (NodeVariable) grandChildOperand;
        }
      }
    }
    return null;
  }
}
