/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgDeclarationStatement;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgStatementType;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.cfg.CfgWaitStatement;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Control flow graph model visitor that detects and transforms wait nodes.
 *
 * <p>The component transforms input model in such a way that the result does not contain
 * {@link CfgWaitStatement} nodes. It substitutes them
 * by sub-processes which are 'guarded' by predicates on new variables. Situations when
 * the same sequence of graph nodes can be executed both after and not after
 * {@link CfgWaitStatement} node are also elaborated.</p>
 *
 * <p>There is Zamia IG feature which is taken into account in this backend. The IG builder
 * creates {@link CfgWaitStatement} objects every time when it
 * detects some process' sensitivity list. The builder puts this node at the end of process'
 * statements sequence. Such nodes are treated by the backend as finalizing and transformed
 * into {@link EventList} sensitivity list components.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgWaitBackend extends CfgDefaultVisitor {

  /* The current module. */
  private CfgModule module;

  /* The current process. */
  private CfgProcess process;

  /* The collection for wait-processes. */
  private Map<CfgModule, List<CfgProcess>> waitProcessMap;

  /* The collection of finalizing wait nodes. */
  private Set<CfgWaitStatement> finalizing;

  /**
   * Default constructor.
   */
  public CfgWaitBackend() {
    super();
  }

  @Override
  public void onCfgModelBegin() {
    this.waitProcessMap = new LinkedHashMap<>();
    this.finalizing = new LinkedHashSet<>();
  }

  @Override
  public void onCfgModelEnd() {

    /* Remove finalizing wait nodes from the model. */
    for (final CfgWaitStatement wait : this.finalizing) {
      final CfgStatement waitParent = wait.getParent();
      waitParent.remove(wait);
    }

    /* Add wait sub-processes. */
    for (final Map.Entry<CfgModule, List<CfgProcess>> entry : this.waitProcessMap.entrySet()) {
      final List<CfgProcess> waitProcesses = entry.getValue();
      for (final CfgProcess waitProcess : waitProcesses) {
        entry.getKey().add(waitProcess);
      }
    }

    this.waitProcessMap.clear();
  }

  @Override
  public void onModuleBegin(final CfgModule module) {
    this.module = module;
    this.waitProcessMap.put(module, new LinkedList<CfgProcess>());
  }

  @Override
  public void onProcessBegin(final CfgProcess process) {
    this.process = process;
  }

  private EventList getSensitivityList(final CfgWaitStatement waitNode) {
    if (waitNode.getEventList().isEmpty()) {
      if (waitNode.getCondition() == null) {
        throw new IllegalArgumentException("Empty node of type: " + waitNode.getClass() + ".");
      } else {
        final Set<Event> events = new LinkedHashSet<>();
        final Collection<NodeVariable> variables = ExprUtils.getVariables(waitNode.getCondition());

        for (final NodeVariable variable : variables) {
          events.add(new Event(new RangedVariable(variable), EventType.ANY_EDGE));
        }
        return new EventList(events);
      }
    } else {
      return waitNode.getEventList().deepCopy();
    }
  }

  @Override
  public void onWaitBegin(final CfgWaitStatement wait) {

    if (isFinalizing(wait)) {

      /* The finalizing wait node is detected. */
      this.finalizing.add(wait);

    } else {

      /* The internal wait node is detected. */

      /* Create new process. */
      final CfgProcess waitProcess = new CfgProcess();
      waitProcess.setSensitivityList(getSensitivityList(wait));

      /* Create & fill process's control flow graph. */

      /* Start CFG building from creating the switch-case pair
       * that checks the related new variable. */

      final CfgDeclarationStatement declarationModule =
          CfgUtils.getNearVariableContainer(wait, CfgStatementType.MODULE);

      final NodeVariable newVariable = declarationModule.declareNewVariable(DataType.BOOLEAN);
      final Node variableCondition = NodeUtils.getEquation(newVariable, SolverUtils.TRUE);
      final Node condition = wait.getCondition() != null
          ? ExprUtils.getConjunction(wait.getCondition(), variableCondition)
          : variableCondition;
      final CfgSwitchStatement mainSwitch = new CfgSwitchStatement(condition);
      waitProcess.add(mainSwitch);

      final CfgCaseStatement aCase = new CfgCaseStatement(SolverUtils.TRUE);
      mainSwitch.add(aCase);

      /* Construct wait-process control flow graph. */

      CfgStatement parent = wait.getParent();
      int fromIndex = parent.indexOf(wait);
      boolean stop = false;

      while (!parent.isType(CfgStatementType.MODULE) && !stop) {

        final List<CfgStatement> toBeRemoved = new LinkedList<>();
        for (int i = fromIndex + 1; i < parent.getStatementNum(); i++) {
          final CfgStatement statement = parent.getStatement(i);

          /* When another wait node is detected, finish graph building. */
          if (statement.isType(CfgStatementType.WAIT)) {
            stop = true;
            break;
          }

          /* Move successor nodes which can be reached only after the specified wait node. */
          if (parent.equals(wait.getParent())) {
            toBeRemoved.add(statement);
          }

          /* Make a copy of the successor because it possibly can be reached other way
           * that does not go through the specified wait node. */
          waitProcess.add(statement.deepCopy());
        }
        for (final CfgStatement removed : toBeRemoved) {
          parent.remove(removed);
        }
        final CfgStatement ancestor = parent.getParent();
        fromIndex = ancestor.indexOf(parent);
        parent = ancestor;
      }

      /* Substitute wait node by block node. */
      final CfgBlockStatement trigger =
          new CfgBlockStatement(newVariable, SolverUtils.TRUE);
      CfgUtils.replace(wait, trigger);
      this.waitProcessMap.get(this.module).add(waitProcess);
    }
  }

  private boolean isFinalizing(final CfgWaitStatement wait) {
    return wait.getCondition() == null
        && wait.getEventList().equals(this.process.getSensitivityList())
        && CfgUtils.getSuccessors(wait).isEmpty();
  }
}
