/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License") {} you may not use this file
 * except in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.verilog;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.*;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.*;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgLoopStatement;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgWaitStatement;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.verilog.parser.model.*;
import ru.ispras.verilog.parser.model.basis.*;
import ru.ispras.verilog.parser.walker.VerilogNodeVisitor;

import java.util.*;
import java.util.logging.Logger;

/**
 * This class creates CfgModel from Verilog AST provided by Verilog Translator.
 *
 * @author <a href="mailto:chupilko@ispras.ru">Mikhail Chupilko</a>
 */
public final class VerilogCfgBuilder extends VerilogNodeVisitor {
  /*
   * {@link VariableTransformer} changes variables' user data format.
   *
   * @author <a href="mailto:chupilko@ispras.ru">Mikhail Chupilko</a>
   */
  private static final class VariableTransformer extends ExprTreeVisitorDefault {
    @Override
    public void onVariable(final NodeVariable variable) {
      InvariantChecks.checkNotNull(variable);

      final VerilogReference reference = (VerilogReference) variable.getUserData();
      InvariantChecks.checkNotNull(reference);
      final VerilogDeclaration declaration = reference.getDeclaration();
      InvariantChecks.checkNotNull(declaration);
      final VerilogElementType elementType = declaration.getElementType();
      InvariantChecks.checkNotNull(elementType);

      final VariableType type = getVariableType(declaration.getType());
      final VariableDescriptor variableDescriptor = new VariableDescriptor(type);

      variableDescriptor.addMetaInfo(MetaInfoType.HDL_TYPE_STR, elementType.getType().toString());
      variable.setUserData(variableDescriptor);
    }
  }

  /*
   * Resulting CFG model is kept here.
   */
  private CfgModel cfgModel;

  /*
   * Current CFG process to add statements.
   */
  private CfgProcess process;

  /*
   * CFG process for initializations.
   */
  private CfgProcess initialProcess;

  /*
   * List of switch types (0 - normal; >0 - casex, casez and levels of their depth).
   */
  private Stack<Integer> switches;

  /*
   * List of currently base block identifiers.
   */
  private Stack<String> currentbases;

  /*
   * List of data type in condition expressions.
   */
  private Stack<DataType> conditions;

  /*
   * Storage of CFG modules.
   */
  private List<CfgModule> modules;

  /*
   * Shows whether it is always process or initial.
   */
  private VerilogActivity.Type activityType;

  /**
   * Returns Retrascope variable type for the given Verilog parser data type.
   *
   * @param type Verilog parser data type.
   * @return Retrascope variable type.
   */
  private static VariableType getVariableType(final VerilogDeclaration.Type type) {
    switch (type) {
      case INPUT:
        return VariableType.IN;
      case OUTPUT:
        return VariableType.OUT;
      case INOUT:
        return VariableType.INOUT;
      case EVENT:
        return VariableType.REG;
      case PARAMETER:
        return VariableType.REG; //exCONST
      case SPECPARAM:
        return VariableType.REG; //exCONST
      case LOCALPARAM:
        return VariableType.REG; //exCONST
      case DEFPARAM:
        return VariableType.REG; //exCONST
      case GENVAR:
        return VariableType.REG;
      case VARIABLE:
        return VariableType.REG;
      default:
        return VariableType.REG;
    }
  }

  /**
   * Returns constructed CFG model.
   *
   * @return CFG model.
   */
  public CfgModel getCfgModel() {
    return cfgModel;
  }


  /**
   * Returns the CFG model reference kept in the storage according to a number of current CFG model.
   *
   * @return Current CFG model.
   */
  private CfgModule getCurrentModule() {
    return modules.get(modules.size() - 1);
  }

  /**
   * Changes userData type kept in the node from Reference to VariableType.
   *
   * @param node Node to be updated.
   * @return Updated node.
   */
  public Node transformNode(final Node node) {
    final Node nodeClone = node.deepCopy();
    final ExprTreeWalker walker = new ExprTreeWalker(new VariableTransformer());
    walker.visitNode(nodeClone);
    return nodeClone;
  }

  /**
   * Returns the logger of the engine.
   *
   * @return Logger of the engine.
   */
  private Logger getLogger() {
    return Log.getLogger(this.getClass());
  }

  private Assignment createAssignment(
      final VerilogAssignment assignment,
      final Node expression,
      final VerilogReference ref,
      final int rightSize) {
    final NodeVariable variable = (NodeVariable) transformNode(ref.getVariable());
    final RangedVariable rangedVariable = new RangedVariable(variable);

    final DataType dataType = getCurrentModule().contains(variable)
        ? getCurrentModule().getVariable(variable.getName()).getDataType()
        : variable.getDataType();

    int leftSize = dataType.getSize();

    if (ref.hasBitsSelection()) {
      final VerilogExpression leftExpression = ref.getBitsSelection().getLeftExpression();

      VerilogExpression rightExpression = ref.getBitsSelection().getRightExpression();
      if (rightExpression == null) {
        rightExpression = leftExpression;
      }
      leftSize =
          Math.abs(leftExpression.evaluateInteger() - rightExpression.evaluateInteger()) + 1;
      final Range range = new Range(leftExpression.getNode(), rightExpression.getNode());
      rangedVariable.setRange(range);
    }

    Node localExpression = expression;
    if (assignment.getReference().size() > 1) {
      final NodeValue lowValue = NodeValue.newInteger(rightSize - leftSize);
      final NodeValue highValue = NodeValue.newInteger(rightSize - 1);
      // TODO: the assignment result is never used
//      rightSize -= leftSize;
      localExpression =
          new NodeOperation(StandardOperation.BVEXTRACT, lowValue, highValue, expression);
    }

    return new Assignment(rangedVariable, localExpression);
  }
  
  /**
   * Processes assignments of different types.
   *
   * @param assignment The assignment expression itself.
   * @param type Blocking, non-blocking, assign (deassign, force, release will not be supported).
   * @param insideProcess Add the assignment to the current process or create a new one.
   */
  public CfgBlockStatement parseAssignment(
      final VerilogAssignment assignment,
      final VerilogAssignStatement.Type type,
      final boolean insideProcess) {
    final EventList eventList = new EventList();
    if (!insideProcess) {
      final Set<NodeVariable> useVars = assignment.getUsedVariables();

      for (final NodeVariable node : useVars) {
        final RangedVariable rangedVariable = new RangedVariable((NodeVariable)transformNode(node));
        final ru.ispras.retrascope.model.basis.Event retrascopeEvent =
            new ru.ispras.retrascope.model.basis.Event(rangedVariable);
        if (getCurrentModule().getDeclarationMap().get(rangedVariable.getVariable()) == null) {
          getLogger().log(
              LogLevel.WARNING,
              String.format(
                  "Variable %s is not a variable in module %s!",
                  rangedVariable.getVariable().getName(),
                  getCurrentModule().getFullName()));
        }
        eventList.add(retrascopeEvent);
      }
    }

    final Node expression = assignment.getExpression().isUsual()
        ? transformNode(assignment.getExpression().getMinExpression().getNode()) : null;
    // NOTICE: for unusual expressions which are not supported
    // System.out.println(assignment.getCondition().getMinExpression());
    // System.out.println(":");
    // System.out.println(assignment.getCondition().getTypExpression());
    // System.out.println(":");
    // System.out.println(assignment.getCondition().getMaxExpression());

    if (expression == null) {
      throw new IllegalStateException();
    }

    final CfgBlockStatement block =
        new CfgBlockStatement(type == VerilogAssignStatement.Type.NON_BLOCKING);

    int rightSize = expression.getDataType().getSize();
    for (final VerilogReference ref : assignment.getReference().getReferences()) {
      final Assignment retraAssignment =
          createAssignment(assignment, expression, ref, rightSize);
      block.addAssignment(retraAssignment);
    }
    addBlock(block, eventList, insideProcess);
    return block;
  }

  private Collection<Assignment> makeAssignment(
      final VerilogAssignment assignment) {
    InvariantChecks.checkTrue(assignment.getExpression().isUsual());

    final Node rhs = transformNode(assignment.getExpression().getMinExpression().getNode());
    final int size = rhs.getDataType().getSize();

    final Collection<Assignment> newAssignments = new ArrayList<>();

    for (final VerilogReference reference : assignment.getReference().getReferences()) {
      newAssignments.add(createAssignment(assignment, rhs, reference, size));
    }

    return newAssignments;
  }

  /**
   * Simplified adding of process nodes into Cfg.
   *
   * <p>Required are only two parameters: the CFGNode itself and flag forcing
   * to add the Node into initial or normal process.
   * The other parameters are set to the following default values:
   * empty eventList and adding into existing process (if it exists).</p>
   *
   * @param block Node to be added.
   */
  public void addBlock(final CfgStatement block) {
    final EventList eventList = new EventList();
    addBlock(block, eventList, true);
  }

  /**
   * Add process node (some block, e.g. assignment) to CFG.
   *
   * @param block Node to be added.
   * @param eventList Should contain all the events touching the process.
   *                  Used in case of continuous assignments.
   * @param insideProcess If {@code true}, the block is added to existing process,
   *                      If {@code false}, new process is created for it.
   */
  public void addBlock(
      final CfgStatement block,
      final EventList eventList,
      final boolean insideProcess) {
    if (insideProcess) {
      attachBlock(block);
    } else {
      final CfgProcess process = new CfgProcess();
      process.setSensitivityList(eventList);
      process.add(block);
      getCurrentModule().add(process);
    }
  }

  /**
   * Attach block into existing process.
   * @param block Node to be added.
   */
  private void attachBlock(CfgStatement block) {
    final CfgProcess localProcess =
        activityType == VerilogActivity.Type.INITIAL ? initialProcess : process;
    if (currentbases.isEmpty()) {
      localProcess.add(block);
    } else {
      CfgStatement statement = localProcess.getStatement(currentbases.peek());
      statement.add(block);
    }
  }

  //------------------------------------------------------------------------------------------------
  // Activity
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onActivityBegin(final VerilogActivity node) {
    activityType = node.getType();
    if (activityType == VerilogActivity.Type.INITIAL) { // or INITIAL or ALWAYS
      if (initialProcess == null) {
        initialProcess = new CfgProcess(true);
        createEmptyCfg(initialProcess);
      }
    } else {
      process = new CfgProcess();
      createEmptyCfg(process);
    }

    return Result.OK;
  }

  private void createEmptyCfg(final CfgProcess process) {
    final List<CfgStatement> nodes = new ArrayList<>();
    process.addAll(nodes);
  }

  @Override
  public Result onActivityEnd(final VerilogActivity node) {
    if (activityType == VerilogActivity.Type.INITIAL) {
      getCurrentModule().add(initialProcess);
      activityType = VerilogActivity.Type.ALWAYS;
      initialProcess = null;
    } else {
      getCurrentModule().add(process);
    }

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Assign
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onAssignBegin(final VerilogAssign node) {
    // node.getDelay() is not used
    parseAssignment(node.getAssignment(), VerilogAssignStatement.Type.ASSIGN, false);

    return Result.OK;
  }

  @Override
  public Result onAssignEnd(final VerilogAssign node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // AssignStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onAssignStatementBegin(final VerilogAssignStatement node) {
    switch (node.getType()) {
      case BLOCKING:
      case NON_BLOCKING:
      case ASSIGN: {
        parseAssignment(node.getAssignment(), node.getType(), true);
        break;
      }
      default: {
        getLogger().log(LogLevel.WARNING, node.getType() + " is unsupported.");
        break;
      }
    }

    return Result.OK;
  }

  @Override
  public Result onAssignStatementEnd(final VerilogAssignStatement node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Assignment
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onAssignmentBegin(final VerilogAssignment node) {
    parseAssignment(node, VerilogAssignStatement.Type.ASSIGN, false);

    return Result.OK;
  }

  @Override
  public Result onAssignmentEnd(final VerilogAssignment node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Attribute
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onAttributeBegin(final VerilogAttribute node) {
    // Nothing
    return Result.OK;
  }

  @Override
  public Result onAttributeEnd(final VerilogAttribute node) {
    // Nothing
    return Result.OK;
  }

  // -------------------------------------------------------------------------
  // BlockGenerate
  // -------------------------------------------------------------------------

  @Override
  public Result onBlockGenerateBegin(final VerilogBlockGenerate node) {
    return Result.OK;
  }

  @Override
  public Result onBlockGenerateEnd(final VerilogBlockGenerate node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CfgBlockStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onBlockStatementBegin(final VerilogBlockStatement node) {

    if (node.getType() == VerilogBlockStatement.Type.FORK) {
      getLogger().log(LogLevel.WARNING, LogMessage.UNSUPPORTED_TYPE + node.getType().name());
    }

    return Result.OK;
  }

  @Override
  public Result onBlockStatementEnd(final VerilogBlockStatement node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CaseGenerate
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onCaseGenerateBegin(final VerilogCaseGenerate node) {
    return Result.OK;
  }

  @Override
  public Result onCaseGenerateEnd(final VerilogCaseGenerate node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CaseGenerateItem
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onCaseGenerateItemBegin(final VerilogCaseGenerateItem node) {
    return Result.OK;
  }

  @Override
  public Result onCaseGenerateItemEnd(final VerilogCaseGenerateItem node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CfgCaseStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onCaseStatementBegin(final VerilogCaseStatement node) {
    switch (node.getType()) {
      case CASE:
        if (node.getExpression().getNode().getDataType() == DataType.UNKNOWN) {
          Log.getLogger().log(
              LogLevel.INFO,
              "[onCaseStatementBegin] DataType of " + node.getExpression().getNode() + " is UNKNOWN.");
        }
        CfgSwitchStatement switchBlock = new CfgSwitchStatement(node.getExpression().getNode());
        attachBlock(switchBlock);
        currentbases.push(switchBlock.getId());
        switches.push(0);
        break;
      case CASEX:
      case CASEZ:
        switches.push(1);
        break;
      default:
        Log.printError(LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, node.getType()));
        break;
    }

    return Result.OK;
  }

  @Override
  public Result onCaseStatementEnd(final VerilogCaseStatement node) {
    currentbases.pop();
    for (int i = (switches.peek() - 1) * 2; i > 1; i--) {
      currentbases.pop();
    }
    switches.pop();

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CaseStatementItem
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onCaseStatementItemBegin(final VerilogCaseStatementItem node) {
    CfgCaseStatement caseBlock =
        new CfgCaseStatement(false);
    for (final VerilogExpression expression : node.getExpressions()) {
      caseBlock.add(expression.getValue());
    }
    if (switches.peek() > 0) {
      if (node.getExpression().getNode().getDataType() == DataType.UNKNOWN) {
        Log.getLogger().log(
            LogLevel.ERROR,
            "[onCaseStatementItemBegin] DataType of "
                + node.getExpression().getNode() + " is UNKNOWN.");
      }
      
      CfgSwitchStatement switchBlock = new CfgSwitchStatement(node.getExpression().getNode());
      attachBlock(switchBlock);
      currentbases.push(switchBlock.getId());

      Integer peekval = switches.peek();
      peekval++;
      switches.pop();
      switches.push(peekval);

      final NodeValue nodeValue = new NodeValue(Data.newBoolean(true));
      caseBlock = new CfgCaseStatement(nodeValue);
    }
    
    attachBlock(caseBlock);
    currentbases.push(caseBlock.getId());
    
    return Result.OK;
  }

  @Override
  public Result onCaseStatementItemEnd(final VerilogCaseStatementItem node) {
    currentbases.pop();

    if (switches.peek() > 0) {
      final NodeValue nodeValue = new NodeValue(Data.newBoolean(false));
      final CfgCaseStatement caseBlock = new CfgCaseStatement(nodeValue);

      attachBlock(caseBlock);
      currentbases.push(caseBlock.getId());
    }

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Code
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onCodeBegin(final VerilogCode node) {
    // Nothing
    return Result.OK;
  }

  @Override
  public Result onCodeEnd(final VerilogCode node) {
    // Creates new empty CFG model and adds it into storage of CFG models.
    cfgModel = new CfgModel();
    if (modules == null) {
      modules = new ArrayList<>();
      final CfgModule module =
          new CfgModule("");
      module.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VERILOG);
      modules.add(module);
    }
    cfgModel.addModules(modules);

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Declaration
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onDeclarationBegin(final VerilogDeclaration node) {
    final Set<VerilogDeclaration> namesakes = node.getNamesakes();

    if (namesakes.isEmpty() || node.isOutput() || node.isInout()) {
      final DataType type = node.getElementType().getDataType();
      final NodeVariable variable = new NodeVariable(node.getName(), type);

      final VariableDescriptor variableDescriptor =
          new VariableDescriptor(getVariableType(node.getType()));
      variableDescriptor.addMetaInfo(MetaInfoType.HDL_TYPE_STR.name(),
          node.getElementType().getType().toString());
      variable.setUserData(variableDescriptor);
      getCurrentModule().declareVariable(variable);

      getLogger().log(LogLevel.DEBUG,
          String.format("add variable %s [%s] (%s) to module %s.",
              variable, type, variable.getData(), getCurrentModule().getFullName()));
    }

    // final ElementType type = node.getElementType();
    // parameter type.isSpecified() is not used

    if (node.getExpression() != null) {
      if (initialProcess == null) {
        initialProcess = new CfgProcess(true);
        createEmptyCfg(initialProcess);
      }

      final VerilogAssignment assignment = new VerilogAssignment(null);
      final VerilogReference reference = new VerilogReference();
      final VerilogPath path = new VerilogPath(node.getName());
      reference.setDeclaration(node);
      reference.setPath(path);
      assignment.addReference(reference);
      assignment.setExpression(node.getExpression());
      activityType = VerilogActivity.Type.INITIAL;
      parseAssignment(assignment, VerilogAssignStatement.Type.ASSIGN, true);
    }

    return Result.OK;
  }

  @Override
  public Result onDeclarationEnd(final VerilogDeclaration node) {
    activityType = VerilogActivity.Type.ALWAYS;
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // DelayedStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onDelayedStatementBegin(final VerilogDelayedStatement node) {
    final VerilogEventControl eventControl = node.getEventControl();

    if (eventControl.isEvents()) { //|| eventControl.isRepeat() eventControl.isDelays() not parsed
      for (final VerilogEvent event : eventControl.getEvents()) {
        EventType eventType = EventType.ANY_EDGE;
        if (event.getEdge().isPositive()) {
          eventType = EventType.POSITIVE_EDGE;
        } else if (event.getEdge().isNegative()) {
          eventType = EventType.NEGATIVE_EDGE;
        } else if (event.getEdge().isUndefined()) {
          eventType = EventType.ANY_EDGE;
        } else {
          assert false;
        }

        final EventList eventList = new EventList();
        for (final NodeVariable nodeVariable : event.getExpression().getVariables()) {
          final RangedVariable rangedVariable =
              new RangedVariable((NodeVariable)transformNode(nodeVariable));

          final Event retrascopeEvent = new Event(rangedVariable, eventType);

          if (!getCurrentModule().contains(nodeVariable)) {
            getLogger().log(
                LogLevel.WARNING,
                String.format(
                    "Variable %s is not a variable in module %s!",
                    nodeVariable.getName(),
                    getCurrentModule().getFullName()));
          }

          final CfgProcess localProcess =
              activityType == VerilogActivity.Type.INITIAL ? initialProcess : process;
          if (activityType == VerilogActivity.Type.INITIAL) {
            eventList.add(retrascopeEvent);
          } else {
            localProcess.getSensitivityList().add(retrascopeEvent);
          }
        }

        if (activityType == VerilogActivity.Type.INITIAL) {
          final CfgWaitStatement waitBlock =
              new CfgWaitStatement(eventList
              /*, transformNode(event.getExpression().getNode())*/);
          addBlock(waitBlock);
        }
      }
    }

    return Result.OK;
  }

  @Override
  public Result onDelayedStatementEnd(final VerilogDelayedStatement node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // DisableStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onDisableStatementBegin(final VerilogDisableStatement node) {
    return Result.OK;
  }

  @Override
  public Result onDisableStatementEnd(final VerilogDisableStatement node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Generate
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onGenerateBegin(final VerilogGenerate node) {
    return Result.OK;
  }

  @Override
  public Result onGenerateEnd(final VerilogGenerate node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // IfGenerate
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onIfGenerateBegin(final VerilogIfGenerate node) {
    return Result.OK;
  }

  @Override
  public Result onIfGenerateEnd(final VerilogIfGenerate node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // IfGenerateBranch
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onIfGenerateBranchBegin(final VerilogIfGenerateBranch node) {
    return Result.OK;
  }

  @Override
  public Result onIfGenerateBranchEnd(final VerilogIfGenerateBranch node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // IfStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onIfStatementBegin(final VerilogIfStatement node) {
    if (node.getExpression().getNode().getDataType() == DataType.UNKNOWN) {
      Log.getLogger().log(
          LogLevel.ERROR,
          "[onIfStatementBegin] DataType of "
          + node.getExpression().getNode() + " is UNKNOWN.");
    }
    
    CfgSwitchStatement switchBlock = new CfgSwitchStatement(node.getExpression().getNode());
    attachBlock(switchBlock);
    currentbases.push(switchBlock.getId());
    conditions.push(node.getExpression().getElaboratedNode().getDataType());
    return Result.OK;
  }

  @Override
  public Result onIfStatementEnd(final VerilogIfStatement node) {
    currentbases.pop();
    conditions.pop();

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // IfStatementBranch
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onIfStatementBranchBegin(final VerilogIfStatementBranch node) {
    if (node.isThen()) {
      final NodeValue nodeValue = getCaseValue(1, true);
      final CfgCaseStatement caseBlock =
          new CfgCaseStatement(nodeValue);
      attachBlock(caseBlock);
      currentbases.push(caseBlock.getId());
    } else if (node.isElse()) {
      final NodeValue nodeValue = getCaseValue(0, false);
      final CfgCaseStatement caseBlock =
          new CfgCaseStatement(nodeValue);
      attachBlock(caseBlock);
      currentbases.push(caseBlock.getId());
    }

    return Result.OK;
  }

  private NodeValue getCaseValue(int bvValue, boolean boolValue) {
    return new NodeValue(conditions.peek().getTypeId() == DataTypeId.BIT_VECTOR
        ? Data.newBitVector(bvValue, conditions.peek().getSize())
        : Data.newBoolean(boolValue));
  }

  @Override
  public Result onIfStatementBranchEnd(final VerilogIfStatementBranch node) {
    currentbases.pop();
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Instantiation
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onInstantiationBegin(final VerilogInstantiation node) {
    getLogger().log(LogLevel.DEBUG,
        String.format(
            "Instantiation of %s type is module %s.", node.getModuleName(), node.getName()));

    final CfgModule module = new CfgModule(node.getModuleName());

    module.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VERILOG);
    module.setInstanceName(node.getName());

    //assert (modules.isEmpty()); TODO
    // Add newly created module as a child to that one.
    // TODO: definition of the module should be found first.
    getCurrentModule().add(module);

    return Result.OK;
  }

  @Override
  public Result onInstantiationEnd(final VerilogInstantiation node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // LoopGenerate
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onLoopGenerateBegin(final VerilogLoopGenerate node) {
    return Result.OK;
  }

  @Override
  public Result onLoopGenerateEnd(final VerilogLoopGenerate node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CfgLoopStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onLoopStatementBegin(final VerilogLoopStatement node) {
    CfgLoopStatement loopStatement = null;

    switch (node.getType()) {
      case FOREVER: {
        loopStatement = new CfgLoopStatement(CfgLoopStatement.Type.FOREVER, null, null, null);
        break;
      }
      case REPEAT: {
        final DataType dataType = node.getExpression().getElaboratedNode().getDataType();
        final NodeVariable nodeVariable = getCurrentModule().declareNewVariable(dataType);

        NodeValue nodeValue;
        Enum<?> operationId;
        if (dataType.getTypeId().equals(DataTypeId.LOGIC_INTEGER)) {
          nodeValue = new NodeValue(Data.newInteger(0));
          operationId = StandardOperation.LESS;
        } else if (dataType.getTypeId().equals(DataTypeId.BIT_VECTOR)) {
          nodeValue = new NodeValue(Data.newBitVector(0, dataType.getSize()));
          operationId = StandardOperation.BVULT;
        } else {
          throw new IllegalArgumentException(
              LogMessage.makeMessage(LogMessage.CANT_PARSE, node.getExpression()));
        }

        final RangedVariable rangedVariable =
            new RangedVariable((NodeVariable)transformNode(nodeVariable));

        final Assignment initAssignment = new Assignment(rangedVariable, nodeValue);

        final NodeOperation expression =
            new NodeOperation(operationId, nodeVariable, node.getExpression().getValue());
        
        if (dataType.getTypeId().equals(DataTypeId.LOGIC_INTEGER)) {
          nodeValue = new NodeValue(Data.newInteger(1));
          operationId = StandardOperation.ADD;
        } else if (dataType.getTypeId().equals(DataTypeId.BIT_VECTOR)) {
          nodeValue = new NodeValue(Data.newBitVector(1, dataType.getSize()));
          operationId = StandardOperation.BVADD;
        } else {
          throw new IllegalArgumentException(
              LogMessage.makeMessage(LogMessage.CANT_PARSE, node.getExpression()));
        }

        final NodeOperation incExpr =
            new NodeOperation(operationId, nodeVariable, nodeValue);
        final RangedVariable incVar =
            new RangedVariable((NodeVariable) transformNode(nodeVariable));
        final Assignment incAssignment = new Assignment(incVar, incExpr);

        loopStatement =
            new CfgLoopStatement(
                CfgLoopStatement.Type.REPEAT,
                expression,
                initAssignment,
                incAssignment);
        break;
      }
      case WHILE: {
        if (node.getExpression().getNode().getDataType() == DataType.UNKNOWN) {
          Log.getLogger().log(
              LogLevel.ERROR,
              "[onLoopStatementBegin] DataType of "
                  + node.getExpression().getNode() + " is UNKNOWN.");
        }

        loopStatement =
            new CfgLoopStatement(
                CfgLoopStatement.Type.WHILE,
                node.getExpression().getNode(),
                null,
                null);
        break;
      }
      case FOR: {
        if (node.getExpression().getNode().getDataType() == DataType.UNKNOWN) {
          Log.getLogger().log(
              LogLevel.ERROR,
              "[onLoopStatementItemBegin2] DataType of "
                  + node.getExpression().getNode() + " is UNKNOWN.");
        }

        final Collection<Assignment> initAssignment = makeAssignment(node.getInitialization());
        InvariantChecks.checkTrue(initAssignment.size() == 1);

        final Collection<Assignment> iterAssignment = makeAssignment(node.getIteration());
        InvariantChecks.checkTrue(iterAssignment.size() == 1);

        loopStatement =
            new CfgLoopStatement(
                CfgLoopStatement.Type.FOR,
                node.getExpression().getNode(),
                initAssignment.iterator().next(),
                iterAssignment.iterator().next());
        break;
      }
      default:
        Log.printError(LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, node.getType()));
    }
    attachBlock(loopStatement);
    currentbases.push(loopStatement.getId());

    return Result.OK;
  }

  @Override
  public Result onLoopStatementEnd(final VerilogLoopStatement node) {
    currentbases.pop();
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Module
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onModuleBegin(final VerilogModule node) {
    if (node.isMacromodule()) {
      getLogger().log(
          LogLevel.ERROR,
          LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, node.getType()));

    } else if (node.isPrimitive()) {
      getLogger().log(
          LogLevel.ERROR,
          LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, node.getType()));
    } else {
      final CfgModule module =
          new CfgModule(node.getName());
      module.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VERILOG);

      getLogger().log(LogLevel.DEBUG,
          String.format("Start observing module %s.", module.getFullName()));

      if (modules == null) {
        modules = new ArrayList<>();
      }
      if (switches == null) {
        switches = new Stack<>();
      }
      if (conditions == null) {
        conditions = new Stack<>();
      }
      if (currentbases == null) {
        currentbases = new Stack<>();
      }

      modules.add(module);
    }

    return Result.OK;
  }

  @Override
  public Result onModuleEnd(final VerilogModule node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // NullStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onNullStatementBegin(final VerilogNullStatement node) {
    return Result.OK;
  }

  @Override
  public Result onNullStatementEnd(final VerilogNullStatement node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // PathDeclaration
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onPathDeclarationBegin(final VerilogPathDeclaration node) {
    return Result.OK;
  }

  @Override
  public Result onPathDeclarationEnd(final VerilogPathDeclaration node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Port
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onPortBegin(final VerilogPort node) {
    // Nothing
    return Result.OK;
  }

  @Override
  public Result onPortEnd(final VerilogPort node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // PortConnection
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onPortConnectionBegin(final VerilogPortConnection node) {
    final VerilogExpression portExpr = node.getExpression();

    if (portExpr == null) {
      return Result.OK;
    }
    if (portExpr.getNode().getDataType() == DataType.UNKNOWN) {
      Log.getLogger().log(
          LogLevel.ERROR,
          "[onPortConnectionBegin] DataType of " + portExpr.getNode() + " is UNKNOWN.");
    }

    final Collection<CfgModule> modules = getCurrentModule().getModules();
    /*if (!modules.isEmpty()) { TODO think about bindings
      final CfgModule module = modules.get(modules.size() - 1);
      Map<String, Node> bindings = module.getBindings();
      if (bindings == null) {
        bindings = new HashMap<>();
      }

      final String portName = node.getName();

      // A binding is not required to be present.
      if (portName != null) {
        // TODO: find this module declaration and check variables
        bindings.put(portName, portExpr.getNode());
      }

      module.setBindings(bindings);
    }*/
    return Result.OK;
  }

  @Override
  public Result onPortConnectionEnd(final VerilogPortConnection node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Procedure
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onProcedureBegin(final VerilogProcedure node) {
    /*final ElementType type = node.getElementType();

    if (type.isSpecified()) { // TODO: implement functions and tasks.
      text("function");
    } else {
      text("task");
    }

    if (node.isAutomatic()) {
      text(" ");
      text("automatic");
    }

    if (type.isSpecified()) {
      text(" ");
      text(type);
    }

    text(" ");
    text(node.getName());
    endl();*/

    return Result.OK;
  }

  @Override
  public Result onProcedureEnd(final VerilogProcedure node) {
    /*final ElementType type = node.getElementType();

    if (type.isSpecified()) {
      text("endfunction");
    } else {
      text("endtask");
    }

    endl();*/

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // PulseStyle
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onPulseStyleBegin(final VerilogPulseStyle node) {
    return Result.OK;
  }

  @Override
  public Result onPulseStyleEnd(final VerilogPulseStyle node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // ShowCancelled
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onShowCancelledBegin(final VerilogShowCancelled node) {
    return Result.OK;
  }

  @Override
  public Result onShowCancelledEnd(final VerilogShowCancelled node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Specify
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onSpecifyBegin(final VerilogSpecify node) {
    return Result.OK;
  }

  @Override
  public Result onSpecifyEnd(final VerilogSpecify node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Table
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onTableBegin(final VerilogTable node) {
    /*text("table"); TODO: implement tables.
    endl();*/

    return Result.OK;
  }

  @Override
  public Result onTableEnd(final VerilogTable node) {
    /*text("endtable");
    endl();*/

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // TableEntry
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onTableEntryBegin(final VerilogTableEntry node) {
    /*text(" ", node.getInputs1());
    text(" ");
    text(" ", node.getEdge()); TODO: implement tables.
    text(" ");
    text(" ", node.getInputs2());
    text(" : ");
    text(" ", node.getOutputs());
    text(";");
    endl();*/

    return Result.OK;
  }

  @Override
  public Result onTableEntryEnd(final VerilogTableEntry node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // TaskStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onTaskStatementBegin(final VerilogTaskStatement node) {
    /*text(node.getPath());
    text("(");

    boolean comma = false;
    for (final Expression expression : node.getArguments()) {
      if (comma) {
        text(", ");
      }
      comma = true;

      text(expression); TODO: implement task statements
    }

    text(");");
    endl();*/

    return Result.OK;
  }

  @Override
  public Result onTaskStatementEnd(final VerilogTaskStatement node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // TriggerStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onTriggerStatementBegin(final VerilogTriggerStatement node) {
    return Result.OK;
  }

  @Override
  public Result onTriggerStatementEnd(final VerilogTriggerStatement node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CfgWaitStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onWaitStatementBegin(final VerilogWaitStatement node) {

    final NodeVariable nodeVariable = (NodeVariable)transformNode(node.getExpression().getNode());
    NodeValue nodeValue;
    if (nodeVariable.getDataTypeId().equals(DataTypeId.LOGIC_INTEGER)) {
      nodeValue = new NodeValue(Data.newInteger(0));
    } else {
      if (nodeVariable.getDataTypeId().equals(DataTypeId.BIT_VECTOR)) {
        nodeValue = new NodeValue(Data.newBitVector(0, nodeVariable.getDataType().getSize()));
      } else {
        throw new IllegalArgumentException(
            LogMessage.makeMessage(LogMessage.CANT_PARSE, node.getExpression()));
      }
    }

    final NodeOperation nodeExpr =
        new NodeOperation(StandardOperation.NOTEQ, nodeVariable, nodeValue);

    final CfgWaitStatement waitBlock =
        new CfgWaitStatement(nodeExpr);
    addBlock(waitBlock);

    return Result.OK;
  }

  @Override
  public Result onWaitStatementEnd(final VerilogWaitStatement node) {
    return Result.OK;
  }
}
