/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.SourceLocation;
import org.zamia.ZamiaException;
import org.zamia.instgraph.IGDefaultOperationVisitor;
import org.zamia.instgraph.IGItem;
import org.zamia.instgraph.IGObject;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGOperationAlias;
import org.zamia.instgraph.IGOperationAllocate;
import org.zamia.instgraph.IGOperationArrayAggregate;
import org.zamia.instgraph.IGOperationAttribute;
import org.zamia.instgraph.IGOperationBinary;
import org.zamia.instgraph.IGOperationDeref;
import org.zamia.instgraph.IGOperationError;
import org.zamia.instgraph.IGOperationIndex;
import org.zamia.instgraph.IGOperationInvokeSubprogram;
import org.zamia.instgraph.IGOperationLiteral;
import org.zamia.instgraph.IGOperationObject;
import org.zamia.instgraph.IGOperationPhi;
import org.zamia.instgraph.IGOperationRange;
import org.zamia.instgraph.IGOperationRecordAggregate;
import org.zamia.instgraph.IGOperationRecordField;
import org.zamia.instgraph.IGOperationTypeConversion;
import org.zamia.instgraph.IGOperationTypeQualification;
import org.zamia.instgraph.IGOperationUnary;
import org.zamia.instgraph.IGRange;
import org.zamia.instgraph.IGSequenceOfStatements;
import org.zamia.instgraph.IGSequentialLoop;
import org.zamia.instgraph.IGSequentialStatement;
import org.zamia.instgraph.IGStaticValue;
import org.zamia.instgraph.IGSubProgram;
import org.zamia.instgraph.IGType;
import org.zamia.instgraph.IGTypeStatic;
import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.Radix;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.datamap.DataMap;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.parser.vhdl.mapper.BitVectorOpCodeMapper;
import ru.ispras.retrascope.parser.vhdl.mapper.IgBinaryMapper;
import ru.ispras.retrascope.parser.vhdl.mapper.IgSubprogramMapper;
import ru.ispras.retrascope.util.TransformerUtils;

/**
 * IG operation objects parser.
 *
 * <p>Component that transforms IG (Instantiation Graph) objects
 * to internal representation.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class IgOperationVisitor extends IGDefaultOperationVisitor {

  /* The current result of object processing. */
  private IgOperationResult result;

  private IgStructureVisitor parent;

  public static final String SHIFT = "shift";

  /**
   * Default constructor.
   */
  public IgOperationVisitor() {
    super();
  }

  /**
   * Sets visitor for parent IG structure.
   *
   * @param visitor Parent structure visitor.
   */
  public void setParentVisitor(final IgStructureVisitor visitor) {
    this.parent = visitor;
  }

  /**
   * Returns parent structure visitor.
   *
   * @return Parent structure visitor.
   */
  public IgStructureVisitor getStructureVisitor() {
    return this.parent;
  }

  /**
   * Returns result that is formed by object on the current argument.
   *
   * @see IgOperationResult
   *
   * @return Result of processing that is available at the current moment.
   */
  public IgOperationResult getResult() {
    return this.result;
  }

  private VariableBuilder getBuilder() {
    return getStructureVisitor().getBuilder();
  }

  @Override
  public void visitOperation(final IGOperation igOperation) {
    super.visitOperation(igOperation);
  }

  @Override
  public void visitStaticValue(final IGStaticValue igStaticValue) {

    NodeValue value = getNodeValue(igStaticValue);

    this.result = new IgOperationResult(value);
  }

  private NodeValue getNodeValue(IGStaticValue igStaticValue) {

    final IGType type = igStaticValue.getType();

    if (type == null) {
      return NodeValue.newBoolean(igStaticValue.isTrue());
    }

    if (type.isBit()) {

      /* One-bit-size vector value. */
      return NodeValue.newBitVector(BitVector.valueOf(igStaticValue.isTrue()));
    } else if (type.isBool()) {

      /* Boolean value. */
      return NodeValue.newBoolean(igStaticValue.isTrue());
    } else if (type.isInteger()) {

      /* Integer value. */
      return NodeValue.newInteger(igStaticValue.getInt());
    } else if (type.isReal()) {

      /* Real value. */
      return NodeValue.newReal(igStaticValue.getReal().doubleValue());
    } else if (type.isLogic()) {

      //TODO: implement 9-sign VHDL logic
      final IGType indexType = type.getIndexType();
      BitVector initVector;
      if (indexType == null) {
        /* 1-bit size vector */
        initVector =
            BitVector.valueOf(
                (igStaticValue.getCharLiteral() == '1')
                    || (igStaticValue.getCharLiteral() == '-'));

      } else {

        final int typeSize = getBuilder().getTypeSize((IGTypeStatic) indexType);
        initVector = BitVector.valueOf(0, typeSize);
      }
      return NodeValue.newBitVector(initVector);
    } else if (type.isArray()) {
      final SourceLocation location = igStaticValue.computeSourceLocation();
      try {
        String arrayString = getSmtLibArrayString(igStaticValue, location);

        return new NodeValue(Data.newArray(
            DataMap.valueOf(arrayString,
            getBuilder().getDataType(type.getIndexType()),
            getBuilder().getDataType(type.getElementType()))));
      } catch (ZamiaException e) {
        throw new RetrascopeRuntimeException(e);
      }
    }
    throw new UnsupportedOperationException(igStaticValue.toHRString());
  }

  private String getSmtLibArrayString(IGStaticValue igStaticValue, SourceLocation location)
    throws ZamiaException {
    final int entryNum = igStaticValue.getNumArrayEntries(location);

    final StringBuilder builder = new StringBuilder("(");
    for (int i = 0; i < entryNum; i++) {
      final IGStaticValue entryValue = igStaticValue.getValue(i, location);
      builder.append(
          String.format("(%s:%s)", Integer.toString(i + 1), getNodeValue(entryValue).toString()));
    }
    builder.append(")");
    return builder.toString();
  }

  @Override
  public void visitRange(final IGRange igRange) {
    /* Empty. */
  }

  @Override
  public void visitUnary(final IGOperationUnary igUnary) {
    StandardOperation opId;
    switch (igUnary.getUnaryOp()) {
      case ABS:
        opId = StandardOperation.ABS;
        break;
      case NEG:
        opId = StandardOperation.MINUS;
        break;
      case BUF:
        opId = StandardOperation.ADD;
        break;
      case NOT:
        opId = StandardOperation.NOT;
        break;
      default:
        throw new UnsupportedOperationException(igUnary.toHRString());
    }
    final Node operandElem = getNode(igUnary.getA());
    this.result = new IgOperationResult(new NodeOperation(opId, operandElem));
  }

  @Override
  public void visitTypeQualification(final IGOperationTypeQualification igQualification) {
    throw new UnsupportedOperationException(igQualification.toHRString());
  }

  @Override
  public void visitTypeConversion(final IGOperationTypeConversion igConversion) {

    /* The conversion object contains exactly one operand. */
    visitOperation(igConversion.getOperand(0));
  }

  @Override
  public void visitRecordField(final IGOperationRecordField igRecordField) {
    throw new UnsupportedOperationException(igRecordField.toHRString());
  }

  @Override
  public void visitRecordAggregate(final IGOperationRecordAggregate igAggregate) {
    throw new UnsupportedOperationException(igAggregate.toHRString());
  }

  @Override
  public void visitOpRange(final IGOperationRange igOperationRange) {

    final Node operand = getNode(igOperationRange.getOperand());

    InvariantChecks.checkNotNull(operand);

    final IGOperation range = igOperationRange.getRange();

    if (range instanceof IGRange) {
      final IGOperation isAscending = ((IGRange) range).getAscending();

      final Node left = getNode(((IGRange) range).getLeft());
      final Node right = getNode(((IGRange) range).getRight());

      final boolean isTrue = ((IGStaticValue)isAscending).isTrue();
      final Node firstIdx = isTrue ? right : left;
      final Node secondIdx = isTrue ? left : right;

      /* Check bvextract argument shift */
      final VariableDescriptor descriptor = (VariableDescriptor) operand.getUserData();
      checkBitVectorMinIndex(
          igOperationRange.getOperand(),
          operand,
          descriptor);

      final boolean shifted = descriptor != null && descriptor.hasMetaInfo(SHIFT);
      final Node shiftNode =
          shifted ? NodeValue.newInteger(Integer.parseInt(descriptor.getMetaInfo(SHIFT))) : null;

      final Node first =
          shifted ? new NodeOperation(StandardOperation.SUB, firstIdx, shiftNode) : firstIdx;
      final Node second =
          shifted ? new NodeOperation(StandardOperation.SUB, secondIdx, shiftNode) : secondIdx;

      final Node expression =
          TransformerUtils.standardize(
              new NodeOperation(StandardOperation.BVEXTRACT, first, second, operand));

      this.result = new IgOperationResult(expression);
    } else {
      throw new UnsupportedOperationException(igOperationRange.toHRString());
    }
  }

  @Override
  public void visitPhi(final IGOperationPhi igPhi) {
    final Node condition = getNode(igPhi.getOperand(0));
    final Node firstVariant = getNode(igPhi.getOperand(1));
    final Node secondVariant = getNode(igPhi.getOperand(2));

    final Node expression =
        new NodeOperation(StandardOperation.ITE, condition, firstVariant, secondVariant);

    this.result = new IgOperationResult(expression);
  }

  @Override
  public void visitObject(final IGOperationObject igOperationObject) {

    Node node;
    final IGObject.IGObjectCat category = igOperationObject.getObject().getCat();
    final IGType opType = igOperationObject.getType();

    if (isVariable(category)) {

      /* Object keeps a variable. */

      final DataType dataType = getBuilder().getDataType(opType);
      final VariableType variableType = VariableBuilder.getVariableType(igOperationObject);

      node = new NodeVariable(igOperationObject.getObject().getId(), dataType);

      /* Set full type meta-info. */
      final VariableDescriptor data = new VariableDescriptor(variableType);
      final String fullType = VariableBuilder.getFullType(igOperationObject.getObject());
      data.addMetaInfo(MetaInfoType.HDL_TYPE_STR, fullType);
      node.setUserData(data);
      getBuilder().fillVariableDescriptor(data, igOperationObject);

      checkBitVectorMinIndex(igOperationObject, node, data);

    } else if (isConstant(category)) {

      /* Object keeps a constant. */

      final IGOperation initValue = igOperationObject.getObject().getInitialValue();
      node = getNode(initValue);
    } else {
      throw new IllegalArgumentException(igOperationObject.toHRString());
    }
    this.result = new IgOperationResult(node);
  }

  public void checkBitVectorMinIndex(
      final IGObject varObject,
      final Node variable,
      final VariableDescriptor varDescriptor) {

    if (variable.isType(DataTypeId.BIT_VECTOR)) {
      final NodeValue minIndex = getMinIndex(varObject);

      /* Minimal index of vector's bit is not 0. */
      if (minIndex != null && !minIndex.equals(NodeValue.newInteger(0))) {

        varDescriptor.addMetaInfo(SHIFT, minIndex.toString());
      }
    }
  }

  public void checkBitVectorMinIndex(
      final IGOperation varObject,
      final Node variable,
      final VariableDescriptor variableDescriptor) {

    if (variable.isType(DataTypeId.BIT_VECTOR)) {
      final NodeValue minIndex = getMinIndex(varObject);

      /* Minimal index of vector's bit is not 0. */
      if (minIndex != null && !minIndex.equals(NodeValue.newInteger(0))) {

        variableDescriptor.addMetaInfo(SHIFT, minIndex.toString());
      }
    }
  }

  private NodeValue getMinIndex(final IGObject object) {

    NodeValue minIndex = null;

    final IGType objectType = object.getType();
    final IGType indexType = objectType.getIndexType();
    if (indexType instanceof IGTypeStatic) {
      final IGStaticValue staticRange = ((IGTypeStatic) indexType).getStaticRange();
      final boolean ascending = staticRange.getAscending().isTrue();
      final IGStaticValue left = staticRange.getLeft();
      final IGStaticValue right = staticRange.getRight();
      minIndex = getNodeValue(ascending ? left : right);
    }
    return minIndex;
  }

  private NodeValue getMinIndex(final IGOperation object) {

    NodeValue minIndex = null;

    final IGType objectType = object.getType();
    final IGType indexType = objectType.getIndexType();
    if (indexType instanceof IGTypeStatic) {
      final IGStaticValue staticRange = ((IGTypeStatic) indexType).getStaticRange();
      final boolean ascending = staticRange.getAscending().isTrue();
      final IGStaticValue left = staticRange.getLeft();
      final IGStaticValue right = staticRange.getRight();
      minIndex = getNodeValue(ascending ? left : right);
    }
    return minIndex;
  }

  private static boolean isConstant(final IGObject.IGObjectCat category) {
    return category.equals(IGObject.IGObjectCat.CONSTANT);
  }

  private static boolean isVariable(final IGObject.IGObjectCat cat) {
    return cat.equals(IGObject.IGObjectCat.SIGNAL) || cat.equals(IGObject.IGObjectCat.VARIABLE);
  }

  @Override
  public void visitLiteral(final IGOperationLiteral igLiteral) {

    NodeValue value;
    try {

      /* Compute static value. */
      final IGStaticValue opStaticValue = igLiteral.computeStaticValue(null);

      switch (opStaticValue.getStaticType().getCat()) {
        case INTEGER:
          value = new NodeValue(Data.newInteger(opStaticValue.getInt()));
          break;
        case REAL:
          value =  new NodeValue(Data.newReal(opStaticValue.getReal().doubleValue()));
          break;
        case ARRAY:

          // TODO: change when 'Z' values will be supported
          final String strVector = substituteUnsupported(opStaticValue.toString());

          value =
              new NodeValue(
                  Data.newBitVector(
                      strVector,
                      Radix.BIN.value(),
                      opStaticValue.toHRString().length()));
          break;
        default:
          throw new UnsupportedOperationException(igLiteral.toHRString());
      }
    } catch (ZamiaException e) {
      throw new RetrascopeRuntimeException(e);
    }
    this.result = new IgOperationResult(value);
  }

  @Override
  public void visitIndex(final IGOperationIndex igIndex) {

    Node index = getNode(igIndex.getIndex());
    final Node operand = getNode(igIndex.getOperand());

    Node resultNode;
    if (operand.isType(DataTypeId.BIT_VECTOR)) {

      /* For parametric operations their parameters should go earlier than operands. */

      final VariableDescriptor descriptor = (VariableDescriptor) operand.getUserData();
      checkBitVectorMinIndex(
          igIndex.getOperand(),
          operand,
          descriptor);

      index = descriptor.hasMetaInfo(SHIFT)
          ? new NodeOperation(
              StandardOperation.SUB,
              index,
              NodeValue.newInteger(Integer.parseInt(descriptor.getMetaInfo(SHIFT))))
          : index;


      resultNode =
          TransformerUtils.standardize(
              new NodeOperation(StandardOperation.BVEXTRACT, index, index, operand));
    } else if (operand.isType(DataTypeId.MAP)) {
      resultNode = new NodeOperation(StandardOperation.SELECT, operand, index);
    } else {
      throw new IllegalArgumentException(operand.toString());
    }
    this.result = new IgOperationResult(resultNode);
  }

  @Override
  public void visitError(final IGOperationError igError) {
    throw new UnsupportedOperationException(igError.toHRString());
  }

  @Override
  public void visitDeref(final IGOperationDeref igDeref) {
    throw new UnsupportedOperationException(igDeref.toHRString());
  }

  @Override
  public void visitAttribute(final IGOperationAttribute igAttribute) {

    /* Object keeps event-based attribute. */
    final IGItem item = igAttribute.getItem();

    if (item != null) {
      final RangedVariable rangedVariable =
          getBuilder().getRangedVariable((IGOperation) item);

      switch (igAttribute.getAttrOp()) {
        case EVENT:

          final Event event = new Event(rangedVariable);
          this.result = new IgOperationResult(event);
          break;
        case LENGTH:
          final Node length =
              ((VariableDescriptor)rangedVariable.getVariable().getUserData()).getLength();
          this.result = new IgOperationResult(length);
          break;
        default:
          throw new UnsupportedOperationException(igAttribute.toHRString());
      }
    }
  }

  @Override
  public void visitAggregate(final IGOperationArrayAggregate igAggregate) {

    final IGType igType = igAggregate.getType();
    final IGType indexType = igType.getIndexType();
    final IGType elementType = igType.getElementType();
    final DataType keyType = getBuilder().getDataType(indexType);
    final DataType valueType = getBuilder().getDataType(elementType);
    final DataMap dataMap = new DataMap(keyType, valueType);

    /* Skip 'others' element if exists. */
    int offset = 0;
    if (igAggregate.getOperand(0) == null) {
      offset = 1;
    }

    for (int i = offset; i < igAggregate.getNumOperands(); i++) {

      final Node node = getNode(igAggregate.getOperand(i));
      final Node value = Transformer.reduce(node);

      if (value.getKind() == Node.Kind.VALUE) {
        dataMap.put(Data.newInteger(i - 1), ((NodeValue) value).getData());
      } else {
        throw new UnsupportedOperationException(value.toString());
      }
    }
    this.result = new IgOperationResult(new NodeValue(Data.newArray(dataMap)));
  }

  @Override
  public void visitAllocate(final IGOperationAllocate igAllocate) {
    throw new UnsupportedOperationException(igAllocate.toHRString());
  }

  @Override
  public void visitAlias(final IGOperationAlias igAlias) {
    throw new UnsupportedOperationException(igAlias.toHRString());
  }

  @Override
  public void visitSubprogram(final IGOperationInvokeSubprogram igSubProgram) {

    final String subProgramName = igSubProgram.getSub().getId();

    final String nonQuotedName = getNonQuotedString(subProgramName);

    if (IgSubprogramMapper.hasOperation(nonQuotedName)) {

      StandardOperation opCode = IgSubprogramMapper.getOperationCode(nonQuotedName);

      extractStandardOperation(igSubProgram, opCode);
    } else {

//      final IgStructureVisitor subProgramVisitor =
//          new IgStructureVisitor(
//              getStructureVisitor().getManager(),
//              getStructureVisitor().getRuntimeEnv(),
//              getStructureVisitor().getTopLevelModule());
//
//      final CfgProcess subProgramProcess = new CfgProcess();
//      final Source source = new Source();
//      final Sink sink = new Sink();
//      subProgramProcess.addChild(source);
//      source.addChild(sink);
//      subProgramProcess.addNode(source);
//      subProgramProcess.addNode(sink);
//
//      subProgramVisitor.setProcess(subProgramProcess);
//
//      /* Get formal-actual mappings. */
//      final Map<String, Node> bindings = getBindings(igSubProgram);
//
//      final IGSubProgram subProgram = igSubProgram.getSub();
//      final IGSequenceOfStatements subProgramCode = subProgram.getCode();
//      subProgramVisitor.visit(subProgramCode);
//      List<CfgStatement> subProgramNodes = subProgramVisitor.getNodes();
      final IGSubProgram subProgram = igSubProgram.getSub();
      final IGSequenceOfStatements code = subProgram.getCode();

      if (noLoopOrRecursionSubprogram(igSubProgram)) {
        this.result = new IgOperationResult(getSubProgramNode(code));
      } else {
        throw new UnsupportedOperationException(igSubProgram.toHRString());
      }
    }
  }

  private Node getSubProgramNode(IGSequenceOfStatements statements) {
    final IgSubprogramElaborator subprogramElaborator = new IgSubprogramElaborator(this);
    return subprogramElaborator.elaborate(statements);
  }

  private boolean noLoopOrRecursionSubprogram(final IGOperationInvokeSubprogram invokeSubprogram) {

    final IGSubProgram subProgram = invokeSubprogram.getSub();
    final String id = subProgram.getId();
    final IGSequenceOfStatements code = subProgram.getCode();

    return !hasLoopOrRecursion(id, code);
  }

  private static boolean hasLoopOrRecursion(
      final String programId,
      final IGSequenceOfStatements code) {

    for (int i = 0; i < code.getNumStatements(); i++) {
      IGSequentialStatement stmt = code.getStatement(i);

      if (stmt instanceof IGSequenceOfStatements
          && hasLoopOrRecursion(programId, (IGSequenceOfStatements) stmt)) {

        return true;
      }

      if (stmt instanceof IGSequentialLoop) {

        /* Loop */
        return true;
      }

      if (stmt.getNumChildren() > 0) {
        for (int j = 0; j < stmt.getNumChildren(); j++) {
          final IGItem child = stmt.getChild(j);
          if (child != null
              && child instanceof IGOperationInvokeSubprogram
              && ((IGOperationInvokeSubprogram) child).getSub().getId().equals(programId)) {

          /* Recursion */
            return true;
          }
        }
      }
    }
    return false;
  }

  private void extractStandardOperation(
      final IGOperationInvokeSubprogram igSubProgram,
      final StandardOperation opCode) {

    final int operandNum = igSubProgram.getNumOperands();
    final Node[] nodes = new Node[operandNum];
    for (int i = 0; i < operandNum; i++) {
      final Node node = getNode(igSubProgram.getOperand(i));
      if (node != null) {
        nodes[i] = node;
      }
    }

    final StandardOperation resultOpCode = modifyOpCode(opCode, nodes);

    this.result = new IgOperationResult(new NodeOperation(resultOpCode, nodes));
  }

  private static StandardOperation modifyOpCode(
      final StandardOperation opCode,
      final Node[] nodes) {

    StandardOperation resultOpCode = opCode;
      /* HACK: substitute binary operation identifiers to unary if needed. */
    if (nodes.length == 1) {
      if (opCode == StandardOperation.ADD) {
        resultOpCode = StandardOperation.PLUS;
      } else if (opCode == StandardOperation.SUB) {
        resultOpCode = StandardOperation.MINUS;
      }
    }

    /* Substitute operation to bit-vector-analogue if at least one operand is bit vector. */
    boolean hasBvOperand = containsType(DataTypeId.BIT_VECTOR, nodes);

    if (hasBvOperand && BitVectorOpCodeMapper.hasOperation(resultOpCode)) {
      resultOpCode = BitVectorOpCodeMapper.getOperation(resultOpCode);
    }
    return resultOpCode;
  }

//  private Map<String, Node> getBindings(IGOperationInvokeSubprogram subProgram) {
//    final Map<String, Node> bindings = new LinkedHashMap<>();
//
//    final int maxNum = subProgram.getNumMappings();
//
//    for (int i = 0; i < maxNum; i++) {
//      final IGMapping mapping = subProgram.getMapping(i);
//
//      /* Extract the binding's target name. */
//      final IGOperation formal = mapping.getFormal();
//
//      final String varName = getBindFormalName(formal);
//
//      /* Extract the binding's value expression. */
//      final Node expression = getNode(mapping.getActual());
//
//      bindings.put(varName, expression);
//    }
//    return bindings;
//  }

  private static boolean containsType(final DataTypeId typeId, final Node ... nodes) {
    boolean hasBvOperand = false;

    for (final Node node : nodes) {
      if (node.isType(typeId)) {
        hasBvOperand = true;
        break;
      }
    }
    return hasBvOperand;
  }

  private static String getNonQuotedString(final String opName) {
    final char dQuote = '"';
    final int dqIndex = opName.indexOf(dQuote);
    return dqIndex < 0 ? opName : opName.substring(dqIndex + 1, opName.lastIndexOf(dQuote));
  }

  @Override
  public void visitBinary(final IGOperationBinary igBinary) {
    final Enum<?> type = IgBinaryMapper.getOperationCode(igBinary);

    final Node left = getNode(igBinary.getA());
    final Node right = getNode(igBinary.getB());

    this.result = new IgOperationResult(new NodeOperation(type, left, right));
  }

  /**
   * Returns syntax node that is extracted from the specified IG object.
   *
   * @param object IG object to be processed.
   * @return Syntax node that is generated.
   */
  public Node getNode(final IGOperation object) {
    visitOperation(object);
    return getResult().getNode();
  }

//  /**
//   * Returns syntax node that is extracted from the specified IG object.
//   *
//   * @param operation IG object
//   * @return syntax node
//   */
//  public static Node getStaticNode(final IGOperation operation) {
//    final IgOperationVisitor visitor = getIgOperationCfgVisitor(operation);
//
//    return visitor.getResult().getNode();
//  }

//  private static IgOperationVisitor getIgOperationCfgVisitor(final IGOperation object) {
//
//    final IgOperationVisitor visitor = new IgOperationVisitor();
//
//    visitor.visitOperation(object);
//
//    return visitor;
//  }

  /**
   * Returns data object that is extracted from th specified IG object.
   *
   * @param object IG object
   * @return data object
   */
  public Object getData(final IGOperation object) {
    visitOperation(object);
    return getResult().getData();
  }

  private String substituteUnsupported(final String vectorValue) {

    /* Substitutes 'Z'-values and '-'-values by zeros. */
    final char[] array = vectorValue.toCharArray();
    for (int i = 0; i < vectorValue.length(); i++) {
      if (array[i] == '-' || array[i] == 'Z') {
        array[i] = '0';
      }
    }
    return new String(array);
  }

  /**
   * Returns name of the specified 'formal' parameter of binding.
   *
   * @param formal Formal parameter of binding.
   * @return Name of the specified parameter.
   */
  public static String getBindFormalName(final IGOperation formal) {
    return formal instanceof IGOperationObject
      ? ((IGOperationObject)formal).getObject().getId()
      : formal.getId();
  }
}
