/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.SourceLocation;
import org.zamia.ZamiaException;
import org.zamia.instgraph.IGConcurrentStatement;
import org.zamia.instgraph.IGContainer;
import org.zamia.instgraph.IGContainerItem;
import org.zamia.instgraph.IGDefaultSequentialStatementVisitor;
import org.zamia.instgraph.IGInstantiation;
import org.zamia.instgraph.IGItem;
import org.zamia.instgraph.IGManager;
import org.zamia.instgraph.IGMapping;
import org.zamia.instgraph.IGModule;
import org.zamia.instgraph.IGObject;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGOperationAttribute;
import org.zamia.instgraph.IGProcess;
import org.zamia.instgraph.IGRange;
import org.zamia.instgraph.IGSequenceOfStatements;
import org.zamia.instgraph.IGSequentialAssert;
import org.zamia.instgraph.IGSequentialAssignment;
import org.zamia.instgraph.IGSequentialExit;
import org.zamia.instgraph.IGSequentialIf;
import org.zamia.instgraph.IGSequentialLoop;
import org.zamia.instgraph.IGSequentialNext;
import org.zamia.instgraph.IGSequentialProcedureCall;
import org.zamia.instgraph.IGSequentialReport;
import org.zamia.instgraph.IGSequentialReturn;
import org.zamia.instgraph.IGSequentialStatement;
import org.zamia.instgraph.IGSequentialWait;
import org.zamia.instgraph.IGStaticValue;
import org.zamia.instgraph.IGStructure;
import org.zamia.instgraph.IGType;
import org.zamia.instgraph.IGTypeStatic;
import org.zamia.instgraph.interpreter.IGInterpreterRuntimeEnv;
import org.zamia.util.PathName;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableDeclaration;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.CfgAssertStatement;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgLoopStatement;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgWaitStatement;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * IG structure objects parser.
 *
 * <p>Component of VHDL parser that transforms IG (Instantiation Graph)
 * structure objects representing processes, modules, instances to
 * control flow graph {@link ru.ispras.retrascope.model.cfg.CfgModel} model representation.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class IgStructureVisitor extends IGDefaultSequentialStatementVisitor {

  /* The IG representation manager. */
  private final IGManager manager;

  /* The IG interpreter runtime environment. */
  private final IGInterpreterRuntimeEnv runtimeEnv;

  /* The parent top-level module. */
  private final CfgModule topLevel;

  /* The list of current structure nodes. */
  private final List<CfgStatement> nodes;

  /* The current process. */
  private CfgProcess process;

  private VariableBuilder builder;

  /* Constants that keep severity string values. */
  private static final String SEVERITY_NOTE = "NOTE";
  private static final String SEVERITY_WARN = "WARNING";
  private static final String SEVERITY_ERROR = "ERROR";
  private static final String SEVERITY_FAIL = "FAILURE";

  /**
   * Creates object with the specified IG manager, IG runtime environment and the top-level module.
   *
   * @param manager IG representation manager.
   * @param runtimeEnv IG runtime environment.
   * @param module Top-level module.
   *
   * @throws IllegalArgumentException when {@code module} argument is {@code null}.
   */
  public IgStructureVisitor(
      final IGManager manager,
      final IGInterpreterRuntimeEnv runtimeEnv,
      final CfgModule module) {

    super(new IgOperationVisitor());
    InvariantChecks.checkNotNull(module);

    getOperationCfgVisitor().setParentVisitor(this);
    setBuilder(new VariableBuilder(this));

    this.manager = manager;
    this.runtimeEnv = runtimeEnv;
    this.topLevel = module;
    this.nodes = new LinkedList<>();
  }

  /**
   * Returns sub-visitor for IGOperation objects.
   *
   * @return IGOperation objects sub-visitor.
   */
  public IgOperationVisitor getOperationCfgVisitor() {
    return (IgOperationVisitor) this.fOperationVisitor;
  }

  /**
   * Returns top-level module.
   *
   * @return Top-level module that is processed at the current moment.
   */
  public CfgModule getTopLevelModule() {
    return this.topLevel;
  }

  /**
   * Returns control flow graph model nodes collection.
   *
   * @return Collection of nodes that are constructed at the current moment.
   */
  public List<CfgStatement> getNodes() {
    return this.nodes;
  }

  /**
   * Returns IG interpreter runtime environment.
   *
   * @return Runtime environment.
   */
  public IGInterpreterRuntimeEnv getRuntimeEnv() {
    return this.runtimeEnv;
  }

  /**
   * Returns IG manager.
   *
   * @return IG manager.
   */
  public IGManager getManager() {
    return this.manager;
  }

  private CfgProcess getProcess() {
    return this.process;
  }

  /**
   * Sets the specified process to be constructed by this visitor.
   *
   * @param process Process to be constructed.
   */
  public void setProcess(final CfgProcess process) {
    this.process = process;
  }

  /**
   * Returns inner variable and input\output signal builder.
   *
   * @return Variable builder.
   */
  public VariableBuilder getBuilder() {
    return this.builder;
  }

  private void setBuilder(final VariableBuilder builder) {
    this.builder = builder;
  }

  private void visitProcess(final IGProcess igProcess) {

    final CfgProcess newProcess = new CfgProcess();
    setProcess(newProcess);

    /* Get variable declarations. */
    final Map<NodeVariable, VariableDeclaration> declarations =
        getDeclarations(igProcess.getContainer());
    newProcess.declareVariables(declarations);

    /* Extract process sensitivity list events. */
    final EventList sensList = getBuilder().getProcessEventList(igProcess);
    newProcess.setSensitivityList(sensList);

    /* Elaborate internal nodes. */
    final IGSequenceOfStatements sos = igProcess.getSequenceOfStatements();
    visit(sos);

    final List<CfgStatement> processNodes = new LinkedList<>(this.nodes);
    newProcess.addAll(processNodes);

    /* Link process node with parent & child nodes. */
    getTopLevelModule().add(newProcess);

    this.nodes.clear();
    setProcess(null);
  }

  private void visitInstantiation(
      final IGInstantiation instantiation,
      final PathName pathName) {

    /* Extract the module name. */
    final IGModule igModule = getManager().findModule(instantiation.getSignature());
    final String moduleName = igModule.getDUUID().toCompactString();

    /* Extract module. */
    final IgStructureVisitor visitor =
        new IgStructureVisitor(getManager(), getRuntimeEnv(), new CfgModule(moduleName));

    visitor.visit(igModule.getStructure(), pathName);

    final CfgModule module = visitor.getTopLevelModule();

    /* Extract module's variables declarations. */
    final Map<NodeVariable, VariableDeclaration> declarations =
        getDeclarations(igModule.getContainer());
    module.declareVariables(declarations);

    /* Instantiate the extracted module. */
    final Map<String, Node> bindings = getBindings(instantiation);
    final String instanceName = instantiation.getLabel();
    module.instantiate(instanceName, bindings);

    getTopLevelModule().add(module);
  }

  private Map<NodeVariable, VariableDeclaration> getDeclarations(final IGContainer igContainer) {

    final Map<NodeVariable, VariableDeclaration> map = new LinkedHashMap<>();

    for (int i = 0; i < igContainer.getNumLocalItems(); i++) {

      final IGContainerItem item = igContainer.getLocalItem(i);

      /* Elaborate variables (not-a-constants). */
      if (item instanceof IGObject && !isConstant((IGObject) item)) {

        final NodeVariable variable = getBuilder().getObjectVariable((IGObject) item);
        final Node initValue = getInitialValue((IGObject) item);

        final Node invariant = getBuilder().getInvariant(variable, ((IGObject) item).getType());

        map.put(variable, new VariableDeclaration(initValue, invariant));
      }
    }

    return map;
  }

  private Node getInitialValue(final IGObject igObject) {
    Node initValue;

    final IGOperation igValue = igObject.getInitialValue();

    /* Try to calculate LEFT expression since it is initial for VHDL variable. */
    if (igValue == null) {
      try {

        final IGType type = igObject.getType();
        final IGTypeStatic staticType = type.computeStaticType(getRuntimeEnv(), null, null);
        final SourceLocation sourceLocation = igObject.computeSourceLocation();
        final IGStaticValue staticLeft =
            staticType.getStaticRange() == null ? null : staticType.getStaticLeft(sourceLocation);
        final IGStaticValue leftValue =
            staticLeft == null ? IGStaticValue.generateZ(staticType, sourceLocation) : staticLeft;
        initValue = getOperationNode(leftValue);

      } catch (ZamiaException e) {
        throw new RetrascopeRuntimeException(e);
      }
    } else {
      initValue = getOperationNode(igValue);
    }

    if (initValue == null) {
      throw new RetrascopeRuntimeException(
          String.format("Can't calculate initial value of: %s.", igObject.toString()));
    }
    return initValue;
  }

  private boolean isConstant(final IGObject item) {
    return item.getCat().equals(IGObject.IGObjectCat.CONSTANT);
  }

  private Map<String, Node> getBindings(final IGInstantiation igInst) {

    final Map<String, Node> bindings = new LinkedHashMap<>();

    final int maxNum = igInst.getNumMappings();

    for (int i = 0; i < maxNum; i++) {
      final IGMapping mapping = igInst.getMapping(i);

      final String varName = IgOperationVisitor.getBindFormalName(mapping.getFormal());
      final Node expression = getOperationNode(mapping.getActual());

      bindings.put(varName, expression);
    }
    return bindings;
  }

  @Override
  public void visit(
      final IGStructure igStructure,
      final PathName pathName) {

    final Map<NodeVariable, VariableDeclaration> declarations =
        getDeclarations(igStructure.getContainer());

    for (int i = 0; i < igStructure.getNumStatements(); i++) {
      final IGConcurrentStatement stmt = igStructure.getStatement(i);

      if (stmt instanceof IGInstantiation) {
        visitInstantiation((IGInstantiation) stmt, pathName);
      } else if (stmt instanceof IGProcess) {
        visitProcess((IGProcess) stmt);
      } else if (stmt instanceof IGStructure) {
        visit((IGStructure) stmt, pathName);
      }
    }
    getTopLevelModule().declareVariables(declarations);
  }

  @Override
  public void visit(final IGSequentialStatement statement) {
    super.visit(statement);
  }

  @Override
  public void visitWait(final IGSequentialWait igWait) {

    final IGOperation clause = igWait.getConditionClause();
    final Node condition = (clause == null) ? null : getOperationNode(clause);

    final EventList list = getBuilder().getEventList(igWait);
    final CfgWaitStatement waitNode =
        (condition == null) ? new CfgWaitStatement(list) : new CfgWaitStatement(list, condition);

    addToThis(waitNode);
  }

  @Override
  public void visitReturn(final IGSequentialReturn igReturn) {
    throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, igReturn.toHRString());
  }

  @Override
  public void visitReport(final IGSequentialReport igReport) {
    throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, igReport.toHRString());
  }

  @Override
  public void visitCall(final IGSequentialProcedureCall igCall) {
    throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, igCall.toHRString());
  }

  @Override
  public void visitNext(final IGSequentialNext igNext) {
    throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, igNext.toHRString());
  }

  @Override
  public void visitLoop(final IGSequentialLoop igLoop) {

    /* IG loop node children. */
    final IGRange loopRange = (IGRange)igLoop.getChild(0);
    final IGItem loopCounter = igLoop.getChild(1);
    final IGSequenceOfStatements loopBody = (IGSequenceOfStatements) igLoop.getChild(4);

    /* The range iteration variable. */
    final NodeVariable rangeVariable = getBuilder().getObjectVariable((IGObject) loopCounter);

    /* The loop condition. */
    final Node left = getOperationNode(loopRange.getLeft());
    final Node right = getOperationNode(loopRange.getRight());
    final Node ascending = getOperationNode(loopRange.getAscending());
    final boolean isAscending = ascending.equals(SolverUtils.TRUE);

    final Node range = createRange(rangeVariable, left, right, isAscending);

    /* Create the "i := 0" initialization block. */
    final Assignment init = new Assignment(rangeVariable, left);

    /* Create the incrementing "i++" basic block. */
    checkRangeParam(rangeVariable);

    final StandardOperation addOp = StandardOperation.ADD;
    final NodeValue one = NodeValue.newInteger(1);
    final NodeOperation incrementExpression = new NodeOperation(addOp, rangeVariable, one);
    final Assignment increment = new Assignment(rangeVariable, incrementExpression);

    final CfgLoopStatement loopStatement =
        new CfgLoopStatement(CfgLoopStatement.Type.FOR, range, init, increment);
    loopStatement.addAll(getSubGraph(loopBody));
    addToThis(loopStatement);
  }

  private void checkRangeParam(final NodeVariable rangeVariable) {
    if (!rangeVariable.isType(DataTypeId.LOGIC_INTEGER)) {
      throw RetrascopeRuntimeException.makeException(
          LogMessage.UNSUPPORTED_TYPE,
          rangeVariable.getDataType());
    }
  }

  @Override
  public void visitIf(final IGSequentialIf igIf) {

    final IGOperation ifCondition = igIf.getCond();
    final IGSequenceOfStatements elseSos = igIf.getElseSOS();
    final IGSequenceOfStatements thenSos = igIf.getThenSOS();

    /* Transform operands & detect event-dependent part of "x'event && x='0'" condition .*/
    final boolean hasEvents = hasEvents(ifCondition);

    CfgSwitchStatement switchStatement =
        hasEvents
        ? createEventCondition(ifCondition)
        : new CfgSwitchStatement(getOperationNode(ifCondition));

    final CfgCaseStatement trueCase = new CfgCaseStatement(SolverUtils.TRUE);
    switchStatement.add(trueCase);
    if (thenSos != null) {
      trueCase.addAll(getSubGraph(thenSos));
    }

    final CfgCaseStatement falseCase = new CfgCaseStatement(SolverUtils.FALSE);
    switchStatement.add(falseCase);
    if (elseSos != null) {
      falseCase.addAll(getSubGraph(elseSos));
    }

    addToThis(switchStatement);
  }

  private static boolean hasEvents(IGOperation ifCondition) {

    for (int i = 0; i < ifCondition.getNumOperands(); i++) {
      final IGOperation operand = ifCondition.getOperand(i);
      if (isEvent(operand)) {
        return true;
      }
    }

    return false;
  }

  private static boolean isEvent(IGOperation operand) {
    return operand instanceof IGOperationAttribute
        && ((IGOperationAttribute)operand).getAttrOp().equals(IGOperationAttribute.AttrOp.EVENT);
  }

  private CfgSwitchStatement createEventCondition(final IGOperation ifCondition) {

    RangedVariable eventVariable = null;
    NodeOperation expression = null;

    final int opNum = ifCondition.getNumOperands();

    /* Extract event variable. */
    if (opNum == 2) {

      /* Condition contains "x'event" expression. */
      for (int i = 0; i < opNum; i++) {

        final IGOperation operand = ifCondition.getOperand(i);

        if (operand instanceof IGOperationAttribute) {

          /* Attribute case. */
          final Event eventData = (Event)getOperationCfgVisitor().getData(operand);
          eventVariable = eventData.getRangedVariable();
        } else {

          /* Expression case. */
          expression = (NodeOperation)getOperationNode(operand);
        }
      }
    } else {
      throw RetrascopeRuntimeException.makeException(
          LogMessage.CANT_PARSE,
          ifCondition.toHRString());
    }

    /* Extract event type. */
    EventType type = EventType.ANY_EDGE;

    assert expression != null : LogMessage.CANT_PARSE + ifCondition.toHRString();

    final Enum<?> opId = expression.getOperationId();

    if (opId == StandardOperation.EQ) {
      final int opCount = expression.getOperandCount();
      for (int i = 0; i < opCount; i++) {
        final Node operand = expression.getOperand(i);
        final EventType operandEventType = getEventType(ifCondition, eventVariable, operand);
        type = operandEventType != null ? operandEventType : type;
      }
    } else {
      throw RetrascopeRuntimeException.makeException(
          LogMessage.CANT_PARSE,
          ifCondition.toHRString());
    }

    return new CfgSwitchStatement(new Event(eventVariable, type), expression);
  }

  private EventType getEventType(
      final IGOperation ifCondition,
      final RangedVariable eventVariable,
      final Node operand) {

    EventType type = null;

    if (operand.getKind() == Node.Kind.VALUE) {

      if (equalsToBitVector(operand, BitVector.FALSE)) {

        type = EventType.NEGATIVE_EDGE;

      } else if (equalsToBitVector(operand, BitVector.TRUE)) {

        type = EventType.POSITIVE_EDGE;
      }
    } else if (!equalParams(eventVariable, operand)) {
      throw RetrascopeRuntimeException.makeException(
          LogMessage.CANT_PARSE,
          ifCondition.toHRString());
    }

    return type;
  }

  private static boolean equalParams(
      final RangedVariable rangedVariable,
      final Node operand) {

    final NodeVariable variable = rangedVariable.getVariable();
    return operand.getKind() == Node.Kind.VARIABLE && operand.equals(variable);
  }

  private static boolean equalsToBitVector(
      final Node operand,
      final BitVector bitVector) {

    return ExprUtils.isType(DataTypeId.BIT_VECTOR, operand)
        && operand.equals(NodeValue.newBitVector(bitVector));
  }

  @Override
  public void visitExit(final IGSequentialExit igExit) {
    throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, igExit.toHRString());
  }

  @Override
  public void visitAssert(final IGSequentialAssert igAssert) {

    final IGOperation assertOp = igAssert.getOp();
    final Node condition = getOperationNode(assertOp);
    CfgAssertStatement.Severity severity = getSeverity(igAssert);
    final String report = igAssert.getReport().toHRString();

    final CfgAssertStatement assertStatement = new CfgAssertStatement(condition, severity, report);

    addToThis(assertStatement);
  }

  private CfgAssertStatement.Severity getSeverity(final IGSequentialAssert igAssert) {

    final String sevString = igAssert.getSeverity().toHRString();

    if (SEVERITY_NOTE.equalsIgnoreCase(sevString)) {

      return CfgAssertStatement.Severity.NOTE;
    } else if (SEVERITY_WARN.equalsIgnoreCase(sevString)) {

      return CfgAssertStatement.Severity.WARNING;
    } else if (SEVERITY_ERROR.equalsIgnoreCase(sevString)) {

      return CfgAssertStatement.Severity.ERROR;
    } else if (SEVERITY_FAIL.equalsIgnoreCase(sevString)) {

      return CfgAssertStatement.Severity.FAILURE;
    } else {
      Log.printWarn(LogMessage.CANT_PARSE + sevString + ".");
    }
    return null;
  }

  @Override
  public void visitAssignment(final IGSequentialAssignment igAssignment) {

    final IGOperation target = igAssignment.getTarget();
    final IGOperation value = igAssignment.getValue();
    final Node valueExpr = getOperationNode(value);
    CfgBlockStatement block = getBasicBlock(target, valueExpr);

    addToThis(block);
  }

  private Node getOperationNode(final IGOperation node) {
    return getOperationCfgVisitor().getNode(node);
  }

  private CfgBlockStatement getBasicBlock(
      final IGOperation target,
      final Node valueExpr) {

    final RangedVariable rangedVariable = getBuilder().getRangedVariable(target);

    final Assignment assignment = new Assignment(rangedVariable, valueExpr);

    /* If the variable is of signal type, then basic block is concurrent. */
    final boolean isConcurrent = isSignal(rangedVariable);

    return new CfgBlockStatement(assignment, isConcurrent);
  }

  private boolean isSignal(final RangedVariable rangedVariable) {

    final NodeVariable variable = rangedVariable.getVariable();
    final Object userData = variable.getUserData();

    return VariableType.isSignal(((VariableDescriptor) userData).getVariableType());
  }

  private List<CfgStatement> getSubGraph(final IGSequenceOfStatements stmt) {
    final IgStructureVisitor visitor =
        new IgStructureVisitor(this.manager, this.runtimeEnv, getTopLevelModule());
    visitor.setProcess(getProcess());
    visitor.visit(stmt);
    return visitor.getNodes();
  }

  private Node createRange(
      final NodeVariable var,
      final Node left,
      final Node right,
      final boolean isAscending) {

    /* Select operation identifiers depending on data types. */
    final boolean areBitVector = ExprUtils.isType(DataTypeId.BIT_VECTOR, left, right);

    final StandardOperation greaterId =
        areBitVector ? StandardOperation.BVUGE : StandardOperation.GREATEREQ;
    final StandardOperation lessId =
        areBitVector ? StandardOperation.BVULE : StandardOperation.LESSEQ;

    Node expr;
    if (isAscending) {

      /* The result is: left <= var <= right. */
      expr =
          ExprUtils.getConjunction(
            new NodeOperation(greaterId, var, left),
            new NodeOperation(lessId, var, right));
    } else {

      /* The result is: left >= var >= right. */
      expr =
          ExprUtils.getConjunction(
              new NodeOperation(lessId, var, left),
              new NodeOperation(greaterId, var, right));
    }

    return expr;
  }

  private void addToThis(final CfgStatement statement) {
    this.nodes.add(statement);
  }
}
