/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGSequenceOfStatements;
import org.zamia.instgraph.IGSequentialAssignment;
import org.zamia.instgraph.IGSequentialIf;
import org.zamia.instgraph.IGSequentialStatement;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.RangedVariableUtils;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.util.LogMessage;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * VHDL parser component that elaborates subprograms (functions).
 *
 * @see VhdlParser
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class IgSubprogramElaborator {

  private final Map<NodeVariable, Node> assignments;
  private Node subProgramExpr;
  private final IgOperationVisitor operationVisitor;

  /**
   * Constructs component with the specified IG operation visitor.
   *
   * @param visitor IG operation objects visitor.
   */
  public IgSubprogramElaborator(IgOperationVisitor visitor) {
    this.operationVisitor = visitor;
    this.assignments = new LinkedHashMap<>();
  }

  /**
   * Elaborates the specified sequence of IG statements resulting to expression node.
   *
   * @param statements IG statements that are to be elaborated.
   * @return Expression node as a result of the specified object's elaboration.
   */
  public Node elaborate(final IGSequenceOfStatements statements) {

    final int stmtNum = statements.getNumStatements();

    for (int i = 0; i < stmtNum; i++) {
      this.subProgramExpr = elaborate(statements.getStatement(i));
    }
    return this.subProgramExpr;
  }

  private Node elaborate(final IGSequentialStatement statement) {
    if (statement instanceof IGSequenceOfStatements) {
      return elaborate(statement);
    } else if (statement instanceof IGSequentialAssignment) {
      return elaborateAssignment((IGSequentialAssignment)statement);
    } else if (statement instanceof IGSequentialIf) {
      return elaborateIf((IGSequentialIf)statement);
    } else {
      throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, statement.toHRString());
    }
  }

  private Node elaborateAssignment(final IGSequentialAssignment igAssignment) {

    final IGOperation target = igAssignment.getTarget();
    final VariableType targetType = VariableBuilder.getVariableType(target);

    if (targetType != null) {

      final IGOperation igAssignmentValue = igAssignment.getValue();
      final Node valueExpr = getOperationNode(igAssignmentValue);
      final RangedVariable rangedVariable = getBuilder().getRangedVariable(target);
      final Node value =
          rangedVariable.isRanged()
              ? RangedVariableUtils.transformNodeWithRanged(rangedVariable, valueExpr) : valueExpr;

      this.assignments.put(rangedVariable.getVariable(), applyPreviousAssignments(value));
    } else {
      throw RetrascopeRuntimeException.makeException(
          LogMessage.CANT_PARSE,
          igAssignment.toHRString());
    }

    return this.subProgramExpr;
  }

  private Node applyPreviousAssignments(final Node expr) {
    Node result = expr;

    for (final Map.Entry<NodeVariable, Node> assignment : assignments.entrySet()) {
      final String variableName = assignment.getKey().getName();
      result = Transformer.substitute(result, variableName, assignment.getValue());
    }

    return result;
  }

  private Node elaborateIf(final IGSequentialIf statement) {
    final Node condition = getCondition(statement);
    final IGSequenceOfStatements thenStatements = statement.getThenSOS();
    final IGSequenceOfStatements elseStatements = statement.getElseSOS();

    return new NodeOperation(
        StandardOperation.ITE, condition, elaborate(thenStatements), elaborate(elseStatements));
  }

  private Node getCondition(final IGSequentialIf statement) {
    final IGOperation igCondition = statement.getCond();

    return getOperationNode(igCondition);
  }

  private IgOperationVisitor getVisitor() {
    return this.operationVisitor;
  }

  private Node getOperationNode(final IGOperation node) {
    return getVisitor().getNode(node);
  }

  private VariableBuilder getBuilder() {
    return getVisitor().getStructureVisitor().getBuilder();
  }
}
