/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.apache.commons.cli.ParseException;
import org.apache.log4j.Level;
import org.zamia.BuildPath;
import org.zamia.SourceFile;
import org.zamia.ZamiaException;
import org.zamia.ZamiaLogger;
import org.zamia.ZamiaProject;
import org.zamia.ZamiaProjectBuilder;
import org.zamia.instgraph.IGManager;
import org.zamia.instgraph.IGModule;
import org.zamia.instgraph.interpreter.IGInterpreterRuntimeEnv;
import org.zamia.zdb.ZDBException;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.FileCreator;
import ru.ispras.retrascope.basis.FileNames;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.MetaInfoValue;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.parser.basis.HdlParser;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;
import ru.ispras.retrascope.util.LogMessage;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.util.Map;

/**
 * Parser for VHDL hardware models.
 *
 * <p>Uses ZamiaCAD toolkit for making an intermediate representation (Instantiation Graph, IG)
 * before control flow graph {@link ru.ispras.retrascope.model.cfg.CfgModel} building.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class VhdlParser extends HdlParser {

  /**
   * The engine identifier.
   */
  public static final String ENGINE_ID = "vhdl-parser";

  /**
   * The engine's input entity type.
   */
  public static final EntityType INPUT_TYPE = EntityType.get("vhd");

  /**
   * The VHDL top-level module name.
   */
  public static final Parameter TOP_LEVEL =
      new Parameter("toplevel", true, "Top-level VHDL modules names");

  /**
   * Default constructor.
   */
  public VhdlParser() {
    super(VhdlParser.ENGINE_ID);

    addInputType(VhdlParser.INPUT_TYPE);
    setOutputType(CfgModel.TYPE);

    addParameter(TOP_LEVEL);
  }

  @Override
  public Entity parse(final Map<EntityType, Entity> inputs) {

    final Arguments arguments = (Arguments)inputs.get(Arguments.TYPE);
    final FileNames fileNames = (FileNames)inputs.get(VhdlParser.INPUT_TYPE);

    final CfgModel cfgModel = new CfgModel();

    try {

      /* Get top level modules names from command line. */
      final Arguments engineArgs = parseCommandLine(arguments.getCommandLine());

      final String[] topLevels = getTopLevelNames(engineArgs, fileNames);

      if (topLevels.length > 0) {
        for (int i = 0; i < topLevels.length; i++) {

          final String fileName = fileNames.getList().get(i);

          if (new File(fileName).exists()) {

            /*
            * Create BuildPath.txt
            */

            /* Get full path to VHDL file directory. */
            final String path = getSourceDirPath(fileName);

            final File buildFile = FileCreator.newFile(path + File.separator + "BuildPath.txt");

            final BufferedWriter writer = new BufferedWriter(
                new OutputStreamWriter(
                    new FileOutputStream(buildFile), Charset.defaultCharset()));

            writer.flush();
            writer.write("local WORK " + "\"" + path + "\"\n");
            writer.write("toplevel " + topLevels[i].toUpperCase() + "\n");
            writer.close();

            /* Create Zamia IG representation of the source code. */

            ZamiaLogger.setup(Level.OFF);
            final ZamiaProject project = new ZamiaProject("", path, new SourceFile(buildFile));

            final ZamiaProjectBuilder builder = project.getBuilder();
            builder.build(true, true, null);

            /* Parse Zamia IG, create CfgModel. */

            final IGManager igm = project.getIGM();

            final BuildPath buildPath = project.getBuildPath();

            /* Elaborate top level modules. */
            final int topLevelNum = buildPath.getNumToplevels();

            if (topLevelNum > 0) {
              for (int j = 0; j < topLevelNum; j++) {

                final IGModule topLevel = igm.findModule(buildPath.getToplevel(j));

                if (topLevel == null) {
                  Log.printError(LogMessage.CANT_FIND + topLevels[i] + " module");
                  return null;
                }

                final CfgModule module = new CfgModule(topLevel.getDUUID().getId());
                module.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);
                module.addMetaInfo(MetaInfoType.FILE_NAME, fileName);

                final IGInterpreterRuntimeEnv runtimeEnv =
                    new IGInterpreterRuntimeEnv(null, project);

                final IgStructureVisitor structureVisitor =
                    new IgStructureVisitor(igm, runtimeEnv, module);

                topLevel.accept(structureVisitor, Integer.MAX_VALUE);

                cfgModel.addModule(structureVisitor.getTopLevelModule());
              }
            }
            project.shutdown();
          } else {
            this.getLogger().log(LogLevel.ERROR, LogMessage.CANT_FIND + fileName);
          }
        }
      }
    } catch (ParseException | ZamiaException | IOException | ZDBException e) {
      throw new RetrascopeRuntimeException(e);
    }
    return cfgModel;
  }

  private String[] getTopLevelNames(final Arguments engineArgs, final FileNames fileNames) {

    if (engineArgs.contains(VhdlParser.TOP_LEVEL)) {
      final String names = engineArgs.getValue(TOP_LEVEL);

      return names.split(File.pathSeparator);
    } else if (fileNames.getList().size() == 1) {
      final String onlyFile = fileNames.getList().get(0);
      return new String[] {
          onlyFile.substring(
              onlyFile.lastIndexOf(File.separator) + 1,
              onlyFile.lastIndexOf(VhdlParser.INPUT_TYPE.getId()) - 1)};
    } else {
      return new String[]{};
    }
  }

  private String getSourceDirPath(final String fileName) {

    String sourceDirPath;

    final int separatorIndex = fileName.lastIndexOf(File.separatorChar);

    if (separatorIndex >= 0) {
      sourceDirPath = fileName.substring(0, separatorIndex);
    } else {

      /* The source file is in the current directory. */
      sourceDirPath = System.getProperty("user.dir");
    }
    return sourceDirPath;
  }
}
