/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.result.assertion;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Entity that contains assertions that are extracted from models.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class AssertionEntity extends Entity {

  public static final EntityType TYPE = EntityType.get("assertion");

  private final Map<String, AssertionContainer> containerMap;

  /**
   * Default constructor.
   */
  public AssertionEntity() {
    super(TYPE);

    this.containerMap = new LinkedHashMap<>();
  }

  /**
   * Returns assertion containers mapping.
   * <p>Container names are stored in mapping keys.</p>
   * @return Mapping of names to assertion containers.
   */
  public Map<String, AssertionContainer> getContainerMap() {
    return Collections.unmodifiableMap(this.containerMap);
  }

  /**
   * Adds non-null assertion container.
   *
   * @param container Assertion container to be added.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addContainer(final AssertionContainer container) {
    InvariantChecks.checkNotNull(container);

    this.containerMap.put(container.getName(), container);
  }

  @Override
  public void add(Entity entity) {
    if (entity instanceof AssertionEntity) {
      this.containerMap.putAll(((AssertionEntity) entity).getContainerMap());
    } else {
      throw RetrascopeRuntimeException.makeException(LogMessage.WRONG_TYPE, entity.getEntityType());
    }
  }
}
