/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.result.assertion;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;

/**
 * Named group of independent assertions.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class AssertionGroup extends AssertionVariableContainer {

  private final String name;

  private final AssertionContainer parentContainer;

  private final List<Node> assertions;

  /**
   * Constructs object with the specified name and parent container.
   *
   * @param name Name of this object.
   * @param parent Parent container for this object.
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public AssertionGroup(final String name, final AssertionContainer parent) {
    super();
    InvariantChecks.checkNotNull(name);
    InvariantChecks.checkNotNull(parent);

    this.name = name;
    this.parentContainer = parent;
    this.assertions = new LinkedList<>();
  }

  /**
   * Returns object's name.
   *
   * @return Name of this object.
   */
  public String getName() {
    return this.name;
  }

  /**
   * Returns parent container.
   *
   * @return Parent container for this object.
   */
  public AssertionContainer getParent() {
    return this.parentContainer;
  }

  /**
   * Adds the specified expression that uses variables
   * that are already declared at the parent container.
   *
   * @param assertion Assertion to be added to this object.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeRuntimeException when the specified expression
   *         contains at least one variable that is undeclared
   *         at the parent container.
   */
  public void addAssertion(final Node assertion) {
    InvariantChecks.checkNotNull(assertion);

    Collection<NodeVariable> variables = ExprUtils.getVariables(assertion);

    AssertionVariableContainer container = this;

    do {
      variables =
          CollectionUtils.complementSets(
              new LinkedHashSet<>(variables),
              container.getDeclaredVariables());
      if (container instanceof AssertionGroup) {
        container = ((AssertionGroup)container).getParent();
      } else {
        container = ((AssertionContainer)container).getParent();
      }
    } while (container != null && !variables.isEmpty());

    if (!variables.isEmpty()) {
      throw new RetrascopeRuntimeException(
          String.format(
              "Cannot add %s assertion: %s variables undeclared.",
              assertion.toString(),
              variables.toString()));
    }

    this.assertions.add(assertion);
  }

  /**
   * Returns a collection of assertions.
   *
   * @return Assertions of this object.
   */
  public List<Node> getAssertions() {
    return Collections.unmodifiableList(this.assertions);
  }
}
