/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.result.assertion;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Container for variables that are used in assertions.
 * <p>
 * A simplified version of AssertionVariableContainer class.
 * </p>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class AssertionVariableContainer {

  private final Set<NodeVariable> declared;

  /* Reset-like variable (should be declared here). */
  private NodeVariable resetVariable;

  public AssertionVariableContainer() {
    this.declared = new LinkedHashSet<>();
  }

  /**
   * Checks whether this object contains the specified variable as declared.
   *
   * @param variable Variable that is checked if declared at this object.
   * @return {@code true} if this object contains the specified variable as declared,
   *         {@code false} otherwise.
   */
  public boolean hasDeclared(final NodeVariable variable) {
    return getDeclaredVariables().contains(variable);
  }

  /**
   * Add all the variables of the specified collection to internal declared variables.
   *
   * @param variables Variables to be declared.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void declareVariables(final Collection<NodeVariable> variables) {
    InvariantChecks.checkNotNull(variables);

    for (final NodeVariable variable : variables) {
      declareVariable(variable);
    }
  }

  /**
   * Adds the specified variable to the internal collection of declared variables.
   *
   * @param variable Variable to be declared.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void declareVariable(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    this.declared.add(variable);
  }

  /**
   * Returns a collection of declared variables.
   *
   * @return Declared variables.
   */
  public Set<NodeVariable> getDeclaredVariables() {
    return Collections.unmodifiableSet(this.declared);
  }

  /**
   * Checks whether this object declares reset-like variable.
   *
   * @return {@code true} if this object declares reset-like variable, {@code false} otherwise.
   */
  public boolean hasResetVariable() {
    return resetVariable != null;
  }

  /**
   * Returns reset-like variable for this object.
   *
   * @return reset-like variable.
   */
  public NodeVariable getResetVariable() {
    return this.resetVariable;
  }

  /**
   * Sets the specified variable as reset-like one.
   * <p>Note that the specified variable should be declared at this object
   * before set as reset-like one.</p>
   * @param variable reset-like variable.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeRuntimeException when the specified variable has not been already
   *         declared at this object.
   */
  public void setResetVariable(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);
    checkIfDeclared(variable);

    this.resetVariable = variable;
  }

  private void checkIfDeclared(final NodeVariable variable) {
    if (!getDeclaredVariables().contains(variable)) {
      throw new RetrascopeRuntimeException(
          String.format("%s: '%s' should be declared.", LogMessage.CANT_FIND, variable.getName()));
    }
  }
}
