/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.result.assertion;

import ru.ispras.fortress.util.InvariantChecks;

/**
 * Assertion walker.
 * <p>Provides methods for {@link AssertionEntity} assertion entity and it's
 * sub-components walking.</p>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class AssertionWalker {

  private final AssertionVisitor visitor;

  /**
   * Constructs walker with the specified visitor object.
   *
   * @param visitor Entity visitor to be used through walking.
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public AssertionWalker(final AssertionVisitor visitor) {
    InvariantChecks.checkNotNull(visitor);

    this.visitor = visitor;
  }

  /**
   * Starts walking of the specified entity.
   *
   * @param entity Entity to be traversed.
   */
  public void start(final AssertionEntity entity) {
    InvariantChecks.checkNotNull(entity);

    this.visitor.onRootBegin();
    visitAssertionEntity(entity);
    this.visitor.onRootEnd();
  }

  private void visitAssertionEntity(final AssertionEntity entity) {

    this.visitor.onEntityBegin(entity);
    visitContainers(entity.getContainerMap().values());
    this.visitor.onEntityEnd(entity);
  }

  private void visitContainers(final Iterable<AssertionContainer> containers) {

    for (final AssertionContainer container : containers) {
      visitContainer(container);
    }
  }

  private void visitContainer(AssertionContainer container) {

    this.visitor.onContainerBegin(container);
    visitContainers(container.getContainerMap().values());
    visitGroups(container.getGroupMap().values());
    this.visitor.onContainerEnd(container);
  }

  private void visitGroups(Iterable<AssertionGroup> groups) {
    for (final AssertionGroup group : groups) {
      visitAssertionGroup(group);
    }
  }

  private void visitAssertionGroup(final AssertionGroup group) {
    this.visitor.onAssertionGroup(group);
  }
}
