/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.result.cfginterface;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;

/**
 * Signal part of control flow graph model interface.
 *
 * <p>Signal part of {@link ru.ispras.retrascope.result.cfginterface.CfgInterface} interface.
 * This part contains interface's input and output signals.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgSignalInterface {

  /* Input signals. */
  private final Collection<NodeVariable> inputs;

  /* Output signals. */
  private final Collection<NodeVariable> outputs;

  /**
   * Constructs object with the specified input and output signals.
   *
   * @param inputs Input signals of the interface.
   * @param outputs Output signals of the interface.
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  public CfgSignalInterface(
      final Collection<NodeVariable> inputs,
      final Collection<NodeVariable> outputs) {
    InvariantChecks.checkNotNull(inputs);
    InvariantChecks.checkNotNull(outputs);

    this.inputs = inputs;
    this.outputs = outputs;
  }

  /**
   * Returns input signals.
   *
   * @return Input signals of this interface.
   */
  public Collection<NodeVariable> getInputs() {
    return Collections.unmodifiableCollection(this.inputs);
  }

  /**
   * Returns output signals.
   *
   * @return Output signals of this interface.
   */
  public Collection<NodeVariable> getOutputs() {
    return Collections.unmodifiableCollection(this.outputs);
  }

  /**
   * Checks whether object contains input variable with the specified name.
   *
   * @param name Name of input variable.
   * @return {@code true} if object contains an input variable with the specified name,
   *         {@code false} otherwise.
   */
  public boolean containsInput(final String name) {
    for (final NodeVariable input : this.inputs) {
      if (input.getName().equals(name)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Checks whether interface object contains output variable with the specified name.
   *
   * @param name Name of output variable.
   * @return {@code true} if interface object contains an output variable with the specified name,
   *         {@code false} otherwise.
   */
  public boolean containsOutput(final String name) {
    for (final NodeVariable output : this.outputs) {
      if (output.getName().equals(name)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Returns signal variable that is in this interface object and has the specified name.
   *
   * @param signalName Name of signal variable to be returned.
   * @return variable that this interface object contains and that has the specified name
   *         or {@code null} if there is no such variable in this interface object.
   */
  public NodeVariable getSignal(final String signalName) {

    final Collection<NodeVariable> signals = new LinkedList<>();
    signals.addAll(this.inputs);
    signals.addAll(this.outputs);

    for (final NodeVariable signal : signals) {
      if (signal.getName().equals(signalName)) {
        return signal;
      }
    }
    return null;
  }
}
