/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.test;

import ru.ispras.retrascope.model.basis.EventList;

import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;

/**
 * The representation of a test sequence. A sequence is consisted of a list of input vectors as
 * well as list of attendant events (for example clock or reset signal change).
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see Vector
 * @see EventList
 */
@XmlType(propOrder = {"vectors"}, name = "sequence")
@XmlRootElement
public class Sequence {
  /**
   * The entity type ID of all {@code Sequence} instances.
   */
  public static final String ENTITY_TYPE_ID = "sequence".intern();

  private static final String LEFT_STRING_BOUND = "{";
  private static final String VECTOR_STRING_HEADER = "vector: ";
  private static final String VECTOR_STRING_DELIMITER = "; ";
  private static final String RIGHT_STRING_BOUND = "}";

  @XmlElement(name = "vector")
  private List<Vector> vectors;

  /**
   * Constructs an isEmpty sequence.
   */
  public Sequence() {
    this.vectors = new ArrayList<>();
  }

  /**
   * Constructs an isEmpty sequence with the specified initial capacity of an underlying collection
   * of vectors.
   *
   * @param capacity the initial capacity
   */
  public Sequence(int capacity) {
    this.vectors = new ArrayList<>(capacity);
  }

  /**
   * Constructs a sequence of the specified vectors with the specified events. An original order of
   * the vectors is preserved.
   *
   * @param eventList a list of the events
   * @param vectors a list of the vectors
   */
  public Sequence(EventList eventList, List<Vector> vectors) {
    this.vectors = new ArrayList<>(vectors);
  }

  /**
   * Adds the specified vector to the end of this sequence.
   *
   * @param vector the vector to be added
   */
  public void addVector(Vector vector) {
    this.vectors.add(vector);
  }

  /**
   * Adds the specified vectors to the end of this sequence. An original order is preserved.
   *
   * @param vectors a list of the vectors to be added
   */
  public void addVectors(List<Vector> vectors) {
    this.vectors.addAll(vectors);
  }

  /**
   * Returns a vector with the specified index from this sequence.
   *
   * @param vectorIndex the index of the requested vector
   *
   * @return the vector with the specified index
   *
   * @throws IndexOutOfBoundsException if the index is out of range
   */
  public Vector getVector(int vectorIndex) {
    return this.vectors.get(vectorIndex);
  }

  /**
   * Returns a list containing all vectors of this sequence.
   *
   * @return the list of the vectors representing this sequence
   */
  public List<Vector> getVectors() {
    return new ArrayList<>(this.vectors);
  }

  /**
   * Returns an amount of vectors in this sequence.
   *
   * @return the amount of vectors in this sequence
   */
  public int getLength() {
    return this.vectors.size();
  }

  @Override
  public int hashCode() {
    return 31 * super.hashCode() + this.vectors.hashCode();
  }

  /**
   * Compares the specified object with this sequence for equality. The object is considered as
   * equal to this sequence if and only if it is also a sequence and it contains the same vectors
   * in the same order and the same events as this sequence does, or more formally:
   * <ul>
   *   <li>it is also an instance of the {@code Sequence} class or its subclass;
   *   <li>its {@code EventList} is equal to this sequence's one;
   *   <li>it contains the same amount of vectors as this test does;
   *   <li>each its vector is equal to the corresponding vector in this sequence.
   * </ul>
   *
   * @param obj the object to be compared for equality with this sequence
   *
   * @return {@code true} if the specified object is equal to this sequence. Otherwise returns
   * {@code false}
   */
  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }

    if ((obj == null) || !(obj instanceof Sequence)) {
      return false;
    }

    Sequence sequenceToCompare = (Sequence) obj;

    return this.vectors.equals(sequenceToCompare.vectors);
  }

  @Override
  public String toString() {
    StringBuilder stringBuilder = new StringBuilder(Sequence.LEFT_STRING_BOUND);

    int preVectorsStringBuilderLength = stringBuilder.length();

    for (Vector vector : this.vectors) {
      if (stringBuilder.length() > preVectorsStringBuilderLength) {
        stringBuilder.append(Sequence.VECTOR_STRING_DELIMITER);
      }

      stringBuilder.append(Sequence.VECTOR_STRING_HEADER);
      stringBuilder.append(vector);
    }

    stringBuilder.append(Sequence.RIGHT_STRING_BOUND);

    return stringBuilder.toString();
  }
}
