/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.test;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.util.LogMessage;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * The ordered container for sequences.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see Sequence
 * @see Entity
 */
public final class SequenceModel extends Entity {
  /**
   * The entity type of all {@code SequenceModel} instances.
   */
  public static final EntityType TYPE = EntityType.get("sequence");

  private final List<Sequence> sequences;

  /**
   * Constructs an empty sequence model.
   */
  public SequenceModel() {
    super(SequenceModel.TYPE);

    this.sequences = new ArrayList<>();
  }

  /**
   * Adds the specified sequence to this sequence model.
   *
   * @param sequence the sequence to be added
   *
   * @throws IllegalArgumentException if {@code sequence} is {@code null}
   */
  public void addSequence(Sequence sequence) {
    InvariantChecks.checkNotNull(sequence);

    this.sequences.add(sequence);
  }

  /**
   * Returns sequences of this sequence model.
   *
   * @return the sequences
   */
  public List<Sequence> getSequences() {
    return Collections.unmodifiableList(this.sequences);
  }

  /**
   * Adds the specified sequence model to this one.
   *
   * @param entity the sequence model to be added
   *
   * @throws IllegalArgumentException if {@code entity} is either {@code null} or not an instance of
   *     {@code SequenceModel}
   */
  @Override
  public void add(Entity entity) {
    InvariantChecks.checkNotNull(entity);

    if (!(entity instanceof SequenceModel)) {
      throw new IllegalArgumentException(LogMessage.UNSUPPORTED_TYPE + entity.getClass().getName());
    }

    this.sequences.addAll(((SequenceModel) entity).sequences);
  }

  @Override
  public int hashCode() {
    return 31 * super.hashCode() + this.sequences.hashCode();
  }

  @Override
  public boolean equals(Object object) {
    if (this == object) {
      return true;
    }

    if (!super.equals(object) || !(object instanceof SequenceModel)) {
      return false;
    }

    SequenceModel otherSequenceModel = (SequenceModel) object;

    return this.sequences.equals(otherSequenceModel.sequences);
  }
}
