/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.test;

import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;

/**
 * The representation of a test. A test consisted of a list of input sequences.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see Sequence
 */
@XmlType(name = "test")
@XmlRootElement
public class Test {
  private static final String LEFT_STRING_BOUND = "{";
  private static final String SEQUENCE_STRING_HEADER = "sequence: ";
  private static final String SEQUENCE_STRING_DELIMITER = "; ";
  private static final String RIGHT_STRING_BOUND = "}";

  @XmlElement(name = "testSequence")
  private List<Sequence> sequences;

  /**
   * Constructs an empty test.
   */
  public Test() {
    this.sequences = new ArrayList<>();
  }

  /**
   * Constructs a test containing the specified list of sequences. An original order of the
   * sequences is preserved.
   *
   * @param sequences the list of sequences
   */
  public Test(List<Sequence> sequences) {
    this.sequences = new ArrayList<>(sequences);
  }

  /**
   * Adds the specified sequence to this test. The sequence is appended to the end of this test.
   *
   * @param sequence the sequence to be added.
   */
  public void addSequence(Sequence sequence) {
    this.sequences.add(sequence);
  }

  /**
   * Adds the specified sequences to this test. The sequences is appended to the end of this test,
   * an original order is preserved.
   *
   * @param sequences a list of the sequences to be added
   */
  public void addSequences(List<Sequence> sequences) {
    this.sequences.addAll(sequences);
  }

  /**
   * Returns a list of sequences which this test consists of.
   *
   * @return the list of sequences
   */
  public List<Sequence> getSequences() {
    return new ArrayList<>(this.sequences);
  }

  /**
   * Returns an amount of sequences which this test consists of.
   *
   * @return the amount of sequences in this test.
   */
  public int getSequencesAmount() {
    return this.sequences.size();
  }

  /**
   * Returns an amount of vectors which this test consists of.
   *
   * @return the amount of vectors in this test.
   */
  public int getVectorsAmount() {
    int vectorsAmount = 0;

    for (Sequence sequence: this.sequences) {
      vectorsAmount = vectorsAmount + sequence.getLength();
    }

    return vectorsAmount;
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = super.hashCode();
    result = prime * result + ((this.sequences == null) ? 0 : this.sequences.hashCode());

    return result;
  }

  /**
   * Compares the specified object with this test for equality. The object is considered as equal to
   * this test if and only if it is also a test and it contains the same sequences in the same order
   * as this test does, or more formally:
   * <ul>
   *    <li>it is an instance of the {@code Test} class or its subclass;
   *    <li>it contains the same amount of sequences as this test does;
   *    <li>each its sequence is equal to the corresponding sequence in this test.
   * </ul>
   *
   * @param obj the object to be compared for equality with this test
   *
   * @return {@code true} if the specified object is equal to this test. Otherwise returns
   * {@code false}
   */
  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }

    if ((obj == null) || !(obj instanceof Test)) {
      return false;
    }

    Test testToCompare = (Test) obj;

    return this.sequences.equals(testToCompare.sequences);
  }

  @Override
  public String toString() {
    StringBuilder stringBuilder = new StringBuilder(Test.LEFT_STRING_BOUND);

    int initialStringBuilderLength = stringBuilder.length();

    for (Sequence sequence : this.sequences) {
      if (stringBuilder.length() > initialStringBuilderLength) {
        stringBuilder.append(Test.SEQUENCE_STRING_DELIMITER);
      }

      stringBuilder.append(Test.SEQUENCE_STRING_HEADER);
      stringBuilder.append(sequence);
    }

    stringBuilder.append(Test.RIGHT_STRING_BOUND);

    return stringBuilder.toString();
  }
}
