/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.test;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.util.LogMessage;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;

/**
 * The ordered container for tests.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see Test
 * @see Entity
 */
@XmlType(name = "testModel")
@XmlRootElement
public final class TestModel extends Entity {

  /**
   * The entity type of all {@code TestModel} instances.
   */
  public static final EntityType TYPE = EntityType.get("test");

  @XmlElement(name = "tests")
  private final List<Test> tests;

  /**
   * Constructs an empty test model.
   */
  public TestModel() {
    super(TestModel.TYPE);

    this.tests = new ArrayList<>();
  }

  /**
   * Adds the specified test to this test model.
   *
   * @param test the test
   *
   * @throws IllegalArgumentException if {@code test} is {@code null}
   */
  public void addTest(Test test) {
    InvariantChecks.checkNotNull(test);

    this.tests.add(test);
  }

  /**
   * Returns tests of this test model.
   *
   * @return the tests
   */
  public List<Test> getTests() {
    return Collections.unmodifiableList(this.tests);
  }

  /**
   * Adds the specified test model to this one.
   *
   * @param entity the test model to be added
   *
   * @throws IllegalArgumentException if {@code entity} is either {@code null} or not an instance of
   *     {@code SequenceModel}
   */
  @Override
  public void add(Entity entity) {
    InvariantChecks.checkNotNull(entity);

    if (!(entity instanceof TestModel)) {
      throw new IllegalArgumentException(LogMessage.UNSUPPORTED_TYPE + entity.getClass().getName());
    }

    this.tests.addAll(((TestModel) entity).tests);
  }

  @Override
  public int hashCode() {
    return 31 * super.hashCode() + this.tests.hashCode();
  }

  @Override
  public boolean equals(Object object) {
    if (this == object) {
      return true;
    }

    if (!super.equals(object) || !(object instanceof TestModel)) {
      return false;
    }

    TestModel otherTestModel = (TestModel) object;

    return this.tests.equals(otherTestModel.tests);
  }
}
