/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.test;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;

/**
 * The walker for the {@link Test} class.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public final class TestWalker {
  private final TestVisitor visitor;

  /**
   * Constructs a walker for the specified visitor.
   *
   * @param visitor the visitor
   *
   * @throws IllegalArgumentException if {@code visitor} is {@code null}
   */
  public TestWalker(TestVisitor visitor) {
    InvariantChecks.checkNotNull(visitor);

    this.visitor = visitor;
  }

  /**
   * Performs walking across the specified test.
   *
   * @param test the test to be walked across
   *
   * @throws IllegalArgumentException if {@code test} is {@code null}
   */
  public void start(Test test) {
    InvariantChecks.checkNotNull(test);

    this.visitor.onTestBegin(test);

    for (Sequence sequence: test.getSequences()) {
      this.visitSequence(sequence);
    }

    this.visitor.onTestEnd(test);
  }

  /**
   * Visits the specified sequence.
   *
   * @param sequence the sequence to be visited
   *
   * @throws IllegalArgumentException if {@code sequence} is {@code null}
   */
  public void visitSequence(Sequence sequence) {
    InvariantChecks.checkNotNull(sequence);

    this.visitor.onSequenceBegin(sequence);

    for (Vector vector: sequence.getVectors()) {
      this.visitVector(vector);
    }

    this.visitor.onSequenceEnd(sequence);
  }

  /**
   * Visits the specified vector.
   *
   * @param vector the vector to be visited
   *
   * @throws IllegalArgumentException if {@code vector} is {@code null}
   */
  public void visitVector(Vector vector) {
    InvariantChecks.checkNotNull(vector);

    this.visitor.onVectorBegin(vector);
    this.visitEventList(vector.getEventList());

    for (Transaction transaction: vector.getTransactions()) {
      this.visitTransaction(transaction);
    }

    this.visitor.onVectorEnd(vector);
  }

  /**
   * Visits the specified transaction.
   *
   * @param transaction the transaction to be visited
   *
   * @throws IllegalArgumentException if {@code transaction} is {@code null}
   */
  public void visitTransaction(Transaction transaction) {
    InvariantChecks.checkNotNull(transaction);

    this.visitor.onTransactionBegin(transaction);

    for (String variableName: transaction.getValues().keySet()) {
      this.visitVariable(variableName, transaction.getValues().get(variableName));
    }

    this.visitor.onTransactionEnd(transaction);
  }

  /**
   * Visits the specified eventList.
   *
   * @param eventList the eventList to be visited
   *
   * @throws IllegalArgumentException if {@code eventList} is {@code null}
   */
  public void visitEventList(EventList eventList) {
    InvariantChecks.checkNotNull(eventList);

    this.visitor.onEventListBegin(eventList);

    for (Event event: eventList.getEvents()) {
      this.visitEvent(event);
    }

    this.visitor.onEventListEnd(eventList);
  }

  /**
   * Visits the specified event.
   *
   * @param event the event to be visited
   *
   * @throws IllegalArgumentException if {@code event} is {@code null}
   */
  public void visitEvent(Event event) {
    InvariantChecks.checkNotNull(event);

    this.visitor.onEvent(event);
  }

  /**
   * Visits a variable with the specified name and value.
   *
   * @param name the name of the variable to be visited
   * @param value the value of the variable to be visited
   *
   * @throws IllegalArgumentException if either {@code name} or {@code value} is {@code null}
   */
  public void visitVariable(String name, Node value) {
    InvariantChecks.checkNotNull(name);
    InvariantChecks.checkNotNull(value);

    this.visitor.onVariable(name, value);
  }
}
