/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.bitvector.BitVectorMath;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;

/**
 * Utility methods for {@code Node} objects.
 *
 * @see Node
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class NodeUtils {

  /**
   * Returns an equation {@code (nodes[0] == ... == nodes[n - 1])} expression
   * of the specified nodes.
   *
   * @param nodes Expressions to be combined.
   *
   * @return An equation of the specified nodes.
   *
   * @throws IllegalArgumentException when argument is either {@code null} or empty.
   */
  public static Node getEquation(final Node ... nodes) {
    InvariantChecks.checkNotNull(nodes);
    InvariantChecks.checkNotEmpty(nodes);

    return new NodeOperation(StandardOperation.EQ, nodes);
  }

  /**
   * Casts the specified node to the specified data type.
   *
   * @param node Expression to be casted.
   * @param dataType Type to which the expression to be casted to.
   *
   * @return An expression which is equivalent to the specified node and has the specified type.
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  public static Node castToDataType(final Node node, final DataType dataType) {
    InvariantChecks.checkNotNull(node);
    InvariantChecks.checkNotNull(dataType);

    if (node.getDataType() == dataType) {
      return node;
    }

    switch (dataType.getTypeId()) {
      case BIT_VECTOR:
        return castToBitVector(node, dataType.getSize());
      case LOGIC_BOOLEAN:
        return castToBool(node);
      case LOGIC_INTEGER:
        return castToInteger(node);
      case LOGIC_REAL:
        return castToReal(node);
      case MAP:
        return castToMap(node);
      default:
        unsupported(dataType.getTypeId());
    }
    return null;
  }

  private static Node castToMap(final Node node) {
    unsupported(node);
    return null;
  }

  private static Node castToReal(final Node node) {
    unsupported(node);
    return null;
  }

  private static Node castToInteger(final Node node) {

    final Node zeroValue = NodeValue.newInteger(0);
    final Node oneValue = NodeValue.newInteger(1);

    switch (node.getDataTypeId()) {
      case BIT_VECTOR:
        final int bvSize = node.getDataType().getSize();
        final BitVector oneBitVector = BitVector.valueOf(1, bvSize);
        Node intExpression = zeroValue;
        BitVector current = oneBitVector;
        for (int i = 1; i < Math.pow(2, bvSize); i++) {

          final Node newInt = NodeValue.newInteger(i);
          intExpression =
              ifThenElse(getEquation(node, NodeValue.newBitVector(current)), newInt, intExpression);
          current = BitVectorMath.add(current, oneBitVector);
        }
        return intExpression;
      case LOGIC_BOOLEAN:
        final Node zeroBool = NodeValue.newBoolean(false);
        return ifThenElse(getEquation(node, zeroBool), zeroValue, oneValue);
      default:
        unsupported(node);
    }
    return null;
  }

  private static Node castToBool(final Node node) {

    final Node trueValue = NodeValue.newBoolean(true);
    final Node falseValue = NodeValue.newBoolean(false);

    switch (node.getDataTypeId()) {
      case BIT_VECTOR:
        final Node zeroBitVector = NodeValue.newBitVector(BitVector.FALSE);
        return ifThenElse(getEquation(node, zeroBitVector), falseValue, trueValue);
      case LOGIC_INTEGER:
        final Node zeroInt = NodeValue.newInteger(0);
        return ifThenElse(getEquation(node, zeroInt), falseValue, trueValue);
      default:
        unsupported(node.getDataTypeId());
    }
    return null;
  }

  private static Node castToBitVector(
      final Node node,
      final int size) {

    final Node trueValue = NodeValue.newBitVector(BitVector.TRUE);
    final Node falseValue = NodeValue.newBitVector(BitVector.FALSE);

    switch (node.getDataTypeId()) {
      case LOGIC_BOOLEAN:
        final Node zeroBool = NodeValue.newBoolean(false);
        return ifThenElse(getEquation(node, zeroBool), falseValue, trueValue);
      case LOGIC_INTEGER:
        Node bitVector = NodeValue.newBitVector(BitVector.valueOf(0, size));
        for (int i = 1; i < Math.pow(2, size); i++) {
          final Node intValue = NodeValue.newInteger(i);
          final Node bvValue = NodeValue.newBitVector(BitVector.valueOf(i, size));
          bitVector = ifThenElse(getEquation(node, intValue), bvValue, bitVector);
        }
        return bitVector;
      default:
        unsupported(node.getDataTypeId());
    }
    return null;
  }

  private static Node ifThenElse(final Node ... nodes) {
    return new NodeOperation(StandardOperation.ITE, nodes);
  }

  private static void unsupported(final Object obj) {
    throw RetrascopeRuntimeException.makeException(LogMessage.UNSUPPORTED_TYPE, obj);
  }
}
