/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.constraint.ConstraintUtils;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;

import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Embedded (project-level) satisfiability (SAT) solver.
 *
 * <p>The internal solver calls external one for every constraint it is unable to solve.
 * It is able to solve constraints of the following kinds:
 * <ul>
 * <li>{@code true}
 * <li>{@code false}
 * <li>single boolean variable or binding
 * <li>a conjunction that include "x AND !x" sub-expression
 * </ul></p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class SatSolver {

  private static SatSolver instance;
  private static Map<Node, SolverResult> solvedMap = new LinkedHashMap<>();

  private SatSolver() {
    /* Empty. */
  }

  /**
   * Returns an instance of the satisfiability solver.
   *
   * @return Solver instance.
   */
  public static synchronized SatSolver getInstance() {
    if (instance == null) {
      instance = new SatSolver();
    }
    return instance;
  }

  private Map<Node, SolverResult> getSolved() {
    return solvedMap;
  }

  /**
   * Solves the specified node constraint.
   *
   * @param constraint Constraint to be solved.
   *
   * @return Result of solving the constraint.
   */
  public SolverResult solve(final Node constraint) {

    final Node node = TransformerUtils.standardize(constraint);

    if (getSolved().containsKey(node)) {
      return getSolved().get(node);
    } else {

      SolverResult result;
      if (canBeSolved(node)) {

        switch (node.getKind()) {
          case VALUE:
            final boolean isSatValue = isSatValue((NodeValue) node);

            final SolverResult.Status status =
              isSatValue ? SolverResult.Status.SAT : SolverResult.Status.UNSAT;

            result = createResult(status, new LinkedList<Variable>());
            break;
          case VARIABLE:
            final Variable variable =
                new Variable(((NodeVariable) node).getName(), Data.newBoolean(true));
            result = createResult(SolverResult.Status.SAT, Collections.singletonList(variable));
            break;
          case OPERATION:
            /* Solve controversial operations only that are unsatisfiable. */
            result = createResult(SolverResult.Status.UNSAT, new LinkedList<Variable>());
            break;
          default:
            throw RetrascopeRuntimeException.makeException(
                LogMessage.UNSUPPORTED_TYPE,
                node.getKind());
        }
      } else {
        result = ConstraintUtils.solve(ConstraintUtils.newConstraint(node));
      }

      getSolved().put(node, result);
      return result;
    }
  }

  /**
   * Checks whether the specified expression is satisfiable.
   *
   * @param expr Expression to be checked.
   * @return {@code true} if the expression is satisfiable, {@code false} otherwise.
   *
   */
  public boolean isSat(final Node expr) {

    final Node node = TransformerUtils.standardize(expr);

    if (getSolved().containsKey(node)) {
      return getSolved().get(node).getStatus() == SolverResult.Status.SAT;
    } else {
      final SolverResult result = solve(node);
      getSolved().put(node, result);
      final List<String> errors = result.getErrors();
      if (!errors.isEmpty() && result.getStatus().equals(SolverResult.Status.ERROR)) {
        throw new IllegalArgumentException(
            "Constraint contains errors: " + node + "; errors=" + errors);
      }
      return result.getStatus() == SolverResult.Status.SAT;
    }
  }

  /**
   * Checks whether the specified logical expressions are compatible
   * {@code (getConjunction(nodes[0], ..., nodes[n-1]) is SAT)}.
   *
   * @param expressions Logical expressions to be checked.
   *
   * @return {@code true} if the specified logical expressions are compatible,
   *         {@code false} otherwise.
   */
  public boolean areCompatible(final Node... expressions) {
    return isSat(ExprUtils.getConjunction(expressions));
  }

  /**
   * Check whether the specified logical expressions are complete
   * {@code (!getComplement(nodes[0], ..., nodes[n-1]) is SAT)}.
   *
   * @param expressions Logical expressions to be checked.
   *
   * @return {@code true} if the specified logical expressions are complete,
   *         {@code false} otherwise.
   */
  public boolean areComplete(final Node... expressions) {
    return !isSat(ExprUtils.getComplement(expressions));
  }

  private static boolean canBeSolved(final Node node) {

    final Node.Kind kind = node.getKind();

    if (kind == Node.Kind.VALUE || kind == Node.Kind.VARIABLE) {
      return ExprUtils.isType(DataType.BOOLEAN, node);
    }
    return kind == Node.Kind.OPERATION && isControversial((NodeOperation) node);
  }

  private static boolean isSatValue(final NodeValue expr) {

    boolean isSatValue;
    if (ExprUtils.isType(DataType.BOOLEAN, expr)) {
      if (expr.equals(SolverUtils.TRUE)) {
        isSatValue = true;
      } else if (expr.equals(SolverUtils.FALSE)) {
        isSatValue = false;
      } else {
        throw new RetrascopeRuntimeException(
            new IllegalArgumentException("Incorrect boolean argument: " + expr.toString()));
      }
    } else {
      isSatValue = ExprUtils.isSAT(expr);
    }
    return isSatValue;
  }

  private static boolean isControversial(final NodeOperation expr) {

    boolean controversial = false;

    if (isOperationId(expr, StandardOperation.AND)) {

      final int operandCount = expr.getOperandCount();
      final Map<Node, Boolean> operandSignMap = new HashMap<>();

      for (int i = 0; i < operandCount; i++) {

        final Node operand = expr.getOperand(i);

        /* Check either object contains at least one controversial operand. */
        final boolean hasControversial =
            operand.getKind() == Node.Kind.OPERATION && isControversial((NodeOperation) operand);

        if (hasControversial) {
          controversial = true;
          break;
        }

        /* Mark as controversial if "x AND !x" is found in the expression. */
        if (containDifferentSignClauses(operandSignMap, operand)) {
          controversial = true;
          break;
        }
      }
    }
    return controversial;
  }

  private static boolean containDifferentSignClauses(
      final Map<Node, Boolean> map,
      final Node operand) {

    boolean result;

    final boolean isNotOperation = isOperationId(operand, StandardOperation.NOT);
    final Node clause = getClause(operand);
    final boolean mapContainClause = map.containsKey(clause);

    result = mapContainClause && map.get(clause) != isNotOperation;

    if (!result) {
      map.put(clause, isNotOperation);
    }
    return result;
  }

  private static Node getClause(final Node operand) {

    final boolean isNotClause = isOperationId(operand, StandardOperation.NOT);

    return isNotClause ? ((NodeOperation)operand).getOperand(0) : operand;
  }

  private static boolean isOperationId(
      final Node operation,
      final Enum<?> operationId) {

    return operation.getKind() == Node.Kind.OPERATION
        && ((NodeOperation)operation).getOperationId().equals(operationId);
  }

  private static SolverResult createResult(
      final SolverResult.Status status,
      final List<Variable> variables) {

    return new SolverResult(status, new LinkedList<String>(), variables);
  }
}
