/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.util;

import ru.ispras.retrascope.basis.FileCreator;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.LinkedList;

/**
 * This class represents the NuXmv model checker wrapper.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SmvChecker {

  /**
   * Result of checker execution.
   */
  public enum Result {
    ERROR, OK, UNKNOWN
  }

  private static final String nuXmvPath = System.getenv("NUXMV_PATH");
  private static final String nuSmvPath = System.getenv("NUSMV_PATH");
  private static SmvChecker checker;

  private Result solveResult;
  private String currentCheckerPath;
  private File logFile;

  private SmvChecker() {
    solveResult = Result.UNKNOWN;
    currentCheckerPath = nuXmvPath;
    logFile = null;
  }

  /**
   * Returns the checker instance.
   *
   * @return The checker instance.
   */
  public static synchronized SmvChecker getChecker() {
    if (checker == null) {
      checker = new SmvChecker();
    }
    return checker;
  }

  /**
   * Enables NuXMV checker.
   */
  public void setNuXmv() {
    currentCheckerPath = nuXmvPath;
  }

  /**
   * Enables NuSMV checker.
   */
  public void setNuSmv() {
    currentCheckerPath = nuSmvPath;
  }

  /**
   * Returns the result of the last checker execution.
   *
   * @return The result of the last execution.
   */
  public Result getResult() {
    return solveResult;
  }

  /**
   * Returns the log of the last checker execution.
   *
   * @return The log file if exists, <code>null</code> otherwise.
   */
  public File getLogFile() {
    return logFile;
  }

  /**
   * Runs the checker using the specified arguments.
   *
   * @param fileName - the specified *.smv file name.
   * @param args - the specified arguments.
   */
  public void runChecker(final String fileName, final String... args) {
    solveResult = Result.UNKNOWN;
    logFile = null;

    final LinkedList<String> arguments = new LinkedList<>();
    arguments.add(currentCheckerPath);
    arguments.addAll(Arrays.asList(args));
    arguments.add(fileName);

    try {
      logFile =
          FileCreator.newFile(fileName.substring(0, fileName.lastIndexOf(".")) + ".smvlog");
      final ProcessBuilder pb = new ProcessBuilder(arguments);
      pb.redirectErrorStream(true);
      pb.redirectOutput(logFile);
      final Process process = pb.start();
      try {
        final int exitValue = process.waitFor();
        if (exitValue == 0) {
          solveResult = Result.OK;
        } else {
          solveResult = Result.ERROR;
          Log.printError(
              String.format("Process with args '%s' exited with '%d'.", arguments, exitValue));
        }
      } catch (InterruptedException e) {
        Log.printError("Failed to exit process");
        e.printStackTrace();
      }
    } catch (IOException | NullPointerException e) {
      Log.printError("Failed to start the checker");
      e.printStackTrace();
    }
  }

}
