/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.transformer.TransformerRule;

import java.util.List;

/**
 * Utility methods for transformation of expressions.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class TransformerUtils {

  private TransformerUtils() {
    /* Empty. */
  }

  /**
   * Replace operations in expression with standard counterparts.
   * <p>Transforms nodes of the following kinds:
   * <ul>
   *   <li> EQ operation with more than 2 operands - to conjunction of 2-operand EQs;
   * </ul>
   * Wrapping method for {@code Transformer.standardize} method.
   * Uses it's internal rules for additional replacements.</p>
   *
   * @see Transformer
   * @param node Expression to be transformed.
   * @return Expression with non-standard operations being replaced.
   */
  public static Node standardize(final Node node) {

    final Node standardizedNode = Transformer.standardize(node);

    final TransformerRule expandEq = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {
        return node.getKind() == Node.Kind.OPERATION
            && ((NodeOperation) node).getOperationId().equals(StandardOperation.EQ)
            && ((NodeOperation) node).getOperandCount() > 2;
      }

      @Override
      public Node apply(final Node node) {
        final List<Node> operands = ((NodeOperation)node).getOperands();
        final Node[] equations = new NodeOperation[operands.size() - 1];

        for (int i = 1; i < operands.size(); i++) {
          equations[i - 1] = NodeUtils.getEquation(operands.get(i), operands.get(0));
        }
        return ExprUtils.getConjunction(equations);
      }
    };

    return Transformer.transform(standardizedNode, StandardOperation.EQ, expandEq);
  }
}
