/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.generator.test.random;

import org.junit.Test;
import ru.ispras.retrascope.HdlTestFiles;
import ru.ispras.retrascope.HdlTestUtils;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.result.test.TestModel;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Test for Random test generator based on Control Flow Graph Model.
 * The test runs test generator on every separate HDL file and the engine produces
 * {@link TestModel} representation of test.
 *
 * @see CfgRandomTestGenerator
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgRandomTestGeneratorTestCase {

  static class FileDescriptor {
    private final String clkName;
    private final String rstName;
    private final String inputValues;

    FileDescriptor(
      final String clkName,
      final String rstName) {
      this.clkName = clkName;
      this.rstName = rstName;
      this.inputValues = "";
    }

    FileDescriptor(
        final String clkName,
        final String rstName,
        final String inputValues) {
      this.clkName = clkName;
      this.rstName = rstName;
      this.inputValues = inputValues;
    }

    public String getClkName() {
      return this.clkName;
    }

    public String getRstName() {
      return this.rstName;
    }

    public String getInputValues() {
      return this.inputValues;
    }
  }

  protected static final int TEST_LENGTH = 10;
  private static final String ITC99 = "itc99";
  private static final String MINIMIPS = "minimips";
  private static final String PLASMA = "plasma";

  protected static final Map<String, FileDescriptor> map = new LinkedHashMap<>();

  static {

    /* === VHDL test projects === */

    /* sample */
    map.put("/bvsample/bvsample.vhd", new FileDescriptor("CLK", ""));

    /* sample */
    map.put("/sample/sample.vhd", new FileDescriptor("CLK", "RST"));

    /* count */
    map.put("/count/count.vhd", new FileDescriptor("CLK", "RESET"));

    /* example */
    map.put("/example/test.vhd", new FileDescriptor("CLK", "RESET"));
    map.put("/example/driver.vhd", new FileDescriptor("CLK", "RESET"));
    map.put("/example/example.vhd", new FileDescriptor("CLK", "RESET"));

    /* ITC'99  benchmark */
    map.put("/" + ITC99 + "/b01/b01.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b02/b02.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b03/b03.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b04/b04.vhd", new FileDescriptor("CLOCK", "RESET", "DATA_IN=[-128,127,10]"));
    map.put("/" + ITC99 + "/b05/b05.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b06/b06.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b07/b07.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b08/b08.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b09/b09.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b10/b10.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b11/b11.vhd", new FileDescriptor("CLOCK", "RESET", "X_IN=[0,63,10]"));
    map.put("/" + ITC99 + "/b12/b12.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b13/b13.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b14/b14.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b14_1/b14_1.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b15/b15.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b15_1/b15_1.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b17/b17.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b17_1/b17_1.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b18/b18.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b18_1/b18_1.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b19/b19.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b19_1/b19_1.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b20/b20.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b20_1/b20_1.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b21/b21.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b21_1/b21_1.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b22/b22.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + ITC99 + "/b22_1/b22_1.vhd", new FileDescriptor("CLOCK", "RESET"));

    /* MiniMIPS */
    map.put("/" + MINIMIPS + "/alu.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + MINIMIPS + "banc.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + MINIMIPS + "bus_ctrl.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + MINIMIPS + "minimips.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + MINIMIPS + "pps_di.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + MINIMIPS + "pps_ei.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + MINIMIPS + "pps_ex.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + MINIMIPS + "pps_mem.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + MINIMIPS + "pps_pf.vhd", new FileDescriptor("CLOCK", "RESET"));
    map.put("/" + MINIMIPS + "ram.vhd", new FileDescriptor("", ""));
    map.put("/" + MINIMIPS + "renvoi.vhd", new FileDescriptor("", ""));
    map.put("/" + MINIMIPS + "rom.vhd", new FileDescriptor("", ""));
    map.put("/" + MINIMIPS + "syscop.vhd", new FileDescriptor("CLOCK", "RESET"));

    /* Plasma */
    map.put("/" + PLASMA + "alu.vhd", new FileDescriptor("", ""));
    map.put("/" + PLASMA + "bus_mux.vhd", new FileDescriptor("", ""));
    map.put("/" + PLASMA + "control.vhd", new FileDescriptor("", ""));
    map.put("/" + PLASMA + "mem_ctrl.vhd", new FileDescriptor("CLK", "RESET_IN"));
    map.put("/" + PLASMA + "mlite2sram.vhd", new FileDescriptor("CLK", ""));
    map.put("/" + PLASMA + "mlite2uart.vhd", new FileDescriptor("CLK", "RESET"));
    map.put("/" + PLASMA + "mlite_cpu.vhd", new FileDescriptor("CLK", "RESET_IN"));
    map.put("/" + PLASMA + "mult.vhd", new FileDescriptor("CLK", ""));
    map.put("/" + PLASMA + "pc_next.vhd", new FileDescriptor("CLK", "RESET_IN"));
    map.put("/" + PLASMA + "pipeline.vhd", new FileDescriptor("CLK", "RESET"));
    map.put("/" + PLASMA + "plasma.vhd", new FileDescriptor("BOARD_CLK", "BOARD_RES"));
    map.put("/" + PLASMA + "ram.vhd", new FileDescriptor("CLK", ""));
    map.put("/" + PLASMA + "reg_bank.vhd", new FileDescriptor("CLK", "RESET_IN"));
    map.put("/" + PLASMA + "shifter.vhd", new FileDescriptor("", ""));
    map.put("/" + PLASMA + "sram2mlite.vhd", new FileDescriptor("CLK", ""));
    map.put("/" + PLASMA + "uart.vhd", new FileDescriptor("CLK", "RESET"));

    /* === Verilog test projects === */

    /* adder */
    map.put("/adder/adder4.v", new FileDescriptor("", ""));
    map.put("/adder/adder4_testbench.v", new FileDescriptor("", ""));

    /* empty */
    map.put("/empty/empty.v", new FileDescriptor("", ""));

    /* fifo */
    map.put("/fifo/fifo.v", new FileDescriptor("clk", "rst"));
    map.put("/fifo/fifo_testbench.v", new FileDescriptor("clk", "rst"));

    /* fifo_mod */
    map.put("/fifo_mod/fifo.v", new FileDescriptor("clk", "rst"));

    /* for_loop */
    map.put("/for_loop/for_loop.v", new FileDescriptor("clk", ""));

    /* incompat_cond */
    map.put("/incompat_cond/incompat_cond.v", new FileDescriptor("clk", ""));

    /* fifo0 */
    map.put("/fifo0/mem_2p.v", new FileDescriptor("rd_clk", ""));
    map.put("/fifo0/mul_fifo.v", new FileDescriptor("clk", "rst"));
    map.put("/fifo0/myfifo.v", new FileDescriptor("clk", "rst"));

    /* mux */
    map.put("/mux/mux4_1.v", new FileDescriptor("", ""));
    map.put("/mux/mux4_1_testbench.v", new FileDescriptor("", ""));

    /* ram */
    map.put("/ram/ram.v", new FileDescriptor("clk", "rst"));
    map.put("/ram/ram_testbench.v", new FileDescriptor("clk", "rst"));
  }

  @Test
  public void runTest() {
    final Map<String, Map<String, String>> hdlProjects = HdlTestFiles.getProjects();

    for (final String projectName : hdlProjects.keySet()) {

      final Map<String, String> projectFiles = hdlProjects.get(projectName);

      for (final Map.Entry<String, String> entry : projectFiles.entrySet()) {

        final String filePath = entry.getKey();
        final Map<String, String> oneFile = new LinkedHashMap<>();
        oneFile.put(filePath, entry.getValue());
        final FileDescriptor descriptor = getDescriptor(filePath);

        HdlTestUtils.runHdl(
            oneFile,
            TestModel.TYPE.getId(),
            "--" + Retrascope.ENGINE_PARAM.getLongName(),
            CfgRandomTestGenerator.ENGINE_ID,
            "--" + CfgRandomTestGenerator.TEST_LEN.getLongName(),
            String.valueOf(TEST_LENGTH),
            "--" + CfgRandomTestGenerator.CLK_NAME.getLongName(),
            descriptor.getClkName(),
            "--" + CfgRandomTestGenerator.RST_NAME.getLongName(),
            descriptor.getRstName(),
            "--" + CfgRandomTestGenerator.INPUT_VALUES.getLongName(),
            descriptor.getInputValues());
      }
    }
  }

  protected FileDescriptor getDescriptor(String filePath) {
    for (final Map.Entry<String, FileDescriptor> entry : map.entrySet()) {
      if (filePath.endsWith(entry.getKey())) {
        return entry.getValue();
      }
    }
    return new FileDescriptor("","");
  }
}
