/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import org.apache.commons.io.Charsets;
import org.junit.Assert;
import org.junit.Test;

import ru.ispras.retrascope.HdlTestUtils;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;

/**
 * Test for transformer from clocked guarded atomic actions model
 * {@link ru.ispras.retrascope.model.cgaa.CgaaModel} to extended finite state machine model
 * {@link ru.ispras.retrascope.model.efsm.EfsmModel}.
 *
 * <p>Test runs the transformer on every top-level module
 * of every HDL file from project test suite.
 * After that it compares statistical info that is stored in debug info file
 * with the reference one.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CgaaEfsmTransformerTestCase {

  @Test
  public void runTest() {

    final String debugFileName = CgaaEfsmTransformer.ENGINE_ID + ".log";
    final String[] referenceDebug = {
      "states=1 transitions=4 efsms=1", // bvsample
      "states=2 transitions=4 efsms=1", // sample
      "states=1 transitions=2 efsms=1", // test
      "states=1 transitions=1 efsms=1", // driver
      "states=5 transitions=8 efsms=5", // example
      "states=8 transitions=24 efsms=1", // b01
      "states=7 transitions=17 efsms=1", // b02
      "states=23 transitions=323 efsms=1", // b03
      "states=3 transitions=29 efsms=1", // b04
//      "states=9 transitions=700 efsms=3", // b05
      "states=7 transitions=33 efsms=1", // b06
      "states=10 transitions=26 efsms=1", // b07
      "states=4 transitions=12 efsms=1", // b08
      "states=4 transitions=12 efsms=1", // b09
      "states=11 transitions=38 efsms=1", // b10
      "states=13 transitions=44 efsms=1", // b11
//      "states=47 transitions=235 efsms=4", // b12
//      "states=30 transitions=96 efsms=5", // b13
      "states=9 transitions=9 efsms=9", // adder4
      "states=0 transitions=0 efsms=0", // empty
      "states=20 transitions=96 efsms=5", // fifo
      "states=5 transitions=4 efsms=5", // fifo_testbench
//      "states=22 transitions=978 efsms=7", // fifo_mod
      "states=1 transitions=1 efsms=1", // for_loop
      "states=1 transitions=4 efsms=1", // incompat_cond
      "states=1 transitions=1 efsms=1", // mux_4_1
      "states=7 transitions=14 efsms=5", // ram
      "states=5 transitions=4 efsms=5" // ram_testbench
    };

    /* If reference debug file exists, clear it. */
    final File debugFile = new File(debugFileName);
    if (debugFile.exists()) {
      try {
        final PrintWriter writer = new PrintWriter(debugFile);
        writer.write("");
        writer.close();
      } catch (FileNotFoundException e) {
        Log.printError(LogMessage.makeMessage(LogMessage.CANT_FIND, debugFileName));
      }
    }

    HdlTestUtils.runHdl(EfsmModel.TYPE.getId(),
        new String[] {
            "--" + Retrascope.ENGINE_PARAM.getLongName(),
            CgaaEfsmTransformer.ENGINE_ID,
            "--" + Retrascope.TOOL_DEBUG_PARAM.getLongName(),
            debugFileName});

    try {
      final FileInputStream inputStream = new FileInputStream(new File(debugFileName));
      final DataInputStream dataStream = new DataInputStream(inputStream);

      try (final BufferedReader reader = getBufferedReader(dataStream)) {

        String line;
        int lineIndex = 0;
        while ((line = reader.readLine()) != null) {
          if (!line.equalsIgnoreCase(referenceDebug[lineIndex])) {
            Assert.fail(referenceDebug[lineIndex] + " should be equal to: " + line);
          }
          lineIndex++;
        }
      }
    } catch (IOException e) {
      throw new RetrascopeRuntimeException(e.getMessage(), e.getCause());
    }
  }

  private BufferedReader getBufferedReader(final DataInputStream dataStream) {
    return new BufferedReader(new InputStreamReader(dataStream, Charsets.UTF_8));
  }
}
