/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm.examples;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.VariableContainer;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.SolverUtils;

/**
 * Class providing utility methods for manual construction of example EFSMs.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
final class ExampleEfsmUtils {
  /**
   * A guard whose condition is always true.
   */
  public static final Guard ALWAYS_TRUE_GUARD = new Guard(SolverUtils.TRUE);

  /**
   * Constructs an input variable on the basis of HDL description and declares it in the specified
   * variable container.
   *
   * @param variableContainer the variable container where the variable should be declared in
   * @param name a name of the variable
   * @param hdlType a string containing HDL type of the variable
   *
   * @return the constructed variable
   */
  static NodeVariable declareInput(VariableContainer variableContainer, String name,
      String hdlType) {
    return ExampleEfsmUtils.declareVariable(variableContainer, name, VariableType.IN, hdlType);
  }

  /**
   * Constructs an output variable on the basis of HDL description and declares it in the specified
   * variable container.
   *
   * @param variableContainer the variable container where the variable should be declared in
   * @param name a name of the variable
   * @param hdlType a string containing HDL type of the variable
   *
   * @return the constructed variable
   */
  static NodeVariable declareOutput(VariableContainer variableContainer, String name,
      String hdlType) {
    return ExampleEfsmUtils.declareVariable(variableContainer, name, VariableType.OUT, hdlType);
  }

  /**
   * Constructs a register variable on the basis of HDL description and declares it in the specified
   * variable container.
   *
   * @param variableContainer the variable container where the variable should be declared in
   * @param name a name of the variable
   * @param hdlType a string containing HDL type of the variable
   *
   * @return the constructed variable
   */
  static NodeVariable declareRegister(VariableContainer variableContainer, String name,
      String hdlType) {
    return ExampleEfsmUtils.declareVariable(variableContainer, name, VariableType.REG, hdlType);
  }

  /**
   * Constructs a variable with the specified parameters and declares it in the specified variable
   * container. The variable has no condition on value which is initially set to zero.
   *
   * @param variableContainer the variable container where the variable should be declared in
   * @param name a name of the variable
   * @param dataType a data type of the variable
   * @param variableType a type of the variable
   *
   * @return the constructed variable
   */
  static NodeVariable declareVariable(VariableContainer variableContainer, String name,
      DataType dataType, VariableType variableType) {
    NodeVariable variable = new NodeVariable(name, dataType);
    variable.setUserData(new VariableDescriptor(variableType));

    variableContainer.declareVariable(variable, new NodeValue(dataType.valueOf("0", 10)), null);

    return variable;
  }

  /**
   * Constructs a variable on the basis of HDL description and declares it in the specified
   * variable container.
   *
   * @param variableContainer the variable container where the variable should be declared in
   * @param name a name of the variable
   * @param variableType a type of the variable
   * @param hdlType a string containing HDL type of the variable
   *
   * @return the constructed variable
   */
  static NodeVariable declareVariable(VariableContainer variableContainer, String name,
      VariableType variableType, String hdlType) {
    VariableDescriptor variableDescriptor = new VariableDescriptor(variableType);

    String[] splittedHdlType = hdlType.replace('(', ' ').replace(')', ' ').split(" ");

    DataType dataType;
    String decimalInitialValue;

    String minValue = null;
    String maxValue = null;

    switch (splittedHdlType[0].toLowerCase()) {
      case "bit": {
        dataType = DataType.BIT_VECTOR(1);
        decimalInitialValue = "0";

        break;
      }

      case "bit_vector": {
        if (splittedHdlType.length == 4) {
          int firstIndex = Integer.valueOf(splittedHdlType[1]);
          int secondIndex = Integer.valueOf(splittedHdlType[3]);

          switch (splittedHdlType[2]) {
            case "to": {
              dataType = DataType.BIT_VECTOR(secondIndex - firstIndex);

              break;
            }

            case "downto": {
              dataType = DataType.BIT_VECTOR(secondIndex - firstIndex);

              break;
            }

            default: {
              throw new IllegalArgumentException("Unsupported HDL type: " + hdlType);
            }
          }
        } else {
          throw new IllegalArgumentException("Unsupported HDL type: " + hdlType);
        }

        decimalInitialValue = "0";

        break;
      }

      case "integer": {
        dataType = DataType.INTEGER;

        switch (splittedHdlType.length) {
          case 1: {
            minValue = "-2147483648";
            maxValue = "2147483647";

            decimalInitialValue = minValue;

            break;
          }

          case 5: {
            switch (splittedHdlType[3].toLowerCase()) {
              case "to": {
                minValue = splittedHdlType[2];
                maxValue = splittedHdlType[4];

                break;
              }

              case "downto": {
                minValue = splittedHdlType[4];
                maxValue = splittedHdlType[2];

                break;
              }

              default: {
                throw new IllegalArgumentException("Unsupported HDL type: " + hdlType);
              }
            }

            decimalInitialValue = splittedHdlType[2];

            break;
          }

          default: {
            throw new IllegalArgumentException("Unsupported HDL type: " + hdlType);
          }
        }

        break;
      }

      default: {
        throw new IllegalArgumentException("Unsupported HDL type: " + splittedHdlType[0]);
      }
    }

    variableDescriptor.addMetaInfo(MetaInfoType.HDL_TYPE_STR, hdlType);

    NodeVariable variable = new NodeVariable(name, dataType);
    variable.setUserData(variableDescriptor);

    Node valueConstraint = null;

    if (minValue != null) {
      NodeValue value = new NodeValue(dataType.valueOf(minValue, 10));
      valueConstraint = new NodeOperation(StandardOperation.GREATEREQ, variable, value);
    }

    if (maxValue != null) {
      NodeValue value = new NodeValue(dataType.valueOf(maxValue, 10));
      NodeOperation condition = new NodeOperation(StandardOperation.LESSEQ, variable, value);

      if (valueConstraint == null) {
        valueConstraint = condition;
      } else {
        valueConstraint = ExprUtils.getConjunction(valueConstraint, condition);
      }
    }

    variableContainer.declareVariable(variable,
        new NodeValue(dataType.valueOf(decimalInitialValue, 10)), valueConstraint);

    return variable;
  }

  /**
   * Constructs a value of the same type as the specified variable has.
   *
   * @param variable the variable which value should be constructed for
   * @param decimalValue a string representation of the value
   *
   * @return the constructed value
   */
  static NodeValue constructValue(NodeVariable variable, String decimalValue) {
    return new NodeValue(variable.getDataType().valueOf(decimalValue, 10));
  }

  /**
   * Constructs an expression of the form "variable == value".
   *
   * @param variable the variable
   * @param decimalValue the value
   *
   * @return the constructed operation
   */
  static Node constructEquality(NodeVariable variable, String decimalValue) {
    return NodeUtils.getEquation(variable,
        ExampleEfsmUtils.constructValue(variable, decimalValue));
  }

  /**
   * Constructs an expression of the form "variable != value".
   *
   * @param variable the variable
   * @param decimalValue the value
   *
   * @return the constructed operation
   */
  static Node constructInequality(NodeVariable variable, String decimalValue) {
    return ExprUtils.getNegation(constructEquality(variable, decimalValue));
  }

  /**
   * Constructs an assignment of the form "variable := value".
   *
   * @param variable the variable
   * @param decimalValue the value
   *
   * @return the constructed operation
   */
  static Assignment constructAssignment(NodeVariable variable, String decimalValue) {
    return new Assignment(variable, ExampleEfsmUtils.constructValue(variable, decimalValue));
  }

  private ExampleEfsmUtils() {
  }
}
