/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm.examples;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;

import java.util.Arrays;
import java.util.Collections;

/**
 * The test class which contains manually created simple EFSM.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public final class SimpleEfsm {
  /**
   * An EFSM model which contains a simple EFSM module.
   */
  public static final EfsmModel EFSM_MODEL = new EfsmModel();

  static {
    EfsmModule efsmModule = new EfsmModule(SimpleEfsm.class.getSimpleName());

    NodeVariable clockVariable = ExampleEfsmUtils.declareVariable(efsmModule, "clock",
        DataType.BIT_VECTOR(1), VariableType.IN);

    NodeVariable i0Variable = ExampleEfsmUtils.declareVariable(efsmModule, "i0",
        DataType.BIT_VECTOR(4), VariableType.IN);
    NodeVariable i1Variable = ExampleEfsmUtils.declareVariable(efsmModule, "i1",
        DataType.BIT_VECTOR(1), VariableType.IN);
    NodeVariable i2Variable = ExampleEfsmUtils.declareVariable(efsmModule, "i2",
        DataType.BIT_VECTOR(4), VariableType.IN);

    NodeVariable varX = ExampleEfsmUtils.declareVariable(efsmModule, "x",
        DataType.BIT_VECTOR(4), VariableType.REG);
    NodeVariable varY = ExampleEfsmUtils.declareVariable(efsmModule, "y",
        DataType.BIT_VECTOR(4), VariableType.REG);
    NodeVariable varZ = ExampleEfsmUtils.declareVariable(efsmModule, "z",
        DataType.BIT_VECTOR(4), VariableType.REG);

    NodeVariable stateVariable = ExampleEfsmUtils.declareVariable(efsmModule, "state",
        DataType.BIT_VECTOR(4), VariableType.REG);

    String state0Value = "0";
    String state1Value = "1";
    String state2Value = "2";
    String state3Value = "3";
    String state4Value = "4";

    Efsm efsm = new Efsm(efsmModule, "EFSM1");

    EfsmState state0 =
        efsm.createState(ExampleEfsmUtils.constructEquality(stateVariable, state0Value));
    EfsmState state1 =
        efsm.createState(ExampleEfsmUtils.constructEquality(stateVariable, state1Value));
    EfsmState state2 =
        efsm.createState(ExampleEfsmUtils.constructEquality(stateVariable, state2Value));
    EfsmState state3 =
        efsm.createState(ExampleEfsmUtils.constructEquality(stateVariable, state3Value));
    EfsmState state4 =
        efsm.createState(ExampleEfsmUtils.constructEquality(stateVariable, state4Value));

    efsm.setInitialState(state0);

    Event clockEvent = new Event(new RangedVariable(clockVariable), EventType.POSITIVE_EDGE);

    {
      Guard guard01 = ExampleEfsmUtils.ALWAYS_TRUE_GUARD;
      Action action01 = new Action(Arrays.asList(new Assignment(varZ, i0Variable),
          ExampleEfsmUtils.constructAssignment(stateVariable, state1Value)));

      GuardedAction guardedAction01 = new GuardedAction(guard01, action01);
      guardedAction01.addClock(clockEvent);

      efsm.createTransition(state0, state1, guardedAction01);
    }

    {
      Guard guard12 = new Guard(ExampleEfsmUtils.constructEquality(i1Variable, "1"));
      Action action12 = new Action(Arrays.asList(new Assignment(varX, varZ),
          ExampleEfsmUtils.constructAssignment(stateVariable, state2Value)));

      GuardedAction guardedAction12 = new GuardedAction(guard12, action12);
      guardedAction12.addClock(clockEvent);

      efsm.createTransition(state1, state2, guardedAction12);
    }

    {
      Guard guard13 = new Guard(ExampleEfsmUtils.constructInequality(i1Variable, "1"));
      Action action13 = new Action(Arrays.asList(new Assignment(varX, varZ),
          ExampleEfsmUtils.constructAssignment(stateVariable, state3Value)));

      GuardedAction guardedAction13 = new GuardedAction(guard13, action13);
      guardedAction13.addClock(clockEvent);

      efsm.createTransition(state1, state3, guardedAction13);
    }

    {
      Guard guard23 = ExampleEfsmUtils.ALWAYS_TRUE_GUARD;
      Action action23 = new Action(Arrays.asList(new Assignment(varY, i2Variable),
          ExampleEfsmUtils.constructAssignment(stateVariable, state3Value)));

      GuardedAction guardedAction23 = new GuardedAction(guard23, action23);
      guardedAction23.addClock(clockEvent);

      efsm.createTransition(state2, state3, guardedAction23);
    }

    {
      Node guardPredicate34 =
          ExprUtils.getConjunction(ExampleEfsmUtils.constructEquality(varX, "4"),
          ExampleEfsmUtils.constructEquality(varY, "2"));
      Action action34 =
          new Action(ExampleEfsmUtils.constructAssignment(stateVariable, state4Value));

      GuardedAction guardedAction34 = new GuardedAction(new Guard(guardPredicate34), action34);
      guardedAction34.addClock(clockEvent);

      efsm.createTransition(state3, state4, guardedAction34);
    }

    GuardedAction resetGuardedAction = new GuardedAction(ExampleEfsmUtils.ALWAYS_TRUE_GUARD,
        new Action(Collections.<Assignment>emptyList()));
    resetGuardedAction.addClock(clockEvent);
    efsm.addResetGuardedAction(resetGuardedAction);

    efsmModule.addNonInitialEfsm(efsm);
    SimpleEfsm.EFSM_MODEL.addModule(efsmModule);
  }

  private SimpleEfsm() {
  }
}
