/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm.examples;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;

import java.util.Arrays;
import java.util.Collections;

/**
 * The test class which contains a manually created simple pipelined EFSM.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public final class SimplePipelinedEfsm {
  /**
   * An EFSM model which contains a simple pipelined EFSM module.
   */
  public static final EfsmModel EFSM_MODEL = new EfsmModel();

  static {
    EfsmModule efsmModule = new EfsmModule(SimplePipelinedEfsm.class.getSimpleName());

    NodeVariable clockVariable = ExampleEfsmUtils.declareVariable(efsmModule, "clock",
        DataType.BIT_VECTOR(1), VariableType.IN);

    NodeVariable i0Variable = ExampleEfsmUtils.declareVariable(efsmModule, "i0",
        DataType.BIT_VECTOR(4), VariableType.IN);
    NodeVariable i1Variable = ExampleEfsmUtils.declareVariable(efsmModule, "i1",
        DataType.BIT_VECTOR(1), VariableType.IN);

    NodeVariable x0Variable = ExampleEfsmUtils.declareVariable(efsmModule, "x0",
        DataType.BIT_VECTOR(4), VariableType.REG);
    NodeVariable x1Variable = ExampleEfsmUtils.declareVariable(efsmModule, "x1",
        DataType.BIT_VECTOR(4), VariableType.REG);
    NodeVariable x2Variable = ExampleEfsmUtils.declareVariable(efsmModule, "x2",
        DataType.BIT_VECTOR(4), VariableType.REG);
    NodeVariable x3Variable = ExampleEfsmUtils.declareVariable(efsmModule, "x3",
        DataType.BIT_VECTOR(4), VariableType.REG);

    NodeVariable stateVariable = ExampleEfsmUtils.declareVariable(efsmModule, "state",
        DataType.BIT_VECTOR(4), VariableType.REG);

    String state0Value = "0";
    String state1Value = "1";
    String state2Value = "2";
    String state3Value = "3";

    Efsm efsm = new Efsm(efsmModule, "EFSM2");

    EfsmState state0 =
        efsm.createState(ExampleEfsmUtils.constructEquality(stateVariable, state0Value));
    EfsmState state1 =
        efsm.createState(ExampleEfsmUtils.constructEquality(stateVariable, state1Value));
    EfsmState state2 =
        efsm.createState(ExampleEfsmUtils.constructEquality(stateVariable, state2Value));
    EfsmState state3 =
        efsm.createState(ExampleEfsmUtils.constructEquality(stateVariable, state3Value));

    efsm.setInitialState(state0);

    Event clockEvent = new Event(new RangedVariable(clockVariable), EventType.POSITIVE_EDGE);

    {
      Guard guard01 = ExampleEfsmUtils.ALWAYS_TRUE_GUARD;
      Action action01 = new Action(Arrays.asList(new Assignment(x0Variable, i0Variable),
          ExampleEfsmUtils.constructAssignment(stateVariable, state1Value)));

      GuardedAction guardedAction01 = new GuardedAction(guard01, action01);
      guardedAction01.addClock(clockEvent);

      efsm.createTransition(state0, state1, guardedAction01);
    }

    {
      Guard guard11 = new Guard(ExampleEfsmUtils.constructEquality(i1Variable, "0"));
      Action action11 = new Action(Arrays.asList(new Assignment(x3Variable, x2Variable),
          new Assignment(x2Variable, x1Variable), new Assignment(x1Variable, x0Variable)));

      GuardedAction guardedAction11 = new GuardedAction(guard11, action11);
      guardedAction11.addClock(clockEvent);

      efsm.createTransition(state1, state1, guardedAction11);
    }

    {
      Node guardPredicate12 =
          ExprUtils.getConjunction(ExampleEfsmUtils.constructInequality(i1Variable, "0"),
              ExampleEfsmUtils.constructEquality(x3Variable, "4"));
      Action action12 =
          new Action(ExampleEfsmUtils.constructAssignment(stateVariable, state2Value));

      GuardedAction guardedAction12 = new GuardedAction(new Guard(guardPredicate12), action12);
      guardedAction12.addClock(clockEvent);

      efsm.createTransition(state1, state2, guardedAction12);
    }

    {
      Node guardPredicate13 =
          ExprUtils.getConjunction(ExampleEfsmUtils.constructInequality(i1Variable, "0"),
              ExampleEfsmUtils.constructInequality(x3Variable, "4"));
      Action action13 =
          new Action(ExampleEfsmUtils.constructAssignment(stateVariable, state3Value));

      GuardedAction guardedAction13 = new GuardedAction(new Guard(guardPredicate13), action13);
      guardedAction13.addClock(clockEvent);

      efsm.createTransition(state1, state3, guardedAction13);
    }

    GuardedAction resetGuardedAction = new GuardedAction(ExampleEfsmUtils.ALWAYS_TRUE_GUARD,
        new Action(Collections.<Assignment>emptyList()));
    resetGuardedAction.addClock(clockEvent);
    efsm.addResetGuardedAction(resetGuardedAction);

    efsmModule.addNonInitialEfsm(efsm);
    SimplePipelinedEfsm.EFSM_MODEL.addModule(efsmModule);
  }

  private SimplePipelinedEfsm() {
  }
}
