/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import org.junit.Assert;
import org.junit.Test;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.SolverUtilTest;

/**
 * Test for {@link SolverUtils} class.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class SolverUtilTestCase extends SolverUtilTest {

  /**
   * Default constructor.
   */
  public SolverUtilTestCase() {
    super();
  }

  @Test
  public void runTest() {

    runIsSatTest();
  }

  private void runIsSatTest() {

    /* Test constraint is: true == true.*/
    final Node satNode = NodeUtils.getEquation(SolverUtils.TRUE, SolverUtils.TRUE);

    checkExpression(satNode);

    /* Test constraint is: true == false. */
    final Node unSatNode = NodeUtils.getEquation(SolverUtils.TRUE, SolverUtils.FALSE);

    checkExpression(unSatNode);

    /* Test constraint is: (var1 == 1) AND (var1 != 1). */
    final NodeVariable var1 = new NodeVariable("var1", DataType.INTEGER);
    final NodeValue one = NodeValue.newInteger(1);
    final Node unSatIntegerNode =
        ExprUtils.getConjunction(
            NodeUtils.getEquation(var1, one),
            new NodeOperation(StandardOperation.NOTEQ, var1, one));
    checkExpression(unSatIntegerNode);

    /* Test constraint is: (var1 == 1) AND ((var2 == 0) OR (var2 != 0)). */
    final NodeVariable var2 = new NodeVariable("var2", DataType.INTEGER);
    final NodeValue zero = NodeValue.newInteger(0);
    final Node satIntegerNode =
        ExprUtils.getConjunction(
            NodeUtils.getEquation(var1, one),
            ExprUtils.getDisjunction(
                NodeUtils.getEquation(var2, zero),
                new NodeOperation(StandardOperation.NOTEQ, var2, zero)));
    checkExpression(satIntegerNode);

    /* Test constraint is: (0 >= 0) AND (0 <= 2) AND (0 >= -128) AND (0 <= 127)
     * AND (0 >= -128) AND (0 <= 127) AND (0 == 0) */
    final NodeValue two = NodeValue.newInteger(2);
    final NodeValue max = NodeValue.newInteger(127);
    final NodeValue min = NodeValue.newInteger(-128);
    final Node constantNode =
        ExprUtils.getConjunction(
            ExprUtils.getConjunction(
                new NodeOperation(StandardOperation.GREATEREQ, zero, zero),
                new NodeOperation(StandardOperation.LESSEQ, zero, two)),
            ExprUtils.getConjunction(
                new NodeOperation(StandardOperation.GREATEREQ, zero, min),
                new NodeOperation(StandardOperation.LESSEQ, zero, max)),
            ExprUtils.getConjunction(
                new NodeOperation(StandardOperation.GREATEREQ, zero, min),
                new NodeOperation(StandardOperation.LESSEQ, zero, max)),
            NodeUtils.getEquation(zero, zero));
    checkExpression(constantNode);

    /* Test constraint is: (var2 <= 2) AND (var1 > 0) AND (var2 > 0) */
    final Node positiveNode =
        ExprUtils.getConjunction(
            new NodeOperation(StandardOperation.LESSEQ, var2, two),
            new NodeOperation(StandardOperation.GREATER, var1, zero),
            new NodeOperation(StandardOperation.GREATER, var2, zero));
    checkExpression(positiveNode);

    /* Test constraint is: (bool1 AND bool2) AND (bool3 -> bool2) AND (NOT(bool1 AND bool2)) */
    final NodeVariable bool1 = new NodeVariable("bool1", DataType.BOOLEAN);
    final NodeVariable bool2 = new NodeVariable("bool2", DataType.BOOLEAN);
    final NodeVariable bool3 = new NodeVariable("bool3", DataType.BOOLEAN);
    final Node forthOrderNode =
        ExprUtils.getConjunction(
            new NodeOperation(StandardOperation.AND, bool1, bool2),
            new NodeOperation(StandardOperation.IMPL, bool3, bool2),
            new NodeOperation(StandardOperation.NOT,
                new NodeOperation(StandardOperation.AND, bool1, bool2)));
    checkExpression(forthOrderNode);

    /* Test constraint is: (NOT(bool1 AND bool2)) AND (bool3 -> bool2) AND (bool1 AND bool2) */
    final Node backOrderNode =
        ExprUtils.getConjunction(
            new NodeOperation(StandardOperation.NOT,
                new NodeOperation(StandardOperation.AND, bool1, bool2)),
            new NodeOperation(StandardOperation.IMPL, bool3, bool2),
            new NodeOperation(StandardOperation.AND, bool1, bool2));
    checkExpression(backOrderNode);
  }

  private void checkExpression(final Node satNode) {
    if (differentSolverResults(satNode)) {
      Assert.fail();
    }
  }

  private boolean differentSolverResults(final Node satNode) {
    return SolverUtils.isSat(satNode) != ExprUtils.isSAT(satNode);
  }
}
